/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2002 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Lang
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "globals.h"
#include "mvParse.h"
/*
  Function Subassign() implements assignment to subscripts
            arg1[list] <- arg2

  Modified by kb to allow the subscripts to be a single matrix, with each
  row providing a full set of subscripts.  Unlike Element() in Lang.c, it
  requires a matrix (2 dimensional array) and does not work with an
  array with more than 2 dimensions.

  Modified by kb to allow negative subscripts, specifying positions not to
  be replaced
 
  Modified 920608 by the addition of an additonal argument.  It is now
  called twice by yypars, once with checkLhs != 0 to check the l.h.s, before
  parsing the r.h.s., and a second time with checkLhs == 0 after parsing
  the r.h.s.

  Modified 920728 so that it returns an object of the expected shape, that is
  for example  'b<- a[1,] <- 3' sets the first row of a to 3 and sets
  b to a row vector of 3's with dim(b)[2] == dim(a)[2].  Previously it
  simply returned arg1.

  950825 modified so as to recognize macros LEFTHANDSIDE and RIGHTHANDSIDE
  The last argument is LEFTHANDSIDE if only checking the lefthand side of
  the assignment.

  980219 a[..., j,...] <- NULL, a[...,j,...] <- scalar now allowed when
  (a) j is NULL
  (b) j is rep(F,n)
  (c) j is -run(n)
  where n is the length of the dimension.  In all cases, no change is made
  to a and the result returned is NULL In particular, x[x>10] <- 10 is
  allowed, even when max(x) <= 10, as also is x[x>10] <- y[x>10]
  This change is based on a suggestion of Peter Fortini
   (Peter_Fortini@ST.cytec.com)

  980223 Fixed bug so that a[] <- b is allowed if b is a scalar or has
         the same length as a.

  990429 Modified to allow assignment to structure components specified
         by subscripts

  010726 Modified to prevent assignment to subscript of locked variable

  020824 Modified to reuse LHS CHARACTER string is all replacement items
         match in size

  020826 Moved code out of Subassign() into new static functions
         AssignToSpecial(), AssignToChar() and AssignToStructure()

  030913 moved definitions of constants LEFTHANDSIDE and RIGHTHANDSIDE to
         new header mvParse.h
*/


#define Trash          NAMEFORTRASH
static Symbolhandle    Trash = SymbolZero;

enum SubassgnScratch
{
	GITEMLISTH = 0,
	GCACHEDCOMPSH,
	GALLSUBSH,
	GCNEWARG1H,
	NTRASH
};

/*
  020826 New function comprised of code from Subassign()
*/

static Symbolhandle AssignToSpecial(Symbolhandle arg1, Symbolhandle arg2,
									long **itemListH, long nitems)
{
	ShortSymbol     symbol;
	ShortSymbol    *symbolP = &symbol;
	Symbolhandle    symbolH = (Symbolhandle) &symbolP;

	/*
	  Initialize a fake Symbolhandle to be used as argument to SETSPECIAL
	*/
	setTYPE(symbolH, LONG);
	markFakeSymbol(symbolH);
	setLONGDATA(symbolH, itemListH);
	setNDIMS(symbolH, 1);
	setDIM(symbolH, 1, nitems);

	return (SETSPECIAL(arg1, arg2, symbolH, ASSIGNSPECIAL));
} /*AssignToSpecial()*/

/*
  020825 New function comprised of code taken from Subassign()
      arg1 and arg2 are Symbolhandles for LHS and RHS
  	  *itemListH contains all the element numbers in the destination, base 0

*/
static Symbolhandle AssignToChar(Symbolhandle arg1, Symbolhandle arg2,
								 long ** itemListH, long nitems)
{
	int      newHandleNeeded = arg1 == arg2;
	long     space, i, j, jLast;
	long     arg1tot = symbolSize(arg1);
	long     arg2tot = symbolSize(arg2);
	long    *itemListP = *itemListH;
	long     length;
	char    *ch1, *ch2;
	char   **cNewArg1H, *cNewArg1P;

    /*
	  compute space needed for the string handle consisting of
	  the LHS string handle with certain elements replaced
	*/
	space = 0;
	ch1 = STRINGPTR(arg1);
	jLast = -1;

	for (i = 0; i < arg1tot; i++)
	{ /* compute space needed */
		for (j = nitems - 1;j >= 0; j--)
		{ /* see if string i is to be replaced */
			if (itemListP[j] == i)
			{
				break;
			}
		} /*for (j = 0;j < nitems; j++)*/

		length = strlen(ch1) + 1;

		if (j >= 0)
		{ /* string i will be replaced */
			long       newLength;

			if (arg2tot == 1)
			{
				j = 0;
			}
			if (j < jLast || jLast < 0)
			{
				ch2 = STRINGPTR(arg2);
				jLast = 0;
			}

			if (j > jLast)
			{
				ch2 = skipStrings(ch2,j-jLast);
				jLast = j;
			}

			newLength = strlen(ch2) + 1;
			if (!newHandleNeeded && newLength != length)
			{
				newHandleNeeded = 1;
			}
			length = newLength;
		} /*if (j < 0){}else{}*/
		ch1 = skipStrings(ch1,1L);
		space += length;
	} /*for (i = 0; i < arg1tot; i++)*/

	ch1 = STRINGPTR(arg1);
	if (newHandleNeeded)
	{
		/* 
		   Either LHS and RHS the same, or at least one element of LHS
		   to be replaced by element of different number of characters
		*/
		if (!getScratch(cNewArg1H,GCNEWARG1H,space,char))
		{
			goto errorExit;
		}
	}
	else
	{
		/*All replacement items are the right size*/
		cNewArg1H = STRING(arg1);
	}
	cNewArg1P = *cNewArg1H;

	/*
	  Now replace LHS elements with RHS elements;  if it's into a
	  new handle, copy the elements not replaced

	  When a new handle was not needed, ch1 should remain the same as
	  cNewArgP; otherwise they both advance in the original arg1 (ch1)
	  and new arg1 (cNewArg1P)
	*/
	jLast =- 1;
	for (i = 0; i < arg1tot; i++)
	{
		for (j = nitems - 1;j >= 0; j--)
		{
			if (itemListP[j] == i)
			{
				break;
			}
		} /*for (j = nitems-1;j >= 0; j--)*/

		if (j < 0)
		{ /* string i not to be replaced and hence not part of result*/
			if (newHandleNeeded)
			{
				cNewArg1P = copyStrings(ch1, cNewArg1P, 1);
				ch1 = skipStrings(ch1, 1);
			}
			else
			{
				cNewArg1P = ch1 = skipStrings(ch1, 1);
			}
		} /*if (j < 0)*/
		else
		{ /* string i to be replaced and new piece added to result*/
			if (arg2tot == 1)
			{
				j = 0;
			}

			if (j < jLast || jLast < 0)
			{
				ch2 = STRINGPTR(arg2);
				jLast = 0;
			}

			if (j > jLast)
			{
				ch2 = skipStrings(ch2,j-jLast);
				jLast = j;
			}

			cNewArg1P = copyStrings(ch2, cNewArg1P, 1);
			if (newHandleNeeded)
			{
				ch1 = skipStrings(ch1,1L); /* skip string being replaced */
			}
			else
			{
				ch1 = cNewArg1P;
			}
		} /*if (j < 0){}else{}*/
	} /*for (i = 0; i < arg1tot; i++)*/

	if (newHandleNeeded)
	{
		/* discard old contents of arg1 */
		mydisphandle((char **) STRING(arg1));
		unTrash(GCNEWARG1H);
		setSTRING(arg1, cNewArg1H);
	}

	return (arg1);
	
  errorExit:
	return (SymbolError);

} /*AssignToChar()*/

/*
  020825 New function comprised of code taken from Subassign()
      dest is Symbolhandle for new LHS variable
      arg1 and arg2 are Symbolhandles for LHS and RHS
  	  *itemListH contains all the element numbers in the destination, base 0
      nitems is the number of items on LHS being replaced
      dest is the same as arg1 unless arg1 and arg2 are the same

*/

static Symbolhandle AssignToStructure(Symbolhandle dest, Symbolhandle arg2,
									  Symbolhandle result,
									  long ** itemListH, long nitems)
{
	Symbolhandle    symhTemp;
	Symbolhandle  **cachedCompsH;
	long           *itemListP = *itemListH;
	long            i, j;
	int             arg2tot = symbolSize(arg2);
	int             rightIsScratch = isscratch(NAME(arg2));
	int             rightIsStructure = TYPE(arg2) == STRUC;
	
	if (rightIsStructure && (nitems != arg2tot || arg2tot == 1))
	{
		rightIsStructure = 0;
	}

	/* Save the components to be replaced in case we have to abort*/
	if (!getScratch(cachedCompsH, GCACHEDCOMPSH, nitems, Symbolhandle))
	{
		goto errorExit;
	}

	for (j = nitems - 1; j >= 0; j--)
	{
		/*
		  Save the handle to be replaced and zero its value
		  in dest.
		*/
		(*cachedCompsH)[j] = COMPVALUE(dest, itemListP[j]);
		COMPVALUE(dest, itemListP[j]) = SymbolZero;
	} /*for (j = 0; j < nitems; j++)*/

	/*
	  Fill in from the end only in slots that are
	  EmptyArgument so as to avoid making copies that
	  will later be thrown away.  This makes a difference only
	  when there are duplicate subscripts
	*/
	for (j = nitems-1; j >= 0; j--)
	{
		Symbolhandle       arg2Temp =
		  (rightIsStructure) ? COMPVALUE(arg2, j) : arg2;
		long position = itemListP[j];

		if (COMPVALUE(dest, position) == EmptyArgument)
		{
			/*
			  Destination component not yet set
			  Copy r.h.s.  If it is scratch, we will use it later
			  in creating result
			*/
			symhTemp = COMPVALUE(dest, position) = Makesymbol(REAL);

			if (symhTemp == SymbolError ||	!Copy(arg2Temp, symhTemp))
			{
				/* Restore LHS to original condition */
				itemListP = *itemListH;
				/* Delete components previously copied*/
				for (i = 0; i < nitems; i++)
				{
					Delete(COMPVALUE(dest, itemListP[i]));
					COMPVALUE(dest, itemListP[i]) = SymbolZero;
				}
				/*
				  Restore original handles
				*/
				for (i = 0; i < nitems; i++)
				{
					COMPVALUE(dest, itemListP[i]) = (*cachedCompsH)[i];
				}
				goto errorExit;
			} /*if (symhTemp == SymbolError || !Copy(arg2Temp, symhTemp))*/

			/*
			  Make sure l.h.s. name gets retained
			*/
			if ((*cachedCompsH)[j] != SymbolZero)
			{
				setNAME(symhTemp, NAME((*cachedCompsH)[j]));
			}
			else if (!rightIsStructure)
			{
				char           compname[NAMELENGTH + 1];

				sprintf(compname, "Comp%ld", j+1);
				setNAME(symhTemp, compname);
			}
		} /*if (COMPVALUE(dest, (*itemListH)[j]) == SymbolZero)*/
	} /*for (j = 0; j < nitems; j++)*/

	/*
	  If we get here, destination has been successfully assigned
	  Delete all the cached components that were replaced
	*/

	for (j = 0; j < nitems; j++)
	{
		Delete((*cachedCompsH)[j]);
	}
	discardScratch(cachedCompsH, GCACHEDCOMPSH, Symbolhandle);

	/*
	  Now add components to result
	*/
	if (rightIsScratch && rightIsStructure)
	{
		/* just move Symbolhandles to result */
		for (j = 0; j < nitems; j++)
		{
			COMPVALUE(result, j) = COMPVALUE(arg2, j);
			COMPVALUE(arg2, j) = SymbolZero;
		}
	} /*if (rightIsScratch && rightIsStructure)*/
	else
	{
		/*
		  RHS is not scratch or is not a structure
		  Copy Symbolhandles to result.
		*/
		for (j = 0; j < nitems; j++)
		{
			Symbolhandle       arg2Temp =
			  (rightIsStructure) ? COMPVALUE(arg2, j) : arg2;

			if (!rightIsScratch || j < nitems - 1)
			{
				/* Have to make a copy */
				symhTemp = COMPVALUE(result, j) = Makesymbol(REAL);
				if (symhTemp == SymbolError ||
					!Copy(arg2Temp, symhTemp))
				{
					Removesymbol(result);
					result = NULLSYMBOL;
					break;
				}
			}
			else
			{
				/* just move scratch RHS to result on last item */
				Cutsymbol(arg2);
				symhTemp = COMPVALUE(result, j) = arg2;
				arg2 = SymbolZero;
			}

			if (!rightIsStructure)
			{
				char        compname[NAMELENGTH+1];

				sprintf(compname, "Comp%ld", j+1);
				setNAME(symhTemp, compname);
			}
		} /*for (j = 0; j < nitems; j++)*/
	} /*if (isscratch(NAME(arg2)) && rightIsStructure){}else{}*/
	return (dest);
	
  errorExit:
	return (SymbolError);

} /*AssignToStructure()*/

/*
   Routine to replace elements of arrays

   961101 Changed code so that, for example,
   x[vector(1,1)] <- a, where length(a) == 2 assigns a[2] to x[1],
   for REAL, LOGICAL and CHARACTER variables.  Previously a[1] was
   assigned for CHARACTER variables.

   980312 Added argument completeDim to buildSubscript() call; not needed here
          but used by extractIt() in Lang.c
          Added awareness of NOMISSING flag.  Left hand side will have flag
          set only if both r.h.s and subscripted variable have it set.  Even
          if all MISSING values in subscripted variable are replaced by non-
          MISSING values, the flag will not be set.

   980616 Modified slightly so that a[,] <- b, a[,,] <- b, etc are treated
          the same as a[] <- b, viz., it's an error unless b is a scalar
          or length(a) = length(b)

   990419 Modified so that it handles special variables in a standard way.
          When arg1 is special, after checking, SETSPECIAL is called
		  with third argument a fake symbolhandle for a LONG vector
		  containing all the destination positions.  It is up to the
		  setSpecial function to do the assignment and create a value
		  to return.  The setSpecial function should print any error messages.
		  Subscripts are checked against the dimensions of the special
		  symbol
   990422 Modified so that assignment to subscripts of a structure is
          defined.  If the r.h.s is not a structure or if it is a
          structure for which the number of top level components does
          not match nitems, the total number of subscripts, then
          A copy of the r.h.s. is replaces each of the targetted components.
          Otherwise, a component of the r.h.s replaces a targetted
          component.  If there are duplicates among the targetted components,
          only the last gets retained.  The value returned is a
          structure with nitems components even if there are duplicates
          among the targetted components.
   990602 Fixed bug in assignment of scratch variable to structure component
   000222 Str1[j] < Str2 always puts all of rhs in Str1 when j selects
          1 component; formerly, if Str was a structure with 1 component,
          that component was assigned.
   000229 Fixed bug causing crash with x[i] <- NULL
   010726 Illegal to assign to subscript of locked variable
   020824 With CHARACTER variables, if LHS is not the same as RHS and
          all LHS elements are replaced by strings of the same size,
          new items are inserted into the original LHS handle instead
          if a new handle being created by merging the old LHS with the
          RHS elements
   020825 Moved some code from Subassign() to new functions AssignToSpecial(),
          AssignToChar() and AssignToStructure()
*/
Symbolhandle    Subassign(Symbolhandle arg1, Symbolhandle list,
						  Symbolhandle arg2, long whichSide)
{
	Symbolhandle    subscrSymh, result = SymbolZero;
	Symbolhandle    dest = SymbolZero;
	long            i, i2, i2inc, j, k;
	long            arg1tot, arg2tot, nitems, factor;
	long            length;
	long            maxSubscr;
	long            ii[MAXDIMS], resdim[MAXDIMS];
	long            ndims, nsubs;
	long          **itemListH = (long **) 0;
	long          **allsubsH = (long **) 0;
	long           *subValues[MAXDIMS+1];
	long           *itemListP;
	long            dims[2];
	long            needed = 0;
	long            position;
	long            subLimit;
	long            arg1Type, arg2Type, subscrType;
	int             matrixSubscr = 0, sameSymbol = 0;
	int             foundNullSubscr = 0;
	int             leftIsStructure;
	int             rightIsNull = 0, rightIsScalar, rightIsStructure;
	int             leftIsSpecial;
	int             useItemList = 0;
	char           *ch2;
	double          nclass;
	double          y;
	double         *subscript;
	char          **cNewArg1H = (char **) 0;
	char           *cResultP = (char *) 0;
	char            namearg1[NAMELENGTH+1];
	WHERE("Subassign");
	
	*OUTSTR = '\0';
	nsubs = NCOMPS(list);
	
	if (arg1 == SymbolZero)
	{ /* paranoia; should not happen */
		sprintf(OUTSTR, "ERROR: no variable in subscript assignment");
		goto yyerrorMsg;
	}
	/* check to make sure subscripts are of proper type and number */

	arg1Type = TYPE(arg1);
	leftIsSpecial = isSpecial(arg1);
	leftIsStructure = arg1Type == STRUC;
	
	if (whichSide == LEFTHANDSIDE)
	{
		/*
		  Haven't yet reached RHS.  Are subscripts OK and
		  is left side OK for assignment?
		*/
		if (isLocked(arg1))
		{
			sprintf(OUTSTR,
					"ERROR: illegal to assign to subscript of a locked variable");
		}
		else if (nsubs > MAXDIMS)
		{
			sprintf(OUTSTR,
					"ERROR: more than %ld subscripts", (long) MAXDIMS);
		}
		else if (nsubs != 1 && nsubs < NDIMS(arg1))
		{
			sprintf(OUTSTR, "ERROR: too few subscripts");
		}
		else if (leftIsSpecial)
		{
			/*
			  Left side is special.  Ask it if it can have subscripts that
			  can be assigned to.

			  Note:  Because this check is before a type check, arg1
			  potentially can be of any type
			*/
			if (SETSPECIAL(arg1, SymbolZero, list, CHECKLEFTSPECIAL) ==
				SymbolZero)
			{
				/* The setSpecial function should print any error messages */
				return (0);
			}
		}
		else if (!isFakeSymbol(arg1) && !myvalidhandle((char **) arg1) ||
				 TYPE(arg1) == LIST)
		{
			sprintf(OUTSTR,
					"ERROR: use of subscripts with damaged (deleted?) variable");
		}
		else if (!isDefined(arg1))
		{
			sprintf(OUTSTR,
					"ERROR: use of subscripts with undefined variable %s",
					NAME(arg1));
		}
#if (0) /* 990423 can now assign to components of a structure */
		else if (arg1Type == STRUC)
		{
			sprintf(OUTSTR,
					"ERROR: illegal to assign to a structure component");
		}
#endif /*0*/
		else if (isscratch(NAME(arg1)))
		{
			sprintf(OUTSTR,
					"ERROR: assignment to subscripted number, expression or function result");
		}
		else if (arg1Type != REAL && arg1Type != LOGIC && arg1Type != CHAR &&
				 arg1Type != STRUC)
		{
			sprintf(OUTSTR,"ERROR: use of subscripts with %s variable",
					typeName(arg1Type));
		}

		if (*OUTSTR)
		{
			yyerror(OUTSTR);
			arg1 = SymbolZero;
		}
		return (arg1);
	} /*if (whichSide == LEFTHANDSIDE)*/
	
	/* whichSide must be RIGHTHANDSIDE.  Now carry out assignment */
	
	arg2Type = TYPE(arg2);

	if (isNull(arg2) && arg2Type != NULLSYM || !isDefined(arg2))
	{
		sprintf(OUTSTR, "ERROR: replacement elements are %s",
				(isNull(arg2)) ? "null" : "undefined");
	}
	else if (leftIsSpecial)
	{
		if (SETSPECIAL(arg1, arg2, list, CHECKRIGHTSPECIAL) == SymbolError)
		{
			goto yyerrorMsg;
		}
	}
	else if (arg2Type == NULLSYM)
	{
		rightIsNull = 1;
	}
	else if (!leftIsStructure && arg2Type != arg1Type)
	{
		sprintf(OUTSTR,"ERROR: replacement elements of wrong type");
	}

	if (*OUTSTR)
	{
		goto yyerrorMsg;
	}
	
	ndims = NDIMS(arg1);
	
	arg1tot = symbolSize(arg1);
	arg2tot = symbolSize(arg2);

	rightIsStructure = arg2Type == STRUC;
	
	if (leftIsStructure && !rightIsStructure)
	{
		/* Entire RHS will be replicated for each LHS element being replaced */
		arg2tot = 1;
		rightIsScalar = 1;
	}
	else
	{
		rightIsScalar = isScalar(arg2);
	}
	
	/* check for empty subscrtips */
	for (i = 0; i < nsubs && COMPVALUE(list, i) == EmptyArgument; i++)
	{
		;
	}

	if (i == nsubs)
	{ /* all empty subscripts*/
		if (arg1 == arg2)
		{
			/* a[] <- a do nothing but return a as result*/
			return (arg1);
		}
		if (arg1tot != arg2tot && !rightIsScalar && !leftIsStructure)
		{
			sprintf(OUTSTR,"ERROR: wrong number of replacement items");
			goto yyerrorMsg;
		}
		subscrSymh = SymbolZero;
	} /*if (i == nsubs)*/
	else
	{
		for (i = 0; i < nsubs; i++)
		{
			subscrSymh = COMPVALUE(list, i);
			if (subscrSymh == EmptyArgument)
			{/* empty subscript */
				continue;
			}

			if (!isDefined(subscrSymh))
			{
				sprintf(OUTSTR,"ERROR: undefined subscript");
			}
			else if ((subscrType = TYPE(subscrSymh)) != LOGIC &&
					 subscrType != REAL && subscrType != NULLSYM)
			{
				sprintf(OUTSTR,"ERROR: subscript not REAL, LOGICAL or NULL");
			}
			else if (subscrType == NULLSYM)
			{
				if (!rightIsNull && !rightIsScalar)
				{
					sprintf(OUTSTR,
							"ERROR: assigning non-NULL non-scalar to NULL subscript");
				}
				else
				{
					foundNullSubscr = 1;
				}
			}
			else if (anyMissing(subscrSymh))
			{
				sprintf(OUTSTR,
						"ERROR: MISSING values not allowed as subscripts");
			}
			else if (!isVector(subscrSymh))
			{
				/* must be matrix subscript; not legal when arg1 is structure */
				if (nsubs > 1)
				{
					sprintf(OUTSTR,"ERROR: matrix used as marginal subscript");
				}
				else if (subscrType != REAL)
				{
					sprintf(OUTSTR,"ERROR: matrix of subscripts must be REAL");
				}
				else if (!isMatrix(subscrSymh, dims))
				{
					sprintf(OUTSTR,
							"ERROR: array of subscripts illegal on assignment");
				}
				else if (dims[1] != ndims)
				{
					sprintf(OUTSTR,
							"ERROR: matrix of subscripts has wrong dimensions");
				}
				else if (arg2tot > 1 && arg2tot != dims[0])
				{
					sprintf(OUTSTR,
							"ERROR: number of rows of matrix of subscripts != number of values");
				}
				else
				{
					matrixSubscr = 1;
				}
			} /*else if (!isVector(subscrSymh))*/
			if (*OUTSTR)
			{
				goto yyerrorMsg;
			}
		} /*for (i = 0; i < nsubs; i++)*/

		if (foundNullSubscr)
		{		
			/* no further check of subscripts; do nothing */
			goto returnNull;
		} /*if (foundNullSubscr)*/
	}  /*if (i == nsubs){}else{}*/
	
	useItemList = (arg1Type == CHAR || leftIsStructure || leftIsSpecial);

	if ((Trash = GarbInstall(NTRASH)) == SymbolError)
	{
		goto errorExit;
	}

	strcpy(namearg1, NAME(arg1));

	for (i = 0; i < nsubs;i++)
	{
		ii[i] = 0;
		subValues[i] = (long *) 0;
	}

	if (matrixSubscr)
	{/* matrix of subscripts */
		k = 0;
		/* check subscripts for validity */
		nitems = dims[0];
		for (j = 0; j < ndims;j++)
		{
			subLimit = DIMVAL(arg1,j+1);
			for (i = 0;i < nitems;i++, k++)
			{
				y = DATAVALUE(subscrSymh, k);
				if (y <= 0 || y > subLimit)
				{
					sprintf(OUTSTR,"ERROR: subscript out of range");
				}
				else if (y != floor(y))
				{
					sprintf(OUTSTR,"ERROR: non-integral subscript");
				}
				if (*OUTSTR)
				{
					goto yyerrorMsg;
				}
			} /*for (i = 0;i < nitems;i++, k++)*/
		} /*for (j = 0;j < ndims;j++)*/

		if (useItemList && !getScratch(itemListH,GITEMLISTH,nitems,long))
		{
			goto errorExit;
		}

		resdim[0] = 1;
		for (j = 1; j < ndims;j++)
		{
			resdim[j] = DIMVAL(arg1, j)*resdim[j-1];
		}

		if (!leftIsSpecial)
		{
			/* Note arg1 cannot be a structure when matrixSubscr != 0*/
			if (arg1Type == CHAR)
			{
				result = Install(SCRATCH, CHAR);
			}
			else
			{
				result = RInstall(SCRATCH, nitems);
			}

			if (result == SymbolError)
			{
				goto errorExit;
			}
			setTYPE(result, arg1Type);
		} /*if (!leftIsSpecial)*/
	} /*if (matrixSubscr)*/
	else
	{
		/* List of subscripts */

		if (arg1Type == REAL && NCLASS(arg1) > 0)
		{	/* destination is factor; check values for validity */
			nclass = (double) NCLASS(arg1);

			for (i = 0; i < arg2tot;i++)
			{
				y = DATAVALUE(arg2, i);
				if (!isMissing(y) && (y != floor(y) || y < 1.0 || y > nclass))
				{
					sprintf(OUTSTR,
							"ERROR: attempt to assign non-integer or out of range values to a factor");
					goto yyerrorMsg;
				}
			} /*for (i = 0; i < arg2tot;i++)*/
		} /*if (arg1Type == REAL && NCLASS(arg1) > 0)*/

		factor = nitems = 1;
		for (i = 0; i < nsubs;i++)
		{
			subscrSymh = COMPVALUE(list,i);
			/* we know that subscrSymh cannot be NULL*/
			if (nsubs == 1)
			{
				maxSubscr = arg1tot;
			}
			else
			{
				maxSubscr = (i < ndims) ? DIMVAL(arg1,i+1) : 1;
			}
			resdim[i] = subscriptLength(subscrSymh ,maxSubscr);
			
			if (resdim[i] == 0)
			{
				goto errorExit;
			}
			needed += resdim[i];
		} /*for (i = 0; i < nsubs;i++)*/

		if (!getScratch(allsubsH,GALLSUBSH,needed,long))
		{
			goto errorExit;
		}
		/*
		  subValues is a vector of pointers to positions in long
		  handle allsubsH.

		  For each subscript component successively, buildSubscript()
		  fills vector subValues[i] with decoded subscripts for subscript
		  i, returning the number of such subscripts which is saved in
		  resdim[i].  The decoded subscripts are then multiplied by
		  the product of the preceding dimensions of the destination so
		  that the final position can be found by adding one value
		  from each of the subValue vectors

		  nitems is computed as the total number of items being assigned

		  For CHARACTER, STRUCT and Special variables, the positions
		  are put in *itemListH for later use.  For LOGICAL and REAL
		  variables, the replacements are made immediately.
		*/
		subValues[0] = *allsubsH;
		for (i = 0; i < nsubs;i++)
		{
			subscrSymh = COMPVALUE(list,i);
			if(nsubs == 1)
			{
				maxSubscr = symbolSize(arg1);
			}
			else
			{
				maxSubscr = (i < ndims) ? DIMVAL(arg1,i+1) : 1;
			}
			resdim[i] = buildSubscript(subscrSymh, maxSubscr, resdim[i], 
									   subValues[i], (int *) 0);
			if (resdim[i] < 0)
			{
				goto errorExit;
			} /*if (resdim[i] < 0)*/

			if (resdim[i] == 0)
			{
				/* all F or -run(maxSubscr) */
				foundNullSubscr = 1;
				nitems = 0;
				break;
			}
			
			if (i > 0)
			{
				long     *subi = subValues[i];
		
				for (j = 0; j < resdim[i];j++)
				{
					subi[j] *= factor;
				} /*for (j = 0; j < resdim[i];j++)*/
			} /*if (i > 0)*/

			factor *= maxSubscr;
			nitems *= resdim[i];
			subValues[i+1] = subValues[i] + resdim[i];
		} /*for (i = 0; i < nsubs;i++)*/

		/* nitems is number of elements specified by subscript(s) */

		if (leftIsStructure && rightIsStructure &&
			(nitems != arg2tot || arg2tot == 1))
		{
			/*
			  Don't treat RHS as structure; just as variable to be
			  replicated for every component on left being assigned
			  to
			  Note: It is impossible to have the wrong number of
			  replacement items when leftIsStructure != 0
			*/
			rightIsStructure = 0;
			rightIsScalar = 1;
			arg2tot = 1;
		} /*if (leftIsStructure && rightIsStructure && nitems != arg2tot)*/
		
		if (foundNullSubscr && (rightIsNull || rightIsScalar))
		{
			emptyTrash();
			goto returnNull;
		}

		if (arg2tot != 1 && nitems != arg2tot)
		{
			sprintf(OUTSTR, "ERROR: wrong number of replacement items");
			goto yyerrorMsg;
		} /*if (arg2tot > 1 && nitems != arg2tot)*/

		if (useItemList && !getScratch(itemListH,GITEMLISTH,nitems,long))
		{
			goto errorExit;
		}

		if (!leftIsSpecial)
		{
			/* Note: result is the value returned, not the variable changed */
			if (arg1Type == CHAR)
			{
				result = Install(SCRATCH, CHAR);
			}
			else if (leftIsStructure)
			{
				result = StrucInstall(SCRATCH, nitems);
			}
			else
			{
				result = RInstall(SCRATCH, nitems);
			}

			if (result == SymbolError)
			{
				goto errorExit;
			}
			else if (arg1Type != STRUC)
			{
				setTYPE(result, arg1Type);
			}
		}
	} /*if (matrixSubscr){}else{}*/

	/*
	  dest is symbol for the subscripted symbol
	  result is symbol to be returned

	  If arg1 and arg2 are the same and are not CHARACTER or Special, arg2 is
	  copied to dest and is removed only in the case of success.  This
	  is to ensure arg2 is not corrupted if the assignment can't
	  be completed.
	*/

	sameSymbol = (arg1 == arg2) && arg1Type != CHAR && !leftIsSpecial;

	if (!sameSymbol)
	{
		/* This case includes CHAR and special on left */
		dest = arg1;
	} /*if (!sameSymbol)*/
	else
	{/* allows, e.g., a[run(length(a),1)] <- a to work correctly */
		dest = Install(namearg1, REAL);
		if (dest == SymbolError || !Copy(arg1, dest))
		{
			goto errorExit;
		}
		Cutsymbol(arg1); /* move arg1 to Scratch symbol table */
		setNAME(arg1, SCRATCH);
		Addsymbol(arg1);
	} /*if (!sameSymbol{}else{})*/

	/*
	  For REAL or LOGIC insert values; for CHAR, STRUCT and
	  special fill (*itemListH)
	*/
	if (useItemList)
	{
		itemListP = *itemListH;
	}

	i2inc = (rightIsScalar) ? 0 : 1;
	if (matrixSubscr)
	{
		/*
		  Matrix of subscripts
		  Left can't be structure
		*/
		for (i = i2 = 0; i < nitems;i++)
		{
			subscript = DATAPTR(subscrSymh);
			position = (long) subscript[i] - 1;

			for (j = 1;j< ndims;j++)
			{
				subscript += nitems;
				position += resdim[j]*((long) subscript[i] - 1);
			}
			if (useItemList)
			{
				itemListP[i] = position;
			}
			else
			{
				DATAVALUE(result, i) = DATAVALUE(dest, position) =
				  DATAVALUE(arg2, i2);
				i2 += i2inc;
			} /*for (i = 0; i < nitems;i++)*/
		} /*for (i = i2 = 0; i < nitems;i++)*/
	} /*if (matrixSubscr)*/
	else
	{ /* list of subscripts */
		long      tail = -1;

		for (i = i2 = 0;i < nitems;i++)
		{
			position = (subValues[0])[ii[0]];
			if (tail < 0)
			{
				tail = 0;
				for (j = 1;j < nsubs;j++)
				{
					tail += (subValues[j])[ii[j]];
				}
			} /*if (tail < 0)*/
			position += tail;

			if (useItemList)
			{
				itemListP[i] = position;
			}
			else
			{
				DATAVALUE(result, i) = DATAVALUE(dest, position) =
				  DATAVALUE(arg2, i2);
				i2 += i2inc;
			}

			for (j = 0; j < nsubs;j++)
			{ /* turn over odometer */
				if (++ii[j] < resdim[j])
				{
					break;
				}
				ii[j] = 0;
				tail = -1;
			} /*for (j = 0; j < nsubs;j++)*/
		} /*for (i = 0;i < nitems;i++)*/
	} /*if (matrixSubscr){}else{}*/

	if (leftIsSpecial)
	{
		result = AssignToSpecial(arg1, arg2, itemListH, nitems);
		
		if (result == SymbolError)
		{
			goto errorExit;
		}
	} /*if (leftIsSpecial)*/
	else if (arg1Type == CHAR)
	{
		/*
		  At this point *itemListH contains all the offsets in the destination
		*/
		
		int      newHandleNeeded = arg1 == arg2;

		/*
		  first create handle for result, either a copy of the RHS 
		  or rep(RHS,nitems)
		*/
		if (arg2tot == nitems)
		{
			TMPHANDLE = myduphandle(STRING(arg2));
		}
		else
		{ /* r.h.s a scalar */
			length = strlen(STRINGPTR(arg2)) + 1;

			TMPHANDLE = mygethandle(nitems * length);
		}
		setSTRING(result, TMPHANDLE);
		if (TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
		if (arg2tot != nitems)
		{
			cResultP = *TMPHANDLE;
			ch2 = STRINGPTR(arg2);

			/* Duplicate RHS nitems times */
			for (i = 0; i < nitems; i++)
			{
				cResultP = copyStrings(ch2, cResultP, 1);
			}
		}

		/*
		  At this point, the result handle has been created, either
		  a copy of RHS or rep(RHS,nitems)

		  Now compute space needed for the string handle consisting of
		  the LHS string handle with certain elements replaced
		*/
		if (AssignToChar(arg1, arg2, itemListH, nitems) == SymbolError)
		{
			goto errorExit;
		}
	}
	else if (leftIsStructure)
	{
		int            rightIsScratch = isscratch(NAME(arg2));
	
		if (AssignToStructure(dest, arg2, result, itemListH, nitems) ==
			SymbolError)
		{
			goto errorExit;
		}
		if (rightIsScratch)
		{
			arg2 = SymbolZero;
		}
	}

	if (!leftIsSpecial && !leftIsStructure)
	{
		if (!matrixSubscr)
		{
			setNDIMS(result, nsubs);
			for (j = 0; j < nsubs;j++)
			{
				setDIM(result, j+1, resdim[j]);
			}
		}
		else
		{ /* result is vector or structure*/
			setNDIMS(result, 1);
			setDIM(result, 1, nitems);
		}

#if (USENOMISSING)
		if (NOMISSING(arg2))
		{
			setNOMISSING(result);
		}
		else
		{
			clearNOMISSING(result);
			clearNOMISSING(dest);
		}
#endif /*USENOMISSING*/
	} /*if (!leftIsSpecial && !leftIsStructure)*/
	
	if (sameSymbol || arg2 != SymbolZero && isscratch(NAME(arg2)))
	{
		Removesymbol(arg2);
	}

	emptyTrash();

	return (result);

  returnNull:
	/*
	  Right side was either null or
	  right side was a scalar and one or more subscripts were NULL or
	  selected no elements (all F or -run(dimlength))
	*/
	if (isscratch(NAME(arg2)))
	{
		/* re-use arg2 */
		result = arg2;
		if (rightIsScalar)
		{
			DeleteContents(arg2);
			setTYPE(arg2, NULLSYM);
		}
		setNAME(arg2, NULLSCRATCH);
	}
	else
	{
		result = Install(SCRATCH, NULLSYM);
		if (result == SymbolError)
		{
			goto errorExit;
		}
	}
	return (result);

  yyerrorMsg:
	yyerror(OUTSTR);
	/* fall through */
  errorExit:
	putOUTSTR();

	if (sameSymbol && dest != SymbolError)
	{
		Removesymbol(dest);
		Cutsymbol(arg1); /* preserve arg1 */
		setNAME(arg1,namearg1);
		Addsymbol(arg1);
	}
	Removesymbol(result);
	emptyTrash();
	return (0);
} /*Subassign()*/


