/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 5.00 or later
*(C)*
*(C)* Copyright (c) 2005 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Anovacoe
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/


#include "globals.h"

#define Trash  NAMEFORTRASH


/* 
  This routine computes the model coefficients for linear models and glms.
  It does the simple thing for balanced linear models, and does some funny
  things for unbalanced models
  
  Usage: coefs() prints linear model coefficients, secoefs() also prints their
    std errors

  	coefs() or secoefs() : Coefficients for all terms
  	coefs(Term) or secoefs(Term) : Coefficients associated with given term
	secoefs(,varno) : All coefficients, variable varno (multivariate models)
	secoefs(Term,varno) : Coefficients for Term, variable varno (multivariate
	models)

	Keyword phrases se:F or coefs:F suppress the computation of either the
	standard errors or coefficients.  To have them both F is an error.
	The default values of se and coefs are F and T on coefs() and T and T on
	secoefs().

	When no specific term is specifed, keyword phrase byterm:F on secoefs()
	changes the form of the output from a structure with one component per
	term, each with components 'coefs' and 'se' to a structure with two
	componenents 'coefs' and 'se', each with one component for each term.
	byterm is ignored if a term is specified or if only coefs or only se are
	to be returned.

	For all usages of secoefs, one can use an extra argument of the form
	error:errorTerm where errorTerm is a CHAR variable which is the name of a
	term in the model that will be used to compute the MSE needed for SE's.
	This is ignored on iterative models (poisson(), logistic(), robust(),
	for which MSE is taken as 1.

	Term and errorTerm can either be CHARACTER variables (e.g., "a.b" or 
	"ERROR1" or a term number such as 1 (for CONSTANT).

	For reasons of backwards compatibility, 'anovacoefs' is synonymous with
    'coefs'.

	950828 fixed bug causing crashes on secoefs(coefs:F) when a uninitialized
	handle was being disposed of (in doBalanced()).
	
	971126 Changed calls to appendLabels to use enum codes in labels.h

    990212 Changed most uses of putOUTSTR() to putErrorOUTSTR() and
           some use of myerrorout to putOutErrorMsg()
    990215 Remaining uses of myerrorout() changed to putOutMsg()
	000210 Made changes to use TERMNAMES_AT when it exists to fix
           bug in finding a numerical term.  The term names used
           in labelling the components of the results are still
           taken from TERMNAMES
	000511 Added keyword phrase silent:T, removed some unnecessary externs,
           did other cleaning up
    020318 NVARS, NREGX, NY, etc. now assumed to be long
    021022 stripped code from calcCoefs() to create calcWts()
	041025 variates don't contribute dimensions to interaction terms
	050818 modified calcCoef() to suppress warning message with silent:T
*/

static Symbolhandle    Trash;

enum coefsTrash
{
	GTMPY = 0,
	GTMPMEANS,
	GTMPALPHA,
	GTMPWTS,
	GCOLPTR,
	NTRASH
};


/* get coords for the jth value in table */
static void getcoords(long j, long dim, long coords[], long *offsets)
{
	long            i;
	
	for (i = dim - 1; i >= 0; i--)
	{
		coords[i] = j / offsets[i];
		j -= coords[i] * offsets[i];
	}
} /*getcoords()*/

/*
   Set dimensions appropriately for coefficients or  standard error
   symbol for a particular term. 

   Dimension for each variable in term are the number of factor levels for
   a factor or 1 for a variate.  If y is multidimensional, we need an extra
   final dimension of ny

   It returns the product of dimensions, excluding ny if multidimensional.
*/
static long setSymbolDims(Symbolhandle symh, long term[])
{
	long          i, ndims = 0, ncells = 1, nvars = NVARS;
	long          nclass, ny = NY;
	
	setDIM(symh,1,1);

	for (i = 0; i < nvars; i++)
	{
		if (term[i] > 0)
		{
			nclass = (NCLASSES[i] > 0) ? NCLASSES[i] : 1;
			ncells *= nclass;
			if (nclass > 1)
			{
				setDIM(symh,++ndims,nclass); /* get dimensions for results */
			}
		}
	} /*for (i = 0; i < nvars; i++)*/
	ndims = (ndims) ? ndims : 1;
	if (ny > 1)
	{
		setDIM(symh, ++ndims, ny);
	}
	setNDIMS(symh,ndims);
	return (ncells);
} /*setSymbolDims()*/

/*
   Set dimensions and get space for coefficients & std errors if needed
   Returns the product of dimensions if no problems; otherwise -1
   result is either a REAL symbol (just 1 term, only coefs or ses) or
   a structure (> 1 term or both coefs and ses requested)

   On entry, space has not been allocated

   980817 changed labelling in multivariate case so that the first coordinates
          get numeric labels rather than no labels
*/

static long setSymbols(Symbolhandle result, long jterm, long term[],
					   long do_coefs,long do_ses, long byterm,
					   Symbolhandle * pCoefsSH, Symbolhandle * pSesSH)
{
	Symbolhandle         symh;
	int                  oneTerm = (jterm < 0);
	long                 ny = NY, ncells;
	WHERE("setSymbols");
	
	/* first set *pCoefsSH and *pSesSH */
	if (do_coefs && do_ses)
	{ /* both coefficients and standard errors */
		if (oneTerm)
		{
			*pCoefsSH = COMPVALUE(result, 0);
			*pSesSH = COMPVALUE(result, 1);
		} /*if (oneTerm)*/
		else
		{
			if (byterm)
			{
				symh = COMPVALUE(result, jterm);
				*pCoefsSH = COMPVALUE(symh, 0);
				*pSesSH = COMPVALUE(symh, 1);
			} /*if (byterm)*/
			else
			{
				symh = COMPVALUE(result,0);
				*pCoefsSH = COMPVALUE(symh, jterm);
				symh = COMPVALUE(result,1);
				*pSesSH = COMPVALUE(symh, jterm);
			} /*if (byterm){}else{}*/
		} /*if (oneTerm){}else{}*/
	} /*if (do_coefs && do_ses)*/
	else
	{ /* either coefficients or standard errors, but not both */
		if (oneTerm)
		{
			*pCoefsSH = result;
		}
		else
		{
			*pCoefsSH = COMPVALUE(result, jterm);
		}
		if (do_ses)
		{
			*pSesSH = *pCoefsSH;
			*pCoefsSH = (Symbolhandle) 0;
		}
		else
		{
			*pSesSH = (Symbolhandle) 0;
		}
	} /*if (do_coefs && do_ses){}else{}*/

	/* now set dimensions and allocate space */
		
	if (do_coefs)
	{
		ncells = setSymbolDims(*pCoefsSH, term);
		TMPHANDLE = mygethandle(ny * ncells * sizeof(double));
		if (TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
		setDATA(*pCoefsSH, (double **) TMPHANDLE);
	}

	if (do_ses)
	{
		ncells = setSymbolDims(*pSesSH, term);
		TMPHANDLE = mygethandle(ny * ncells * sizeof(double));
		if (TMPHANDLE == (char **) 0)
		{
			goto errorExit;
		}
		setDATA(*pSesSH, (double **) TMPHANDLE);
	}

	if (ny > 1 && HASLABELS(MODELVARS[0]))
	{
		Symbolhandle symh1 = (do_coefs) ? *pCoefsSH : *pSesSH;
		long         ndims = NDIMS(symh1);
		long         dims[MAXDIMS];
		long         lengths[MAXDIMS];
		char        *labels[MAXDIMS];
		char         numericLabel[2];
		long         idim;
		char        *depvLabs[2];
		long         depvLabLengths[2];
		
		numericLabel[0] = NUMERICLABEL;
		numericLabel[1] = '\0';
		
		for (idim = 0; idim < ndims; idim++)
		{
			labels[idim] = numericLabel;
			dims[idim] = DIMVAL(symh1, idim+1);
			lengths[idim] = 2*dims[idim];
		}
		getMatLabels(MODELVARS[0], depvLabs, depvLabLengths);
		lengths[ndims-1] = depvLabLengths[1];
		if (do_coefs)
		{
			TMPHANDLE = createLabels(ndims, lengths);
			if (TMPHANDLE == (char **) 0 || !setLabels(*pCoefsSH, TMPHANDLE)) 
			{
				mydisphandle(TMPHANDLE);
				goto errorExit;
			}
		} /*if (do_coefs)*/

		if (do_ses)
		{
			TMPHANDLE = createLabels(ndims, lengths);
			if (TMPHANDLE == (char **) 0 || !setLabels(*pSesSH, TMPHANDLE)) 
			{
				mydisphandle(TMPHANDLE);
				goto errorExit;
			}
		} /*if (do_ses)*/
		/* Get column labels again since they may have moved */
		getMatLabels(MODELVARS[0], depvLabs, depvLabLengths);
		labels[ndims-1] = depvLabs[1];
		if (do_coefs)
		{
			for (idim = 0; idim < ndims; idim++)
			{
				appendLabels(*pCoefsSH, labels[idim], idim,
							 (idim < ndims-1) ? doExpand : dontExpand);
			}
		}
		if (do_ses)
		{
			for (idim = 0; idim < ndims; idim++)
			{
				appendLabels(*pSesSH, labels[idim], idim,
							 (idim < ndims-1) ? doExpand : dontExpand);
			}
		}
	} /*if (ny > 1 && HASLABELS(MODELVARS[0]))*/ 

	return (ncells);

  errorExit:
	return (-1);
} /*setSymbols()*/

/*
	Code to work with (pseudo) balanced case.  Removed from anovacoefs()
	so as to shorten that code
	We just work through the "hand computations"
*/
static Symbolhandle doBalanced(long do_coefs, long do_ses, long byterm,
							   long icurrmin, long icurrmax,
							   long errorterm, Symbolhandle result)
{
	Symbolhandle    coefsSH, sesSH;
	long            ncells = 1;
	long            ndata = NDATA, nvars = NVARS;
	long            term[MAXVARS];
	long            oneTerm = (icurrmin == icurrmax);
	long            i, icurr;
	double          dfe = (*DF)[errorterm], mse;
	double        **tmpyH = (double **) 0, **tmpMeansH = (double **) 0;
	double        **coefsH, **sesH;
	double         *y, *residuals, *tmpy, *ses;
	WHERE("doBalanced");
	
	ncells = 1;
	for (i = 0; i < nvars; i++)
	{
		ncells *= NCLASSES[i];
	}
	if (do_coefs)
	{
		if (!getScratch(tmpyH,GTMPY,ndata,double) ||
			!getScratch(tmpMeansH,GTMPMEANS,ncells,double))
		{
			goto errorExit;
		}
		y = *Y;
		residuals = *RESIDUALS;
		tmpy = *tmpyH;
		for (i = 0; i < ndata; i++)
		{
			tmpy[i] =  (PREVMODELTYPE & IPF) ?
				log(y[i] - residuals[i]) : y[i];
		} /*for (i = 0; i < ndata; i++)*/

		for (icurr = 0; icurr < icurrmin; icurr++)
		{/* sweep out prior stuff */
			getterm(modelterm(MODEL,icurr), term, nvars);
			/* means */
			(void) cellstat(tmpyH, term, tmpMeansH,
							(double **) 0, (double **) 0, CELLMEANSANDSWEEP);
		}
		discardScratch(tmpMeansH,GTMPMEANS,double);
	} /*if (do_coefs)*/
	
	if (do_ses > 0)
	{
		mse = (*SS)[errorterm]/dfe;
	}
	
	for (icurr = icurrmin; icurr <= icurrmax; icurr++)
	{
		getterm(modelterm(MODEL,icurr), term, nvars);
	/* set dimensions and get space for coefficients & std errors if needed */
		ncells = setSymbols(result, (oneTerm) ? -1 : icurr - icurrmin, term,
							do_coefs, do_ses, byterm, &coefsSH, &sesSH);
		if (ncells < 0)
		{
			goto errorExit;
		}
		

		/* coefsSH has coefs, sesSH has se's, if requested */

		if (do_coefs)
		{
			coefsH = DATA(coefsSH);
	/* tmpyH now has residuals after sweeping out terms 0 through icurr-1 */
			(void) cellstat(tmpyH, term, coefsH, (double **) 0,
							(double **) 0, CELLMEANSANDSWEEP);
	/* tmpyH now has residuals after sweeping out terms 0 through icurr */
		} /*if (do_coefs)*/
		
		if (do_ses)
		{
			sesH = DATA(sesSH);
			ses = *sesH;
			if (do_ses > 0)
			{
				doubleFill(ses, sqrt((*DF)[icurr] * mse / (double) NDATA),
						   ncells);
			}
			else
			{
				for (i = 0; i < ncells; i++)
				{
					setMissing(ses[i]);
				}
			}
		} /*if (do_ses)*/
	} /* for (icurr = icurrmin; icurr <= icurrmax; icurr++) */

	discardScratch(tmpyH,GTMPY,double);
	return (result);

  errorExit:
	discardScratch(tmpyH,GTMPY,double);
	return (0);
} /*doBalanced()*/

/*
  function to compute vector of weights for converting from coefficients
  derived in the Gram-Schmidt orthogonalization to main or interaction effects

  cellno   is the cell number in the table of main effects or interactions
           for which we need the weights
  ncells   is the size of the table
  ndims    is the number of dimensions
  dims     vector of ndims dimensions
  offsets  is a vector derived from the dimensions
  wts      the vector of weights of length ncells; values are -1, +1 or 0

  021022 created with code from calcCoefs()
*/

static void calcWts(long  cellno, long ncells, long ndims, long dims [],
					long offsets [], double * wts)
{
	long          newcoords[MAXVARS];
	long          icell;
	int           idim;
	
	getcoords(cellno, ndims, newcoords, offsets);
	for (idim = 0; idim < ndims; idim++)
	{
		newcoords[idim]++;
	}
/*
   Loop over original coefficients to create weights in wts
   for the linear combination that will yield the cellno-th
   coefficient
*/
	for (icell = 0; icell < ncells; icell++)
	{
		long              oldcoords[MAXVARS];
		int               addin = 1;
		
		/* first index changes fastest*/
		getcoords(icell, ndims, oldcoords, offsets);

		/*
		   Note that old coordinates run from 0 to dims[icell] - 1 and
		   new coordinates run from 1 to dims[icell].

		   Include with weight +- 1 any old coefficient each of whose
		   coordinates "matches" the corresponding coordinate of the
		   new coefficient.  If there are any non-matches, the old
		   coefficient gets weight 0.

		   Coordinates "match" if
			(a) they are identical;
			(b) the old coordinate is 0;
		   or
			(c) the new coordinate is at its maximum level.
		   They are included with coefficient (-1)^k where k = number
		   of coordinates for which (c) is true but not (b).

		   Note that (a) is incompatible with either (b) or (c)

		   Coordinates do not "match" if the old coordinate is not 0
		   and the new coordinate < dim[icell] but is not equal to the old
		   coordinate.
		  */

		for (idim = 0; idim < ndims; idim++)
		{
			if (oldcoords[idim] > 0)
			{ /* possible non-"match" */
				/* check for max coords */
				if (newcoords[idim] == dims[idim])
				{ /* "match" */
					addin *= -1;
				}
				else if (newcoords[idim] != oldcoords[idim])
				{ /* non "match" */
					addin = 0;
					break;
				}
			} /*if (oldcoords[idim] > 0)*/
		} /*for (idim = 0; idim < ndims; idim++)*/
		wts[icell] = (double) addin;
	} /*for (icell = 0; icell < ncells; icell++)*/
} /*calcWts()*/

/*
   function to compute the coefficients for term termno of the 
   current model.  If standard errors are required, variances are
   taken from term errorterm.

   050818 added argument silent so that 'silent:T' actually suppresses
          the warning message
*/
static int calcCoefs(long varno, long termno, long errorterm, long colno, 
					 long do_ses, Symbolhandle coefsSH, Symbolhandle sesSH,
					 int silent)
{
	Symbolhandle     symh;
	double         **tmpWtsH = (double **) 0;
	double         **tmpAlphaH = (double **) 0;
	double          *tmpWts, *tmpAlpha;
	double          *coefs;
	double          *ses;
	double          *xtxinv, *ss;
	double           tmpWtsi, tmpWtsj;
	double           cjj, mse, scale = PREVGLMSCALE;
	double           dfe = (*DF)[errorterm];
	long           **colptrH = (long **) 0, *colptr;
	long             term[MAXVARS];
	long             offsets[MAXVARS];
	long             dims[MAXDIMS];
	long             i, j, j2, ij, ik, ikk, jk, k, k1;
	long             colptri, colptrj;
	long             df;
	long             ndims, ndims1, ncells;
	long             ny = (varno == 0) ? NY : 1;
	long             nterms = NTERMS;
	long             nvars = NVARS, nregx = NREGX;
	int              do_coefs = (coefsSH != (Symbolhandle) 0);
	int              useScale =
		(do_ses > 0) && (PREVGLMCONTROL & NONNORMALDIST) != 0;
	char            *termname;
	WHERE("calcCoefs");

	symh = (do_coefs) ? coefsSH : sesSH;

	getterm(modelterm(MODEL, termno), term, nvars);
	offsets[0] = 1;
	ncells = 1;  /* maximum number of coefficients per variable */
	ndims = 0;
	ndims1 = 1;
	for (j = 0; j < nvars; j++)
	{
		if (term[j] > 0)
		{
				offsets[ndims] = ncells;
				dims[ndims] = (NCLASSES[j] > 0) ?
				  DIMVAL(symh, ndims1++) : 1;
				ncells *= dims[ndims++];
		} /*if (term[j] > 0)*/
	} /*for (j = 0; j < nvars; j++)*/
	ndims = (ndims == 0) ? 1 : ndims;
	if (!getScratch(tmpWtsH,GTMPWTS,ncells,double))
	{
		goto errorExit;
	}

	/*
	   get coefficients of X-variables
	*/
	df = getAlpha(termno, colno, (do_coefs) ? &tmpAlphaH : (double ***) 0,
				  &colptrH, TRANSPOSED);
	toTrash(tmpAlphaH,GTMPALPHA);
	toTrash(colptrH,GCOLPTR);
	if (df < 0)
	{
		goto errorExit;
	}

	colno += df;
	/* no more storage to be allocated; safe to dereference handles */

	/* 
	  Now rearrange and fix up marginal terms;
	  warn if pivoted term included
	  Note: (*colptrH)[0] contains ncells
	*/
	colptr = *colptrH + 1;
	/* 
	   colptr[i] has column number of i-th coefficient if it has not
	   been aliased (colptr[i] = ALIASED) or pivoted out (colptr[i] =
	   PIVOTED)
	*/
	termname = skipStrings((TERMNAMES_AT != (char **) 0) ?
						   *TERMNAMES_AT : *TERMNAMES, termno);

	if (!silent && PRINTWARNINGS) 
	{
		for (j = 0; j < ncells; j++)
		{
			if (colptr[j] == PIVOTED)
			{
				sprintf(OUTSTR,
						"WARNING: Missing df(s) in term %s", termname);
				putErrorOUTSTR();
				putOutMsg("Missing coefficient or effects set to zero");
				break;
			} /*if (colptr[j] == PIVOTED)*/
		} /*for (j = 1; j <= ncells; j++)*/
	} /*if (!silent) */
	
	/* start by filling in body with zeros */

	if (do_coefs)
	{
		coefs = DATAPTR(coefsSH);
		doubleFill(coefs, 0.0, ncells * ny);
	}
	else
	{
		coefs = (double *) 0;
	}

	if (do_ses)
	{
		ses = DATAPTR(sesSH);
		doubleFill(ses, 0.0, ncells * ny);
	}
	else
	{
		ses = (double *) 0;
	}

	/*
		Now look at each value, add it in where it is marginal, and
		subtract off where it is internal
	*/
	xtxinv = *XTXINV;
	ss = *SS;
	tmpWts = *tmpWtsH;
	if (do_coefs)
	{
		tmpAlpha = *tmpAlphaH;
	}

/*
   loop over coefficients (or vectors of coefficients if ny > 0) to be
   computed
*/
	for (j = 0; j < ncells; j++)
	{
		calcWts(j, ncells, ndims, dims, offsets, tmpWts);

		if (do_ses > 0)
		{
			cjj = 0.;
			for (i = 0; i < ncells; i++)
			{
				colptri = colptr[i];
				if (colptri >= 0)
				{ /* not aliased or pivoted */
					tmpWtsi = tmpWts[i];
					for (j2 = 0; j2 < ncells; j2++)
					{/* not aliased or pivoted */
						colptrj = colptr[j2];
						if (colptrj >= 0)
						{
							tmpWtsj = tmpWts[j2];
							ij = colptri + colptrj * nregx;			
							cjj += tmpWtsi * tmpWtsj * xtxinv[ij];
						}
					} /*for (j2 = 0; j2 < ncells; j2++)*/
				} /*if (colptri >= 0)*/
			} /*for (i = 0; i < ncells; i++)*/
		} /*if (do_ses > 0)*/

		/* compute linear combination  of elements in tmpAlpha */
		jk = j; /* j + k*ncells */
		for (k = 0; k < ny; k++)
		{
			k1 = (varno != 0) ? varno - 1 : k;
			if (do_coefs)
			{
				ik = k1 * ncells; /* i+k1*ncells */
				for (i = 0; i < ncells; i++)
				{
					coefs[jk] += tmpAlpha[ik] * tmpWts[i];
					ik++;
				} /*for (i = 0; i < ncells; i++)*/
			} /*if (do_coefs)*/

			if (do_ses)
			{
#if (0)
				if (PREVMODELTYPE & ROBUSTREG)
				{
					ses[jk] = sqrt(fabs(cjj * RobBeta)) *
						ROBUSTVARK * CURRENTRSC;
					/* variance formula from page 40 of Huber 1977 */
				}
				else if (PREVMODELTYPE & ITERGLMS)
#endif
				if (useScale)
				{
					ses[jk] = scale*sqrt(fabs(cjj));
				}
				else
				{
					if (do_ses > 0)
					{
						ikk = k1*(nterms+1)*(ny+1) + errorterm;
						mse = ss[ikk]/dfe;
						ses[jk] = sqrt(fabs(cjj * mse));
					}
					else
					{
						setMissing(ses[jk]);
					}
				}
			} /*if (do_ses)*/
			jk += ncells;
		} /*for (k = 0; k < ny; k++)*/
	} /*for (j = 0; j < ncells; j++)*/

	if (do_coefs)
	{
		discardScratch(tmpAlphaH,GTMPALPHA,double);
	}
	discardScratch(tmpWtsH,GTMPWTS,double);
	discardScratch(colptrH,GCOLPTR,long);

	return (colno);
	
  errorExit:
	return (-1);
	
} /*calcCoefs()*/

/*
   960310  Modified so labels containing full term names are added
           to structure output when they will not fit on the labels
           If memory isn't available for labels, they are not added but
           it isn't an error that causes termination

   990318  Fixed bug in adding labels with byterm:F
   050818  Added silent to calcCoefs() call
*/
Symbolhandle    anovacoefs(Symbolhandle list)
{
	long            i, i2, colct = 0, icurr, j, k;
	long            term[MAXVARS];
	long            type, margs;
	long            do_ses, do_coefs = 1, oneTerm;
	long            icurrmin, icurrmax;
	long            nargs, varno = 0;
	long            ny = NY;
	long            nterms = NTERMS, errorterm = -1;
	long            nvars = NVARS;
	long            ncomps;
	long            byterm = 1;
	long            useScale = 0;
	long            needlabels = 0, labelLength;
	int             silent = !PRINTWARNINGS;
	char           *compNames[2], *keyword = (char *) 0, *glmname, *termname;
	char            compNames2[9];
	char           *compNames2P = compNames2, **compNames2H = &compNames2P;
	Symbolhandle    symh, symh2, symhkey, result = (Symbolhandle) 0;
	Symbolhandle    coefsSH = (Symbolhandle) 0, sesSH = (Symbolhandle) 0;
	double          tmp, dfe;
	WHERE("anovacoefs");

	Trash = GarbInstall(NTRASH);
	if (Trash == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	*OUTSTR = '\0';
	do_ses = (strcmp(FUNCNAME, "secoefs") == 0);
	margs = 6;
	
	if (PREVMODELTYPE == 0)
	{
		sprintf(OUTSTR,"ERROR: no active model for %s()", FUNCNAME);
	}
	else if (PREVMODELTYPE & LEAPS)
	{
		sprintf(OUTSTR,"ERROR: %s() cannot be used after screen()",FUNCNAME);
	}
	else if (!(PREVMODELTYPE & DOCOEFS))
	{
		sprintf(OUTSTR,
				"ERROR: %s() cannot be used after GLM command with coefs:F",
				FUNCNAME);
	}
	if (*OUTSTR)
	{
		goto errorExit;
	}
	
	nargs = NARGS(list);

	if (nargs > margs)
	{
		badNargs(FUNCNAME,-margs);
		goto errorExit;
	}
	
	for (i = 0; i < nargs;i++)
	{
		symhkey = COMPVALUE(list,i);
		if (i > 0 && symhkey == (Symbolhandle) 0)
		{
			(void) argOK(symhkey,0, i + 1);
			goto errorExit;
		}
		
		if ((keyword = isKeyword(symhkey)))
		{
			int       logValue = 0;
			
			type = TYPE(symhkey);
			if (isTorF(symhkey))
			{
				logValue = (DATAVALUE(symhkey, 0) != 0.0);
			}
			else
			{
				logValue = -1;
			}
			
			if (strncmp(keyword,"err",3) == 0)
			{				
				if (!isCharOrString(symhkey) &&
					!isInteger(symhkey,POSITIVEVALUE))
				{
					sprintf(OUTSTR,
							"ERROR: %s not CHARACTER variable or string or term number",
							keyword);
					goto errorExit;
				}

				if (type == REAL)
				{
					tmp = DATAVALUE(symhkey,0);
					if (tmp > (double) (nterms+1))
					{
						sprintf(OUTSTR,
								"ERROR: %s term number is not integer between 1 and %ld",
								keyword,nterms+1);
						goto errorExit;
					}
					errorterm = (long) tmp - 1;
				}
				else
				{
					int         imax = 0, inames;
					char       *termnames[2];
					
					if (TERMNAMES_AT != (char **) 0)
					{
						termnames[imax++] = *TERMNAMES_AT;
					}
					termnames[imax++] = *TERMNAMES;
					
					for (inames = 0; inames < imax; inames++)
					{
						i2 = 0;
						for (errorterm = 0;errorterm <= nterms;errorterm++)
						{			/* try to match it */
							if (strcmp(termnames[inames]+i2,STRINGPTR(symhkey)) == 0)
							{
								break;
							}
							i2 += strlen(termnames[inames]+i2) + 1;
						}

						if (errorterm <= nterms)
						{
							break;
						}
					} /*for (inames = 0; inames < imax; inames++)*/

					if (inames == imax)
					{
						sprintf(OUTSTR,
								"ERROR: %s is not a term in the model %s",
								STRINGPTR(symhkey), *STRMODEL);
						goto errorExit;
					}
				}
			} /*if (strncmp(keyword,"err",3) == 0)*/
			else if (strncmp(keyword,"se",2) == 0 ||
					 strncmp(keyword,"coef",4) == 0)
			{
				if (logValue < 0)
				{
					notTorF(keyword);
					goto errorExit;
				}
				if (*keyword == 's')
				{
					do_ses = logValue;
				}
				else
				{
					do_coefs = logValue;
				}
			}
			else if (strcmp(keyword,"byterm") == 0)
			{
				if (logValue < 0)
				{
					notTorF(keyword);
					goto errorExit;
				}
				byterm = logValue;
			}
			else if (strcmp(keyword, "silent") == 0)
			{
				if (logValue < 0)
				{
					notTorF(keyword);
					goto errorExit;
				}
				silent = logValue;
			}
			else
			{
				badKeyword(FUNCNAME,keyword);
				goto errorExit;
			}
			setNAME(symhkey,USEDKEYWORD);
		} /*if ((keyword = isKeyword(symhkey)))*/
	} /*for (i = 0; i < nargs;i++)*/

	if (!silent && strcmp(FUNCNAME, "anovacoefs") == 0)
	{
		putOutMsg("NOTE: anovacoefs() is obsolete; use coefs() instead");
	}

	if (PREVMODELTYPE & (IPF | FASTANOVA))
	{
		glmname = (PREVMODELTYPE & IPF) ? "ipf" : "fastanova";
		if (!do_coefs)
		{
			sprintf(OUTSTR,
					"ERROR: coefs:F is illegal with %s() after %s()",
					FUNCNAME, glmname);
		}
		else if (do_ses)
		{
			sprintf(OUTSTR,
					"ERROR: standard errors not available to %s() after %s()",
					FUNCNAME, glmname);
		}		
		if (*OUTSTR)
		{
			goto errorExit;
		}
	} /*if (PREVMODELTYPE & (IPF | FASTANOVA))*/
	
	if (nargs > 1 && !isKeyword(symh2 = COMPVALUE(list,1)))
	{/* should be variable number */
		if (!argOK(symh2,REAL,2))
		{
			goto errorExit;
		}
		if (!isInteger(symh2, POSITIVEVALUE))
		{
			notPositiveInteger("variable number");
			goto errorExit;
		}
		
		if (DATAVALUE(symh2, 0) > (double) NY)
		{
			sprintf(OUTSTR,
					"ERROR: variable number not positive integer <= %ld", ny);
			goto errorExit;
		}
		varno = (long) DATAVALUE(symh2, 0);
		nargs--;
		ny = 1;
	} /*if (nargs > 1 && !isKeyword(symh2 = COMPVALUE(list,1)))*/
	
	ncomps = do_coefs + do_ses;
	if (ncomps == 0)
	{
		sprintf(OUTSTR,
				"ERROR: neither coefficients or standard errors requested in %s()",
				FUNCNAME);
		goto errorExit;
	} /*if (ncomps == 0)*/
	
	if (errorterm < 0)
	{
		errorterm = nterms;
	}
	else if (!do_ses && !silent)
	{
		sprintf(OUTSTR,
				"WARNING: keyword 'error' ignored when no standard errors requested by %s()",
				FUNCNAME);
		putErrorOUTSTR();
	}
	
	symh = COMPVALUE(list, 0);
	/* icurrmin and icurrmax are the first and last terms for which
		we want coefficients */

	if (symh == (Symbolhandle) 0 || isKeyword(symh))
	{
		icurrmin = 0;
		icurrmax = nterms - 1;	/* do all terms */
	}
	else
	{
		long        iterm;
		
		type = TYPE(symh);
		if (isscratch(NAME(symh)) &&
			(!isScalar(symh) || type != REAL && type != CHAR))
		{
			sprintf(OUTSTR,
					"ERROR: 1st argument to %s() does not specify a term in %s",
					FUNCNAME, *STRMODEL);
			goto errorExit;
		}
		else if (isScalar(symh) && type == REAL)
		{
			tmp = DATAVALUE(symh,0);
			if (tmp != floor(tmp) || tmp <= 0.0 || tmp > (double) nterms)
			{
				sprintf(OUTSTR,
						"ERROR: term number is not an integer between 1 and %ld",
						nterms);
				goto errorExit;
			}
			iterm = (long) tmp - 1;
		}
		else
		{
			int          imax = 0;
			char        *termnames[2];
			
			termname = (isScalar(symh) && type == CHAR) ?
				STRINGPTR(symh) : NAME(symh);
			if (TERMNAMES_AT != (char **) 0)
			{
				termnames[imax++] = *TERMNAMES_AT;
			}
			termnames[imax++] = *TERMNAMES;
			for (i = 0; i < imax; i++)
			{
				i2 = 0;
				for (iterm = 0; iterm < nterms; iterm++)
				{					/* try to match term */
					if (strcmp(termname, termnames[i] + i2) == 0)
					{
						break;
					}
					i2 += strlen(termnames[i] + i2) + 1;
				}
				if (iterm < nterms)
				{
					break;
				}
			} /*for (i = 0; i < imax; i++)*/
			
			if (i == imax)
			{
				sprintf(OUTSTR, "ERROR: %s is not in model %s",
						termname, *STRMODEL);
				goto errorExit;
			}
		}
		icurrmin = iterm;
		icurrmax = iterm;
	}
	oneTerm = (icurrmin == icurrmax);
	
	compNames[0] = compNames2;
	strcpy(compNames[0],(do_coefs) ? "coefs" : "se");
	if (ncomps == 2)
	{
		compNames[1] = compNames2+6;
		strcpy(compNames[1], "se");
	}
	
	dfe = (*DF)[errorterm];

	if (do_ses)
	{
		/* if this is called by the function secoefs, we want to compute se's
		for the coefs.  do_ses is a flag that says to do se's */

		useScale = (PREVGLMCONTROL & NONNORMALDIST) != 0;

		if (dfe <= 0.0 && (!useScale || PREVGLMSCALE <= 0.0))
		{
			if (!silent)
			{
				putOutErrorMsg("WARNING: no DF for error; standard errors set MISSING");
			}
			do_ses = -1;
			useScale = 0;
		}
			
		if (!silent && useScale)
		{
			strcpy(OUTSTR,
				   "NOTE: standard errors assume scale parameter is ");
			formatDouble(OUTSTR + strlen(OUTSTR), PREVGLMSCALE, TRIMLEFT);
			putOUTSTR();
		} /*if (useScale)*/
	} /*if (do_ses)*/
	
	if (!oneTerm)
	{
		char        *pos = *TERMNAMES;
		
		for (i = 0; i < nterms; i++)
		{
			if (strlen(pos) > NAMELENGTH)
			{
				break;
			}
			pos = skipStrings(pos, 1);
		} /*for (i = 0; i < nterms; i++)*/
		if (i < nterms)
		{
			needlabels = 1;
			labelLength = skipStrings(pos, nterms - i) - *TERMNAMES;
		}
	} /*if (!oneTerm)*/

	/* Symbols for results; actual space not yet allocated */
	if (ncomps == 2)
	{
		if (oneTerm)
		{/* single term with standard errors */
			result = RSInstall(SCRATCH, 2L, compNames, 0L);
		}
		else
		{/* all terms with standard errors */
			if (byterm)
			{
				result = RSInstall3(SCRATCH,nterms, TERMNAMES, 
									2L, compNames, 0L);
				if (needlabels && result != (Symbolhandle) 0)
				{
					TMPHANDLE = mygethandle(labelLength);

					if (TMPHANDLE != (char **) 0 &&
						setLabels(result, TMPHANDLE))
					{
						copyStrings(*TERMNAMES, *TMPHANDLE, nterms);
					}
					else
					{
						mydisphandle(TMPHANDLE);
					}
				}
			} /*if (byterm)*/
			else
			{
				result = RSInstall3(SCRATCH, 2, compNames2H,
									nterms, (char **) 0, 0);
				if (result != (Symbolhandle) 0)
				{
					for (i = 0; i < 2;i++)
					{
						k=0;
						symh = COMPVALUE(result,i);
						for (j = 0; j < nterms;j++)
						{
							setNAME(COMPVALUE(symh,j), *TERMNAMES + k);
							k += strlen(*TERMNAMES + k) + 1;
						} /*for (j = 0; j < nterms;j++)*/
						if (needlabels)
						{
							TMPHANDLE = mygethandle(labelLength);
							if (TMPHANDLE != (char **) 0 &&
								setLabels(symh, TMPHANDLE))
							{
								copyStrings(*TERMNAMES, *TMPHANDLE, nterms);
							}
							else
							{
								mydisphandle(TMPHANDLE);
							}
						}
					} /*for (i = 0; i < 2;i++)*/
				} /*if (result != (Symbolhandle) 0)*/
			} /*if (byterm)else{}*/
		} /* if (oneTerm){}else{}*/
	} /*if (ncomps == 2)*/
	else
	{
		if (oneTerm)
		{/* single term with only coefs or only standard errors */
			result = Install(SCRATCH, REAL);
		}
		else
		{/* all terms with only coefs or only standard errors */
			result = RSInstall2(SCRATCH, nterms, TERMNAMES, 0L);
			if (needlabels && result != (Symbolhandle) 0)
			{
				TMPHANDLE = mygethandle(labelLength);
				
				if (TMPHANDLE != (char **) 0 && setLabels(result, TMPHANDLE))
				{
					copyStrings(*TERMNAMES, *TMPHANDLE, nterms);
				}
				else
				{
					mydisphandle(TMPHANDLE);
				}
			}
		}
	} /*if (ncomps == 2){}else{}*/

	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	if (!(PREVGLMCONTROL & UNBALANCED))
	{/* (pseudo) balanced data */
		
		result = doBalanced(do_coefs, do_ses, byterm, icurrmin, icurrmax,
							errorterm, result);
		if (result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		emptyTrash();
	
		return (result);
	} /*if (!(PREVGLMCONTROL & UNBALANCED))*/

	/* now the unbalanced case */

	/* find number of columns in on REGX matrix */
	colct = 0;
	for (i = 0; i < icurrmin; i++)
	{		/* for each term */
		colct += buildXmatrix(i, (double *) 0, MODELINFO);
	}

	for (icurr = icurrmin; icurr <= icurrmax; icurr++)
	{
		getterm(modelterm(MODEL, icurr), term, nvars);

		if (setSymbols(result, (oneTerm) ? -1 : icurr - icurrmin, term, 
					   do_coefs, do_ses, byterm, &coefsSH, &sesSH) < 0)
		{
			goto errorExit;
		}
		/* coefsSH and/or sesSH are already part of result */
		colct = calcCoefs(varno, icurr, errorterm, colct, do_ses,
						  coefsSH, sesSH, silent);

		if (colct < 0)
		{
			goto errorExit;
		}
	} /*for (icurr = icurrmin; icurr <= icurrmax; icurr++)*/

	emptyTrash();
	
	return (result);
	
  errorExit:
	putErrorOUTSTR();
	
	emptyTrash();
	
	Removesymbol(result);

	return (0);
} /*anovacoefs()*/

