/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.12 or later
*(C)*
*(C)* Copyright (c) 2002 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Blas
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "blas.h"

/*
  Level 1 BLAS
   The (level 1) blas routines in this file were translated to C from files
   dasum.f, daxpy.f, dcopy.f, ddot.f, dnrm2.f, drot.f, drotg.f, drotm.f,
   drotmg.f, dscal.f, dsdot.f, dswap.f, and idamax.f in cmlib and are
   in the public domain.

   In blas.h, the following are defined as entry names to the Level 1
   blas routines

     DAXPY, DDOT, DASUM, DCOPY, DNRM2, DROT, DROTG, DROTM, DROTMG,
	 DSCAL, DSWAP, and IDAMAX

   The default definitions are of the form daxpy_, ddot_, ... but may
   be overridden by Make defines (e.g., -DDAXPY=daxpy__ )
   A routine will be compiled only if the corresponding entry name
   macro is defined.  Several are disabled here.

   All follow the Fortran calling conventions, with all arguments
   pointers so any library of blas routines following this convention
   can be substituted.  If it is optimized for the machine architecture
   it might provide substantial improvement.  On Suse Linux, using
   libblas.a provided no improvement on matrix multiplication (heavy user
   of ddot()); compiling blas.c with gcc -O3 gave about a 20% improvement.

   All integer arguments and most integer variables have type INT.  The
   definition in blas.h and here translates INT to long, but it can be
   overridden by make (e.g., -DINT=int)

   000407 fixed bug in translation of drnm2()
   010830 Expanded the above comments; removed extra
          #include "blas.h" lines
   020325 Added dgemm() code, with optional modification to use ddot()
          for inner products
   020329 All MacAnova specific defines moved here from blas.h
   020809 Changed values of DGEMM_MB and DGEMM_KB for BCPP
*/

/*
   The routines referred to in the following set of #undefs are not
   compiled for MacAnova.  Remove #define MACANOVABLAS to compile them
   all
*/

#define MACANOVABLAS
#ifdef MACANOVABLAS
#undef DROT     /*DROT() not needed by MacAnova*/
#undef DROTG    /*DROTG() not needed by MacAnova*/
#undef DROTM    /*DROTM() not needed by MacAnova*/
#undef DROTMG   /*DROTMG() not needed by MacAnova*/

/* Size of block on Macintosh; compilers put limits on size of local variables*/
#if defined(MPW) || defined (powerc)
#define DGEMM_MB    48
#define DGEMM_KB    32
#endif /*MPW || POWERC*/

#if defined(BCPP)
#define DGEMM_MB    32
#define DGEMM_KB    32
#endif /*BCPP*/

#define DONTTESTDGEMMINPUT /*don't check dgemm() arguments*/

#endif /*MACANOVABLAS*/

#define USEDDOT  /*dgemm() and perhaps others will use ddot()*/

#ifndef INT
#define INT      long
#endif /*INT*/

#ifdef DAXPY
/*====> daxpy.c <====*/
/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  Overwrite double precision y with double precision a*x + y.
  for i = 0 to n-1, replace  y(ly+i*incy) with a*x(lx+i*incx) +
    y(ly+i*incy), where lx = 1 if incx .ge. 0, else lx = (-incx)*n,
  and ly is defined in a similar way using incy.
*/

void DAXPY(INT * pn, double * pa, double * x1, INT * pincx, double * y1,
		   INT * pincy)
/*double x[1],y[1],a;*/
{
	INT                  n = *pn, incx = *pincx, incy = *pincy;
	register double     *x = x1, *y = y1, a = *pa;
	register INT         i, ix, iy, m, ns;

	if(n > 0 && a != 0.0)
	{
		if (incx == incy && incx  > 0)
		{
			if (incx == 1)
			{
				/*  code for both increments equal to 1 */

				/*  clean-up loop so remaining vector length is a multiple of 4. */

				m = n % 4;
				for (i = 0; i < m ; i++)
				{
					y[i] += a*x[i];
				}
				for (i = m; i < n; i += 4 )
				{
					y[i]   += a*x[i];
					y[i+1] += a*x[i+1];
					y[i+2] += a*x[i+2];
					y[i+3] += a*x[i+3];
				}
			} /*if (incx == 1)*/
			else
			{
				ns = n*incx;
				for (i = 0; i < ns; i += incx)
				{
					y[i] += a*x[i];
				}
			}
		} /*if (incx == incy && incx  > 0)*/
		else
		{
			/* code for nonequal or nonpositive increments. */

			ix = (incx >= 0) ? 0 : (-n+1)*incx;
			iy = (incy >= 0) ? 0 : (-n+1)*incy;

			for (i = 0; i < n ; i++)
			{
				y[iy] += a*x[ix];
				ix += incx;
				iy += incy;
			}
		}
	} /*if(n > 0 && a != 0.0)*/
} /* DAXPY() */
#endif /*DAXPY*/

#ifdef DDOT
/*====> ddot.c <====*/

/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu */

/*
  Return the dot product of double precision dx and dy.
  ddot = sum for i = 0 to n-1 of  dx(lx+i*incx) * dy(ly+i*incy)
  where lx = 0 if incx .ge. 0, else lx = (-incx)*(n-1), and ly is
  defined in a similar way using incy.
*/

double DDOT(INT * pn, double * x1, INT * pincx, double * y1, INT * pincy)
{
	INT              n = *pn, incx = *pincx, incy = *pincy;
	register double *x = x1, *y = y1, s;
	register INT     i, ix, iy, m, ns;

	s = 0.0;
	if (n > 0)
	{
		if (incx == incy && incx > 0)
		{
			if (incx == 1)
			{
				/* code for both increments equal to 1. */

				/* clean-up loop so remaining vector length is a multiple of 5 */

				m = n % 5;
				for (i = 0; i < m ; i++)
				{
					s += x[i]*y[i];
				}

				for (i = m; i < n; i += 5 )
				{
					s += x[i]*y[i] + x[i+1]*y[i+1] +
						x[i+2]*y[i+2] + x[i+3]*y[i+3] + x[i+4]*y[i+4];
				}
			} /*if (incx == 1)*/
			else
			{/* incx == incy > 1 */
				ns = n*incx;
				for (i = 0; i < ns; i += incx)
				{
					s += x[i]*y[i];
				}
			}
		} /*if (incx == incy && incx > 0)*/
		else
		{
			/* code for unequal or nonpositive increments. */

			ix = (incx >= 0) ? 0 : (-n+1)*incx;
			iy = (incy >= 0) ? 0 : (-n+1)*incy;

			for (i = 0; i < n ; i++)
			{
				s += x[ix]*y[iy];
				ix += incx;
				iy += incy;
			}
		} /*if (incx == incy && incx > 0){}else{}*/
	} /*if (n > 0)*/

	return (s);
} /* DDOT() */
#endif /*DDOT*/

#ifdef DASUM
/*====> dasum.c <====*/
/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  returns sum of magnitudes of double precision dx.
  dasum = sum from 0 to n-1 of dabs(dx(1+i*incx))
*/

double DASUM(INT * pn, double dx[], INT * pincx)
{
	INT        i, m, ns;
	INT        n = *pn, incx = *pincx;
	double     sum = 0.0;

	if (n > 0)
	{
		if (incx != 1)
		{/*  code for increments not equal to 1. */
			ns = n*incx;
			for (i = 0; i < ns; i += incx)
			{
				sum += fabs(dx[i]);
			}
		} /*if (incx!=1)*/
		else
		{/* code for increments equal to 1. */
	/* clean-up loop so remaining vector length is a multiple of 6. */

			m = n % 6;
			for (i = 0; i < m ; i++)
			{
				sum += fabs(dx[i]);
			}
			for (i = m; i < n; i += 6 )
			{
				sum += fabs(dx[i]) + fabs(dx[i+1]) + fabs(dx[i+2]) +
					fabs(dx[i+3]) + fabs(dx[i+4]) + fabs(dx[i+5]);
			}
		} /*if (incx!=1){}else{}*/
	} /*if (n>0)*/
	return (sum);
} /*DASUM() */
#endif /*DASUM*/

#ifdef DCOPY
/*====> dcopy.c <====*/
/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  Copy double precision dx to double precision dy.
  for i = 0 to n-1, copy dx(lx+i*incx) to dy(ly+i*incy),
  where lx = 1 if incx .ge. 0, else lx = (-incx)*n, and ly is
  defined in a similar way using incy.
*/

void DCOPY(INT * pn, double dx[], INT * pincx, double dy[], INT * pincy)
{
	INT        n = *pn, incx = *pincx, incy = *pincy;
	INT        i, m, ns, ix, iy;

	if (n > 0)
	{
		if (incx == incy && incx > 0)
		{
			if (incx != 1)
			{
				ns = n*incx;
				for (i = 0; i < ns; i += incx)
				{
					dy[i] = dx[i];
				}
			} /*if (incx != 1)*/
			else
			{ /* incx == incy == 1 */
			/* clean-up loop so remaining vector length is a multiple of 7. */

				m = n % 7;
				for (i = 0; i < m ; i++)
				{
					dy[i] = dx[i];
				}
				for (i = m; i < n; i += 7 )
				{
					dy[i]   = dx[i];
					dy[i+1] = dx[i+1];
					dy[i+2] = dx[i+2];
					dy[i+3] = dx[i+3];
					dy[i+4] = dx[i+4];
					dy[i+5] = dx[i+5];
					dy[i+6] = dx[i+6];
				}
			}
		} /*if (incx == incy && incx > 0)*/
		else
		{
			/* code for unequal or nonpositive increments. */

			ix = (incx >= 0) ? 0 : (-n+1)*incx;
			iy = (incy >= 0) ? 0 : (-n+1)*incy;
			for (i = 0; i < n ; i++)
			{
				dy[iy] = dx[ix];
				ix += incx;
				iy += incy;
			}
		} /*if (incx == incy && incx > 0){}else{}*/
	} /*if (n>0)*/
} /*DCOPY()*/
#endif /*DCOPY*/


#ifdef DNRM2
/*====> dnrm2.c <====*/
/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  Euclidean norm of the n-vector stored in dx() with storage increment incx.
  If  n <= 0 return with result = 0.
  If n >= 1 then incx must be >= 1

  C.L.Lawson, 1978 Jan 08

  Four phase method     using two built-in constants that are
  hopefully applicable to all machines.
	  cutlo = maximum of  dsqrt(u/eps)  over all known machines.
	  cuthi = minimum of  dsqrt(v)      over all known machines.
  where
	  eps = smallest no. such that eps + 1. .gt. 1.
	  u   = smallest positive no.   (underflow limit)
	  v   = largest  no.            (overflow  limit)

  Brief outline of algorithm..

  Phase 1    scans zero components.
	move to phase 2 when a component is nonzero and .le. cutlo
	move to phase 3 when a component is .gt. cutlo
	move to phase 4 when a component is .ge. cuthi/m
  where m = n for x() real and m = 2*n for complex.

  Values for cutlo and cuthi..
  From the environmental parameters listed in the IMSL converter
  document the limiting values are as follows..
  cutlo, s.p.   u/eps = 2**(-102) for  Honeywell.  Close seconds are
  Univac and Dec at 2**(-103)
  Thus cutlo = 2**(-51) = 4.44089e-16
  cuthi, s.p.   v = 2**127 for Univac, Honeywell, and Dec.
  Thus cuthi = 2**(63.5) = 1.30438e19
  cutlo, d.p.   u/eps = 2**(-67) for Honeywell and Dec.
  Thus cutlo = 2**(-33.5) = 8.23181e-11
  cuthi, d.p.   same as s.p.  cuthi = 1.30438e19
  data cutlo, cuthi / 8.232e-11,  1.304e19 /
  data cutlo, cuthi / 4.441e-16,  1.304e19 /
 */

/*
  000407 Fixed bug that had been here forever.  Fortunately DNRM2()
         is used only in qr() at present
*/
#define CUTLO     8.232e-11
#define CUTHI     1.304e19

double DNRM2(INT * pn, double dx[], INT * pincx)
{
	INT          n = *pn, incx = *pincx;
	INT          nn, i;
	double       dxi, d;
	double       sum = 0.0, xmax;
	double       cutlo = CUTLO, cuthi = CUTHI;
	double       hitest;

	nn = n * incx;
	for (i = 0; i  <  nn && dx[i] == 0.0; i += incx)
	{ /* phase 1.  sum is zero */
		;
	}
	
	if(i >= nn)
	{
		return (sum);
	}

	if (fabs(dx[i]) <= cutlo)
	{
		xmax = fabs(dx[i]);
		sum = 1.0;
		i += incx;
		for (/*i as is*/; i < nn; i += incx)
		{
			/*
			  phase 2.  sum is small.
			  scale to avoid destructive underflow.
			*/
			dxi = fabs(dx[i]);
			if(dxi > cutlo)
			{
				break;
			}
			d = dxi/xmax;
			if(d <= 1.0)
			{
				sum += d*d;
			}
			else
			{
				sum = 1.0 + sum/(d*d);
				xmax = dxi;
			}
		} /*for (; i < nn; i += incx)*/
		if(i >= nn)
		{
			return (xmax * sqrt(sum));
		}
		/* undo scaling, enforcing (sum*xmax) * xmax computation */
		sum *= xmax;
		sum *= xmax;
	} /*if (fabs(dx[i]) <= cutlo)*/
	
	/*
		prepare for phase 3.
		for real or d.p. set hitest = cuthi/n
		for complex      set hitest = cuthi/(2*n)
	*/
	hitest = cuthi / (double) n;

	for (/* i as is */; i < nn; i++)
	{ /* Phase 3.  Sum is mid-range.  No scaling. */
		dxi = fabs(dx[i]);
		if(dxi < hitest)
		{
			sum += dxi*dxi;
		}
		else
		{/* prepare for phase 4. */
			xmax = dxi;
			sum /= xmax; /* rescale, enforcing (sum/xmax)/xmax computation */
			sum /= xmax;
			break;
		}
	} /*for (; i < nn; i++)*/
	if(i >= nn)
	{
		return (sqrt(sum));
	}

	for (/* i as is*/; i < nn; i += incx)
	{/* in phase 4 sum is large.  scale to avoid overflow. */
		dxi = fabs(dx[i]);
		d = dxi/xmax;
		if(d <= 1.0)
		{
			sum += d*d;
		}
		else
		{
			sum = 1.0 + sum/(d*d);
			xmax = dxi;
		}
	} /*for (; i < nn; i += incx)*/
	/* compute square root and adjust for scaling. */
	return (xmax * sqrt(sum));
} /*DNRM2()*/
#endif /*DNRM2*/

#ifdef DROT /* not used in MacAnova */
/*====> drot.c <====*/

/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  Multiply the 2 x 2 matrix  ( dc ds) times the 2 x n matrix (dx**t)
                             (-ds dc)                        (dy**t)
  where **t indicates transpose.    The elements of dx are in
  dx(lx+i*incx), i = 0 to n-1, where lx = 1 if incx .ge. 0, else
  lx = (-incx)*n, and similarly for dy using ly and incy.
*/


void DROT(INT * pn, double dx [], INT * pincx, double dy [], INT * pincy,
		  double * pdc, double * pds)
{
	double      dc = *pdc, ds = *pds;
	double      w,z;
	INT         n = *pn, incx = *pincx, incy = *pincy;
	INT         i, kx, ky, nsteps;

	if (n > 0 && (ds != 0.0 || dc != 1.0))
	{
		if (incx == incy && incx > 0)
		{
			nsteps = incx*n;
			for (i = 0; i < nsteps; i += incx)
			{
				w = dx[i];
				z = dy[i];
				dx[i] = dc*w+ds*z;
				dy[i] = -ds*w+dc*z;
			}
		}
		else
		{
			kx = (incx >= 0) ? 0 : (-n+1)*incx;
			ky = (incy >= 0) ? 0 : (-n+1)*incy;

			for (i = 0; i < n ; i++)
			{
				w = dx[kx];
				z = dy[ky];
				dx[kx] = dc*w+ds*z;
				dy[ky] = -ds*w+dc*z;
				kx += incx;
				ky += incy;
			}
		}
	}
	return;
} /*DROT()*/
#endif /*DROT*/

#ifdef DROTG  /* not used in MacAnova */
/*====> drotg.c <====*/

/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  Designed by C.L.Lawson, JPL, 1977 Sept 08

  Construct the Givens transformation

      ( dc  ds )
  g = (        ) ,    dc**2 + ds**2 = 1 ,
      (-ds  dc )

  which zeros the second entry of the 2-vector  (da,db)**t .

  The quantity r = (+/-)dsqrt(da**2 + db**2) overwrites da in
  storage.  The value of db is overwritten by a value z which
  allows dc and ds to be recovered by the following algorithm:
        if z=1  set  dc=0.e0  and  ds=1.e0
        if dabs(z) .lt. 1  set  dc=dsqrt(1-z**2)  and  ds=z
        if dabs(z) .gt. 1  set  dc=1/z  and  ds=dsqrt(1-dc**2)

  Normally, the subprogram drot(n,dx,incx,dy,incy,dc,ds) will
  next be called to apply the transformation to a 2 by n matrix.

  ------------------------------------------------------------------
*/

void DROTG(double * da, double * db, double * dc, double * ds)
{
	double    u,v,r;
	
	if (fabs(*da) > fabs(*db))
	{
	/*  *** here dabs(da) .gt. dabs(db) *** */
		u = (*da)+(*da);
		v = (*db)/u;
	/*      note that u and r have the sign of da */
		r = sqrt(.25+v*v)*u;
	
	/*      note that dc is positive */
	
		*dc = (*da)/r;
		*ds = v*((*dc)+(*dc));
		*db = *ds;
		*da = r;
	}
	else if (*db == 0.0) 	/*  *** here dabs(da) .le. dabs(db) *** */
	{
	/*  *** here da = db = 0.0 *** */
		*dc = 1.0;
		*ds = 0.0;
	}
	else
	{
		u = (*db)+(*db);
		v = (*da)/u;
	
	/*      note that u and r have the sign of db */
	/*      (r is immediately stored in da) */
	
		*da = sqrt(.25 + v*v)*u;
	
	/*      note that ds is positive */
	
		*ds = (*db)/(*da);
		*dc = v*((*ds)+(*ds));
		if (*dc == 0.0)
		{
			*db = 1.0;
		}
		else
		{
			*db = 1.0/(*dc);
		}
	}
} /*DROTG()*/
#endif /*DROTG*/

#ifdef DROTM
/* ====> drotm.c <==== */  /* not used in MacAnova */

/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  Apply the modified Givens transformation, h, to the 2 by n matrix

  (dx**t) , where **t indicates transpose. The elements of dx are in
  (dy**t)

  dx(lx+i*incx), i = 0 to n-1, where lx = 1 if incx .ge. 0, else
  lx = (-incx)*n, and similarly for sy using ly and incy.

  With dparam[0]=dflag, h has one of the following forms..

    dflag=-1.0      dflag=0.0         dflag=1.0      dflag=-2.0

    (dh11  dh12)    (1.0   dh12)    (dh11  1.0 )    (1.0   0.0 )
  h=(          )    (          )    (          )    (          )
    (dh21  dh22),   (dh21  1.0 ),   (-1.0  dh22),   (0.0   1.0 ).

  See drotmg for a description of data storage in dparam.
*/

void DROTM(INT * pn, double dx [],INT * pincx, double dy [], INT * pincy,
		   double dparam [5])
{
	double      dflag,dh12,dh22,z,dh11,dh21,w;
	INT         n = *pn, incx = *pincx, incy = *pincy;
	INT         kx, ky, i, nsteps;

	dflag = dparam[0];
	if (n>0 && dflag != -2.0)
	{
		if (dflag < 0)
		{
			dh11 = dparam[1];
			dh12 = dparam[3];
			dh21 = dparam[2];
			dh22 = dparam[4];
		}
		else if (dflag == 0)
		{
			dh12 = dparam[3];
			dh21 = dparam[2];
		}
		else
		{
			dh11 = dparam[1];
			dh22 = dparam[4];
		}
		if (incx != incy || incx <= 0)
		{
			kx = (incx >= 0) ? 0 : (-n+1)*incx;
			ky = (incy >= 0) ? 0 : (-n+1)*incy;

			if (dflag < 0)
			{
				for (i = 0; i < n ; i++)
				{
					w = dx[kx];
					z = dy[ky];
					dx[kx] = w*dh11+z*dh12;
					dy[ky] = w*dh21+z*dh22;
					kx += incx;
					ky += incy;
				} /*for (i = 0; i < n ; i++)*/
			} /*if (dflag < 0)*/
			else if (dflag == 0)
			{
				for (i = 0; i < n ; i++)
				{
					w = dx[kx];
					z = dy[ky];
					dx[kx] = w + z*dh12;
					dy[ky] = w*dh21 + z;
					kx += incx;
					ky += incy;
				} /*for (i = 0; i < n ; i++)*/
			}
			else
			{
				for (i = 0; i < n ; i++)
				{
					w = dx[kx];
					z = dy[ky];
					dx[kx] = w*dh11 + z;
					dy[ky] = -w + dh22*z;
					kx += incx;
					ky += incy;
				} /*for (i = 0; i < n ; i++)*/
			}
		} /*if (incx != incy || incx <= 0)*/
		else
		{
			nsteps = n*incx;
			if (dflag<0)
			{
				for (i = 0; i < nsteps; i += incx)
				{
					w = dx[i];
					z = dy[i];
					dx[i] = w*dh11 + z*dh12;
					dy[i] = w*dh21 + z*dh22;
				}
			} /*if (dflag<0)*/
			else if (dflag == 0)
			{
				for (i = 0; i < nsteps; i += incx)
				{
					w = dx[i];
					z = dy[i];
					dx[i] = w + z*dh12;
					dy[i] = w*dh21 + z;
				}
			}
			else
			{
				for (i = 0; i < nsteps; i += incx)
				{
					w = dx[i];
					z = dy[i];
					dx[i] = w*dh11 + z;
					dy[i] = -w + dh22*z;
				}
			}
		} /*if (incx != incy || incx <= 0){}else{}*/
	} /*if (n>0 && dflag != -2.0)*/
} /*DROTM()*/
#endif /*DROTM*/

#ifdef DROTMG
/*====> drotmg.c <====*/ /* not used in MacAnova */

/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  Construct the modified givens transformation matrix h which zeros
  the second component of the 2-vector  (dsqrt(dd1)*dx1,dsqrt(dd2)*
  dy2)**t.
  With dparam(1)=dflag, h has one of the following forms..

  dflag=-1.0      dflag=0.0         dflag=1.0      dflag=-2.0

    (dh11  dh12)    (1.0   dh12)    (dh11  1.0 )    (1.0   0.0 )
  h=(          )    (          )    (          )    (          )
    (dh21  dh22),   (dh21  1.0 ),   (-1.0  dh22),   (0.0   1.0 ).
  Locations 2-4 of dparam contain dh11, dh21, dh12, and dh22
  respectively. (values of 1.0 , -1.0 , or 0.0  implied by the
  value of dparam(1) are not stored in dparam.)

  The values of gamsq and rgamsq set in the data statement may be
  inexact.  This is ok as they are only used for testing the size
  of dd1 and dd2.  All actual scaling of data is done using gam.
*/

void DROTMG(double dd1, double dd2, double dx1, double dy1, double dparam[5])
{
	double        dh11,dh21,dh12,dh22,du;
	double        dp1, dp2, dq1, dq2;
	double        dflag,dtemp;
	double        gam = 4096.0, gamsq = 16777216.0, rgamsq = 5.9604645e-8;
	INT           igo, jgo = 1;
	INT           done;
	
	done = 1;
	while(dd1 >= 0.0)
	{ /* not a real loop; an if block that can be exited by break */
		/* case-dd1-nonnegative */
		dp2 = dd2*dy1;
		if (dp2 == 0.0)
		{
			dparam[0] = -2.0;
			return;
		}
		/* regular case */
		dp1 = dd1*dx1;
		dq2 = dp2*dy1;
		dq1 = dp1*dx1;

		if (fabs(dq1) > fabs(dq2))
		{
			dh21 = -dy1/dx1;
			dh12 = dp2/dp1;

			du = 1.0 - dh12*dh21;

			if (du>0.0)
			{
				dflag = 0.0;
				dd1 /= du;
				dd2 /= du;
				dx1 *= du;
				done = 0;
			}
		}
		else if(dq2 >= 0.0)
		{
			dflag = 1.0;
			dh11 = dp1/dp2;
			dh22 = dx1/dy1;
			du = 1.0 + dh11*dh22;
			dtemp = dd2/du;
			dd2 = dd1/du;
			dd1 = dtemp;
			dx1 = dy1*du;
			done = 0;
		}
		break;
	} /* while (not reached)*/

	if(done)
	{ /* procedure..zero-h-d-and-dx1.. */
		dflag = -1.0;
		dh11 = dh12 = dh21 = dh22 = dd1 = dd2 = dx1 = 0.0;
	}
	else
	{
		igo = 0;

		while(!done)
		{
			if(igo == 0)
			{
				jgo = 1;
			}
			else
			{
				jgo = igo;
				switch (igo)
				{
				  case 1:
					dd1 = dd1*(gam*gam);
					dx1 = dx1/gam;
					dh11 = dh11/gam;
					dh12 = dh12/gam;
					break;
		
				  case 2:
					dd1 = dd1/(gam*gam);
					dx1 = dx1*gam;
					dh11 = dh11*gam;
					dh12 = dh12*gam;
					break;
		
				  case 3:
					dd2 = dd2*gam*gam;
					dh21 = dh21/gam;
					dh22 = dh22/gam;
					break;
		
				  case 4:
					dd2 = dd2/(gam*gam);
					dh21 = dh21*gam;
					dh22 = dh22*gam;
					break;
				} /* switch */
			} /* if ... else */
			
			switch (jgo)
			{ /* procedure..scale-check */
			  case 1:
				if(dd1 != 0.0 && dd1 <= rgamsq)
				{
					igo = 1;
					break;
				}
				/* fall through */
			  case 2:
				if(dd1 >= gamsq)
				{
					igo = 2;
					break;
				}
				/* fall through */
			  case 3:
				if(dd2 != 0.0 && fabs(dd2) <= rgamsq)
				{
					igo = 3;
					break;
				}
				/* fall through */
			  case 4:
				if(fabs(dd2) >= gamsq)
				{
					igo = 4;
					break;
				}
				done = 1;
			} /* switch */
	
			if(!done && dflag>=0.0)
			{ /* procedure..fix-h.. */
				if (dflag!=0.0)
				{
					dh21 = -1.0;
					dh12 = 1.0;
				}
				else
				{
					dh11 = 1.0;
					dh22 = 1.0;
				}
				dflag = -1.0;
			} /* if */
		} /* while */
	}
	/* set dparam[] */
	if (dflag<0)
	{
		dparam[1] = dh11;
		dparam[2] = dh21;
		dparam[3] = dh12;
		dparam[4] = dh22;
	}
	else if (dflag == 0)
	{
		dparam[2] = dh21;
		dparam[3] = dh12;
	}
	else
	{
		dparam[1] = dh11;
		dparam[4] = dh22;
	}

	dparam[0] = dflag;
	return;
} /*DROTMG()*/
#endif /*DROTMG*/

#ifdef DSCAL
/*====> dscal.c <====*/
/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  Replace double precision dx by double precision da*dx.
  For i = 0 to n-1, replace dx(1+i*incx) with  da * dx(1+i*incx)
*/
void DSCAL(INT * pn, double * pda, double dx[], INT * pincx)
{
	INT         n = *pn, incx = *pincx;
	INT         i, m, ns;
	double      da = *pda;

	if (n > 0)
	{
		if (incx != 1)
		{
			/* code for increments not equal to 1. */
			ns = n*incx;
			for (i = 0; i < ns; i += incx)
			{
				dx[i] *= da;
			}
		} /*if (incx!=1)*/
		else
		{
		/* code for increments equal to 1. */

		/* clean-up loop so remaining vector length is a multiple of 5. */
			m = n % 5;
			for (i = 0; i < m ; i++)
			{
				dx[i] *= da;
			}
			for (i = m; i < n; i += 5 )
			{
				dx[i] *= da;
				dx[i+1] *= da;
				dx[i+2] *= da;
				dx[i+3] *= da;
				dx[i+4] *= da;
			}
		} /*if (incx != 1){}else{}*/
	} /*if (n > 0)*/
} /*dscal()*/
#endif /*DSCAL*/

#ifdef DSWAP
/*====> dswap.c <====*/
/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  Interchange double precision dx and double precision dy.
  For i = 0 to n-1, interchange  dx(lx+i*incx) and dy(ly+i*incy),
  where lx = 1 if incx .ge. 0, else lx = (-incx)*n, and ly is
  defined in a similar way using incy.
*/

void DSWAP(INT * pn, double dx[], INT * pincx, double dy[], INT * pincy)
{
	double     dtemp1,dtemp2,dtemp3;
	INT        n = *pn, incx = *pincx, incy = *pincy;
	INT        i, ns, m, ix, iy;

	if (n > 0)
	{
		if (incx == incy && incx >= 1)
		{
			if (incx != 1)
			{/* code for equal, positive, nonunit increments. */
				ns = n*incx;
				for (i = 0; i < ns; i += incx)
				{
					dtemp1 = dx[i];
					dx[i] = dy[i];
					dy[i] = dtemp1;
				}
			} /*if (incx != 1)*/
			else
			{/* code for both increments equal to 1 */
			/* clean-up loop so remaining vector length is a multiple of 3*/
				m = n % 3;
				for (i = 0; i < m ; i++)
				{
					dtemp1 = dx[i];
					dx[i] = dy[i];
					dy[i] = dtemp1;
				}
				for (i = m; i < n; i += 3 )
				{
					dtemp1 = dx[i];
					dtemp2 = dx[i+1];
					dtemp3 = dx[i+2];
					dx[i] = dy[i];
					dx[i+1] = dy[i+1];
					dx[i+2] = dy[i+2];
					dy[i] = dtemp1;
					dy[i+1] = dtemp2;
					dy[i+2] = dtemp3;
				}
			} /*if (incx != 1){}else{}*/
		} /*if (incx == incy && incx >= 1)*/
		else
		{/* code for unequal or nonpositive increments. */
			ix = (incx >= 0) ? 0 : (-n+1)*incx;
			iy = (incy >= 0) ? 0 : (-n+1)*incy;
			for (i = 0; i < n ; i++)
			{
				dtemp1 = dx[ix];
				dx[ix] = dy[iy];
				dy[iy] = dtemp1;
				ix += incx;
				iy += incy;
			}
		} /*if (incx == incy && incx >= 1){}else{}*/
	} /*if (n>0)*/
} /*DSWAP()*/
#endif /*DSWAP*/

#ifdef IDAMAX
/*====> idamax.c <====*/
/* Translated from Fortran by C. Bingham (kb@umnstat.stat.umn.edu) */

/*
  find smallest index of maximum magnitude of double precision dx.
  idamax =  first i, i = 1 to n, to minimize  abs(dx(1-incx+i*incx))
  Note that C version returns value from 1 to n (unless n == 0), just as
  does Fortran.
*/

INT IDAMAX(INT * pn, double dx[], INT * pincx)
{
	double     dmax, xmag;
	INT        n = *pn, incx = *pincx;
	INT        i, ii, ns, idamax;

	if (n > 0)
	{
		idamax = 0;
		if (n>1)
		{
			dmax = fabs(dx[0]);
			if (incx == 1)
			{/* code for increments equal to 1. */
				for (i = 1; i < n ; i++)
				{
					xmag = fabs(dx[i]);
					if (xmag>dmax)
					{
						idamax = i;
						dmax = xmag;
					}
				} /*for (i = 1; i < n ; i++)*/
			} /*if (incx == 1)*/
			else
			{/* code for increments not equal to 1. */
				ns = n*incx;
				ii = 0;
				for (i = incx; i < ns; i += incx)
				{
					xmag = fabs(dx[i]);
					if (xmag > dmax)
					{
						idamax = ii;
						dmax = xmag;
					}
					ii++;
				}
			} /*if (incx == 1){}else{}*/
		} /*if (n > 1)*/
	} /*if (n > 0)*/
	else
	{
		idamax = -1;
	}
	return (idamax + 1);
} /*IDAMAX()*/
#endif /*IDAMAX*/

/*
  Level 3 BLAS
*/
#ifdef DGEMM

#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))

/*
  Purpose
  =======

  dgemm  performs one of the matrix - matrix operations

  c := alpha*op( a )*op( b ) + beta*c,

  where  op( x ) is one of

  op( x ) = x   or   op( x ) = x',

  alpha and beta are scalars,  and a,  b and c are matrices,  with op( a )
  an m by k matrix,   op( b )  a  k by n matrix and  c an m by n matrix.

  Parameters
  ==========

  transa - character*1.
  On entry,  transa specifies the form of op( a ) to be used in
  the matrix multiplication as follows:

  transa = 'n' or 'n',   op( a ) = a.

  transa = 't' or 't',   op( a ) = a'.

  transa = 'c' or 'c',   op( a ) = a'.

  Unchanged on exit.

  transb - character*1.
  On entry,  transb specifies the form of op( b ) to be used in
  the matrix multiplication as follows:

  transb = 'N' or 'n',   op( b ) = b.

  transb = 'T' or 't',   op( b ) = b'.

  transb = 'C' or 'c',   op( b ) = b'.

  Unchanged on exit.

  m - integer.
  On entry,   m  specifies  the number  of rows  of the  matrix
  op( a )  and of the  matrix  c.  m  must  be at least  zero.
  Unchanged on exit.

  n - integer.
  On entry,   n  specifies the number  of columns of the matrix
  op( b ) and the number of columns of the matrix c. n must be
  at least zero.
  Unchanged on exit.

  k - integer.
  On entry,   k  specifies  the number of columns of the matrix
  op( a ) and the number of rows of the matrix op( b ). k must
  be at least  zero.
  Unchanged on exit.

  alpha - double precision.
  On entry,  alpha specifies the scalar alpha.
  Unchanged on exit.

  a - double precision array of dimension ( lda,  ka ),  where ka is
  k  when  transa = 'n' or 'n',   and is  m  otherwise.
  before entry with  transa = 'n' or 'n',   the leading  m by k
  part of the array  a  must contain the matrix  a,   otherwise
  the leading  k by m  part of the array  a  must contain  the
  matrix a.
  Unchanged on exit.

  lda - integer.
  On entry,  lda specifies the first dimension of a as declared
  in the calling (sub) program. when  transa = 'n' or 'n' then
  lda must be at least  max( 1,  m ),  otherwise  lda must be at
  least  max( 1,  k ).
  Unchanged on exit.

  b - double precision array of dimension ( ldb,  kb ),  where kb is
  n  when  transb = 'n' or 'n',   and is  k  otherwise.
  before entry with  transb = 'n' or 'n',   the leading  k by n
  part of the array  b  must contain the matrix  b,   otherwise
  the leading  n by k  part of the array  b  must contain  the
  matrix b.
  Unchanged on exit.

  ldb - integer.
  On entry,  ldb specifies the first dimension of b as declared
  in the calling (sub) program. when  transb = 'n' or 'n' then
  ldb must be at least  max( 1,  k ),  otherwise  ldb must be at
  least  max( 1,  n ).
  Unchanged on exit.

  beta - double precision.
  On entry,   beta  specifies the scalar  beta.  when  beta  is
  supplied as zero then c need not be set on input.
  Unchanged on exit.

  c - double precision array of dimension ( ldc,  n ).
  before entry,  the leading  m by n  part of the array  c must
  contain the matrix  c,   except when  beta  is zero,  in which
  case c need not be set On entry.
  on exit,  the array  c  is overwritten by the  m by n  matrix
  ( alpha*op( a )*op( b ) + beta*c ).

  ldc - integer.
  On entry,  ldc specifies the first dimension of c as declared
  in  the  calling  (sub)  program.   ldc  must  be  at  least
  max( 1,  m ).
  Unchanged on exit.


  Level 3 blas routine.

  -- Written on 8 - february - 1989.
  Jack Dongarra,  Argonne National Laboratory.
  Iain Duff,  AERE Harwell.
  Jeremy Du Croz,  Numerical Algorithms Group Ltd.
  Sven Hammarling,  Numerical Algorithms Group Ltd.

  This code comes from a report entitled:
  The IBM RISC System/6000 and linear algebra operations,  by
  Jack J. Dongarra,  Peter Mayes,  and Giuseppe Radicati di Brozolo,
  University of Tennessee Computer Science Tech Report: CS - 90 - 122.

  Translated to C by C. Bingham, kb@stat.umn.edu, March 24, 2002

  It preserves Fortran calling sequence; all arguments are considered
  pointers; subscript ordering follows Fortran; subscripts are 0 based

  If USEDDOT is not defined, the original coding is followed, doing
  the inner products by for loops derived from Fortran do loops.

  If USEDDOT is defined, DDOT is used for the inner products.  
*/

#ifndef DONTTESTDGEMMINPUT /* if defined, arguments are not checked */
extern void     xerbla(char *, int);
#endif /*DONTTESTDGEMMINPUT*/

#include <ctype.h>
#define lsame(C1, C2) (*(C1) == (C2) || *(C1) == toupper(C2))

#ifndef DGEMM_MB
#define DGEMM_MB 64
#endif /*DGEMM_MB*/

#ifndef DGEMM_KB
#define DGEMM_KB 64
#endif /*DGEMM_KB*/

enum   dgemmParameters
{
	/* These define the size of block used */
	mb = DGEMM_MB,
	nb = mb,
	kb = DGEMM_KB,
	kbmb = kb*mb /* size of scratch array allocated on stack */
};

#define Qa(I,J)    a[(I) + (J)*lda]
#define Qb(I,J)    b[(I) + (J)*ldb]
#define Qc(I,J)    c[(I) + (J)*ldc]
#define Qch(I,J)   ch[(I) + kb*(J)]

#define QaP(I,J)    (a + (I) + (J)*lda)
#define QbP(I,J)    (b + (I) + (J)*ldb)
#define QcP(I,J)    (c + (I) + (J)*ldc)
#define QchP(I,J)   (ch + (I) + kb*(J))

void DGEMM(char * transa, char * transb, INT * m1, INT * n1, INT * k1,
		   double * alpha1, double * a, INT * lda1, double * b, INT * ldb1,
		   double * beta1, double * c, INT * ldc1)
/*
  Original declaration
  double precision a(lda,*),b(ldb,*),c(ldc,*)
*/
{
	/*      .. dereferenced scalar arguments .. */
	double     alpha = *alpha1, beta = *beta1;
	INT        k = *k1, lda = *lda1, ldb = *ldb1, ldc = *ldc1;
	INT        m = *m1, n = *n1;
	/*      .. local scalars .. */
#ifndef USEDDOT
	double     t11, t12, t21, t22;
#else /*USEDDOT*/
	INT        one = 1;
#endif /*USEDDOT*/
    INT        i, idepth, ii, ilen, ispan;
	INT        j, jdepth, jj, jlen, jspan;
	INT        l,ll, lspan, ncola, nrowa, nrowb;
#ifndef DONTTESTDGEMMINPUT
	INT        info
#endif /*DONTTESTDGEMMINPUT*/
	int         nota, notb;
	/*      .. local arrays .. */
	/* 
	   Original declaration
	   double precision ch(kb,mb), ch1(kb), ch2(kb)
	*/

	double     ch[kbmb], ch1[kb], ch2[kb];
	/*  .. executable statements .. */

	/*
	  Set  nota  and  notb  as  true if  a  and  b  respectively are not
	  transposed and set  nrowa,  ncola and  nrowb  as the number of rows
	  and  columns of  a  and the  number of  rows  of  b  respectively.
	*/
	nota = lsame(transa, 'n');
	notb = lsame(transb, 'n');

	if (!nota) 
	{
		nrowa = k;
		ncola = m;
	}
	else 
	{
		nrowa = m;
		ncola = k;
	}

	nrowb = (notb) ? k : n;
	
#ifndef DONTTESTDGEMMINPUT
	/*  Test the input parameters. */

	info = 0;
	if (!nota && !lsame(transa, 'c') && !lsame(transa, 't'))
	{
		info = 1;
	}
	else if (!notb && !lsame(transb, 'c') && !lsame(transb, 't'))
	{
		info = 2;
	}
	else if (m < 0)
	{
		info = 3;
	}
	else if (n < 0)
	{
		info = 4;
	}
	else if (k < 0)
	{
		info = 5;
	}
	else if (lda < ((nrowa > 1) ? nrowa : 1))
	{
		info = 8;
	}
	else if (ldb < ((nrowb > 1) ? nrowb : 1))
	{
		info = 10;
	}
	else if (ldc < ((m > 1) ? m : 1))
	{
		info = 13;
	}

	if (info != 0)
	{
		goto errorExit;
	}
#endif /*DONTTESTDGEMMINPUT*/

	/*  Quick return if possible. */

	if (m == 0 || n == 0 || beta == 1.0 && (alpha == 0.0 || k == 0))
	{
		return;
	}

	if (beta != 0.0)
	{
		for (j = 0; j < n ; j++)
		{
			for (i = 0; i < m ; i++)
			{
				Qc(i,j) *= beta;
			}
		}
	}
	else
	{
		for (j = 0; j < n ; j++)
		{
			for (i = 0; i < m ; i++)
			{
				Qc(i,j) = 0.0;
			}
		}
	}

	/*  And if  alpha.eq.0.0. */

	if (alpha == 0.0)
	{
		return;
	}
	
	/*  start the operations. */

	if (!notb)
	{
		if (!nota)
		{
			/* form  c := c + alpha*a'*b' */
			for (j = 0;j < n;j += nb )
			{ 
				jdepth = 2;
				jspan = MIN(nb, n - j);
				jlen = jdepth*(jspan/jdepth);
				for (l = 0;l < k;l += kb )
				{ 
					lspan = MIN(kb, k - l);
					for (jj = j;jj < j + jspan ;jj++)
					{
						for (ll = l;ll < l + lspan ;ll++)
						{
							Qch(ll - l,jj - j) = alpha*Qb(jj, ll);
						}
					}

					for (i = 0;i < m;i += mb )
					{ 
						idepth = 2;
						ispan = MIN(mb, m - i);
						ilen = idepth*(ispan/idepth);
						for (ii = i;ii < i + ilen;ii += idepth )
						{ 
							for (jj = j;jj < j + jlen;jj += jdepth )
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t21 = 0.0;
								t12 = 0.0;
								t22 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ll, ii)*Qch(ll - l+1, jj - j);
									t21 += Qa(ll, ii + 1)*Qch(ll - l+1, jj - j);
									t12 += Qa(ll, ii)*Qch(ll - l+1, jj - j+1);
									t22 += Qa(ll, ii + 1)*Qch(ll - l+1, jj - j+1);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
								Qc(ii,jj + 1) += t12;
								Qc(ii + 1,jj + 1) += t22;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QaP(l,ii), &one,
									   QchP(0,jj-j), &one);
								Qc(ii + 1,jj) +=
								  DDOT(&lspan, QaP(l,ii+1), &one,
									   QchP(0,jj-j), &one);
								Qc(ii,jj + 1) +=
								  DDOT(&lspan, QaP(l,ii), &one,
									   QchP(0,jj+1-j), &one);
								Qc(ii + 1,jj + 1) +=
								  DDOT(&lspan, QaP(l,ii + 1), &one,
									   QchP(0,jj+1-j), &one);
#endif /*USEDDOT*/
							}

							for (jj = j + jlen;jj < j + jspan ;jj++)
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t21 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ll, ii)*Qch(ll - l, jj - j);
									t21 += Qa(ll, ii + 1)*Qch(ll - l, jj - j);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QaP(l,ii), &one,
									   QchP(0,jj-j), &one);
								Qc(ii + 1,jj) +=
								  DDOT(&lspan, QaP(l,ii+1), &one,
									   QchP(0,jj-j), &one);
#endif /*USEDDOT*/
							}
						}

						for (ii = i + ilen;ii < i + ispan ;ii++)
						{ 
							for (jj = j;jj < j + jlen;jj += jdepth )
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t12 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ll, ii)*Qch(ll - l, jj - j);
									t12 += Qa(ll, ii)*Qch(ll - l, jj - j + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii,jj + 1) += t12;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QaP(l,ii), &one,
									   QchP(0,jj-j), &one);
								Qc(ii,jj + 1) +=
								  DDOT(&lspan, QaP(l,ii), &one,
									   QchP(0,jj+1-j), &one);
#endif /*USEDDOT*/
							} /*for (jj = j;jj < j + jlen;jj += jdepth )*/

							for (jj = j + jlen;jj < j + jspan ;jj++)
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{
									t11 += Qa(ll, ii)*Qch(ll - l, jj - j);
								}
								Qc(ii,jj) += t11;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QaP(l,ii), &one,
									   QchP(0,jj-j), &one);
#endif /*USEDDOT*/
							} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
						} /*for (ii = i + ilen;ii < i + ispan ;ii++)*/
					} /*for (i = 0;i < m;i += mb )*/
				} /*for (l = 0;l < k;l += kb )*/
			} /*for (j = 0;j < n;j += nb )*/
		} /*if (!nota)*/
		else
		{
			/* form  c := c + alpha*a*b' */
			for (j = 0;j < n;j += nb )
			{ 
				jdepth = 2;
				jspan = MIN(nb, n - j);
				jlen = jdepth*(jspan/jdepth);
				for (l = 0;l < k;l += kb )
				{ 
					lspan = MIN(kb, k - l);
					for (jj = j;jj < j + jspan ;jj++)
					{
						for (ll = l;ll < l + lspan ;ll++)
						{
							Qch(ll - l,jj - j) = alpha*Qb(jj, ll);
						}
					}

					for (i = 0;i < m;i += mb )
					{ 
						idepth = 2;
						ispan = MIN(mb, m - i);
						ilen = idepth*(ispan/idepth);
						for (ii = i;ii < i + ilen;ii += idepth )
						{ 
							for (ll = l;ll < l + lspan ;ll++)
							{ 
								ch1[ll - l] = Qa(ii, ll);
								ch2[ll - l] = Qa(ii + 1, ll);
							}
							for (jj = j;jj < j + jlen;jj += jdepth )
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t21 = 0.0;
								t12 = 0.0;
								t22 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += ch1[ll - l]*Qch(ll - l, jj - j);
									t21 += ch2[ll - l]*Qch(ll - l, jj - j);
									t12 += ch1[ll - l]*Qch(ll - l, jj - j+1);
									t22 += ch2[ll - l]*Qch(ll - l, jj - j+1);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
								Qc(ii,jj + 1) += t12;
								Qc(ii + 1,jj + 1) += t22;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, ch1, &one,
									   QchP(0,jj-j), &one);
								Qc(ii + 1,jj) +=
								  DDOT(&lspan, ch2, &one,
									   QchP(0,jj-j), &one);
								Qc(ii,jj + 1) +=
								  DDOT(&lspan, ch1, &one,
									   QchP(0,jj+1-j), &one);
								Qc(ii + 1,jj + 1) +=
								  DDOT(&lspan, ch2, &one,
									   QchP(0,jj+1-j), &one);
#endif /*USEDDOT*/
							} /*for (jj = j;jj < j + jlen;jj += jdepth )*/

							for (jj = j + jlen;jj < j + jspan ;jj++)
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t21 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ii, ll)*Qch(ll - l, jj - j);
									t21 += Qa(ii + 1, ll)*Qch(ll - l, jj - j);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QaP(ii,l), &lda,
									   QchP(0,jj-j), &one);
								Qc(ii + 1,jj) +=
								  DDOT(&lspan, QaP(ii+1,l), &lda,
									   QchP(0,jj-j), &one);
#endif /*USEDDOT*/
							} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
						} /*for (ii = i;ii < i + ilen;ii += idepth )*/

						for (ii = i + ilen;ii < i + ispan ;ii++)
						{ 
							for (jj = j;jj < j + jlen;jj += jdepth )
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t12 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ii, ll)*Qch(ll - l, jj - j);
									t12 += Qa(ii, ll)*Qch(ll - l, jj - j+1);
								}
								Qc(ii,jj) += t11;
								Qc(ii,jj + 1) += t12;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QaP(ii,l), &lda,
									   QchP(0,jj-j), &one);
								Qc(ii,jj + 1) +=
								  DDOT(&lspan, QaP(ii,l), &lda,
									   QchP(0,jj+1-j), &one);
#endif /*USEDDOT*/
							} /*for (jj = j;jj < j + jlen;jj += jdepth )*/

							for (jj = j + jlen;jj < j + jspan ;jj++)
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{
									t11 += Qa(ii, ll)*Qch(ll - l, jj - j);
								}
								Qc(ii,jj) += t11;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QaP(ii,l), &lda,
									   QchP(0,jj-j), &one);
#endif /*USEDDOT*/
							} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
						} /*for (ii = i + ilen;ii < i + ispan ;ii++)*/
					} /*for (i = 0;i < m;i += mb )*/
				} /*for (l = 0;l < k;l += kb )*/
			} /*for (j = 0;j < n;j += nb )*/
		} /*if (!nota){}else{}*/
	} /*if (!notb)*/
	else
	{
		if (!nota)
		{
			/*        form  c := c + alpha*a'*b */
			for (i = 0;i < m;i += mb )
			{ 
				idepth = 2;
				ispan = MIN(mb, m - i);
				ilen = idepth*(ispan/idepth);
				for (l = 0;l < k;l += kb )
				{ 
					lspan = MIN(kb, k - l);
					for (ii = i;ii < i + ispan ;ii++)
					{
						for (ll = l;ll < l + lspan ;ll++)
						{
							Qch(ll - l,ii - i) = alpha*Qa(ll, ii);
						}
					}

					for (j = 0;j < n;j += nb )
					{ 
						jdepth = 2;
						jspan = MIN(nb, n - j);
						jlen = jdepth*(jspan/jdepth);
						for (jj = j;jj < j + jlen;jj += jdepth )
						{ 
							for (ii = i;ii < i + ilen;ii += idepth )
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t21 = 0.0;
								t12 = 0.0;
								t22 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t21 += Qch(ll - l, ii - i+1)*Qb(ll, jj);
									t12 += Qch(ll - l, ii - i)*Qb(ll, jj + 1);
									t22 += Qch(ll - l, ii - i+1)*Qb(ll, jj + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
								Qc(ii,jj + 1) += t12;
								Qc(ii + 1,jj + 1) += t22;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj), &one);
								Qc(ii + 1,jj) +=
								  DDOT(&lspan, QchP(0,ii+1-i), &one,
									   QbP(l, jj), &one);
								Qc(ii,jj + 1) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj + 1), &one);
								Qc(ii + 1,jj + 1) +=
								  DDOT(&lspan, QchP(0,ii+1-i), &one,
									   QbP(l, jj + 1), &one);
#endif /*USEDDOT*/
							} /*for (ii = i;ii < i + ilen;ii += idepth )*/

							for (ii = i + ilen;ii < i + ispan ;ii++)
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t12 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t12 += Qch(ll - l, ii - i)*Qb(ll, jj + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii,jj + 1) += t12;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj), &one);
								Qc(ii,jj + 1) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj + 1), &one);
#endif /*USEDDOT*/
							} /*for (ii = i + ilen;ii < i + ispan ;ii++)*/
						} /*for (jj = j;jj < j + jlen;jj += jdepth )*/
						for (jj = j + jlen;jj < j + jspan ;jj++)
						{ 
							for (ii = i;ii < i + ilen;ii += idepth )
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t21 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t21 += Qch(ll - l, ii - i+1)*Qb(ll, jj);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj), &one);
								Qc(ii + 1,jj) +=
								  DDOT(&lspan, QchP(0,ii+1-i), &one,
									   QbP(l, jj), &one);
#endif /*USEDDOT*/
							} /*for (ii = i;ii < i + ilen;ii += idepth )*/

							for (ii = i + ilen;ii < i + ispan ;ii++)
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
								}
								Qc(ii,jj) += t11;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj), &one);
#endif /*USEDDOT*/
							} /*for (ii = i + ilen;ii < i + ispan ;ii++)*/
						} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
					} /*for (j = 0;j < n;j += nb )*/
				} /*for (l = 0;l < k;l += kb )*/
			} /*for (i = 0;i < m;i += mb )*/
		} /*if (!nota)*/ 
		else
		{
			/* form  c := c + alpha*a*b. */
			for (l = 0;l < k;l += kb )
			{ 
				lspan = MIN(kb, k - l);
				for (i = 0;i < m;i += mb )
				{ 
					idepth = 2;
					ispan = MIN(mb, m - i);
					ilen = idepth*(ispan/idepth);
					for (ii = i;ii < i + ispan ;ii++)
					{
						for (ll = l;ll < l + lspan ;ll++)
						{
							Qch(ll - l,ii - i) = alpha*Qa(ii, ll);
						}
					}

					for (j = 0;j < n;j += nb )
					{ 
						jdepth = 2;
						jspan = MIN(nb, n - j);
						jlen = jdepth*(jspan/jdepth);
						for (jj = j;jj < j + jlen;jj += jdepth )
						{ 
							for (ii = i;ii < i + ilen;ii += idepth )
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t21 = 0.0;
								t12 = 0.0;
								t22 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t21 += Qch(ll - l, ii - i+1)*Qb(ll, jj);
									t12 += Qch(ll - l, ii - i)*Qb(ll, jj + 1);
									t22 += Qch(ll - l, ii - i+1)*Qb(ll, jj + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
								Qc(ii,jj + 1) += t12;
								Qc(ii + 1,jj + 1) += t22;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj), &one);
								Qc(ii + 1,jj) +=
								  DDOT(&lspan, QchP(0,ii+1-i), &one,
									   QbP(l, jj), &one);
								Qc(ii,jj + 1) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj + 1), &one);
								Qc(ii + 1,jj + 1) +=
								  DDOT(&lspan, QchP(0,ii+1-i), &one,
									   QbP(l, jj + 1), &one);
#endif /*USEDDOT*/
							}

							for (ii = i + ilen;ii < i + ispan ;ii++)
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t12 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t12 += Qch(ll - l, ii - i)*Qb(ll, jj + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii,jj + 1) += t12;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj), &one);
								Qc(ii,jj + 1) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj + 1), &one);
#endif /*USEDDOT*/
							}
						}

						for (jj = j + jlen;jj < j + jspan ;jj++)
						{ 
							for (ii = i;ii < i + ilen;ii += idepth )
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								t21 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t21 += Qch(ll - l, ii - i+1)*Qb(ll, jj);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj), &one);
								Qc(ii + 1,jj) +=
								  DDOT(&lspan, QchP(0,ii+1-i), &one,
									   QbP(l, jj), &one);
#endif /*USEDDOT*/
							}

							for (ii = i + ilen;ii < i + ispan ;ii++)
							{ 
#ifndef USEDDOT
								t11 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
								}
								Qc(ii,jj) += t11;
#else /*USEDDOT*/
								Qc(ii,jj) +=
								  DDOT(&lspan, QchP(0,ii-i), &one,
									   QbP(l, jj), &one);
#endif /*USEDDOT*/
							}
						} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
					} /*for (j = 0;j < n;j += nb )*/
				} /*for (i = 0;i < m;i += mb )*/
			} /*for (l = 0;l < k;l += kb )*/
		}  /*if (!nota){}else{}*/
	} /*if (!notb){}else{}*/

#ifndef DONTTESTDGEMMINPUT
	return;

  errorExit:
	xerbla("dgemm ", info);
#endif /*DONTTESTDGEMMINPUT*/
} /*dgemm()*/

#endif /*DGEMM*/
