/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2003 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Stemleaf
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/


#ifdef MPW
#include <Memory.h>
#endif /*MPW*/

#include "globals.h"

#include "mvIface.h"

#include "mvGraphs.h"
#include "mvIo.h"

enum boxPlotConsts
{
	/* boxtypes*/
	STANDARDBOX        =     0,
	FIVENUMBOX,
	NBOXTYPES,
	MAXBOX             =     100, /* maximum number of boxes */

	MAXBOXDUMB         =     25, /* maximum number of boxes with dumb plots, ignored*/
	MAXBOXTICKS        =     20, /* max number automatic ticks to number boxes*/
/*
	"standard" numbers of boxes; used to determine width or height for
	small number of boxes.	
*/
	HORSTANDARD         =    4,
	VERTSTANDARD        =    5,

	STARTSMALL          =    21, /*start of use of small symbols*/
	STARTSMALLER        =    41, /*start of use of smaller symbols*/

	INNERPOINT          =    STARPOINT + 1,  /* star */
	INNERCHAR           =    INNERPOINT, /*for dumb terminal*/
	OUTERPOINT          =    DIAMONDPOINT + 1,  /* diamond */
	OUTERCHAR           =    OUTERPOINT, /*for dumb terminal*/
	SMALLINNERPOINT     =    SMALLSTARPOINT + 1,  /* small star */
	SMALLOUTERPOINT     =    SMALLDIAMONDPOINT + 1,  /* small diamond */
	SMALLERINNERPOINT   =    SMALLERSTARPOINT + 1,  /* smaller star */
	SMALLEROUTERPOINT   =    SMALLERDIAMONDPOINT + 1  /* smaller diamond */
};

#define DEFAULTBOXSIZE       0.66
#define	INNERFENCEMULTIPLE   1.5
#define	OUTERFENCEMULTIPLE   3.0

#if (0) /*old definitions*/
#define INNERCHAR     '*'
#define OUTERCHAR     'o' /* for dumb terminal */
#endif /*0*/

enum boxPlotScratch
{
	GYH = 0,
	NTRASH
};

/*
   Make box plots of data

   Uses function boxprep(y, n, fivenum, nout)
   where y[] is a vector of n not missing values
   fivenum[0] is set to smallest value >= than lower innter fence
   fivenum[1] is set to Q1 (lower quartile)
   fivenum[2] is set to Q2 (median)
   fivenum[3] is set to Q3 (upper quartile)
   fivenum[4] is set to smallest value <= than upper innter fence
   lower inner and outer fences are Q1 - 1.5*IQR and Q1 - 3.0*IQR, respectively
   upper inner and outer fences are Q1 + 1.5*IQR and Q1 + 3.0*IQR, respectively

   nout[0] = # of cases < lower outer fence
   nout[1] = # of case >= lower outer fence and < lower inner fence
   nout[2] = # of case <= upper outer fence and > upper inner fence
   nout[3] = # of cases > upper outer fence

   Version 951206
	 Tidied up
	 box is now drawn as 4 connected line segments
	 De-referenced handles
   960705 moved boxprep() here from boxprep.c
   970105 added support for keyword height and width

   970307 moved preparation of the GRAPH symbol to new function buildBoxplot()

   971215 changed call to getPlotKeys() and modified code accordingly
   980724 Implemented notes:charVector
   980824 When only one box is drawn and the variable is not a scratch
          variable, its name is used instead of "Values" as an axis label
   981007 Minor changes to error messages
   981206 Changed default sizes for boxes when nbox is small, to avoid big
          fat boxes.  For horizontal boxes with nbox < 4, their heights are
          the same as the standard with nbox = 4, while for vertical boxes
          with nbox < 5 the widths are the same as the standard with nbox = 5.
          For horizontal boxes, including either ymin:? or ymax:? gives
          previous behavior, while for vertical boxes, xmin:? or xmax:?
          gives previous behavior.
   990212 Replaced putOUTSTR() by putErrorOUTSTR()
   990522 Now includes mvGraphs.h
   000523 All WARNING messages are protected by if (!keyValues->silent)
          to implement 'silent:T' as a keyword phrase.
   010125 boxprep() has new argument and boxplot() is aware of new
          option excludeM
   010522 Changed call to macSetFile()
   010523 Replaced macSetInfo() by macSetFInfo()
   010524 Changed calls to findPlotFile(), openPlotFile()
          Moved setting of Macintosh creator and type to openPlotFile()
		  in plotutil.c
   010601 Replaced includes of wxIface.h and macIface.h by include mvIface.h
          changed other references to MACINTOSH
   020220 Moved actual box drawing to drawBoxWhiskers() with location and
          width arguments.
   020222 Added keywords 'vs' and 'boxsize'.  With vs:x, boxes are plotted
          against x rather than box number.  With boxsize:w, box
          widths are specified by w instead of a default
   020223 Maximum number of ticks labelling boxes is 20, except possible with
          'vs'
          Keyword 'boxtype' implemented. boxtype:1 is default; boxtype:2
          gets 5 number boxplots
   030311 Keyword phrase 'symbols:vector(innerSym,outerSym)' sets symbols
          for moderate and extreme outliers
*/

#define TICKLENGTHNOTSET (4*TICKUNIT)

/*
  Function to do preparatory computation for boxplots and stem and leaf
  displays.  Originally in boxprep.c
  On output:
    array xsc has been sorted
    fivenum[0] = adjlo   nout[0] = number < lower outer fence
    fivenum[1] = Q1      nout[1] = number < lower inner fence, >= outer fence
    fivenum[2] = median  nout[2] = number > upper inner fence, <= outer fence
    fivenum[3] = Q3      nout[3] = number > upper outer fence
    fivenum[4] = adjhi

    where adjlo = whisker end = smallest number >= lower inner fence 
          adjhi = whisker end = largest number <= upper inner fence
*/

void boxprep(double * xsc, long n, double fivenum [], long nout [],
			 int excludeM)
{
	long            i, j, k;
	double          adjlo, adjhi, lf1, lf2, uf1, uf2;
	double          iqr, q1,q2,q3;

	sortquick(xsc, n);

	i = (n + 1) / 2;
	j = (n / 2) + 1;
	q2 = (xsc[i - 1] + xsc[j - 1]) / 2.;

	j = ((excludeM) ? n - 2 : n - 1)/4;
	k = ((excludeM) ? n : n + 1)/4;
	q1 = (xsc[j] + xsc[k]) / 2.;
	q3 = (xsc[n - 1 - j] + xsc[n - 1 - k]) / 2.;

	iqr = q3 - q1;

	/* compute fences*/
	lf1 = q1 - INNERFENCEMULTIPLE * iqr;
	lf2 = q1 - OUTERFENCEMULTIPLE * iqr;
	uf1 = q3 + INNERFENCEMULTIPLE * iqr;
	uf2 = q3 + OUTERFENCEMULTIPLE * iqr;

	nout[0] = 0; /* number < lower outer fence */
	nout[1] = 0; /* number >= lower outer fence  && < lower inner fence*/
	nout[2] = 0; /* number > upper outer fence */
	nout[3] = 0; /* number <= upper outer fence  && > upper inner fence*/

	i = 0;
	while (xsc[i] < lf2)
	{
		i++;
		nout[0]++;
	}
	while (xsc[i] < lf1)
	{
		i++;
		nout[1]++;
	}
	adjlo = xsc[i]; /* smallest >= lower inner fence */

	i = n - 1;
	while (xsc[i] > uf2)
	{
		i--;
		nout[3]++;
	}
	while (xsc[i] > uf1)
	{
		i--;
		nout[2]++;
	}
	adjhi = xsc[i]; /* largest <= upper inner fence */

	fivenum[0] = adjlo;
	fivenum[1] = q1;
	fivenum[2] = q2;
	fivenum[3] = q3;
	fivenum[4] = adjhi;

} /*boxprep()*/

#define OFFSETFROMTICK   0.0  /* was 0.05 */

#undef SWAP
#define SWAP(X,Y,T) (T = X, X = Y, Y = T)

static void set_box_point(curve_points ** this_plot, long pointNo,
						  double x, double y, char * pointLab,
						  unsigned char colors [], long vertical)
{
	if (vertical)
	{
		double         tmp;

		SWAP(x, y, tmp);
	} /*if (vertical)*/
	set_point(this_plot, pointNo, x, y, pointLab, colors[0], colors[1]);
} /*set_box_point()*/

/*
	980826 moved lines in following from the call to setGraphStuff() because
	CodeWarrior crashed on the GRAPHLOGY line
*/
static void setGraphStuff(whole_graph  ** graph, plotKeyValuesPtr keyValues)
{
	strcpy(GRAPHTITLE(graph), keyValues->title);
	strcpy(GRAPHXLAB(graph), keyValues->xlab);
	strcpy(GRAPHYLAB(graph), keyValues->ylab);
	GRAPHXMIN(graph) = keyValues->xmin;
	GRAPHXMAX(graph) = keyValues->xmax;
	GRAPHYMIN(graph) = keyValues->ymin;
	GRAPHYMAX(graph) = keyValues->ymax;
	GRAPHLOGX(graph) = keyValues->logx;
	GRAPHLOGY(graph) = keyValues->logy;
	GRAPHXAXIS(graph) = keyValues->xaxis;
	GRAPHYAXIS(graph) = keyValues->yaxis;
	GRAPHFRAME(graph) = keyValues->frame;
} /*setGraphStuff()*/

enum boxCounts
{
	WHISKERSIZE     =   3, /* # 2 pts to define whisker plus 1 undef pt*/
	WHISKERBARSIZE  =   3, /* # 2 pts to define bars at ends, plus 1 undef*/
	BOXPOINTS       =   8 /* # pts to define box and median line, incld undef*/
};

static curve_points   **drawBoxWhiskers(double boxloc, double boxSize,
										double **yH, long n,
										int vertical, int boxType,
										unsigned char colors [2], 
										char innerPoint[2], char outerPoint[2],
										double fivenum[5], long nout[4],
										curve_points **last_plot,
										whole_graph **graph)
{
	curve_points    **this_plot = (curve_points **) 0;
	double            side1 = boxloc - 0.5*boxSize;
	double            side2 = boxloc + 0.5*boxSize;
	double            loQuartile = fivenum[1];
	double            median = fivenum[2];
	double            hiQuartile = fivenum[3];
	double            whiskerEnd;
	double           *y;
	int               whiskerBoxSize;
	long              pointNo;
	WHERE("drawBoxWhiskers");
	
	/* always draw whiskers */
	this_plot = (curve_points **)
		mygethandle(sizeof(curve_points));
	if (this_plot == (curve_points **) 0)
	{
		goto errorExit;
	}
	NEXTGRAPH(this_plot) = (curve_points **) 0;
	GRAPHPOINTS(this_plot) = (coordinate **) 0;
	GRAPHNPLOTS(graph)++;

	if (GRAPHFIRST_PLOT(graph) == (curve_points **) 0)
	{
		GRAPHFIRST_PLOT(graph) = this_plot;
	}
	else
	{
		NEXTGRAPH(last_plot) = this_plot;
	}
	mylockhandle((char **) this_plot);
	last_plot = this_plot;

	whiskerBoxSize = 2*WHISKERSIZE + BOXPOINTS;
	if (boxType == FIVENUMBOX)
	{
		whiskerBoxSize += 2*WHISKERBARSIZE;
	}
	
	TMPHANDLE = mygethandle(whiskerBoxSize * sizeof(coordinate));
	if (TMPHANDLE == (char **) 0)
	{
		goto errorExit;
	}
	GRAPHPOINTS(this_plot) = (coordinate **) TMPHANDLE;
	mylockhandle(TMPHANDLE); /* needed in Mac version */

	PLOTSTYLE(this_plot) = LINES;
	POINTTYPE(this_plot) = INNERPOINT; /* ignored */
	LINETYPE(this_plot) = BORDERLINETYPE;

	PCOUNT(this_plot) = whiskerBoxSize;

	/* whiskers (2 segments) */

	/* lower whisker */
	pointNo = 0;
	y = *yH;
	whiskerEnd = (boxType == STANDARDBOX) ? fivenum[0] : y[0];
	
	set_box_point(this_plot, pointNo++, whiskerEnd, boxloc, (char *) 0,
				  colors, vertical);
	set_box_point(this_plot, pointNo++, loQuartile, boxloc, (char *) 0,
				  colors, vertical);
	undef_point(this_plot, pointNo++);
	if (boxType == FIVENUMBOX)
	{/*draw bar at end of whisker*/
		set_box_point(this_plot, pointNo++, whiskerEnd, side1, (char *) 0,
					  colors, vertical);
		set_box_point(this_plot, pointNo++, whiskerEnd, side2, (char *) 0,
					  colors, vertical);
		undef_point(this_plot, pointNo++);
	}

	/* upper whisker */
	whiskerEnd = (boxType == STANDARDBOX) ? fivenum[4] : y[n-1];

	set_box_point(this_plot, pointNo++, hiQuartile, boxloc, (char *) 0,
				  colors, vertical);
	set_box_point(this_plot, pointNo++, whiskerEnd, boxloc, (char *) 0,
				  colors, vertical);
	undef_point(this_plot, pointNo++);
	if (boxType == FIVENUMBOX)
	{ /*draw bar at end of whisker*/
		set_box_point(this_plot, pointNo++, whiskerEnd, side1, (char *) 0,
					  colors, vertical);
		set_box_point(this_plot, pointNo++, whiskerEnd, side2, (char *) 0,
					  colors, vertical);
		undef_point(this_plot, pointNo++);
	}

	/* draw box with median line (5 segments) */

	set_box_point(this_plot, pointNo++, loQuartile, side2, (char *) 0,
				  colors, vertical);
	set_box_point(this_plot, pointNo++, hiQuartile, side2, (char *) 0,
				  colors, vertical);
	set_box_point(this_plot, pointNo++, hiQuartile, side1, (char *) 0,
				  colors, vertical);
	set_box_point(this_plot, pointNo++, loQuartile, side1, (char *) 0,
				  colors, vertical);
	set_box_point(this_plot, pointNo++, loQuartile, side2, (char *) 0,
				  colors, vertical);
	undef_point(this_plot, pointNo++);

	/* mid line */
	set_box_point(this_plot, pointNo++, median, side1, (char *) 0,
			  colors, vertical);
	set_box_point(this_plot, pointNo++, median, side2, (char *) 0,
			  colors, vertical);

	/* 020222 allocate space for all points in one handle */
	if (boxType == STANDARDBOX)
	{
		long          npoints = nout[0] + nout[1] + nout[2] + nout[3];

		/* outside values */
		if (npoints > 0)
		{
			long         j, k;
		
			this_plot = (curve_points **) mygethandle(sizeof(curve_points));
			if (this_plot == (curve_points **) 0)
			{
				goto errorExit;
			}
			NEXTGRAPH(this_plot) = (curve_points **) 0;
			GRAPHPOINTS(this_plot) = (coordinate **) 0;
			NEXTGRAPH(last_plot) = this_plot;
			GRAPHNPLOTS(graph)++;
			mylockhandle((char **) this_plot);
			last_plot = this_plot;

			TMPHANDLE = mygethandle(npoints*sizeof(coordinate));
			if (TMPHANDLE == (char **) 0)
			{
				goto errorExit;
			}
			GRAPHPOINTS(this_plot) = (coordinate **) TMPHANDLE;
			mylockhandle(TMPHANDLE); /* needed in Mac version */

			PLOTSTYLE(this_plot) = CHARS;
			POINTTYPE(this_plot) = INNERPOINT; /* ignored */
			LINETYPE(this_plot) = BORDERLINETYPE;

			PCOUNT(this_plot) = npoints;

			pointNo = 0;

			/* low outside (between inner and outer fences) */
			k = nout[0];
			y = *yH;
			for (j = 0; j < nout[1]; j++)
			{
				set_box_point(this_plot, pointNo++, y[k++], boxloc,
							  innerPoint, colors, vertical);
			}
			/* hi out */
			k = n - 1 - nout[3];
			for (j = 0; j < nout[2]; j++)
			{
				set_box_point(this_plot, pointNo++, y[k--], boxloc,
							  innerPoint, colors, vertical);
			}
			/*
			  far outside values, beyond outer fences
			*/
			k = 0;
			for (j = 0; j < nout[0]; j++)
			{
				set_box_point(this_plot, pointNo++, y[k++], boxloc,
							  outerPoint, colors, vertical);
			}
			k = n - 1;
			for (j = 0; j < nout[3]; j++)
			{
				set_box_point(this_plot, pointNo++, y[k--], boxloc,
							  outerPoint, colors, vertical);
			}
		} /*if (npoints > 0)*/
	}
	
	return (this_plot);
	
  errorExit:
	return ((curve_points **) 0);

} /*drawBoxWhiskers()*/


/*
  010125 changed type of vertical to int and added argument excludeM
  030311 added arguments innerPoint and outerPoint; these specify plotting
         symbols for moderate and extreme outliers
*/
static Symbolhandle buildBoxplot(Symbolhandle datalist, long nbox,
								 Symbolhandle symhx,
								 Symbolhandle symhBoxSize,
								 char innerPoint [], char outerPoint [],
								 unsigned char colors [],
								 int vertical, int excludeM, int boxType,
								 plotKeyValuesPtr keyValues)
{
	Symbolhandle   graphsymbol;
	whole_graph  **graph = (whole_graph **) 0;
	curve_points **last_plot = (curve_points **) 0;
	double         boxSize;
	double         hugedbl = HUGEDBL;
	double       **defaultTicks = (double **) 0;
	double         defaultXmin = hugedbl, defaultXmax = -hugedbl;
	double         defaultYmin, defaultYmax;
	double         symhxExtremes[2];
	double         firstBoxSize, lastBoxSize;
	long           standardNbox = (vertical) ? VERTSTANDARD : HORSTANDARD;
	unsigned short defaultXlen, defaultYlen;
	char           coordLabels[2][XLABELSIZE+2];
	char          *defaultXlab = coordLabels[0];
	char          *defaultYlab = coordLabels[1];
	char          *defaultTitle;
	long           nxticks = -1, nyticks = -1;
	long           ibox;
	long           widthPlace, widthInc;
	int            logplot;
	WHERE("buildBoxplot");
	TRASH(NTRASH,errorExit);

	if (symhx != SymbolZero)
	{
		symExtremes(symhx, symhxExtremes);
	}
	else
	{
		symhxExtremes[0] = 1.0;
		symhxExtremes[1] = (double) nbox;
	}
	
	if (symhBoxSize != SymbolZero)
	{
		widthPlace = 0;
		widthInc = (isScalar(symhBoxSize)) ? 0 : 1;
	} /*if (symhBoxSize != SymbolZero)*/
	else
	{
		if (symhx == SymbolZero)
		{
			boxSize = DEFAULTBOXSIZE;
		}
		else
		{
			if (nbox > 1)
			{
				boxSize = DEFAULTBOXSIZE*(symhxExtremes[1] - symhxExtremes[0])/
				  (nbox - 1);
			}
			else if (DATAVALUE(symhx, 0) != 0.0)
			{
				boxSize = DEFAULTBOXSIZE*fabs(DATAVALUE(symhx, 0));
			}
			else
			{
				boxSize = DEFAULTBOXSIZE;
			}
		}
	} /*if (symhBoxSize != SymbolZero){}else{}*/
	
	logplot = (vertical) ? keyValues->logy : keyValues->logx;

	if (nbox == 1 && !isscratch(NAME(COMPVALUE(datalist, 0))))
	{
		char      *name = NAME(COMPVALUE(datalist, 0));
		
		strcpy(coordLabels[0], (istempname(name)) ? name + 1 : name);
	}
	else
	{
		strcpy(coordLabels[0], "Values");
	}
	
	if (symhx == SymbolZero)
	{
		strcpy(coordLabels[1], "Box Number");
	}
	else
	{
		coordLabels[1][0] = '\0';
	}

	/*
	  Be very careful to keep the structure of graph up to date.
	  All handles are zeroed prior to attempting to get storage;
	  GRAPHNPLOTS(graph) is updated as soon a a new struct curve_points
	  is allocated, as is GRAPHFIRST_PLOT(graph) or NEXTGRAPH(last_plot);
	  NEXTGRAPH(this_plot) and GRAPHPOINTS(this_plot) are zeroed as soon
	  as this_plot is allocated.
	  This allows cleanup to be done by deleteGraph
	*/
	if ((graphsymbol = GraphInstall(SCRATCH)) == SymbolError)
	{
		goto errorExit;
	}
	graph = GRAPH(graphsymbol);

	for (ibox = 0; ibox < nbox; ibox++)
	{
		Symbolhandle   symhy = COMPVALUE(datalist, ibox);
		long           j;

		if (!isNull(symhy))
		{/* null symbol is not an error */
			double       **yH;
			double        *x, *y;
			long           tot = symbolSize(symhy);
			long           notmiss, nout[4], positive;

			if (!getScratch(yH, GYH, tot, double))
			{
				goto errorExit;
			}
			x = DATAPTR(symhy);
			y = *yH;
			notmiss = positive = 0;
			/* copy non-MISSING values from x to y */

			for (j = 0; j < tot; j++)
			{
				if (!isMissing(x[j]))
				{
					y[notmiss++] = x[j];
					if (logplot && x[j] > 0)
					{
						positive++;
					}
				}
			} /*for (j = 0; j < tot; j++)*/

			if (notmiss < tot && !keyValues->silent)
			{
				sprintf(OUTSTR, "WARNING: %ld missing values in box %ld",
						tot - notmiss, ibox+1);
				putErrorOUTSTR();
			}

			if (logplot && positive < notmiss && !keyValues->silent)
			{
				sprintf(OUTSTR,
						"WARNING: %ld values <= 0.0 in box %ld",
						notmiss - positive, ibox+1);
				putErrorOUTSTR();
			} /*if (logplot && positive < notmiss)*/
			
			if (notmiss > 0 && (!logplot || positive > 0))
			{
				double         boxloc;
				double         fivenum[5];

				if (symhBoxSize != SymbolZero)
				{
					boxSize = DATAVALUE(symhBoxSize, widthPlace);
					widthPlace += widthInc;
				}
				if (boxSize > 0.0)
				{
					boxprep(y, notmiss, fivenum, nout, excludeM);
					if (y[0] < defaultXmin)
					{
						defaultXmin = y[0];
					} /*if (y[0] < defaultXmin)*/
				
					if (y[notmiss - 1] > defaultXmax)
					{
						defaultXmax = y[notmiss - 1];
					} /*if (y[notmiss - 1] > defaultXmax)*/

					if (symhx == SymbolZero)
					{
						boxloc = (double) ibox + 1 + OFFSETFROMTICK;
					}
					else
					{
						boxloc = DATAVALUE(symhx, ibox);
					}
				
					if (symhx != SymbolZero)
					{
						if (boxloc == symhxExtremes[0])
						{
							firstBoxSize = boxSize;
						}
						else if (boxloc == symhxExtremes[1])
						{
							lastBoxSize = boxSize;
						}
					}
				
					last_plot = drawBoxWhiskers(boxloc, boxSize, yH, notmiss,
												vertical, boxType,
												colors, innerPoint, outerPoint,
												fivenum, nout,
												last_plot, graph);
				} /*if (boxSize > 0.0)*/
				
				if (last_plot == (curve_points **) 0)
				{
					goto errorExit;
				}
				
			} /*if (notmiss > 0)*/
			discardScratch(yH,0,double);
		} /* if (!isNull(symhy) */
	} /*for (ibox = 0; ibox < nbox; ibox++)*/

	if (GRAPHNPLOTS(graph) == 0)
	{
		sprintf(OUTSTR,"ERROR: no data!");
		goto errorExit;
	}

/*
	Limits are always set to include all boxes and
	all data although xmin, xmax, ymin and ymax may be
	used to set limits outside these extremes

	981206 Changed default size of boxes to avoid big "fat boxes"
	For vertical plots with nbox < 5, width of boxes is <= width for 5
	box case
	For horizontal plotsi with nbox < 4, height of boxes is <= width for
	4 box case
	This is overridden when the user sets ymin or ymax (horizontal boxes)
	or xmin or xmax (vertical boxes)
*/
#if (0) /*the following would treat horizontal dumb plots specially */
	if (keyValues->dumb && !vertical &&
		((keyValues->height != UNSETLONG) ?
		 keyValues->height : SCREENHEIGHT) <= 25)
	{
		standardNbox--;
	}
#endif /*0*/

	if (symhx != SymbolZero)
	{
		defaultYmin = symhxExtremes[0] - 0.5*firstBoxSize;
		defaultYmax = symhxExtremes[1] + 0.5*lastBoxSize;
	}
	else if (nbox >= standardNbox ||
		vertical && (realIsSet(keyValues->xmin) || realIsSet(keyValues->xmax)) ||
		!vertical && (realIsSet(keyValues->ymin) || realIsSet(keyValues->ymax)))
	{
		defaultYmin = 0.5;
		defaultYmax = defaultYmin + (double) nbox;
	}
	else
	{
		defaultYmin = 0.5 - .5*standardNbox + .5*nbox;
		defaultYmax = defaultYmin + standardNbox;
	}

	keyValues->ymin =
	  (!realIsSet(keyValues->ymin) || isMissing(keyValues->ymin)) ?
	  hugedbl : keyValues->ymin;
	keyValues->ymax =
	  (!realIsSet(keyValues->ymax) || isMissing(keyValues->ymax)) ?
	  -hugedbl : keyValues->ymax;

	keyValues->xmin =
	  (!realIsSet(keyValues->xmin) || isMissing(keyValues->xmin)) ?
	  hugedbl : keyValues->xmin;
	keyValues->xmax =
	  (!realIsSet(keyValues->xmax) || isMissing(keyValues->xmax)) ?
	  -hugedbl : keyValues->xmax;

	defaultXlen = DEFAULTTICKLENGTH;
	defaultYlen = (symhx == SymbolZero) ? 0 : DEFAULTTICKLENGTH;
	if (vertical)
	{
		/* swap extremes, tick information, default labels */
		double         tmpD;
		unsigned short tmpUs;
		char          *tmpCp;
		
		SWAP(defaultXmin, defaultYmin, tmpD);
		SWAP(defaultXmax, defaultYmax, tmpD);

		SWAP(defaultXlen, defaultYlen, tmpUs);
		
		SWAP(defaultXlab, defaultYlab, tmpCp);
	} /*if (vertical)*/

	keyValues->xticklength = (keyValues->xticklength == TICKLENGTHNOTSET) ?
		defaultXlen : keyValues->xticklength;
	keyValues->yticklength = (keyValues->yticklength == TICKLENGTHNOTSET) ?
		defaultYlen : keyValues->yticklength;
	
	if (keyValues->xticks != UNSETSYMH)
	{
		nxticks = symbolSize(keyValues->xticks);
		GRAPHXTICKS(graph) = DATA(keyValues->xticks);
		setDATA(keyValues->xticks, (double **) 0);
	} /*if (keyValues->xticks != UNSETSYMH)*/

	if (keyValues->yticks != UNSETSYMH)
	{
		nyticks = symbolSize(keyValues->yticks);
		GRAPHYTICKS(graph) = DATA(keyValues->yticks);
		setDATA(keyValues->yticks, (double **) 0);
	} /*if (keyValues->yticks != UNSETSYMH)*/

	if (keyValues->xticklabs != UNSETSYMH)
	{
		GRAPHXTICKLABS(graph) = STRING(keyValues->xticklabs);
		setSTRING(keyValues->xticklabs, (char **) 0);
	} /*if (keyValues->xticklabs != UNSETSYMH)*/

	if (keyValues->yticklabs != UNSETSYMH)
	{
		GRAPHYTICKLABS(graph) = STRING(keyValues->yticklabs);
		setSTRING(keyValues->yticklabs, (char **) 0);
	} /*if (keyValues->yticklabs != UNSETSYMH)*/

	
	if (symhx == SymbolZero)
	{
		char     *wrongLabNumber =
		  "WARNING: wrong number of tick labels for %c axis; ignored";
		char     *noTickLocations =
		  "WARNING: '%cticklabs' ignored without '%c' tick locations";

		if (!vertical && keyValues->yticks == UNSETSYMH ||
			vertical && keyValues->xticks == UNSETSYMH)
		{
			int       tickInc = 1 + (nbox - 1)/MAXBOXTICKS;
			int       defaultNticks = 1 + (nbox - 1)/tickInc;
			
			TMPHANDLE = mygethandle(defaultNticks*sizeof(double));
			if (TMPHANDLE == (char **) 0)
			{
				goto errorExit;
			}
			defaultTicks = (double **) TMPHANDLE;
			if (vertical)
			{
				GRAPHXTICKS(graph) = defaultTicks;
				nxticks = defaultNticks;
			}
			else
			{
				GRAPHYTICKS(graph) = defaultTicks;
				nyticks = defaultNticks;
			}
		
			for (ibox = 0; ibox < nbox; ibox += tickInc)
			{
				(*defaultTicks)[ibox/tickInc] = (double) (ibox + 1);
			}
		}
	
		if (GRAPHXTICKLABS(graph) != (char **) 0 &&
			(GRAPHXTICKS(graph) == (double **) 0 ||
			 symbolSize(keyValues->xticklabs) != nxticks))
		{
			if (GRAPHXTICKS(graph) == (double **) 0)
			{
				sprintf(OUTSTR, noTickLocations, 'x', 'x');
			}
			else
			{
				sprintf(OUTSTR, wrongLabNumber, 'x');
			}
			if (!keyValues->silent)
			{
				putErrorOUTSTR();
			}
			else
			{
				*OUTSTR = '\0';
			}
			
			mydisphandle(GRAPHXTICKLABS(graph));
			GRAPHXTICKLABS(graph) = (char **) 0;
		}

		if (GRAPHYTICKLABS(graph) != (char **) 0 &&
			(GRAPHYTICKS(graph) == (double **) 0 ||
			 symbolSize(keyValues->yticklabs) != nyticks))
		{
			if (GRAPHYTICKS(graph) == (double **) 0)
			{
				sprintf(OUTSTR, noTickLocations, 'y', 'y');
			}
			else
			{
				sprintf(OUTSTR, wrongLabNumber, 'y');
			}
		
			if (!keyValues->silent)
			{
				putErrorOUTSTR();
			}
			else
			{
				*OUTSTR = '\0';
			}
			
			mydisphandle(GRAPHYTICKLABS(graph));
			GRAPHYTICKLABS(graph) = (char **) 0;
		}
	} /*if (symhx == SymbolZero)*/
	

	GRAPHNXTICKS(graph) = nxticks;
	GRAPHNYTICKS(graph) = nyticks;
	GRAPHXTICKLENGTH(graph) = keyValues->xticklength;
	GRAPHYTICKLENGTH(graph) = keyValues->yticklength;

	if (boxType == FIVENUMBOX)
	{
		defaultTitle = "5 Number Box Plot";
	}
	else
	{
		defaultTitle = "Box Plot";
	}
	
	if (keyValues->title[0] == '\0')
	{
		strcpy(keyValues->title, defaultTitle);
	}
	if (keyValues->xlab[0] == '\0')
	{
		strcpy(keyValues->xlab, defaultXlab);
	}
	if (keyValues->ylab[0] == '\0')
	{
		strcpy(keyValues->ylab, defaultYlab);
	}
	keyValues->xmin =
	  (keyValues->xmin < defaultXmin) ? keyValues->xmin : defaultXmin;
	keyValues->xmax =
	  (keyValues->xmax > defaultXmax) ? keyValues->xmax : defaultXmax;
	keyValues->ymin =
	  (keyValues->ymin < defaultYmin) ? keyValues->ymin : defaultYmin;
	keyValues->ymax =
	  (keyValues->ymax > defaultYmax) ? keyValues->ymax : defaultYmax;
	
	setGraphStuff(graph, keyValues);

	emptyTrash();
	return (graphsymbol);

  errorExit:
	putErrorOUTSTR();
	
	Removesymbol(graphsymbol);
	emptyTrash();
	return (SymbolError);
	
} /*buildBoxplot()*/

Symbolhandle    boxplot(Symbolhandle list)
{
	/* routine to do boxplot */

	Symbolhandle   symhy, symhKey;
	Symbolhandle   symhx = SymbolZero, symhBoxSize = SymbolZero;
	Symbolhandle   graphsymbol = SymbolZero;
	whole_graph  **graph = (whole_graph **) 0;
	plotKeyValues  keyValuesData;
	plotKeyValues *keyValues = &keyValuesData;
	long           j, ibox, nbox;
	long           nargs = NARGS(list), first;
	long           firstKey = -1;
	long           isStruct = 0;	
	long           windno = FINDEMPTYWINDOW;
	int            vertical = 0, excludeM = EXCLUDEMEDIAN;
	int            boxType = STANDARDBOX;
	long           toFile = 0, termType = STANDARDTERM;
	FILE          *plotfile = (FILE *) 0;
	unsigned char  pointColor = DEFAULTLINECOLOR;
	unsigned char  lineColor = DEFAULTLINECOLOR;
	unsigned char  colors[2];
	char           innerPoint[PLOTSTRINGSIZE+1];
	char           outerPoint[PLOTSTRINGSIZE+1];
	char          *msgs;
	char           filePath[PATHSIZE+1];
#ifdef HASFINDFILE
	char          *tmpFileName;
#endif /*HASFINDFILE*/
	WHERE("boxplot");
	
	if (nargs == 1 && COMPVALUE(list, 0) == EmptyArgument)
	{
		badNargs(FUNCNAME, -1);
		goto errorExit;
	}

	unsetPlotKeys(keyValues, 1); /* 1 means unset all keys */

	keyValues->xticklength = TICKLENGTHNOTSET;
	keyValues->yticklength = TICKLENGTHNOTSET;
	keyValues->logx = keyValues->logy = 0;
	keyValues->xaxis = keyValues->yaxis = 0;
	keyValues->frame = UNSETTICKS | UNSETFRAME;
	
	keyValues->pause = DEFAULTPLOTPAUSE;
	keyValues->landscape = 0;

	OUTSTR[0] = '\0';

	/* 000129 removed fossil code related to fboxplot() */

	/*
		Find first keyword argument
		Decode non-graphics keywords ('excludeM', 'vertical', 'vs',
		'boxsize'), if any
	*/

	for (j = 1; j < nargs; j++)
	{
		char       *keyword;
		
		symhKey = COMPVALUE(list, j);
		keyword = isKeyword(symhKey);
		
		if (keyword != (char *) 0)
		{
			if (firstKey < 0)
			{
				firstKey = j;
			}
			if (mystrncmp(keyword, "excludeM", 8) == 0 ||
				mystrncmp(keyword, "vert", 4) == 0)
			{
				if (!isTorF(symhKey))
				{
					notTorF(keyword);
					goto errorExit;
				}
				if (tolower(keyword[0]) == 'v')
				{
					vertical = (DATAVALUE(symhKey, 0) != 0.0);
				}
				else
				{
					excludeM = (DATAVALUE(symhKey, 0) != 0.0);
				}
			}
			else if (mystrncmp(keyword, "vs", 2) == 0 ||
					 mystrncmp(keyword, "boxsize", 4) == 0)
			{
				if (!isVector(symhKey) || TYPE(symhKey) != REAL ||
					anyMissing(symhKey))
				{
					sprintf(OUTSTR,
							"ERROR: value for %s not REAL vector with no MISSING values",
							keyword);
					goto errorExit;
				}
				if (keyword[0] == 'b')
				{
					double      extremes[2];
					
					symhBoxSize = symhKey;
					symExtremes(symhKey, extremes);
					if (extremes[0] < 0.0)
					{
						sprintf(OUTSTR,
								"ERROR: widths for boxes must not be negative");
						goto errorExit;
					}
					if (extremes[1] == 0.0)
					{
						sprintf(OUTSTR,
								"ERROR: you must have at least one box with non-zero width");
						goto errorExit;
					}
					symhBoxSize = symhKey;
				}
				else
				{
					symhx = symhKey;
				}
			}
			else if (mystrncmp(keyword, "boxtype", 4) == 0)
			{
				if (!isInteger(symhKey, POSITIVEVALUE) ||
					DATAVALUE(symhKey, 0) > NBOXTYPES)
				{
					sprintf(OUTSTR,
							"ERROR: value for '%s' must be integer between 1 and %d",
							keyword, NBOXTYPES);
					goto errorExit;
				}
				boxType = (int) DATAVALUE(symhKey, 0) - 1;
			}
		} /*if (keyword != (char *) 0)*/
	} /*for (j = 1; j < nargs; j++)*/

	if (firstKey >= 0)
	{
#if (0) /* 990503 show always legal */
		keyValues->show = (!toFile) ? keyValues->show : DONTSETCHAR;
#endif /*0*/
		keyValues->add = DONTSETCHAR;
		keyValues->impulses = keyValues->lines = DONTSETCHAR;
		keyValues->linetype = DONTSETLONG;
		keyValues->thickness = DONTSETREAL;
		
		keyValues->x = keyValues->y = DONTSETSYMH;
		keyValues->strings = DONTSETSYMH;

#ifndef EPSFPLOTSOK
		keyValues->epsf = DONTSETCHAR;
#endif /*EPSFPLOTSOK*/
		keyValues->justify = DONTSETCHAR;
#ifndef SCREENDUMP
		keyValues->screendump = DONTSETPOINTER;
#endif /*SCREENDUMP*/

		first = getPlotKeys(list, firstKey, keyValues);
		unsetPlotKeys(keyValues, 0); /* change DONTSETs to UNSETs */

		/* File names are now (951113) expanded by getPlotKeys()*/
		if (first < 0)
		{
			goto errorExit;
		} /*if (first < 0)*/
	} /*if (j < nargs)*/
	else
	{
		first = 0;
	}

	toFile = (keyValues->file != UNSETPOINTER);

	if (!checkPlotKeys(toFile, keyValues, &termType))
	{
		goto errorExit;
	} 

	if (symhIsSet(keyValues->symbols))
	{
		char      *outlierSymbols;
		
		if (TYPE(keyValues->symbols) != CHAR ||
			!isVector(keyValues->symbols) ||
			DIMVAL(keyValues->symbols, 1) > 2)
		{
			sprintf(OUTSTR,
					"ERROR: value for 'symbols' not a CHARACTER vector of length <= 2");
			goto errorExit;
		}
		outlierSymbols = STRINGPTR(keyValues->symbols);
		strncpy(innerPoint, outlierSymbols,PLOTSTRINGSIZE);
		innerPoint[PLOTSTRINGSIZE] = '\0';
		if (DIMVAL(keyValues->symbols, 1) > 1)
		{
			outlierSymbols = skipStrings(outlierSymbols, 1);
			strncpy(outerPoint, outlierSymbols,PLOTSTRINGSIZE);
			outerPoint[PLOTSTRINGSIZE] = '\0';
		}
		else
		{
			outerPoint[0] = '\0';
		}
	}
	else
	{
		innerPoint[0] = outerPoint[0] = '\0';
	}
	
	if ((keyValues->frame & UNSETFRAME) && (keyValues->frame & UNSETTICKS))
	{
		/* neither border or ticks keywords; use values from graph*/
		keyValues->frame = FRAMEALLSIDES | TICKSALLSIDES;
	}
	else if (keyValues->frame & UNSETFRAME)
	{
		/* ticks but not border; use complete border */
		keyValues->frame = (keyValues->frame & TICKSALLSIDES) | FRAMEALLSIDES;
	}
	else if (keyValues->frame & UNSETTICKS)
	{
		/* border but not ticks; set ticks like border*/
		keyValues->frame = (keyValues->frame & FRAMEALLSIDES) |
		  (TICKSFACTOR * (keyValues->frame & FRAMEALLSIDES));
	}

    if (toFile)
    {
		strcpy(filePath, keyValues->file);
		keyValues->file = filePath;
    }

#ifdef SCREENDUMP
	if (keyValues->screendump != UNSETPOINTER)
	{
		if (termType != STANDARDTERM)
		{
			sprintf(OUTSTR,
					"ERROR: screendump:fileName can only be used with high resolution plot");
			goto errorExit;
		}
#ifdef HASFINDFILE
		if (keyValues->screendump[0] != '\0' &&
			!isfilename(keyValues->screendump))
#else  /*HASFINDFILE*/
		if (!isfilename(keyValues->screendump))
#endif /*HASFINDFILE*/
		{
			goto errorExit;
		}
	} /*if (keyValues->screendump != UNSETPOINTER)*/
#endif /*SCREENDUMP*/

	windno = (keyValues->window != UNSETLONG) ?
	  keyValues->window : FINDEMPTYWINDOW;

	nbox = (firstKey < 0) ? nargs : firstKey;
	symhy = COMPVALUE(list, 0);
	if (!argOK(symhy, (long) 0, 1))
	{
		goto errorExit;
	}
	
	if (nbox == 1 && !isStruc(symhy) && !isVector(symhy))
	{
		sprintf(OUTSTR,
				"ERROR: argument %ld to %s() is not vector or structure",
				1L, FUNCNAME);
		goto errorExit;
	}
	if (nbox == 1 && isStruc(symhy))
	{
		isStruct = 1;
		msgs = "structure component";
		list = symhy;	/* if structure form, set list to structure */
		nbox = NCOMPS(list);
	} /*if (nbox == 1 && TYPE(symhy) == STRUC)*/
	else
	{
		msgs = "argument";
	}

	if (nbox > MAXBOX)
	{
		sprintf(OUTSTR,"ERROR: attempting to plot more than %ld boxes",
				(long) MAXBOX);
		goto errorExit;
	}
	if (symhx != SymbolZero && nbox != symbolSize(symhx))
	{
		sprintf(OUTSTR,
				"ERROR: length of 'vs' vector doesn't match the number of boxes");
		goto errorExit;
	}
	if (symhBoxSize != SymbolZero && !isScalar(symhBoxSize) &&
		nbox != symbolSize(symhBoxSize))
	{
		sprintf(OUTSTR,
				"ERROR: length of 'boxwidth' vector doesn't match the number of boxes");
		goto errorExit;
	}
	
	for (ibox = 0; ibox < nbox; ibox++)
	{ /* check all args or components before doing anything */
		int    argno = ibox + 1;

		symhy = COMPVALUE(list, ibox);
		if (isNull(symhy))
		{
			if (!isStruct)
			{
				sprintf(OUTSTR,"ERROR: %d%s %s for %s is missing or null",
						argno, n_th(argno), msgs, FUNCNAME);
			}
		} /*if (isNull(symhy))*/
		else if (!isDefined(symhy))
		{
			sprintf(OUTSTR,"ERROR: %d%s %s (%s) for %s is undefined",
					argno, n_th(argno), msgs, NAME(symhy),FUNCNAME);
		}
		else if (TYPE(symhy) != REAL)
		{
			sprintf(OUTSTR,"ERROR: %d%s %s for %s is %s, not REAL",
					argno, n_th(argno), msgs, FUNCNAME,typeName(TYPE(symhy)));
		}
		else if (!isVector(symhy))
		{
			sprintf(OUTSTR,"ERROR: %d%s %s for %s is not vector",
					argno, n_th(argno), msgs, FUNCNAME);
		}
		if (*OUTSTR)
		{
			goto errorExit;
		}
	} /*for (ibox = 0; ibox < nbox; ibox++)*/
	
	if (toFile)
	{
		/* keyValues->file must be pointing to filePath */
#ifdef HASFINDFILE
		tmpFileName = findPlotFile(keyValues, "Boxplot");
		
		if (tmpFileName == (char *) 0)
		{ /* Canceled */
			return (0);
		}
		if (tmpFileName != keyValues->file)
		{
			strncpy(keyValues->file, tmpFileName, PATHSIZE);
		}
#endif /*HASFINDFILE*/
		if (!isfilename(keyValues->file))
		{
			goto errorExit;
		}
		
		plotfile = openPlotFile(keyValues);
		if (plotfile == (FILE *) 0)
		{
			goto errorExit;
		}
	} /*if (toFile)*/
	else
	{
		plotfile = STDOUT;
	}
	
	
	if (innerPoint[0] == '\0')
	{
		if (keyValues->dumb)
		{
			innerPoint[0] = INNERCHAR;
		}
		else if (nbox < STARTSMALL)
		{
			innerPoint[0] = INNERPOINT;
		}
		else if (nbox < STARTSMALLER)
		{
			innerPoint[0] = SMALLINNERPOINT;
		}
		else
		{
			innerPoint[0] = SMALLERINNERPOINT;
		}
		innerPoint[1] = '\0';
	} /*if (innerPoint[0] == '\0')*/

	if (outerPoint[0] == '\0')
	{
		if (keyValues->dumb)
		{
			outerPoint[0] = OUTERCHAR;
		}
		else if (nbox < STARTSMALL)
		{
			outerPoint[0] = OUTERPOINT;
		}
		else if (nbox < STARTSMALLER)
		{
			outerPoint[0] = SMALLOUTERPOINT;
		}
		else
		{
			outerPoint[0] = SMALLEROUTERPOINT;
		}
		outerPoint[1] = '\0';
	} /*if (outerPoint[0] == '\0')*/
	
	colors[0] = pointColor;
	colors[1] = lineColor;
	graphsymbol = buildBoxplot(list, nbox, symhx, symhBoxSize,
							   innerPoint, outerPoint,
							   colors, vertical, excludeM, boxType, keyValues);
	if (graphsymbol == SymbolError)
	{
		goto errorExit;
	}
	graph = GRAPH(graphsymbol);
	if (keyValues->keep)
	{
		Remove(LastGraphName);
		Cutsymbol(graphsymbol);
		setNAME(graphsymbol, LastGraphName);
		Addsymbol(graphsymbol);
	} /*if (keyValues->keep)*/	

	if (keyValues->show &&
	   !do_whole_graph(graph, plotfile, termType, windno, keyValues))
	{
		if (keyValues->keep)
		{
			graph = (whole_graph **) 0;
		}
		goto errorExit;
	}
	if (plotfile != STDOUT)
	{
		/*
          Moved setting of Macintosh creator and type to openPlotFile()
		  in plotutil.c
		*/
		fclose(plotfile);
	} /*if (plotfile != STDOUT)*/
	
	if (!keyValues->keep)
	{
		Removesymbol(graphsymbol);
		graphsymbol = SymbolZero;
	}
	else if (ThisWindow >= 0 && ThisWindow < getNgraphs())
	{
		Delete(getGraphWindow(ThisWindow));
		attachGraphWindow(graphsymbol, ThisWindow);
	}
	if (graphsymbol != SymbolZero &&
		keyValues->notes != SymbolZero)
	{
		if (setNotes(graphsymbol, STRING(keyValues->notes)))
		{
			setSTRING(keyValues->notes, (char **) 0);
		}
		else if (!keyValues->silent)
		{
			sprintf(OUTSTR, "WARNING: notes not set in %s()", FUNCNAME);
			putErrorOUTSTR();
		}
	} /*if (keyValues->notes != SymbolZero)*/
#ifdef SEGMENTED
	UNLOADSEG(do_whole_graph);
#endif /*SEGMENTED*/
	return (NULLSYMBOL);

  errorExit:
	putErrorOUTSTR();
	
	if (plotfile != STDOUT)
	{
		fclose(plotfile);
	}
	Removesymbol(graphsymbol);
	
#ifdef SEGMENTED
	UNLOADSEG(do_whole_graph);
#endif /*SEGMENTED*/
	return (0);
	
} /*boxplot()*/


