/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2003 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Columnop
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/


/*
  Operations sum, prod, max, min
  modified 12/7/91 so that op(a,b,c,...) is equivalent to
  op(vector(a,b,c,...)), where a, b, c, ... are all scalars.
  931121 op(a,b,c,...) now allows vectors.

  Modified February, 1993 to operate recursively on structures.
  Modified March, 1993 to accept logical arguments
  960405 Minor modification in coding of op codes and initial values
  990222 Fixed initialization bug affecting results when a column is all MISSING
         Removed unnecessary 'extern' statements
  020420 New keyword dimension:j allows performing operation on dimension j
         sum(a,dimension:1) is the same as sum(a)
		 if ndim(a) = 2, sum(a,dimension:2) is sum(a')'
         Removed some fossil code
  020427 Value for keyword 'dimension' now can be a vector of distinct
         positive integers <= ndims(arg) and sums will be over all
         dimensions
  020501 Added enum argOffsets
  021105 Fixed bug so that max(T) is not -HUGEDBL and min(T) is not HUGEDBL
*/

#include "globals.h"

static char       *OpNames[] =
{
	"sum",
	"prod",
	"max",
	"min",
	(char *) 0
};

enum columnOpCodes
{
	ISUM = 1,
	IPROD,
	IMAX,
	IMIN
};

enum argOffsets
{
	squeezeOffset = 0,
	marginalOffset,
	undefOffset,
	ndimsOffset,
	nOffsets
};

static Symbolhandle   doColOps(Symbolhandle arg,double *params,
							unsigned long control, unsigned long *status)
{
	Symbolhandle    result = SymbolZero;
	long            nvalues, ndims, resultNdims, type = TYPE(arg);
	long            resultSize;
	int             nTargetDims = (int) params[ndimsOffset];
	int             targetDims[MAXDIMS];
	long            resultDimVals[MAXDIMS];
	long            startDigits[MAXDIMS], offsetDigits[MAXDIMS];
	long            startFactors[MAXDIMS], offsetFactors[MAXDIMS];
	long            startDims[MAXDIMS], offsetDims[MAXDIMS];
	long            factor;
	double         *x, *resultVals;
	double          undefinedValue = params[undefOffset];
	double          startup;
	long            i, j, k;
	long            I, J;
	int             nStartDigits, nOffsetDigits;
	int             op = control & OPMASK;
	int             extremes = (op == IMAX || op == IMIN);
	int             squeezeOut = (int) params[squeezeOffset];
	int             marginals = (int) params[marginalOffset];
	WHERE("doColOps");
	
	if (type == NULLSYM || isNull(arg))
	{
		result = Install(NULLSCRATCH, NULLSYM);
		if (result == SymbolError)
		{
			goto errorExit;
		}
		return (result);
	} /*if (type == NULLSYM || isNull(arg))*/
	
	ndims = NDIMS(arg);

	if (!marginals)
	{
		/* retrieve dimensions to be operated on */
		for (i = 0; i < MAXDIMS; i++)
		{
			targetDims[i] = (i < nTargetDims) ? (int) params[i+nOffsets] : 0;
		}
	} /*if (!marginals)*/
	else
	{
		/*
		  retrieve marginal dimensions and translate to dimensions
		  to be operated on
		*/
		
		k = j = 0;
		for (i = 1; i <= ndims; i++)
		{
			if ((int) params[j + nOffsets] != i)
			{
				targetDims[k++] = i;
			} 
			else
			{
				j++;
			}
		} /*for (i = 1; i <= ndims; i++)*/
		nTargetDims = k;
	} /*if (!marginals){}else{}*/
	//	PRINT("targetDims[0-3] = %d %d %d %d",targetDims[0],targetDims[1],targetDims[2],targetDims[3]);

	nStartDigits = nOffsetDigits = 0;
	factor = 1;
	/*
	  nvalues is the number of values operated on
	  resultSize is number of values in the result
	*/
	nvalues = resultSize = 1;
	resultNdims = 0;
	for (i = 1; i <= ndims; i++)
	{
		long    dimval = DIMVAL(arg, i);
		
		if (nOffsetDigits < nTargetDims && targetDims[nOffsetDigits] == i)
		{
			/* operating on this dimension */
			nvalues *= dimval;
			offsetDims[nOffsetDigits] = dimval;
			offsetFactors[nOffsetDigits++] = factor;
			if (!squeezeOut)
			{
				resultDimVals[resultNdims++] = 1;
			}
		}
		else
		{
			/* not operating on this dimension */
			resultSize *= dimval;
			startDims[nStartDigits] = dimval;
			startDigits[nStartDigits] = 0;
			startFactors[nStartDigits++] = factor;
			resultDimVals[resultNdims++] = dimval;
		}
		factor *= dimval;
	} /*for (i = 0; i < ndims; i++)*/

	if (nStartDigits == 0)
	{
		/* operating on all dimensions */
		startDims[0] = 1;
		startFactors[0] = 1;
		startDigits[0] = 0;
		nStartDigits = 1;
		if (squeezeOut)
		{
			resultDimVals[0] = 1;
			resultNdims = 1;
		}
	} /*if (nStartDigits == 0)*/

	result = RInstall(SCRATCH, resultSize);

	if (result == SymbolError)
	{
		goto errorExit;
	}

	/*
	   951010 Always keep the same number of dimensions
	   020427 except when squeezing out dimensions operated over
	*/

	Setdims(result, resultNdims, resultDimVals);
	
	if (HASLABELS(arg) && !isScalar(result))
	{
	/*
	   Make label for targeted dimensions targetDims[0] "@" and copy other dim 
	   dimension labels
	*/
		long            labLengths[MAXDIMS];
		char           *labels[MAXDIMS];

		getAllLabels(arg, labels, labLengths, (long *) 0);

		if (squeezeOut)
		{
			j = k = 0;
			for (i = 0; i < ndims; i++)
			{
				if (i != targetDims[j] - 1)
				{
					labLengths[k++] = labLengths[i];
				}
				else
				{
					j++;
				}
			} /*for (i = 0; i < ndims; i++)*/
		} /*if (squeezeOut)*/
		else
		{
			for (i = 0; i < nTargetDims; i++)
			{
				labLengths[targetDims[i] - 1] = 2;
			}
		}

		TMPHANDLE = createLabels(resultNdims, labLengths);

		if (TMPHANDLE != (char **) 0 && setLabels(result, TMPHANDLE))
		{
			char            numericLabel[2];

			numericLabel[0] = NUMERICLABEL;
			numericLabel[1] = '\0';

			getAllLabels(arg, labels, (long *) 0, (long *) 0);

			if (!squeezeOut)
			{
				for (i = 0; i < nTargetDims; i++)
				{
					labels[targetDims[i] - 1] = numericLabel;
				}
			}
			
			j = k = 0;
			for (i = 0; i < ndims; i++)
			{
				if (!squeezeOut || i+1 != targetDims[j])
				{
					appendLabels(result, labels[i], k++, dontExpand);
				}
				else
				{
					j++;
				}
			}
		} /*if (TMPHANDLE != (char **) 0 && setLabels(result, TMPHANDLE))*/
		else
		{
			mydisphandle(TMPHANDLE);
		}
	} /*if (HASLABELS(arg))*/

	x = DATAPTR(arg);
	resultVals = DATAPTR(result);

	if (!extremes)
	{
		startup = (op == ISUM) ? 0.0 : 1.0; /*sum() or prod()*/
	}
	else if (type == LOGIC)
	{
		startup = (op == IMAX) ? 0.0 : 1.0; /* max() or min()*/
	}
	else
	{
		startup = (op == IMAX) ? -HUGEDBL : HUGEDBL; /* max() or min()*/
	}

	for (I = 0; I < resultSize; I++)
	{
		double       s = startup, y;
		double      *xStart;
		int          foundOK = 0;
		
		for (j = 0; j < nOffsetDigits; j++)
		{
			offsetDigits[j] = 0;
		}
		xStart = x + readOdometer(startDigits, startFactors, nStartDigits, 0);
		stepOdometer(startDigits, startDims, nStartDigits, 0, 0);

		for (J = 0; J < nvalues; J++)
		{
			long             offset;
			
			offset = readOdometer(offsetDigits, offsetFactors,
								  nOffsetDigits, 0);
			stepOdometer(offsetDigits, offsetDims, nOffsetDigits, 0, 0);
			
			y = xStart[offset];
			if (!isMissing(y))
			{
				foundOK = 1;
				if (type == REAL)
				{
					switch (op)
					{
					  case ISUM:
						s += y;
						break;

					  case IPROD:
						s *= y;
						break;

					  case IMIN:
						s = (y < s) ? y : s;
						break;

					  case IMAX:
						s = (y > s) ? y : s;
						break;
					} /* switch(op) */
				} /*if (type == REAL)*/
				else
				{/* LOGIC */
					switch (op)
					{
					  case ISUM:
						if (y != 0.0)
						{
							s++;
						}
						break;

					  case IPROD:
					  case IMIN:
						s = (y != 0.0) ? s : 0.0;
						break;

					  case IMAX:
						s = (y != 0.0) ? 1.0 : s;
						break;

					} /* switch(op) */
				} /*if (type == REAL){}else{}*/
			} /*if (!isMissing(y))*/
			else
			{
				*status |= FOUNDMISSING;
			}
		} /*for (J = 0; J < nvalues; J++);*/

		if (!foundOK)
		{
			if (undefinedValue != UNSETREAL)
			{
				resultVals[I] = undefinedValue;
			}
			else if (extremes)
			{
				setMissing(resultVals[I]);
			}
			else
			{
				resultVals[I] = s;
			}
		} /*if (!foundOK)*/
#ifdef HASINFINITY
		else if (isInfinite(s))
		{
			*status |= FOUNDOVERFLOW;
			setMissing(resultVals[I]);
		}
#endif /*HASINFINITY*/
		else
		{
			resultVals[I] = s;
		}
	} /*for (I = 0; I < resultSize; I++)*/

	return (result);
	
  errorExit:
	Removesymbol(result);
	return (SymbolError);
} /*doColOps()*/

				
/*
  Operations sum(), prod(), max(), min()
  930202 These now all operate on structures
  931129 Bug fix: If structure component is empty, so is result
  951010 Changed behavior; results always have same number of dimensions
         as argument, with the first replaced by 1.  Previously, an array
         with more than 2 dimensions resulted in an array with one less
		 dimension
  960311 Add code to doColOps() to copy labels.  If result is not a scalar and
         the argument has labels, so does the result, with the label for the
         first dimension null and the remaining labels copied from arg
  980220 NULL arguments are now acceptable.  If they are all NULL the result
         is NULL
  980308 Keyword phrase 'silent:T' can be the final argument to suppress
         warning messages

  020420 Keyword phrase dimension:j to do operation over dimension j,
         default is dimension:1
  020427 Allow dimensions:J, where J is a vector of distinct positive
         integers; default is still J = 1.
         New keyword phrase squeeze:T, squeezes out from the result
		 dimensions operated over.
		 To get marginal totals for dimensions K = vector(k1,k2,...),
		 sum(table,dimensions:run(ndims(table))[-K])
  020429 Recognize 'margins:K' to specify marginal totals; default for
         'squeeze' is T with 'margins'
  020430 Recognize 'undefval:U', scalar U used in the result where all the
         elements operated on are MISSING
*/

Symbolhandle    columnops(Symbolhandle list)
{
	Symbolhandle    symh, arg1 = SymbolZero, result = SymbolZero;
	long            i, j;
	unsigned long   status = 0, control = 0;
	int             op, foundNonNull = 0;
	long            nargs = NARGS(list), type = NULLSYM, typei, length;
	double         *source, *dest;
	double          params[MAXDIMS + nOffsets];
	double          undefinedValue = UNSETREAL;
	char           *keyword;
	char           *dimKeyName = "dimensions", *margKeyName = "margins";
	long            dimensions[MAXDIMS];
	int             nDimensions = 1, maxDimension = 1;
	int             silent = 0, squeezeOut = -1, marginals = -1;
	WHERE("columnops");

	*OUTSTR = '\0';
	
	dimensions[0] = -1;
	for (i = 1; i < MAXDIMS; i++)
	{
		dimensions[i] = 0;
	}
	
	if (nargs == 1 && COMPVALUE(list, 0) == EmptyArgument)
	{
		nargs = 0;
	}
	op = matchKey(FUNCNAME,OpNames, (long *) 0);

	while (nargs > 0 && (keyword = isKeyword(COMPVALUE(list, nargs - 1))))
	{
		Symbolhandle        symhKey = COMPVALUE(list, nargs - 1);

		if (strcmp(keyword, "silent") == 0 ||
			strcmp(keyword, "squeeze") == 0)
		{
			if (!isTorF(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			if (keyword[1] == 'i')
			{
				silent = DATAVALUE(symhKey, 0) != 0.0;
			}
			else
			{
				squeezeOut = DATAVALUE(symhKey, 0) != 0.0;
			}
		}
		else if (strncmp(keyword, dimKeyName, 3) == 0 ||
				 strncmp(keyword, margKeyName, 4) == 0)
		{
			int              ok = isVector(symhKey) && 
			  TYPE(symhKey) == REAL;
			
			if (ok)
			{
				nDimensions = DIMVAL(symhKey, 1);
				dimensions[0] = 0;
				
				for (i = 0; i < nDimensions; i++)
				{
					double        dDimi = DATAVALUE(symhKey, i);
					int           iDimi;

					if (isMissing(dDimi) || dDimi < 1.0 || 
						dDimi != floor(dDimi) || dDimi > MAXDIMS)
					{
						ok = 0;
						break;
					}
					iDimi = (long) dDimi;
					if (dimensions[iDimi - 1] > 0)
					{
						sprintf(OUTSTR,
								"ERROR: duplicate item in value of '%s'",
								keyword);
						goto errorExit;
					}
					dimensions[iDimi - 1] = 1;
					if (iDimi > maxDimension)
					{
						maxDimension = iDimi;
					}
				} /*for (i = 0; i < nDimensions; i++)*/
				if (marginals < 0)
				{
					marginals = (keyword[0] == 'm');
				}
				else if (marginals == 0 && keyword[0] == 'm' ||
						 marginals > 0 && keyword[0] == 'd')
				{
					sprintf(OUTSTR,
							"ERROR: duplicate use of keyword '%s' in %s()",
							keyword, FUNCNAME);
				}
				else
				{
					sprintf(OUTSTR,
							"ERROR: illegal to use both '%s' and '%s' in %s()",
							dimKeyName, margKeyName, FUNCNAME);
				}
				if (*OUTSTR)
				{
					goto errorExit;
				}
			} /*if (ok)*/
			if (!ok)
			{
				sprintf(OUTSTR,
						"ERROR: Keyword %s not vector of integers between 1 and %d",
						keyword, MAXDIMS);
				goto errorExit;
			}
		}
		else if (strncmp(keyword, "undefval",5) == 0)
		{
			if (!isScalar(symhKey) || TYPE(symhKey) != REAL)
			{
				sprintf(OUTSTR,
						"ERROR: value for '%s' must be REAL scalar or MISSING",
						keyword);
				goto errorExit;
			}
			undefinedValue = DATAVALUE(symhKey, 0);
		}
		else
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		nargs--;
	} /* while (nargs > 1 && (keyword = isKeyword(COMPVALUE(list, nargs - 1))))*/

	if (nargs == 0)
	{
		sprintf(OUTSTR,
				"ERROR: %s() must have at least 1 non-keyword argument",
				FUNCNAME);
		goto errorExit;
	} /*if (nargs == 0)*/

	arg1 = COMPVALUE(list, 0);
	if (nargs == 1)
	{
		if (!argOK(arg1, NULLSYM, 0))
		{
			goto errorExit;
		}
		type = TYPE(arg1);

		if (type != STRUC && type != REAL && type != LOGIC && type != NULLSYM)
		{
			badType(FUNCNAME,-type,0);
			goto errorExit;
		}
		foundNonNull = (type != NULLSYM);
		dimensions[0] = labs(dimensions[0]);

		if (type == STRUC)
		{
			type = getStrucTypes(arg1);
			if (!(type & (REALTYPE1|LOGICTYPE1)) ||
			   (type & ~(REALTYPE1|LOGICTYPE1)))
			{
				sprintf(OUTSTR,
						"ERROR: non-NULL components of argument to %s() must all be REAL or LOGICAL",
						FUNCNAME);
			}
			else if (maxDimension > 1 && maxDimension > getMinNdims(arg1))
			{
				sprintf(OUTSTR,
						"ERROR: argument to %s() has component with ndims < %d",
						FUNCNAME, maxDimension);
			}
			if (*OUTSTR)
			{
				goto errorExit;
			}
		} /*if (type == STRUC)*/
		else if (maxDimension > NDIMS(arg1))
		{
			sprintf(OUTSTR,
					"ERROR: argument to %s() has fewer than %d dimensions",
					FUNCNAME, maxDimension);
			goto errorExit;
		}

		for (i = j = 0; i < MAXDIMS; i++)
		{
			if (dimensions[i])
			{
				dimensions[j++] = i + 1;
			}
		}
	} /*if (nargs == 1)*/
	else
	{ /* allow list of vectors as equivalent to vector */
		if (dimensions[0] > 0)
		{
			sprintf(OUTSTR,
					"ERROR: more than 1 non-keyword argument to %s() with '%s:J'",
					FUNCNAME, (marginals) ? margKeyName : dimKeyName);
			goto errorExit;
		}
		dimensions[0] = 1;
		length = 0;
		for (i = 0; i < nargs;i++)
		{
			symh = COMPVALUE(list,i);
			if (!argOK(symh, NULLSYM, i+1))
			{
				goto errorExit;
			}
			typei = TYPE(symh);

			if (typei != REAL && typei != LOGIC && typei != NULLSYM)
			{
				badType(FUNCNAME,-typei,0);
				goto errorExit;
			}

			if (typei != NULLSYM)
			{
				if (type == NULLSYM)
				{
					type = typei;
				}
				else if (typei != type)
				{
					sprintf(OUTSTR,
							"ERROR: all non-NULL arguments to %s() must have same type",
							FUNCNAME);
					goto errorExit;
				}
			
				if (!isVector(symh))
				{
					sprintf(OUTSTR,
							"ERROR: non-NULL multiple arguments to %s() must all be vectors",
							FUNCNAME);
					goto errorExit;
				}
				length += symbolSize(symh);
			}
		} /*for (i = 0; i < nargs;i++)*/

		if (type == NULLSYM)
		{
			return (NULLSYMBOL);
		}

		symh = COMPVALUE(list,0);

		/* put new arg1 in the list so that it will eventually be deleted */

		arg1 = COMPVALUE(list,0) = RInstall(SCRATCH,length);
		if (arg1 == SymbolError)
		{
			goto errorExit;
		}
		setTYPE(arg1,type);
		dest = DATAPTR(arg1);
		for (i = 0;i < nargs;i++)
		{
			if (i > 0)
			{
				symh = COMPVALUE(list,i);
			}
			if (TYPE(symh) != NULLSYM)
			{
				source = DATAPTR(symh);
				length = symbolSize(symh);
				while (length-- > 0)
				{
					*dest++ = *source++;
				}
			}
			
			if (i == 0 && isscratch(NAME(symh)))
			{
				Removesymbol(symh);
			}
		} /*for (i = 0;i < nargs;i++)*/
	} /*if (nargs == 1){}else{}*/

	if (marginals < 0)
	{
		marginals = 0;
	}
	
	if (squeezeOut < 0)
	{
		squeezeOut = marginals;
	}
	
	params[squeezeOffset] = (double) squeezeOut;
	params[marginalOffset] = (double) marginals;
	params[undefOffset] = undefinedValue;
	params[ndimsOffset] = (double) nDimensions;

	for (i = 0; i < nDimensions; i++)
	{
		params[i + nOffsets] = (double) dimensions[i];
	}
	
	control = (isStruc(arg1)) ? LEFTRECUR : 0;
	
	result = doRecur1(doColOps,arg1, params, control | op, &status);

	/* Note: result is already in symbol table */
	if (result == SymbolError)
	{
		goto errorExit;
	}

	if (!silent && status & FOUNDMISSING)
	{
		sprintf(OUTSTR,"WARNING: MISSING values found by %s()", FUNCNAME);
		putErrorOUTSTR();
	} /*if (status & FOUNDMISSING)*/

#ifdef HASINFINITY
	if (!silent && status & FOUNDOVERFLOW)
	{
		sprintf(OUTSTR,"WARNING: too large result(s) of %s() set to MISSING",
				FUNCNAME);
		putErrorOUTSTR();
	} /*if (status & FOUNDOVERFLOW)*/
#endif /*HASINFINITY*/

	return (result);

  errorExit:
	putErrorOUTSTR();

	return (0);
} /*columnops()*/
