/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.12 or later
*(C)*
*(C)* Copyright (c) 2001 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Print
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "globals.h"
#include "mvIface.h"
#include "mvIo.h" /* includes codes related to reading Mouse */

/*
  Functions for access system related aspects of MacAnova
  As of 971210 only gethistory() is implemented.  This would be a place
  to put functions like setmenu(), setmenuitem(), closewindow(), newwindow()
  etc., if they existed

  990226 replaced putOUTSTR() by putErrorOUTSTR()

  990603 added readMouse() to implement Mouse() for interactive input
         from graphs using a mouse or cross hairs
  
  000704 added keyword 'cancelok' to Mouse()
*/

/*
  usage:
  gethistory()    get entire history list as character VECTOR
  gethistory(n)   get most recent n commands; give warning if fewer than n
*/

#ifdef SAVEHISTORY
Symbolhandle getHistory(Symbolhandle list)
{
	Symbolhandle   result = (Symbolhandle) 0;
	Symbolhandle   arg;
	char         **commands = (char **) 0;
	long           nargs = NARGS(list);
	long           nlines, mlines;
	int            get = FUNCNAME[0] == 'g';

	if (!get)
	{ /*sethistory()*/
		if (nargs > 1 || (arg = COMPVALUE(list, 0)) == (Symbolhandle) 0)
		{
			badNargs(FUNCNAME, 1);
			goto errorExit;
		}
		if (!isVector(arg) || TYPE(arg) != CHAR)
		{
			sprintf(OUTSTR,
					"ERROR: %s() argument is not CHARACTER vector", FUNCNAME);
			goto errorExit;
		}
		nlines = symbolSize(arg);
		if (nlines > HISTORY)
		{
			sprintf(OUTSTR,
					"ERROR: attempt to restore a history of more than %ld commands",
					HISTORY);
			goto errorExit;
		} /*if (symbolSize(arg) > HISTORY)*/

		setSomeHistory(nlines, STRING(arg));

		result = NULLSYMBOL;
	} /*if (!get)*/
	else
	{ /*gethistory()*/
		if (nargs > 1)
		{
			badNargs(FUNCNAME, -1);
			goto errorExit;
		}
		arg = COMPVALUE(list, 0);
		if (arg != (Symbolhandle) 0)
		{
			if (!isInteger(arg, POSITIVEVALUE))
			{
				notPositiveInteger("argument 1");
				goto errorExit;
			}
			nlines = (long) DATAVALUE(arg, 0);
		}
		else
		{
			nlines = 0;
		}

		result = Install(SCRATCH, CHAR);
		if (result == 0)
		{
			goto errorExit;
		}
		
		mlines = getSomeHistory(nlines, &commands);

		setSTRING(result, commands);
		if (commands == (char **) 0)
		{
			Removesymbol(result);
			goto errorExit;
		}
		
		setNDIMS(result, 1);
		setDIM(result, 1, mlines);

		if (nlines > 0)
		{
			if (mlines == 1 && (*commands)[0] == '\0')
			{
				mlines = 0;
			}

			if (mlines < nlines)
			{
				sprintf(OUTSTR,
						"WARNING: %s() retrieved only %ld past command lines",
						FUNCNAME, mlines);
				putErrorOUTSTR();
			}
		} /*if (result != (Symbolhandle) 0 && nlines > 0)*/
	} /*if (!get){}else{}*/

	return (result);

  errorExit:
	putErrorOUTSTR();
	return (0);
} /*Symbolhandle getHistory()*/
#else /*SAVEHISTORY*/
Symbolhandle getHistory(Symbolhandle list)
{
	notImplemented(FUNCNAME);
	return (0);
} /*getHistory()*/
#endif /*SAVEHISTORY*/

/*
  Constants related to reading Mouse
  Other constants are in mvIo.h
*/
enum mouseKeys
{
	keyPoint     = 0,
	keyLine,
	keyBox,
	keyXyonly,
	keyCancelok,
	keyN,
	nMouseKeys
};
/*
  Function to read the mouse

  Usage:
	reply <- Mouse([xyonly:T] [,cancelok:T])
	      or reply <- Mouse(point:T [,xyonly:T] [,cancelok:T])
	reply <- Mouse(line:T [,xyonly:T] [,cancelok:T])
	reply <- Mouse(box:T [,xyonly:T] [,cancelok:T])

	In each case reply has the form
	  structure(x:xvals,y:yvals ,window:windno, add:T [,lines:T])
	   (without 'xyonly:T') or
	  structure(x:xvals,y:yvals) (with 'xyonly:T')

	With cancelok:T, cancelling the operation is not an error; Mouse()
	returns NULL

	For point:T, xvals and yvals are scalars and 'lines:T' is omitted

	For line:T, xvals and yvals are vectors of length 2 with (x[i],y[i]),
	i = 1,2, being the coordinates of the line read

	For box:T, xvals and yvals are vectors of length 5 containing the
	x and y coordinates tracing around the box back to the starting point.

	readFromMouse() is platform specific.  For Macintosh it is in macIo.c and
	for Windows/Motif it is in wxIo.cc

	990603 Installed
	990606 Standard form of output is now
	       structure(x:xvals,y:yvals,window:n,add:T[,lines:T])
	       If keyword phrase xyonly:T is an argument, it returns
	       structure(x:xvals,y:yvals)
	990609 Usage allows any keywords following point:T, line:T and box:T
	       these are copied to the output.  Thus, for example
	       Mouse(line:T,thickness:3) is permitted.  If xyonly:T is an argument
		   no other keywords except 'point', 'line' and 'box' can appear
           Missing arguments become NULL when copied
	990611 If keyword 'add' is an argument, 'add:T' will not automatically
	       be a component of the output.  Similarly, if 'lines' is an argument,
	       'lines:T' will not automatically be a component of the output
           with 'box:T' and 'line:T'
    990620 Changed keywords 'line', 'point', 'box' to 'getlines', 'getpoints',
           'getbox'
    000704 Added keyword phrase 'cancelok:T' 
*/
#if !defined(SKEL_NONE) || defined(TEK)
#define READMOUSE
#endif

struct listOfKeys
{
	char           *keyName;
	int             minKeyLength;
};

static struct listOfKeys MouseKeys[nMouseKeys] =
{
	{"getpoints", 8},
	{"getlines", 7},
	{"getbox", 0},
	{"xyonly", 0},
	{"cancelok",0},
	{"n", 0}
};


Symbolhandle readMouse(Symbolhandle list)
{
	Symbolhandle        result = (Symbolhandle) 0;
	Symbolhandle        symhKey, symhx, symhy;
	Symbolhandle        symhAdd, symhLines, symhWindow;
	double              x[MaxMousePoints+1], y[MaxMousePoints+1];
	long                nargs = NARGS(list);
	long                length, n;
	long                ncomps, mcomps, margs;
	long                npoints = -11;
	int                 op = -1;
	int                 windno, xyonly = 0, cancelok = 0;
	int                 needWindow = -1, needLines = -1, needAdd = -1;
	int                 iarg;
	char               *keyword;
	char               *windowKeyName = "window", *addKeyName = "add";
	WHERE("readMouse");
	
	if (nargs == 1 && COMPVALUE(list, 0) == (Symbolhandle) 0)
	{
		nargs = 0;
	}
	margs = (nargs > 1) ? 4 : nargs;
	
	for (iarg = 0; iarg < margs; iarg++)
	{
		int          iKey;

		symhKey = COMPVALUE(list, iarg);
		if (symhKey == (Symbolhandle) 0)
		{
			/*
			  An empty argument is not an error;
			  it ends keywords intended for Mouse()
			*/
			margs = iarg;
			break;
		}
		if (!argOK(symhKey, 0, (nargs > 1) ? iarg + 1 : 0))
		{
			goto errorExit;
		}

		if (!(keyword = isKeyword(symhKey)))
		{
			sprintf(OUTSTR,
					"ERROR: all non-missing arguments to %s() must be keyword phrases",
					FUNCNAME);
			goto errorExit;
		} /*if (!(keyword = isKeyword(symhKey)))*/

		/*
		  Is it a Mouse() keyword?
		*/
		for (iKey = 0; iKey < nMouseKeys; iKey++)
		{
			if (MouseKeys[iKey].minKeyLength == 0 && 
				strcmp(keyword, MouseKeys[iKey].keyName) == 0 ||
				MouseKeys[iKey].minKeyLength != 0 && 
				strncmp(keyword, MouseKeys[iKey].keyName,
						MouseKeys[iKey].minKeyLength) == 0)
			{
				break;
			}
		} /*for (iKey = 0; iKey < nMouseKeys; iKey++)*/

		if (iKey >= nMouseKeys)
		{
			margs = iarg;
			if (xyonly)
			{
				sprintf(OUTSTR,
						"ERROR: keyword '%s' cannot be used with '%s':T",
						keyword, MouseKeys[keyXyonly].keyName);
				goto errorExit;
			}
			break;
		} /*if (iKey > lastOp)*/

		if (iKey == keyN)
		{
			if (!isInteger(symhKey, POSITIVEVALUE) ||
				DATAVALUE(symhKey, 0) > MaxMousePoints)
			{
				sprintf(OUTSTR,
						"ERROR: value for '%s' on %s() not positive integer <= %d",
						MouseKeys[keyN].keyName, FUNCNAME, MaxMousePoints);
				goto errorExit;
			}
			npoints = (long) DATAVALUE(symhKey, 0);
		} /*if (iKey == keyN)*/
		else if (!isTorF(symhKey))
		{
			notTorF(keyword);
			goto errorExit;
		}
		else if (iKey == keyXyonly)
		{
			xyonly = (DATAVALUE(symhKey, 0) != 0.0);
			if (xyonly)
			{
				needLines = needAdd = needWindow = 0;
			}
		}
		else if (iKey == keyCancelok)
		{
			cancelok = (DATAVALUE(symhKey, 0) != 0.0);
		}
		else if (DATAVALUE(symhKey, 0) == 0.0)
		{
			sprintf(OUTSTR,
					"ERROR: '%s:F' is illegal in %s()", keyword, FUNCNAME);
			goto errorExit;
		}
		else
		{
			op = getPoint + iKey;
		}
	} /*for (iarg = 0; iarg < margs; iarg++)*/

	if (op < 0)
	{
		op = getPoint;
	}

	for (iarg = margs; iarg < nargs; iarg++)
	{
		symhKey = COMPVALUE(list, iarg);
		if (symhKey != (Symbolhandle) 0)
		{
			keyword = isKeyword(symhKey);
			if (keyword == (char *) 0)
			{
				sprintf(OUTSTR,
						"ERROR: all non-missing arguments to %s() must be keyword phrases",
						FUNCNAME);
				goto errorExit;
			}
			if (strcmp(keyword, addKeyName) == 0)
			{
				needAdd = 0;
			}
			else if (strcmp(keyword, "lines") == 0)
			{
				needLines = 0;
			}
			else if (strncmp(keyword, windowKeyName, 4) == 0)
			{
				needWindow = 0;
			}
		} /*if (symhKey != (Symbolhandle) 0)*/
	} /*for (iarg = margs; iarg < nargs; iarg++)*/

	if (op == getBox && npoints >= 0)
	{
		sprintf(OUTSTR,
				"ERROR: %s:N illegal with %s:T on %s()",
				MouseKeys[keyN].keyName, MouseKeys[keyBox].keyName, FUNCNAME);
		goto errorExit;
	}
	npoints = (npoints < 0) ? 1 : npoints;

#ifdef READMOUSE
	if (needLines < 0)
	{
		needLines = (op == getPoint) ? 0 : 1;
	}
	needAdd = (needAdd < 0) ? 1 : needAdd;
	needWindow = (needWindow < 0) ? 1 : needWindow;

	if (op == getPoint)
	{
		length = npoints;
	}
	else if (op == getLine)
	{
		length = npoints + 1;
	}
	else if (op == getBox)
	{
		length = 5;
	}
	
	ncomps = 2 + needAdd + needWindow + needLines + nargs - margs;

	result = StrucInstall(SCRATCH, ncomps);
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}

	mcomps = 0;
	COMPVALUE(result, mcomps++) = symhx = Makereal(length);
	if (symhx == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	setNAME(symhx, "x");

	COMPVALUE(result, mcomps++) = symhy = Makereal(length);
	if (symhy == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	setNAME(symhy, "y");
	
	if (needWindow)
	{
		COMPVALUE(result, mcomps++) = symhWindow = Makereal(1);
		if (symhWindow == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNAME(symhWindow, windowKeyName);
	} /*if (needWindow)*/
	if (needAdd)
	{
		COMPVALUE(result, mcomps++) = symhAdd = Makereal(1);
		if (symhAdd == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNAME(symhAdd, addKeyName);
		setTYPE(symhAdd, LOGIC);
		DATAVALUE(symhAdd, 0) = 1.0;
	} /*if (needAdd)*/
	if (needLines)
	{
		COMPVALUE(result, mcomps++) = symhLines = Makereal(1);
		if (symhLines == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNAME(symhLines, "lines");
		setTYPE(symhLines, LOGIC);
		DATAVALUE(symhLines, 0) = 1.0;
	} /*if (needLines)*/

	/* Copy extra arguments (all keyword phrases), if any, to result */
	
	for (iarg = margs; iarg < nargs; iarg++)
	{
		symhKey = COMPVALUE(list, iarg);
		if (symhKey == (Symbolhandle) 0)
		{
			symhKey = Makesymbol(NULLSYM);
			if (symhKey == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			keyword = NULLNAME;
		} /*if (symhKey == (Symbolhandle) 0)*/
		else
		{
			keyword = isKeyword(symhKey);
			COMPVALUE(list, iarg) = (Symbolhandle) 0;
			Cutsymbol(symhKey);
		}
		COMPVALUE(result, mcomps++) = symhKey;
		setNAME(symhKey, keyword);
	} /*for (iarg = margs; iarg < nargs; iarg++)*/

	n = npoints;
	windno = readFromMouse(x, y, op, &n);

	if (windno < 0)
	{
		if (windno == MouseAbort && cancelok)
		{
			Removesymbol(result);
			result = NULLSYMBOL;
		}
		else
		{
			if (windno == MouseNoGraphs)
			{
				sprintf(OUTSTR,
						"ERROR: %s() can't work when no graph windows are open",
						FUNCNAME);
			}
			else
			{
				sprintf(OUTSTR,"ERROR: %s() %s", FUNCNAME,
					(windno == MouseInterrupt) ? "interrupted" : "cancelled");
			}
			goto errorExit;
		}
	} /*if (windno < 0)*/
	else
	{
		if (n != length)
		{
			sprintf(OUTSTR,
					"ERROR: wrong number (%ld) of points read by %s()",
					n, FUNCNAME);
			goto errorExit;
		}
	
		doubleCopy(x, DATAPTR(symhx), length);
		doubleCopy(y, DATAPTR(symhy), length);
		if (needWindow)
		{
			DATAVALUE(symhWindow, 0) = (double) windno + 1.0;
		}
	}

	return (result);

#else /*READMOUSE*/	
	notImplemented(FUNCNAME);
	/* Fall through */
#endif /*READMOUSE*/

errorExit:
	putErrorOUTSTR();
	Removesymbol(result);

	return (0);

} /*readMouse()*/

