/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 5.03 or later
*(C)*
*(C)* Copyright (c) 2006 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#define Place(I) ALERT("Place %g",(double) I,0,0,0)

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Columnop
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "globals.h"

enum opCodes
{
	OPN     = 0x0001,
	OPMIN   = 0x0002,
	OPQ1    = 0x0004,
	OPMED   = 0x0008,
	OPQ3    = 0x0010,
	OPMAX   = 0x0020,
	OPMEAN  = 0x0040,
	OPVAR   = 0x0080,
	OPSTDEV = 0x0100,
	OPM2    = 0x0200,
	OPM3    = 0x0400,
	OPM4    = 0x0800,
	OPG1    = 0x1000,
	OPG2    = 0x2000,
	OPRANGE = 0x4000,
	OPIQR   = 0x8000,
	DOMEAN  = OPMEAN | OPVAR   | OPSTDEV | OPM2 | OPM3 | OPM4 | OPG1 | OPG2,
	DOS2    = OPVAR  | OPSTDEV | OPM2    | OPG1 | OPG2,
	DOS3    = OPM3   | OPG1,
	DOS4    = OPM4   | OPG2,
	DOMIN   = OPMIN  | OPRANGE,
	DOMAX   = OPMAX  | OPRANGE,
	DOSORT  = OPQ1   | OPMED   | OPQ3    | OPIQR
};
/* 
   These are defines because EXCLUDEM might not be an int which is
   required for an enum
*/
#define  INCLUDEM 0x000000  /*include Median in upper and lower halves*/
#define  EXCLUDEM 0x010000  /*exclude Median in upper and lower halves*/

enum opindex
{
	JN = 0,					/* sample size n */
	JMIN,					/* minimum */
	JQ1,					/* 1st quartile */
	JMEDIAN,				/* median = 2nd quartile */
	JQ3,					/* 3rd quartile */
	JMAX,					/* maximum */
	JMEAN,					/* sample mean */
	JVAR,					/* sample variance, divisor = n - 1 */
	BASICSTATS,				/* number of basic statistics */
	JSTDEV = BASICSTATS,	/* sample SD = sqrt(var) */
	JM2,					/* sample 2nd moment about mean, divisor = n */
	JM3,					/* sample 3rd moment about mean, divisor = n */
	JM4,					/* sample 4th moment about mean, divisor = n */
	JG1,					/* Fisher's g1 = skewness coefficient */
	JG2,					/* Fisher's g2 = kurtosis coefficient */
	JRANGE,					/* Sample range = max - min*/
	JIQR,					/* Sample interquartile range = Q3 - Q1*/
	NSTATS,					/* total number of all statistics */
	StatLabelsSize = NSTATS*12  /*conservative limit for Stat label size */
};

static char          *DesCompNames[] = 
{
	"n",      "min", "q1", "median", "q3", "max", "mean",  "var",
	"stddev", "m2",  "m3", "m4",     "g1", "g2",  "range", "iqr" 
};

static char          *DesRowNames[] = 
{
	"n",       "Minimum", "Q1", "Median", "Q3", "Maximum", "Mean",  "Variance",
	"Std Dev", "m2",      "m3", "m4",     "g1", "g2",      "Range", "IQR"      
};


/*
  These macros define entries into vector answers which is filled
  by compStats() and compStats1() 
*/
#define NVal     answers[JN]
#define MinVal   answers[JMIN]
#define Q1Val    answers[JQ1]
#define MedVal   answers[JMEDIAN]
#define Q3Val    answers[JQ3]
#define MaxVal   answers[JMAX]
#define MeanVal  answers[JMEAN]
#define VarVal   answers[JVAR]
#define SDVal    answers[JSTDEV]
#define M2Val    answers[JM2]
#define M3Val    answers[JM3]
#define M4Val    answers[JM4]
#define G1Val    answers[JG1]
#define G2Val    answers[JG2]
#define RangeVal answers[JRANGE]
#define IQRVal   answers[JIQR]


/* 
   990225 Fixed bug affecting stddev when all values are MISSING
   990226 replaced putOUTSTR() by putErrorOUTSTR()
   010110 Added new statistics 'iqr' and 'range' and rearranged computations
          so that statistics should always come out the same.  Previously
          you could get slightly different results for statistics computed
          from sums depending on whether a sort was required.  Now the
          necessary sums are computed before the data are sorted

   010124 Ignores case of keywords
          New describe() keyword phrase excludeM:F changes way quartiles and
          IQR are computed.  Quartiles are then the medians of the lower
          and upper halves of the data *excluding* the median when n is odd

   011214 New keyword phrase 'silent:T' suppresses warning messages
   020501 New keyword phrase 'fivenum:T' computes min,q1,median,q3,max
   050705 Changed calling sequence of compStats so n is returned in answers[]
          New routine compStats1() for computing statistics when sorting
          is required
   050708 New keywords 'bystats', 'statsfirst'.  Output has labels except
          when an array has no labels and 'bystats:F' is not an argument
          New entry function descriptive() is equivalent to
          describe(x,...,bystats:F,statsfirst:T)
   060502 Fixed bugs in compStats() and compStats1()
*/

/*
   Compute statistics that don't need sorting, used whenever q1, median,
   q3, iqr not wanted so as to avoid need for allocating scratch space

   060502 Fixed bugs
   060503 Fixed further bugs having to do with handling overflow
   060729 Fixed bug; compStats() now computes range
   061103 Fixed bug by initializing array answers.
*/
static void compStats(double *x, long op, long nrows, double answers[NSTATS],
					  long *status)
{
	long       i, n = 0;
	double     y;
	double     s1 = 0.0, s2 = 0.0, s3 = 0.0, s4 = 0.0;
	WHERE("compStats");

	doubleFill(answers, 0.0, NSTATS);

	MinVal = HUGEDBL;
	MaxVal = -MinVal;

	for (i = 0;i < nrows; i++)
	{
		y = x[i];
		if (!isMissing(y))
		{
			n++;
			if (op & DOMEAN)
			{
				s1 += y;
			}
			if (op & DOMIN && y < MinVal)
			{
				MinVal = y;
			}
			if (op & DOMAX && y > MaxVal)
			{
				MaxVal = y;
			}
		} /*if (!isMissing(y))*/
	} /*for (i = 0;i < nrows; i++)*/

	NVal = (double) n;

	if (n < nrows)
	{
		*status |= FOUNDMISSING;
	}
	if (n == 0)
	{
		for (i = 1; i < NSTATS; i++)
		{
			setMissing(answers[i]);
		}
	} /*if (n == 0)*/
	else
	{
		if (op & DOMEAN)
		{
			double         fn = (double) n;

#ifdef HASINFINITY
			if (isInfinite(s1))
			{
				setMissing(MeanVal);
				setMissing(VarVal);
				setMissing(SDVal);
				setMissing(G1Val);
				setMissing(G2Val);
				*status |= FOUNDOVERFLOW;
				goto done;
			} /*if (isInfinite(MeanVal))*/
#endif /*HASINFINITY*/

			MeanVal = s1 / fn;

			if (op & (DOS2 | DOS3 | DOS4))
			{
				if (n == 1)
				{
					VarVal = SDVal = G1Val = G2Val = 0.0;
				}
				else
				{
					int       ngt2 = (n > 2);
					int       ngt3 = (n > 3);

					for (i = 0;i < nrows; i++)
					{
						y = x[i];
						if (!isMissing(y))
						{
							double       yy;

							y -= MeanVal;

							yy = y*y;

							if (op & DOS2)
							{
								s2 += yy;
							}

							if (op & DOS3)
							{
								s3 += y*yy;
							}
							if (op & DOS4)
							{
								s4 += yy*yy;
							}
						} /*if (!isMissing(y))*/
					} /*for (i = 0;i < nrows; i++)*/
#ifdef HASINFINITY
					if (op & DOS2 && isInfinite(s2))
					{
						setMissing(VarVal);
						setMissing(SDVal);
						setMissing(M2Val);
						setMissing(M3Val);
						setMissing(M4Val);
						setMissing(G1Val);
						setMissing(G2Val);
						*status |= FOUNDOVERFLOW;
						goto done;
					} /*if (op & DOS2 && isInfinite(s2))*/

					if (op & DOS3 && isInfinite(s3))
					{
						setMissing(s3);
						setMissing(G1Val);
						setMissing(M3Val);
						*status |= FOUNDOVERFLOW;
					} /*if (op & DOS3 && isInfinite(s3))*/

					if (op & DOS4 && isInfinite(s4))
					{
						setMissing(s4);
						setMissing(G2Val);
						setMissing(M4Val);
						*status |= FOUNDOVERFLOW;
					} /*if (op & DOS4 && isInfinite(s4))*/
#endif /*HASINFINITY*/

					VarVal = s2 / (fn - 1.0);
					SDVal = sqrt(VarVal);
					M2Val = (!isMissing(s2)) ? s2 / fn : s2;
					M3Val = (!isMissing(s3)) ? s3 / fn : s3;
					M4Val = (!isMissing(s4)) ? s4 / fn : s4;

					if (ngt2 && (op & OPG1) && VarVal != 0.0)
					{
						G1Val = (!isMissing(G1Val)) ?
						  (fn*s3/((fn-1.0)*(fn-2.0)))/(SDVal*VarVal) : G1Val;
					}
					else
					{
						G1Val = 0.0;
					}

					if (ngt3 && (op & OPG2) && VarVal != 0.0)
					{
						G2Val = (!isMissing(G2Val)) ?
						  ((fn*(fn+1.0)*s4 -
							3*(fn-1.0)*s2*s2)/((fn-1.0)*(fn-2.0)*(fn-3.0)))/
						  (VarVal*VarVal) : G2Val;
					}
					else
					{
						G2Val = 0.0;
					}
				}
			} /*if (n > 1)*/
		} /*if (op & ~OPMEAN)*/
		if (op & OPRANGE)
		{
			RangeVal = MaxVal - MinVal;
		}

	} /*if (n == 0){} else if (op & DOMEAN){}*/

  done:
	
	return;
} /*compStats()*/

/*
  Function to compute Statistics when sorting is necessary
  Results are put in answers[] which is accessed by macros such as
  NVar, MinVar, MaxVar, ...

  060502 Fixed bug preventing MISSING variance when s2 overflowed
         Fixed bug affecting value of g1 when stddev not specified too
         Rearranged and simplified some code
  060503 Fixed further bugs having to do with handling overflow
  061103 Fixed bug by initializeing answers
*/
static void compStats1(double *x, double * xscratch, long op, long nrows,
					   double answers[NSTATS], long *status)
{
	double           s1, s2, s3, s4;
	double           fn;
	long             foundOverflow = 0;
	long             nmiss = 0, i, n;
	WHERE("compStats1");

	doubleFill(answers, 0.0, NSTATS);
	nmiss = 0;		/* missing count */
	/* Copy nonmissing values to xscratch */
	for (i = 0; i < nrows; i++)
	{
		if (!isMissing(x[i]))
		{
			xscratch[i - nmiss] = x[i];
		}
		else
		{
			nmiss++;
		}
	} /*for (i = 0; i < nrows; i++)*/

	if (nmiss > 0)
	{
		*status |= FOUNDMISSING;
	}

	n = nrows - nmiss;
	NVal = (double) n;
	
	if (n > 0)
	{		
		if (op & DOMEAN)
		{
			s1 = s2 = s3 = s4 = 0.0;
			fn = (double) n;

			for (i = 0; i < n; i++)
			{
				s1 += xscratch[i];
			}
#ifdef HASINFINITY
			if ((foundOverflow = isInfinite(s1)))
			{
				setMissing(MeanVal);
			}
			else
#endif /*HASINFINITY*/
			{
				MeanVal = s1/fn;
			}

			if (op & (DOS2 | DOS3 | DOS4))
			{
#ifdef HASINFINITY
				if (!foundOverflow)
				{
#endif /*HASINFINITY*/
					for (i = 0; i < n; i++)
					{
						double        res, resq;

						res = xscratch[i] - MeanVal;
						resq = res * res;
						if (op & DOS2)
						{
							s2 += resq;
						}

						if (op & DOS3)
						{
							s3 += res*resq;
						}

						if (op & DOS4)
						{
							s4 += resq*resq;
						}
					} /*for (i = 0; i < n; i++)*/
#ifdef HASINFINITY
					if (isInfinite(s2))
					{
						setMissing(s2);
						setMissing(s3);
						setMissing(s4);
						foundOverflow = 1;
					}
					else if (isInfinite(s3))
					{
						setMissing(s3);
						setMissing(s4);
						foundOverflow = 1;
					}
					else if (isInfinite(s4))
					{
						setMissing(s4);
						foundOverflow = 1;
					}
				
					if (!isMissing(s2))
					{
#endif /*HASINFINITY*/
						VarVal = (n > 1) ? s2 / (fn - 1.0) : 0.0;
#ifdef HASINFINITY
					}
				} /*if (!foundOverflow)*/
				if (foundOverflow)
				{
					*status |= FOUNDOVERFLOW;
					if (isMissing(s2))
					{
						setMissing(VarVal);
						setMissing(SDVal);
						setMissing(M2Val);
					}
					if (isMissing(s3))
					{						
						setMissing(M3Val);
						setMissing(G1Val);
					}
					if (isMissing(s4))
					{
						setMissing(M4Val);
						setMissing(G2Val);
					}					
				}
#endif /*HASINFINITY*/						
			} /*if (op & (OPVAR | OPSTDEV | OPG1 | OPG2))*/
		} /*if (op & DOMEAN)*/

		sortquick(xscratch, n);

		MinVal = xscratch[0];
		MaxVal = xscratch[n-1];

		if (op & (OPQ1 | OPQ3 | OPIQR))
		{
			long          j, k;

			if (op & EXCLUDEM)
			{
				j = (n - 2)/4;
				k = n/4;
			}
			else
			{
				j = (n - 1)/4;
				k = (n + 1)/4;
			}
			if (op & (OPQ1 | OPIQR))
			{
				Q1Val = (xscratch[j] + xscratch[k]) / 2.0;
			}
			if (op & (OPQ3 | OPIQR))
			{
				Q3Val = (xscratch[n - 1 - j] + xscratch[n - 1 - k]) / 2.0;
			}
			if (op & OPIQR)
			{
				IQRVal = Q3Val - Q1Val;
			} /*if (op & OPIQR)*/
		} /*if (op & (OPQ1 | OPQ3 | OPIQR))*/

		if (op & OPMED)
		{
			long          j, k;

			j = (n - 1) / 2;
			k = n / 2;
			MedVal = (xscratch[j] + xscratch[k]) / 2.0;
		}

		if (op & OPRANGE)
		{
			RangeVal = MaxVal - MinVal;
		}

		if (op & (OPSTDEV | OPG1))
		{
			SDVal = (!isMissing(VarVal)) ? sqrt(VarVal) : VarVal;
		}
		if (op & OPM2)
		{
			M2Val = (!isMissing(s2)) ? s2/fn : s2;
		}
		if (op & OPM3)
		{
			M3Val = (!isMissing(s3)) ? s3/fn : s3;
		}
		if (op & OPM4)
		{
			M4Val = (!isMissing(s4)) ? s4/fn : s4;
		}

		if (op & OPG1)
		{
			if (n <= 3 || VarVal == 0.0)
			{
				G1Val = 0.0;
			}
			else
#ifdef HASINFINITY
			if (isMissing(s3))
			{
				setMissing(G1Val);
			}
			else
#endif /*HASINFINITY*/
			{
				G1Val = (fn*s3/((fn-1.0)*(fn-2.0)))/(SDVal*VarVal);
			}
		} /*if (op & OPG1)*/
		if (op & OPG2)
		{
			if (n <= 3 || VarVal == 0.0)
			{
				G2Val = 0.0;
			}
			else
#ifdef HASINFINITY
			if (isMissing(s4))
			{
				setMissing(G2Val);
			}
			else
#endif /*HASINFINITY*/
			{
				G2Val = ((fn*(fn+1.0)*s4 -
						  3*(fn-1.0)*s2*s2)/((fn-1.0)*(fn-2.0)*(fn-3.0)))/
				  (VarVal*VarVal);
			}
		} /*if (op & OPG2)*/
	
	} /*if (n > 0)*/
	else
	{/* n == 0 */
		for (i = 1;i < NSTATS;i++)
		{
			setMissing(answers[i]);
		}
	}
} /*compStats1()*/

/*
  Function to control computation of descriptive statistics when
  the result is to be a scalar, vector, matrix or array or a structure
  with such components, one for each component of a structure argument
   arg           Symbolhandle for arg 1 of describe, a structure or
                 a REAL array
   op            bit pattern specifying what statistics are to be
                 computed nu or'ing a selection from OPN, OPMIN,
                 ..., OPIQR
   nstats        The number of statistics requested
   stats         vector of length nstats with j = nstats[i] the index
                 of the i-th statistic, 0 <= j < NSTATS
   statsfirst    when non-zero, first dimension has length nstats corresponding
                 to statistics; otherwise, last dimension has length nstats
   status        pointer to status flag

   Value returned Symbolhandle for array with the same number of
                 dimensions as arg, the 1st or last dimension having length 
                 nstats and the remaining dimensions matching all but
                 dimension 1 of arg.  **Exception** when ndims(arg) == 1
                 and statsfirst != 0, the result is a nstats by 1 column
                 vector with column label ""

                 When arg is a structure, the result is a structure with
                 the same tree of components as arg

   050708 new function called when bystats:F is argument
          result is labelled
*/
static Symbolhandle doDescribe1(Symbolhandle arg, long op, int statsfirst,
								int nstats, int stats [], long *status)
{
	Symbolhandle    result = (Symbolhandle) 0;
	Symbolhandle    scratchH = (Symbolhandle) 0;
	double          answers[NSTATS];
	double         *xj, *xscratch = (double *) 0;
	double         *statsj, *statsjk;
	long            nrows, ncols, i, j, jlab;
	long            ndims, dims[MAXDIMS+1];
	long            inc1, inc2, shift;
#ifdef HASINFINITY
	long            foundOverflow = 0;
#endif /*HASINFINITY*/
	WHERE("doDescribe1");

	if (isStruc(arg))
	{
		long            icomp, ncomps = NCOMPS(arg);
		Symbolhandle    component1 = (Symbolhandle) 0;
		
		result = StrucInstall(SCRATCH, ncomps);
		
		if (result == SymbolError)
		{
			goto errorExit;
		}
		
		for (icomp = 0; icomp < ncomps; icomp++)
		{
			Symbolhandle     component = 
			  doDescribe1(COMPVALUE(arg, icomp), op, statsfirst,
						  nstats, stats, status);

			if (component == SymbolError)
			{
				goto errorExit;
			}
			Cutsymbol(component);
			COMPVALUE(result, icomp) = component;
			setNAME(component, NAME(COMPVALUE(arg, icomp)));
		}
		goto normalExit;
	} /*if (isStruc(arg))*/

	/* Argument not a structure ; must be NULL or REAL */

	if (isNull(arg)) 
	{
		result = Install(SCRATCH, NULLSYM);
		if (result == SymbolError)
		{
			goto errorExit;
		}
		goto normalExit;
	} /*if (isNull(arg)) */

	getDims(dims, arg);

	ndims = dims[0];
	nrows = dims[1];
	ncols = symbolSize(arg)/nrows;

	result = RInstall(SCRATCH, ncols*nstats);
	if (result == SymbolError)
	{
		goto errorExit;
	}

	shift = (statsfirst) ? 1 : 2;
	if (statsfirst && ndims == 1)
	{
		/* Result for vector will be explicit column vector */
		ndims = 2;
		dims[1] = 1;
	}
	else
	{
		for (i = 2; i <= ndims; i++)
		{
			dims[i-shift] = dims[i];
		}
	}
	
	dims[(statsfirst) ? 0 : ndims-1] = nstats;

	Setdims(result, ndims, dims);

	{
		/* Add labels */
		char       **labelH;
		char        *labels[MAXDIMS + 1];
		char         statLabels[StatLabelsSize];
		char        *place;
		char         numericLabel[2];
		long         sizes[MAXDIMS];
		int          sizeofStatLabels;
		
		for (i = 0, place = statLabels; i < nstats; i++)
		{
			place = copyStrings(DesRowNames[stats[i]], place, 1);
		} /*for (i = 0, place = statLabels; i < nstats; i++)*/
		sizeofStatLabels = place - statLabels;

		if (NDIMS(arg) > 1)
		{
			if (HASLABELS(arg))
			{
				getAllLabels(arg, labels, sizes, (long *) 0);
				if (!statsfirst) 
				{
					for (i = 0; i < ndims - 1; i++)
					{
						sizes[i] = sizes[i+1];
					}
				}
			}
			else
			{
				numericLabel[0] = NUMERICLABEL;
				numericLabel[1] = '\0';
				
				for (i = 2; i <= ndims; i++)
				{
					sizes[i - shift] = 2*DIMVAL(arg, i);
				}
			}
		} /*if (NDIMS(arg) > 1)*/
		else if (statsfirst)
		{
			sizes[1] = 1;
		}
		sizes[(statsfirst) ? 0 : ndims-1] = sizeofStatLabels;

		labelH = createLabels(ndims, sizes);
		setLabels(result, labelH);
		if (labelH == (char **) 0)
		{
			goto errorExit;
		}

		jlab = 0;
		if (statsfirst)
		{
			appendLabels(result, statLabels, jlab++, dontExpand);
		}
		
		if (NDIMS(arg) > 1) 
		{
			int      expandit;
			
			if (HASLABELS(arg))
			{
				getAllLabels(arg, labels, (long *) 0, (long *) 0);
				for (i = 0; i < ndims - 1; i++)
				{
					labels[i] = labels[i+1];
				}
				expandit = dontExpand;
			}
			else
			{
				for (i = 0; i < ndims - 1; i++)
				{
					labels[i] = numericLabel;
				}
				expandit = doExpand;
			}
			for (i = 0; i < ndims-1; i++)
			{
				appendLabels(result, labels[i], jlab++, expandit);
			}
		} /*if (NDIMS(arg) > 1) */

		if (!statsfirst)
		{
			appendLabels(result, statLabels, jlab, dontExpand);
		}
		else if (NDIMS(arg) == 1)
		{
			appendLabels(result, NullString, 1, dontExpand);
		}
	}

	if (op & DOSORT)
	{
		scratchH = RInstall(SCRATCH, nrows);

		if (scratchH == (Symbolhandle) 0)
		{
			goto errorExit;;
		}
		xscratch = DATAPTR(scratchH);
	} /*if (op & DOSORT)*/

	xj = DATAPTR(arg);
	if (!statsfirst)
	{
		inc1 = 1;
		inc2 = ncols;
	}
	else
	{
		inc1 = nstats;
		inc2 = 1;
	}
	statsj = DATAPTR(result);
	
	/*
	  The argument is temporarily treated as a nrows by ncols matrix.
	  The result is temporarily treated as a ncols by nstats matrix
	  a nstats by ncols matrix
	*/

	for (j = 0; j < ncols;j++, statsj += inc1, xj += nrows)
	{
		if (op & DOSORT)
		{
			compStats1(xj, xscratch, op, nrows, answers, status);
		}
		else
		{
			compStats(xj, op, nrows, answers, status);
		}
		
		/*
		  Order should be
		  n min q1 median q3 max mean var stddev m2 m3 m4 g1 g2 range iqr
		*/

		statsjk = statsj; /* start of jth row or column of statistics */

		for (i = 0; i < nstats; i++)
		{
			*statsjk = answers[stats[i]];
			statsjk += inc2;
		} /*for (i = 0; i < nstats; i++)*/
	} /* for (j = 0;j < ncols;j++) */

	if (op & DOSORT)
	{
		Removesymbol(scratchH);
	}
	
  normalExit:
	return (result);

  errorExit:
	Removesymbol(result);

	return (0);
} /*doDescribe1()*/	

/*
  function to control computation of descriptive statistics when
  the result is to be a structure (classic case)
   arg           Symbolhandle for arg 1 of describe
   result        Structure with a component for each statistic
                 On return each component contains the statistics.
                 When arg is a structure, each component of result
                 is a structure with the same tree of components as
                 arg
   op            bit pattern specifying what statistics are to be
                 computed nu or'ing a selection from OPN, OPMIN,
                 ..., OPIQR
   nstats        number of statistics computed
   stats         vector with j = stats[i] = index of i-th statistic,
                 0 <= j < NSTATS
   status        pointer to bitstring reporting MISSING value and
                 overflow status
   Building the result proceeds recursively with result always having
   the same number of components.  When arg is a structure, component
   component j of doDescribe(arg[i]) is put in component i of result[j]

   050707 many changes
   050708 result is now labeled if arg is
   061103 array answers is now initialized
*/

static short doDescribe(Symbolhandle arg, Symbolhandle result, long op,
						int nstats, int stats [], long *status)
{
	Symbolhandle    symh, result1 = (Symbolhandle) 0;
	Symbolhandle    components[NSTATS];
	double         *x, *xscratch = (double *) 0;
	long            nrows, ncols, i, j, l, jj;
	long            ndims, dims[MAXDIMS+1];
#ifdef HASINFINITY
	long            foundOverflow = 0;
#endif /*HASINFINITY*/
	WHERE("doDescribe");

	if (isStruc(arg))
	{
		long            icomp, ncomps = NCOMPS(arg);
		Symbolhandle    component1 = (Symbolhandle) 0;
		
		for (i = 0;i < nstats;i++)
		{
			components[i] = COMPVALUE(result, i) = Makestruc(ncomps);
			if (components[i] == (Symbolhandle) 0)
			{
				goto errorExit;
			}
		} /*for (i = 0;i < nstats;i++)*/
		
		result1 = StrucInstall(SCRATCH, nstats);
		if (result1 == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		
		for (icomp = 0;icomp < ncomps;icomp++)
		{
			symh = COMPVALUE(arg, icomp);
			if (doDescribe(symh, result1, op, nstats, stats, status) == 0)
			{
				goto errorExit;
			}
			
			for (i = 0;i < nstats;i++)
			{
				component1 = COMPVALUE(result1, i);
				COMPVALUE(result1, i) = (Symbolhandle) 0;
				COMPVALUE(components[i], icomp) = component1;
				setCompName(component1, NAME(symh));
			}
		} /*for (icomp = 0;icomp < ncomps;icomp++)*/
		Removesymbol(result1);
		return (1);
	} /* if (isStruc(arg)) */

	ncols = 1;
	ndims = NDIMS(arg);
	if (ndims <= 1)
	{
		dims[0] = 1;
	}
	else
	{
		ndims--;
		for (i = 0;i < ndims;i++)
		{
			dims[i] = DIMVAL(arg, i+2);
			ncols *= dims[i];
		}
	}
	
	nrows = DIMVAL(arg, 1);

	for (i = 0;i < nstats;i++)
	{
		components[i] = COMPVALUE(result,i) = Makereal(ncols);
		if (components[i] == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		Setdims(components[i], ndims, dims);
	} /*for (i = 0;i < nstats;i++)*/

	if (HASLABELS(arg) && NDIMS(arg) > 1)
	{
		char           *labels[MAXDIMS];
		long            sizes[MAXDIMS];
		char          **labelH;

		getAllLabels(arg, labels, sizes, (long *) 0);

		for (i = 0; i < nstats; i++)
		{
			labelH = createLabels(ndims, sizes+1);
			if (labelH == (char **) 0)
			{
				goto errorExit;
			}
			if (!setLabels(components[i], labelH))
			{
				mydisphandle(labelH);
				goto errorExit;
			}
			getAllLabels(arg, labels, (long *) 0, (long *) 0);
			for (j = 0; j < ndims; j++)
			{
				appendLabels(components[i], labels[j+1], j, dontExpand);
			}
		}
	} /*if (HASLABELS(arg))*/
	
	if (nrows > 0)
	{
		Symbolhandle    scratchH = (Symbolhandle) 0;
		double          answers[NSTATS];

		if (op & DOSORT)
		{
			scratchH = RInstall(SCRATCH, nrows);

			if (scratchH == (Symbolhandle) 0)
			{
				goto errorExit;;
			}
			xscratch = DATAPTR(scratchH);
		} /*if (op & DOSORT)*/
		x = DATAPTR(arg);

		l = 0;
		for (jj = 0;jj < ncols;jj++)
		{
			if (op & DOSORT)
			{
				compStats1(x, xscratch, op, nrows, answers, status);
			}
			else
			{
				compStats(x, op, nrows, answers, status);
			}
				
			for (i = 0; i < nstats; i++)
			{
				DATAVALUE(components[i], jj) = answers[stats[i]];
			} /*for (i = 0; i < nstats; i++)*/			
			x += nrows;
		} /* for (jj = 0;jj < ncols;jj++) */

		if (op & DOSORT)
		{
			Removesymbol(scratchH);
		}
	} /* if (nrows > 0) */
	else
	{/* component is empty */
		for (i = 0;i < nstats;i++)
		{
			symh = components[i];
			setMissing(DATAVALUE(symh, 0));
			setNDIMS(symh, 1);
			setDIM(symh, 1, 1);
		}
	}
	
	return (1);

  errorExit:
	Removesymbol(result1);

	return (0);
} /*doDescribe()*/	

/*
   Function to compute variable statistics, returning a structure
   describe(x) returns n, min, q1, median, q3, max, mean, var
   describe(x,mean:T,var:T), say, computes only mean and var
   describe(x,all:T,mean:F,var:F), say, computes all except mean and var
   
   If x is matrix or array, describe computes the statistics over the
   first dimensions for all combinations of later dimensions.

   If only one statistic is requested, describe() returns a REAL variable,
   not a structure.

   If x is a structure, each component of describe(x) is a structure whose
   components are the result of computing the each statistic for each 
   component of x.

   950823 added keyword all

   990225 Fixed bug affecting stddev when all values are MISSING

   010110 Added 'range' and 'iqr'
   010125 New keyword excludeM with default set by option excludeM
   060503 Minor change to warning about overflow
*/

#define DEFAULTSTATSFIRST  1
#define DEFAULTBYSTATS     1
Symbolhandle    describe(Symbolhandle list)
{
	Symbolhandle    result = (Symbolhandle) 0, symh, symhkey;
	long            i, j, nargs = NARGS(list);
	long            status = 0, op;
	long            excludeM = (EXCLUDEMEDIAN) ? EXCLUDEM : INCLUDEM;
	int             nstats, stats[NSTATS];
	int             foundKey = 0;
	int             silent = 0, statsfirst = -1, bystats = DEFAULTBYSTATS;
	int             descript = strcmp(FUNCNAME, "descriptive") == 0;
	char           *keyword;
	WHERE("describe");

	*OUTSTR = '\0';
	
	if (descript)
	{
		statsfirst = 1;
		bystats = 0;
	}
	
	symh = COMPVALUE(list, 0);
	if (!argOK(symh, 0, 0))
	{
		goto errorExit;
	}
	if (!isReal(symh))
	{
		sprintf(OUTSTR,
				"ERROR: argument%s to %s() must be REAL or structure of REALs",
				(nargs == 1) ? "" : " 1", FUNCNAME);
		goto errorExit;
	}
	
	for (j = 0; j < NSTATS; j++)
	{
		stats[j] = -1;
	}

	/*
	  first check to see whether 'silent:T or F' is an argument
	  since it is needed to control warnings on other keywords
	*/
	keyword = "silent";
	i = findKeyword(list, keyword, 1);
	if (i < 0)
	{
		keyword = "quiet";
		i = findKeyword(list, keyword, 1);
	}
	
	if (i > 0)
	{
		symhkey = COMPVALUE(list, i);
		if (!isTorF(symhkey))
		{
			notTorF(keyword);
		}
		silent = DATAVALUE(symhkey, 0) != 0;
	} /*if (i > 0)*/	

	for (i = 1;i < nargs;i++)
	{
		int         logValue;
		
		symhkey = COMPVALUE(list,i);
		if (!argOK(symhkey,0,i+1))
		{
			goto errorExit;
		}

		if (!(keyword = isKeyword(symhkey)))
		{
			sprintf(OUTSTR,
					"ERROR: all but first argument to %s() must be keyword phrases",
					FUNCNAME);
			goto errorExit;
		}

		if (!isTorF(symhkey))
		{
			notTorF(keyword);
			goto errorExit;
		}

		logValue = DATAVALUE(symhkey, 0) != 0.0;

		if (mystrncmp(keyword, "excludeM", 0) == 0) 
		{
			excludeM = (logValue) ? EXCLUDEM : INCLUDEM;
		}
		else if (mystrncmp(keyword, "silent", 0) == 0 ||
				 mystrncmp(keyword, "quiet", 0) == 0)
		{
			silent = logValue;
		}
		else if (mystrncmp(keyword, "bystats", 6) == 0)
		{
			bystats = logValue;
		}
		else if (mystrncmp(keyword, "statsfirst", 6) == 0)
		{
			statsfirst = logValue;
		}
		else if (mystrncmp(keyword, "all", 0) == 0)
		{
			foundKey = 1;
			if (logValue)
			{
				for (j = 0; j < NSTATS; j++)
				{
					if (stats[j] < 0)
					{
						stats[j] = 1;
					}
				} /*for (j = 0; j < NSTATS; j++)*/
			} /*if (logValue)*/ 
		} /*if (mystrncmp(keyword, "all", 0) == 0)*/
		else if (mystrncmp(keyword, "fivenum", 4) == 0)
		{
			foundKey = 1;
			if (logValue)
			{
				if (stats[JMIN] < 0)
				{
					stats[JMIN] = 1;
				}
				if (stats[JQ1] < 0)
				{
					stats[JQ1] = 1;
				}
				if (stats[JMEDIAN] < 0)
				{
					stats[JMEDIAN] = 1;
				}
				if (stats[JQ3] < 0)
				{
					stats[JQ3] = 1;
				}
				if (stats[JMAX] < 0)
				{
					stats[JMAX] = 1;
				}
			} /*if (logValue)*/ 
		} /*if (mystrncmp(keyword, "fivenum", 4) == 0)*/
		else
		{
			foundKey = 1;
			if (mystrncmp(keyword, "q2", 0) == 0)
			{ /* alternate keyword for median */
				j = JMEDIAN;
			} /*if (mystrncmp(keyword, "q2", 0) == 0)*/
			else if (mystrncmp(keyword, "m1", 0) == 0)
			{
				j = JMEAN;
			}
			else
			{
				for (j = 0; j < NSTATS; j++)
				{
					if (mystrncmp(keyword,DesCompNames[j],
								  strlen(DesCompNames[j])) == 0)
					{
						break;
					}
				} /*for (j = 0; j < NSTATS; j++)*/
				if (j == NSTATS)
				{
					if (!silent)
					{
						sprintf(OUTSTR,
								"%sUnrecognized keyword %s ignored by %s()",
								WarningStart, keyword, FUNCNAME);
						putErrorOUTSTR();
					}
					j = -1;
				}
			} /*if (mystrncmp(keyword, "q2", 0) == 0){}else{}*/			
			if (j >= 0)
			{
				stats[j] = (DATAVALUE(symhkey,0) != 0.0);
			}
		} /*if (mystrncmp(keyword, "all", 0) == 0)else{}*/ 
	} /*for (i = 1;i < nargs;i++)*/

	op = excludeM;
	nstats = 0;
	for (j = 0; j < NSTATS; j++)
	{
		if (!foundKey && j < BASICSTATS)
		{
			stats[j] = 1;
		}
		
		if (stats[j] > 0)
		{
			op |= (1L << j);
			stats[nstats++] = j;
		}
	} /*for (j = 0; j < NSTATS; j++)*/

	/* 
	   At this point, stats[0],..., stats[nstats-1] contain the index
	   of the statistics wanted
	*/
	if (op == 0)
	{
		sprintf(OUTSTR,
				"%sNo statistics specified for %s()", ErrorStart, FUNCNAME);
		goto errorExit;
	} /*if (op == 0)*/

	if (bystats) 
	{
		if (!silent && statsfirst >= 0 && !descript)
		{
			sprintf(OUTSTR, 
					"%s Keyword statsfirst ignored by %s() without bystats:F",
					WarningStart, FUNCNAME);
			putErrorOUTSTR();
		}
		
		result = StrucInstall(SCRATCH, nstats);
		if (result == SymbolError ||
			doDescribe(symh, result, op, nstats, stats, &status) == 0)
		{
			goto errorExit;
		}

		if (nstats > 1)
		{
			for (i = 0;i < nstats;i++)
			{
				setNAME(COMPVALUE(result, i), DesCompNames[stats[i]]);
			}
		} /*if (nstats > 1)*/
		else
		{
			symh = COMPVALUE(result,0);
			COMPVALUE(result, 0) = (Symbolhandle) 0;
			setNAME(symh, SCRATCH);
			Addsymbol(symh);
			Removesymbol(result);
			result = symh;
		} /*if (nstats > 1){}else{}*/
	} /*if (bystats) */
	else
	{
		statsfirst = (statsfirst < 0) ? DEFAULTSTATSFIRST : statsfirst;
		result = doDescribe1(symh, op, statsfirst, nstats, stats, &status);
		if (result == SymbolError)
		{
			goto errorExit;
		}
	} /*if (bystats) {}else{}*/

	if (!silent)
	{
		if (status & FOUNDMISSING)
		{
			sprintf(OUTSTR, "%sMissing values in input to %s()",
					WarningStart, FUNCNAME);
			putErrorOUTSTR();
			sprintf(OUTSTR, "%sOutput reflects non-missing data only",
					WarningStart);
			putErrorOUTSTR();
		} /*if (status & FOUNDMISSING)*/	

		if (status & FOUNDOVERFLOW)
		{
			sprintf(OUTSTR,
					"%soverflow computing mean or variance%s; result(s) set MISSING",
					WarningStart,
					(op & (DOS3 | DOS4)) ? " or higher moment" : NullString);
			putErrorOUTSTR();
		}
	} /*if (!silent)*/
	
	return (result);
	
  errorExit:
	putErrorOUTSTR();

	Removesymbol(result);
	
	return (0);
} /*describe()*/
