#include <stdio.h>
#include <math.h>

#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))

/*
  Purpose
  =======

  dgemm  performs one of the matrix - matrix operations

  c := alpha*op( a )*op( b ) + beta*c,

  where  op( x ) is one of

  op( x ) = x   or   op( x ) = x',

  alpha and beta are scalars,  and a,  b and c are matrices,  with op( a )
  an m by k matrix,   op( b )  a  k by n matrix and  c an m by n matrix.

  Parameters
  ==========

  transa - character*1.
  On entry,  transa specifies the form of op( a ) to be used in
  the matrix multiplication as follows:

  transa = 'n' or 'n',   op( a ) = a.

  transa = 't' or 't',   op( a ) = a'.

  transa = 'c' or 'c',   op( a ) = a'.

  Unchanged on exit.

  transb - character*1.
  On entry,  transb specifies the form of op( b ) to be used in
  the matrix multiplication as follows:

  transb = 'N' or 'n',   op( b ) = b.

  transb = 'T' or 't',   op( b ) = b'.

  transb = 'C' or 'c',   op( b ) = b'.

  Unchanged on exit.

  m - integer.
  On entry,   m  specifies  the number  of rows  of the  matrix
  op( a )  and of the  matrix  c.  m  must  be at least  zero.
  Unchanged on exit.

  n - integer.
  On entry,   n  specifies the number  of columns of the matrix
  op( b ) and the number of columns of the matrix c. n must be
  at least zero.
  Unchanged on exit.

  k - integer.
  On entry,   k  specifies  the number of columns of the matrix
  op( a ) and the number of rows of the matrix op( b ). k must
  be at least  zero.
  Unchanged on exit.

  alpha - double precision.
  On entry,  alpha specifies the scalar alpha.
  Unchanged on exit.

  a - double precision array of dimension ( lda,  ka ),  where ka is
  k  when  transa = 'n' or 'n',   and is  m  otherwise.
  before entry with  transa = 'n' or 'n',   the leading  m by k
  part of the array  a  must contain the matrix  a,   otherwise
  the leading  k by m  part of the array  a  must contain  the
  matrix a.
  Unchanged on exit.

  lda - integer.
  On entry,  lda specifies the first dimension of a as declared
  in the calling (sub) program. when  transa = 'n' or 'n' then
  lda must be at least  max( 1,  m ),  otherwise  lda must be at
  least  max( 1,  k ).
  Unchanged on exit.

  b - double precision array of dimension ( ldb,  kb ),  where kb is
  n  when  transb = 'n' or 'n',   and is  k  otherwise.
  before entry with  transb = 'n' or 'n',   the leading  k by n
  part of the array  b  must contain the matrix  b,   otherwise
  the leading  n by k  part of the array  b  must contain  the
  matrix b.
  Unchanged on exit.

  ldb - integer.
  On entry,  ldb specifies the first dimension of b as declared
  in the calling (sub) program. when  transb = 'n' or 'n' then
  ldb must be at least  max( 1,  k ),  otherwise  ldb must be at
  least  max( 1,  n ).
  Unchanged on exit.

  beta - double precision.
  On entry,   beta  specifies the scalar  beta.  when  beta  is
  supplied as zero then c need not be set on input.
  Unchanged on exit.

  c - double precision array of dimension ( ldc,  n ).
  before entry,  the leading  m by n  part of the array  c must
  contain the matrix  c,   except when  beta  is zero,  in which
  case c need not be set On entry.
  on exit,  the array  c  is overwritten by the  m by n  matrix
  ( alpha*op( a )*op( b ) + beta*c ).

  ldc - integer.
  On entry,  ldc specifies the first dimension of c as declared
  in  the  calling  (sub)  program.   ldc  must  be  at  least
  max( 1,  m ).
  Unchanged on exit.


  Level 3 blas routine.

  -- Written on 8 - february - 1989.
  Jack Dongarra,  Argonne National Laboratory.
  Iain Duff,  AERE Harwell.
  Jeremy Du Croz,  Numerical Algorithms Group Ltd.
  Sven Hammarling,  Numerical Algorithms Group Ltd.

  This code comes from a report entitled:
  The IBM RISC System/6000 and linear algebra operations,  by
  Jack J. Dongarra,  Peter Mayes,  and Giuseppe Radicati di Brozolo,
  University of Tennessee Computer Science Tech Report: CS - 90 - 122.

  Translated to C by C. Bingham, kb@stat.umn.edu, March 24, 2002
  It preserves fortran calling sequence; all arguments are considered
  pointers; subscript ordering follows fortran; subscripts are 0 based
*/

extern void     xerbla(char *, int);

enum   dgemmParameters
{
	mb = 64,
	nb = mb,
	kb = 64,
	kbmb = kb*mb
};

#define Qa(I,J)    a[(I) + (J)*lda]
#define Qb(I,J)    b[(I) + (J)*ldb]
#define Qc(I,J)    c[(I) + (J)*ldc]
#define Qch(I,J)   ch[(I) + kb*(J)]

#define lsame(C1, C2) (*C1 == C2 || *C1 == toupper(C2))

#define INT  long

void DGEMM(char * transa, char * transb, INT * m1, INT * n1, INT * k1,
		   double * alpha1, double * a, INT * lda1, double * b, INT * ldb1,
		   double * beta1, double * c, INT * ldc1)
/*
  Original declaration
  double precision a(lda,*),b(ldb,*),c(ldc,*)
*/
{
	/*      .. dereferenced scalar arguments .. */
	double     alpha = *alpha1, beta = *beta1;
	INT        k = *k1, lda = *lda1, ldb = *ldb1, ldc = *ldc1;
	INT        m = *m1, n = *n1;
	/*      .. local scalars .. */
	double     t11, t12, t21, t22;
    INT        i, idepth, ii, ilen, info, ispan;
	INT        j, jdepth, jj, jlen, jspan;
	INT        l,ll, lspan, ncola, nrowa, nrowb;
	int         nota, notb;
	/*      .. local arrays .. */
	/* 
	   Original declaration
	   double precision ch(kb,mb), ch1(kb), ch2(kb)
	*/

	double     ch[kbmb], ch1[kb], ch2[kb];
	/*  .. executable statements .. */

	/*
	  Set  nota  and  notb  as  true if  a  and  b  respectively are not
	  transposed and set  nrowa,  ncola and  nrowb  as the number of rows
	  and  columns of  a  and the  number of  rows  of  b  respectively.
	*/
	nota = lsame(transa, 'n');
	notb = lsame(transb, 'n');

	if (!nota) 
	{
		nrowa = k;
		ncola = m;
	}
	else 
	{
		nrowa = m;
		ncola = k;
	}

	nrowb = (notb) ? k : n;
	
#ifdef TESTDGEMMINPUT
	/*  Test the input parameters. */

	info = 0;
	if (!nota && !lsame(transa, 'c') && !lsame(transa, 't'))
	{
		info = 1;
	}
	else if (!notb && !lsame(transb, 'c') && !lsame(transb, 't'))
	{
		info = 2;
	}
	else if (m < 0)
	{
		info = 3;
	}
	else if (n < 0)
	{
		info = 4;
	}
	else if (k < 0)
	{
		info = 5;
	}
	else if (lda < ((nrowa > 1) ? nrowa : 1))
	{
		info = 8;
	}
	else if (ldb < ((nrowb > 1) ? nrowb : 1))
	{
		info = 10;
	}
	else if (ldc < ((m > 1) ? m : 1))
	{
		info = 13;
	}

	if (info != 0)
	{
		goto errorExit;
	}
#endif /*TESTDGEMMINPUT*/

	/*  Quick return if possible. */

	if (m == 0 || n == 0 || beta == 1.0 && (alpha == 0.0 || k == 0))
	{
		return;
	}

	if (beta != 0.0)
	{
		for (j = 0; j < n ; j++)
		{
			for (i = 0; i < m ; i++)
			{
				Qc(i,j) *= beta;
			}
		}
	}
	else
	{
		for (j = 0; j < n ; j++)
		{
			for (i = 0; i < m ; i++)
			{
				Qc(i,j) = 0.0;
			}
		}
	}

	/*  And if  alpha.eq.0.0. */

	if (alpha == 0.0)
	{
		return;
	}
	
	/*  start the operations. */

	if (!notb)
	{
		if (!nota)
		{
			for (j = 0;j < n;j += nb )
			{ 
				jdepth = 2;
				jspan = MIN(nb, n - j);
				jlen = jdepth*(jspan/jdepth);
				for (l = 0;l < k;l += kb )
				{ 
					lspan = MIN(kb, k - l);
					for (jj = j;jj < j + jspan ;jj++)
					{
						for (ll = l;ll < l + lspan ;ll++)
						{
							Qch(ll - l,jj - j) = alpha*Qb(jj, ll);
						}
					}

					for (i = 0;i < m;i += mb )
					{ 
						idepth = 2;
						ispan = MIN(mb, m - i);
						ilen = idepth*(ispan/idepth);
						for (ii = i;ii < i + ilen;ii += idepth )
						{ 
							for (jj = j;jj < j + jlen;jj += jdepth )
							{ 
								t11 = 0.0;
								t21 = 0.0;
								t12 = 0.0;
								t22 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ll, ii)*Qch(ll - l+1, jj - j);
									t21 += Qa(ll, ii + 1)*Qch(ll - l+1, jj - j);
									t12 += Qa(ll, ii)*Qch(ll - l+1, jj - j+1);
									t22 += Qa(ll, ii + 1)*Qch(ll - l+1, jj - j+1);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
								Qc(ii,jj + 1) += t12;
								Qc(ii + 1,jj + 1) += t22;
							}

							for (jj = j + jlen;jj < j + jspan ;jj++)
							{ 
								t11 = 0.0;
								t21 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ll, ii)*Qch(ll - l, jj - j);
									t21 += Qa(ll, ii + 1)*Qch(ll - l, jj - j);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
							}
						}

						for (ii = i + ilen;ii < i + ispan ;ii++)
						{ 
							for (jj = j;jj < j + jlen;jj += jdepth )
							{ 
								t11 = 0.0;
								t12 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ll, ii)*Qch(ll - l, jj - j);
									t12 += Qa(ll, ii)*Qch(ll - l, jj - j + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii,jj + 1) += t12;
							} /*for (jj = j;jj < j + jlen;jj += jdepth )*/

							for (jj = j + jlen;jj < j + jspan ;jj++)
							{ 
								t11 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{
									t11 += Qa(ll, ii)*Qch(ll - l, jj - j);
								}
								Qc(ii,jj) += t11;
							} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
						} /*for (ii = i + ilen;ii < i + ispan ;ii++)*/
					} /*for (i = 0;i < m;i += mb )*/
				} /*for (l = 0;l < k;l += kb )*/
			} /*for (j = 0;j < n;j += nb )*/
		} /*if (!nota)*/
		else
		{
			/* form  c := c + alpha*a*b' */
			for (j = 0;j < n;j += nb )
			{ 
				jdepth = 2;
				jspan = MIN(nb, n - j);
				jlen = jdepth*(jspan/jdepth);
				for (l = 0;l < k;l += kb )
				{ 
					lspan = MIN(kb, k - l);
					for (jj = j;jj < j + jspan ;jj++)
					{
						for (ll = l;ll < l + lspan ;ll++)
						{
							Qch(ll - l,jj - j) = alpha*Qb(jj, ll);
						}
					}

					for (i = 0;i < m;i += mb )
					{ 
						idepth = 2;
						ispan = MIN(mb, m - i);
						ilen = idepth*(ispan/idepth);
						for (ii = i;ii < i + ilen;ii += idepth )
						{ 
							for (ll = l;ll < l + lspan ;ll++)
							{ 
								ch1[ll - l] = Qa(ii, ll);
								ch2[ll - l] = Qa(ii + 1, ll);
							}
							for (jj = j;jj < j + jlen;jj += jdepth )
							{ 
								t11 = 0.0;
								t21 = 0.0;
								t12 = 0.0;
								t22 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += ch1[ll - l]*Qch(ll - l, jj - j);
									t21 += ch2[ll - l]*Qch(ll - l, jj - j);
									t12 += ch1[ll - l]*Qch(ll - l, jj - j+1);
									t22 += ch2[ll - l]*Qch(ll - l, jj - j+1);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
								Qc(ii,jj + 1) += t12;
								Qc(ii + 1,jj + 1) += t22;
							} /*for (jj = j;jj < j + jlen;jj += jdepth )*/

							for (jj = j + jlen;jj < j + jspan ;jj++)
							{ 
								t11 = 0.0;
								t21 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ii, ll)*Qch(ll - l, jj - j);
									t21 += Qa(ii + 1, ll)*Qch(ll - l, jj - j);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
							} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
						} /*for (ii = i;ii < i + ilen;ii += idepth )*/

						for (ii = i + ilen;ii < i + ispan ;ii++)
						{ 
							for (jj = j;jj < j + jlen;jj += jdepth )
							{ 
								t11 = 0.0;
								t12 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qa(ii, ll)*Qch(ll - l, jj - j);
									t12 += Qa(ii, ll)*Qch(ll - l, jj - j+1);
								}
								Qc(ii,jj) += t11;
								Qc(ii,jj + 1) += t12;
							} /*for (jj = j;jj < j + jlen;jj += jdepth )*/

							for (jj = j + jlen;jj < j + jspan ;jj++)
							{ 
								t11 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{
									t11 += Qa(ii, ll)*Qch(ll - l, jj - j);
								}
								Qc(ii,jj) += t11;
							} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
						} /*for (ii = i + ilen;ii < i + ispan ;ii++)*/
					} /*for (i = 0;i < m;i += mb )*/
				} /*for (l = 0;l < k;l += kb )*/
			} /*for (j = 0;j < n;j += nb )*/
		} /*if (!nota){}else{}*/
	} /*if (!notb)*/
	else
	{
		if (!nota)
		{
			/*        form  c := c + alpha*a'*b */
			for (i = 0;i < m;i += mb )
			{ 
				idepth = 2;
				ispan = MIN(mb, m - i);
				ilen = idepth*(ispan/idepth);
				for (l = 0;l < k;l += kb )
				{ 
					lspan = MIN(kb, k - l);
					for (ii = i;ii < i + ispan ;ii++)
					{
						for (ll = l;ll < l + lspan ;ll++)
						{
							Qch(ll - l,ii - i) = alpha*Qa(ll, ii);
						}
					}

					for (j = 0;j < n;j += nb )
					{ 
						jdepth = 2;
						jspan = MIN(nb, n - j);
						jlen = jdepth*(jspan/jdepth);
						for (jj = j;jj < j + jlen;jj += jdepth )
						{ 
							for (ii = i;ii < i + ilen;ii += idepth )
							{ 
								t11 = 0.0;
								t21 = 0.0;
								t12 = 0.0;
								t22 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t21 += Qch(ll - l, ii - i+1)*Qb(ll, jj);
									t12 += Qch(ll - l, ii - i)*Qb(ll, jj + 1);
									t22 += Qch(ll - l, ii - i+1)*Qb(ll, jj + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
								Qc(ii,jj + 1) += t12;
								Qc(ii + 1,jj + 1) += t22;
							} /*for (ii = i;ii < i + ilen;ii += idepth )*/

							for (ii = i + ilen;ii < i + ispan ;ii++)
							{ 
								t11 = 0.0;
								t12 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t12 += Qch(ll - l, ii - i)*Qb(ll, jj + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii,jj + 1) += t12;
							} /*for (ii = i + ilen;ii < i + ispan ;ii++)*/
						} /*for (jj = j;jj < j + jlen;jj += jdepth )*/
						for (jj = j + jlen;jj < j + jspan ;jj++)
						{ 
							for (ii = i;ii < i + ilen;ii += idepth )
							{ 
								t11 = 0.0;
								t21 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t21 += Qch(ll - l, ii - i+1)*Qb(ll, jj);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
							} /*for (ii = i;ii < i + ilen;ii += idepth )*/

							for (ii = i + ilen;ii < i + ispan ;ii++)
							{ 
								t11 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
								}
								Qc(ii,jj) += t11;
							} /*for (ii = i + ilen;ii < i + ispan ;ii++)*/
						} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
					} /*for (j = 0;j < n;j += nb )*/
				} /*for (l = 0;l < k;l += kb )*/
			} /*for (i = 0;i < m;i += mb )*/
		} /*if (!nota)*/ 
		else
		{
			/* form  c := c + alpha*a*b. */
			for (l = 0;l < k;l += kb )
			{ 
				lspan = MIN(kb, k - l);
				for (i = 0;i < m;i += mb )
				{ 
					idepth = 2;
					ispan = MIN(mb, m - i);
					ilen = idepth*(ispan/idepth);
					for (ii = i;ii < i + ispan ;ii++)
					{
						for (ll = l;ll < l + lspan ;ll++)
						{
							Qch(ll - l,ii - i) = alpha*Qa(ii, ll);
						}
					}

					for (j = 0;j < n;j += nb )
					{ 
						jdepth = 2;
						jspan = MIN(nb, n - j);
						jlen = jdepth*(jspan/jdepth);
						for (jj = j;jj < j + jlen;jj += jdepth )
						{ 
							for (ii = i;ii < i + ilen;ii += idepth )
							{ 
								t11 = 0.0;
								t21 = 0.0;
								t12 = 0.0;
								t22 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t21 += Qch(ll - l, ii - i+1)*Qb(ll, jj);
									t12 += Qch(ll - l, ii - i)*Qb(ll, jj + 1);
									t22 += Qch(ll - l, ii - i+1)*Qb(ll, jj + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
								Qc(ii,jj + 1) += t12;
								Qc(ii + 1,jj + 1) += t22;
							}

							for (ii = i + ilen;ii < i + ispan ;ii++)
							{ 
								t11 = 0.0;
								t12 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t12 += Qch(ll - l, ii - i)*Qb(ll, jj + 1);
								}
								Qc(ii,jj) += t11;
								Qc(ii,jj + 1) += t12;
							}
						}

						for (jj = j + jlen;jj < j + jspan ;jj++)
						{ 
							for (ii = i;ii < i + ilen;ii += idepth )
							{ 
								t11 = 0.0;
								t21 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{ 
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
									t21 += Qch(ll - l, ii - i+1)*Qb(ll, jj);
								}
								Qc(ii,jj) += t11;
								Qc(ii + 1,jj) += t21;
							}

							for (ii = i + ilen;ii < i + ispan ;ii++)
							{ 
								t11 = 0.0;
								for (ll = l;ll < l + lspan ;ll++)
								{
									t11 += Qch(ll - l, ii - i)*Qb(ll, jj);
								}
								Qc(ii,jj) += t11;
							}
						} /*for (jj = j + jlen;jj < j + jspan ;jj++)*/
					} /*for (j = 0;j < n;j += nb )*/
				} /*for (i = 0;i < m;i += mb )*/
			} /*for (l = 0;l < k;l += kb )*/
		}  /*if (!nota){}else{}*/
	} /*if (!notb){}else{}*/

#ifdef TESTDGEMMINPUT
	return;

  errorExit:
	xerbla("dgemm ", info);
#endif /*TESTDGEMMINPUT*/
} /*dgemm()*/



