/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.11 or later
*(C)*
*(C)* Copyright (c) 2000 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Findfact
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/


#include "globals.h"
/*
  001123 New functions related to factoring integers
  Usage:
   primefactors(n), n a vector of positive integers
      result is either a vector of factors (length(n) == 1) or
	  a structure of vectors of factors.  When length(n) > 1, the
	  names of the components are the integers being factored.
	  When n is a scalar it must be < MaxInteger = 2^52.
	  When n is a vector, n[i] must be < 10^12 = 10^NAMELENGTH
   primefactors(n,max:T), n a vector of positive integers, returns a
      vector containing the maximum prime factors of the elements of n
   goodfactors(n,p), n a vector of positive integers < MaxInteger, p > 0
      an integer,  returns a vector n1 such that n1[i] is the smallest
	  integer >= n[i] that has no prime factors > p
   goodfactors(n) is equivalent to goodfactors(n,29) 

   Cmd> primefactors(vector(3,15,64094231))
   component: 3
   (1)            3
   component: 15
   (1)            3            5
   component: 64094231
   (1)          641        99991
   Cmd> primefactors(vector(3,15,64094231),max:T)
   (1)            3            5        99991

   Actual factoring is done in factorize() in mathutil.c

  010516 Use atLeastNArgs constant on badNargs() call
*/

Symbolhandle findfactors(Symbolhandle list)
{
	Symbolhandle       arg = COMPVALUE(list,0), result = (Symbolhandle) 0;
	double             maxFactorable;
	double             maxOKfactor = 29.0; /* largest FFT can handle */
	long               nargs = NARGS(list);
	long               i, n;
	int                maxOnly = 0;
	int                primeFactors = (FUNCNAME[0] == 'p');
	WHERE("findfactors");
	
	if (nargs == 1 && arg == (Symbolhandle) 0)
	{
		badNargs(FUNCNAME, -1 + atLeastNArgs);
		goto errorExit;
	}
	if (nargs > 2)
	{
		badNargs(FUNCNAME, -2);
		goto errorExit;
	}
	if (!argOK(arg,REAL,0))
	{
		goto errorExit;
	}
	if (!isVector(arg) || anyMissing(arg))
	{
		sprintf(OUTSTR,
				"ERROR: argument to %s() is not a REAL vector with no MISSING elements",
				FUNCNAME);
		goto errorExit;
	}

	if (nargs == 2)
	{
		if (primeFactors)
		{
			Symbolhandle      symhKey = COMPVALUE(list, 1);
			char             *keyword = isKeyword(symhKey);
	
			if (keyword == (char *) 0 || strncmp(keyword ,"max", 3) != 0)
			{
				sprintf(OUTSTR,
						"ERROR: argument 2 to %s() must be 'max:T' or 'max:F'",
						FUNCNAME);
				goto errorExit;
			}
			maxOnly = DATAVALUE(symhKey, 0) != 0.0;
		}
		else
		{
			Symbolhandle     symhMaxOKfactor = COMPVALUE(list, 1);
			double           biggestfactor = floor(sqrt(MaxInteger));

			if (!isInteger(symhMaxOKfactor, POSITIVEVALUE))
			{
				char          outstr[50];
				
				sprintf(outstr, "argument 2 to %s()", FUNCNAME);
				notPositiveInteger(outstr);
				goto errorExit;
			}
			maxOKfactor = DATAVALUE(symhMaxOKfactor, 0);
			if (maxOKfactor > biggestfactor)
			{
				sprintf(OUTSTR,
						"ERROR: argument 2 to %s() > %.17g",
						FUNCNAME, biggestfactor);
				goto errorExit;
			}
		}
	} /*if (nargs == 2)*/

	n = DIMVAL(arg, 1); /* length of vector*/
	maxFactorable = (n > 1 && primeFactors) ?
		intpow(10.0, (double) NAMELENGTH) - 1.0 : MaxInteger - 1.0;
	for (i = 0; i < n; i++)
	{
		double     x = DATAVALUE(arg, i);

		if (x < 1.0 || x != floor(x) || x > maxFactorable)
		{
			sprintf(OUTSTR,
					"ERROR: argument to %s() not %s <= %.17g",FUNCNAME,
					(n == 1) ? "positive integer" : "all positive integers",
					maxFactorable);
			goto errorExit;
		}
	} /*for (i = 0; i < n; i++)*/
	
	if (maxOnly || !primeFactors)
	{
		result = RInstall(SCRATCH, n);
	}
	else if (n == 1)
	{
		result = RInstall(SCRATCH, 0);
	}
	else
	{
		result = RSInstall(SCRATCH, n, (char **) 0, 0);
	}
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	for (i = 0; i < n; i++)
	{
		Symbolhandle       symh = (n == 1 || maxOnly) ?
		  result : COMPVALUE(result, i);
		double             x = DATAVALUE(arg, i);
		double             factors[60];
		int                nfactors;

		if (primeFactors)
		{
			nfactors = factorize(x, factors);
	
			if (nfactors < 0)
			{
				/* interrupted */
				goto errorExit;
			}
			if (maxOnly)
			{
				DATAVALUE(symh, i) = (double) factors[nfactors-1];
			} /*if (maxOnly)*/
			else
			{
				double       *dFactors;
				int           j;
				
				TMPHANDLE = mygethandle(nfactors * sizeof(double));
				setDATA(symh, (double **) TMPHANDLE);
				if (TMPHANDLE == (char **) 0)
				{
					goto errorExit;
				}
				DIMVAL(symh, 1) = nfactors;
				dFactors = DATAPTR(symh);
				for (j = 0; j < nfactors; j++)
				{
					dFactors[j] = factors[j];
				}
	
				if (n > 1)
				{
					mvSymbolName        nameOfNumber;
	
					sprintf((char *) nameOfNumber, "%.17g", x);
					Setname(symh, nameOfNumber);
				}
			} /*if (maxOnly){}else{}*/
		} /*if (primeFactors)*/
		else
		{
			do
			{
				nfactors = factorize(x++,factors);
				if (nfactors < 0)
				{
					goto errorExit;
				}
			} while (factors[nfactors-1] > maxOKfactor);
			x--;
			DATAVALUE(result, i) = x;
		}
	} /*for (i = 0; i < n; i++)*/
	return (result);

  errorExit:
	putErrorOUTSTR();
	Removesymbol(result);

	return (0);
} /*findfactors()*/
