/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2002 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#ifndef GLMH__
#define GLMH__
/*
  Defines and global declarations related to glm computations
  930811 Modified so that large globals such as MODELINFO are allocated
	dynamically rather than taking up part of the global segment;  this was
	because the Mac version ran over 32K globals
  Appropriate changes where made in the glm modules.

  970819 added macro VARNAME
  980710 add macro BinomClamp
  980804 Renamed BINOMCLAMP and BinomClamp to BINOMPLIMIT and BinomPlimit
  990421 Added global STRMODELNAME
  990619 Changed so that MAXERRTRMS can be set in a platform file
         Removed include "typedefs.h"
  000209 Add global TERMNAMES_AT which includes '@' in temporary variable names
  010512 Added enum variableNameCodes
  020318 Major modifications to allow for stack of model results
  021101 Added enum getAlphaCodes with codes previously defined in
         anovacoe.c and glmutils.c
*/


#undef EXTERN
#undef INIT
#undef INITDIM
#undef INITARRAY
#undef COMMA

#ifdef MAIN__
#define EXTERN
#define INIT(VAL) = VAL
#define INITDIM(N) N
#define INITARRAY(VALS) = { VALS }
#else
#define EXTERN extern
#define INIT(VAL)
#define INITDIM(N)
#define INITARRAY(VALS)
#endif /*MAIN__*/

#define COMMA ,


/*
   Options related to GLMs
   Their default values can be changed in Makefile if desired.
*/

#ifndef DEFAULTPRINTFSTATS
#define DEFAULTPRINTFSTATS   0
#endif /*DEFAULTPRINTFSTATS*/

#ifndef DEFAULTPRINTPVALS
#define DEFAULTPRINTPVALS    0
#endif /*DEFAULTPRINTPVALS*/

EXTERN long       PRINTFSTATS INIT(DEFAULTPRINTFSTATS);
EXTERN long       PRINTPVALS  INIT(DEFAULTPRINTPVALS);

/*
   Constants and globals affecting algorithms
*/

#ifndef REGTOLVAL
#define REGTOLVAL     1.0e-12  /* used in singularity check in gramschmidt */
#endif /*REGTOLVAL*/
EXTERN double  REGTOL INIT(REGTOLVAL); /* made a global 951012*/

/* maximum number of iterations and tolerance for iterative functions */
#define FASTANOVAMAXITER  25
#define FASTANOVAEPSILON  1e-6
#define ITERGLMMAXITER    50
#define ITERGLMEPSILON    1e-6
#define IPFMAXITER        25
#define IPFEPSILON        1e-6

/* Limits related to model representation (formerly in globdefs.h) */
#ifndef WORDSPERTERM
#define WORDSPERTERM  3       /* number of words in a term */
#endif /*WORDSPERTERM*/

#ifndef VARSPERWORD
#define VARSPERWORD  32     /* number of variables per word */
#endif /*VARSPERWORD*/

#if     WORDSPERTERM == 1
#define MAXVARS    VARSPERWORD		/* max number of variables in model */
#else
#define MAXVARS    (WORDSPERTERM*VARSPERWORD)		/* max number of variables in model */
#endif  /*WORDSPERTERM == 1 */

#define OLDMAXVARS   31
#define OLDMAXERRTRMS 10   /* former max number of error terms */
#ifndef MAXERRTRMS
#define MAXERRTRMS    OLDMAXERRTRMS  /* max number of error terms */
#endif /*MAXERRTRMS*/

#define MAXFACTORLEVELS 32767 /* maximum number of levels in a factor */

/* Default maximum depth of model stack used by pushmodel() and popmodel() */
#ifndef MAXMODELSTACKDEPTH /*can be set in Makefile */
#ifndef BCPP
#define MAXMODELSTACKDEPTH 3
#else /*BCPP*/
#define MAXMODELSTACKDEPTH 1
#endif /*BCPP*/
#endif /*MAXMODELSTACKDEPTH*/


/*
  Model related type definitions
*/
typedef unsigned long termWord;
typedef termWord      modelType[WORDSPERTERM];
typedef modelType    *modelPointer, **modelHandle;

typedef struct       modelInfo
{
	long             nvars;
	long             nfactors;
	long             nterms;
	long             nerrorterms;
	modelHandle      model;
	modelType        errorterms[MAXERRTRMS+1];
	Symbolhandle     modelvars[MAXVARS+1];
	long             nclasses[MAXVARS];
} modelInfo, *modelInfoPointer, **modelInfoHandle;

/*
  Pseudo prototypes
  modelPointer        modelterm(modelHandle, long);
  void                zeroTerm(modelPointer);
*/
#define modelterm(M,I) (((modelPointer) *(M)) + (I)) /* type modelPointer */
#define zeroTerm(T)  modeltermAssign((modelPointer) NULLTERM, (modelPointer) T)

#define nSpecialConstants  10 /*length of glmModelInfo.specialConstants*/
typedef struct       glmModelInfo
{
	modelInfo        currentModel;
	mvSymbolName     varNames[MAXVARS+1];
	char           **strmodel;
	char           **depvName;
	char           **termNames;
	char           **termNames_at;
	long             ndata;
	long             notmissing;
	long             ny;
	long             nregx;
	long             modeltype;
	unsigned long    glmcontrol;
	double         **residuals;
	double         **wtdresiduals;
	double         **hii;
	double         **ss;
	double         **df;
	double         **misswts;
	double         **casewts;
	double         **iterwts;
	double         **allwts;
	double         **logitn;
	double         **glmoffset;
	double         **regx;
	double         **xtxinv;
	double         **regcoef;
	double         **regss;
	double           specialConstants[nSpecialConstants];
	char             incremental; /*incremental deviance flag*/
	char             useGlmOffset; /*flag to use offset*/
} glmModelInfo, *glmModelInfoPointer, **glmModelInfoHandle;


/*
  The following are initialized by glmInit()
  ModelInfoStack is pointer to bottom of stack of models.  It is equivalent
  to an array that could be declared as glmModelInfo[MaxModelStackDepth]
  
  ModelStackDepth is an integer >= 1 such that
  ModelInfoStack[ModelStackDepth-1] contains information on the current active
  model
  CurrentModelInfo is pointer to ModelInfoStack[ModelStackDepth-1]
*/
EXTERN int                    MaxModelStackDepth INIT(MAXMODELSTACKDEPTH);
EXTERN glmModelInfoPointer    ModelInfoStack INIT((glmModelInfoPointer) 0);
EXTERN glmModelInfoPointer    CurrentModelInfo INIT((glmModelInfoPointer) 0);
EXTERN int                    ModelStackDepth INIT(0);

/* Global NULLTERM */
#if    WORDSPERTERM == 1
EXTERN modelType       NULLTERM INITARRAY(0);
#elif  WORDSPERTERM == 2
EXTERN modelType       NULLTERM INITARRAY(0 COMMA 0);
#else  /*WORDSPERTERM == 3*/
EXTERN modelType       NULLTERM INITARRAY(0 COMMA 0 COMMA 0);
#endif /*WORDSPERTERM == 3*/
/*
  Global variables related to doing glm
  020321 They are all members of the glmModelInfo structure pointed to
         by CurrentModelInfo, initialized in glmInit().  A stack of
         glmModelInfo structures is allocated and initialized in glmInit()
         and CurrentModelInfo points to one of its elements.

         Individual variables are all accessed using macros so as to
         not break existing code.
         CurrentModelInfo and the space it points to are initialized
         in glmInit()
*/

#define MODELINFO      (&CurrentModelInfo->currentModel)

#if (0) /*superseded declaration; now access through macro*/
EXTERN modelInfoPointer MODELINFO; /* previously global space was allocated */
#endif

/* Matrix of variable names in model*/
#define VARNAMES (CurrentModelInfo->varNames)

#if (0) /*superseded declaration; now accessed through macro*/
EXTERN char           (*VARNAMES)[NAMELENGTH+1];
#endif /*0*/

#if (0)   /* even older declaration allocating global spaced */
EXTERN char            VARNAMES[INITDIM(MAXVARS+1)][NAMELENGTH+1];
#endif /*0*/

/* find pointer to name or to expression computing variable */
#define VARNAME(I) ((strcmp(VARNAMES[I], EVALSCRATCH) != 0) ? \
					VARNAMES[I] : LABELSPTR(MODELVARS[I]))

/* handle to modelTypes indicating which order to enter terms in model */
#define MODEL          (MODELINFO->model)

#if (0) /* superseded declaration; now accessed through macro */
EXTERN modelHandle     MODEL INIT((modelHandle) 0);
#endif /*0*/

#define  INCREMENTAL   (CurrentModelInfo->incremental)
#define  USEGLMOFFSET  (CurrentModelInfo->useGlmOffset)

#if (0) /*superseded declarations; now accessed through macros*/
EXTERN long            INCREMENTAL INIT(0);/*  incremental deviance flag */
EXTERN long	           USEGLMOFFSET INIT(0); /* flag for using glm offsets */
#endif

/* handles to variables used in model, 0 is y */
#define MODELVARS      (MODELINFO->modelvars)

#if (0) /*superseded declaration; now accessed through macro*/
EXTERN Symbolhandle    MODELVARS[INITDIM(MAXVARS+1)];
#endif /*0*/

/*
  020320 These are now all longs; previously equivalent globals were doubles
*/
#define                NVARS (MODELINFO->nvars)
#define                NFACTORS (MODELINFO->nfactors)
#define                NDATA (CurrentModelInfo->ndata)
#define                NOTMISSING (CurrentModelInfo->notmissing)
#define                NTERMS (MODELINFO->nterms)

#if (0) /* superseded declarations; now accessed through macros */
EXTERN double          NVARS INIT(-1);	/* number of variables in model */
EXTERN double          NFACTORS INIT(-1); /* number of factors in model */
EXTERN double          NDATA INIT(0);	/* number of data values */
EXTERN double          NOTMISSING INIT(0);	/* number of nonmissing data */
EXTERN double          NTERMS INIT(0);	/* number of terms in model */
#endif

/* number of classes in each model variable */
#define NCLASSES       (MODELINFO->nclasses)

#if (0) /*superceded declaration; now accessed through macro*/
EXTERN long            NCLASSES[INITDIM(MAXVARS)];
#endif /*0*/

#define NREGX          (CurrentModelInfo->nregx)
#define RESIDUALS      (CurrentModelInfo->residuals)
#define WTDRESIDUALS   (CurrentModelInfo->wtdresiduals)
#define HII            (CurrentModelInfo->hii)

#if (0) /*superseded declarations; now all accessed through macros*/
EXTERN double          NREGX; /* number of columns in regression matrix */
EXTERN double        **RESIDUALS INIT(0);	/* handle to residuals */
EXTERN double        **WTDRESIDUALS INIT(0);	/* handle to residuals */
EXTERN double        **HII INIT(0);	/* handle to leverages */
#endif /*0*/

/* Note: Y should always be identical to DATA(MODELVARS[0]) */
EXTERN double        **Y INIT(0);		/* handle to y data */

/* NY = number of columns in Y; now a long; previously a double*/
#define NY             (CurrentModelInfo->ny)

#if (0) /*superseded declaration; now accessed through a macro*/
EXTERN double          NY INIT(0);		/* number of Y's (for manova) */
#endif /*0*/

/* Note: X[i] should always be identical to DATA(MODELVARS[i+1]) */
EXTERN double       ***X; /*initialized in glmInit()*/

#if (0) /*older declaration directly allocating global space*/
EXTERN double        **X[INITDIM(MAXVARS)];	/* array of handles to x columns */
#endif /* 0 */

/*
  matrices or vectors
*/
#define SS             (CurrentModelInfo->ss)
#define DF             (CurrentModelInfo->df)
#define MISSWTS        (CurrentModelInfo->misswts)
#define CASEWTS        (CurrentModelInfo->casewts)
#define ITERWTS        (CurrentModelInfo->iterwts)
#define ALLWTS         (CurrentModelInfo->allwts)
#define LOGITN         (CurrentModelInfo->logitn)
#define GLMOFFSET      (CurrentModelInfo->glmoffset)
/*
  scalars
*/
#define MODELTYPE      (CurrentModelInfo->modeltype)
#define GLMCONTROL     (CurrentModelInfo->glmcontrol)

#if (0) /*superseded declarations; now all accessed through macros*/
EXTERN double        **SS INIT(0);	/* handle for model term sums of squares */
EXTERN double        **DF INIT(0);		/* handle for model term dfs */
EXTERN double        **MISSWTS INIT(0);	/* handle for missing value weights */
EXTERN double        **CASEWTS INIT(0);
	/* handle for user weights in linear models */
EXTERN double        **ITERWTS INIT(0);	/* handle for iterative weights */
EXTERN double        **ALLWTS INIT(0);	/* combined weights */
EXTERN double        **LOGITN INIT(0);	/* counts for logistic N */
EXTERN double	     **GLMOFFSET INIT(0); 	/* offsets in a glm model */
EXTERN long            MODELTYPE INIT(0);	/* indicator for model type */
EXTERN unsigned long   GLMCONTROL  INIT(0);
#endif /*0*/

/*
  The following should be the same as MODELTYPE, GLMCONTROL and
  GLMSCALE once a new GLM computatation has completed;  they are
  restored from MODELTYPE, GLMCONTROL and GLMSCALE when GLM
  information is popped from the stack
*/
EXTERN long            PREVMODELTYPE INIT(0); /* previous model type */
EXTERN unsigned long   PREVGLMCONTROL  INIT(0);
EXTERN double          PREVGLMSCALE    INIT(1.0);

/*
   Bits in MODELTYPE and PREVMODELTYPE
				0  means no current model (NOMODEL)
				1  means leaps screening (LEAPS)
				2  means anova (ANOVA)
				4  means fast anova (FASTANOVA) 
				8  means manova (MANOVA)
			   16  means standard regression (OLSREG)
               32  means robust regression (ROBUSTREG)
			   64  means general glm regression (GLMREG)
			  128  means logistic regression (LOGITREG)
			  256  means Poisson regression (POISSONREG)
              512  means Probit regression (PROBITREG)
			 1024  means case weights (CASEWEIGHTS)
			 2048  means missing value weights (MISSWEIGHTS)
*/

#define GlmGlobals   (CurrentModelInfo->specialConstants)

/* Constants and macros related to robust()*/
#define   ROBTRUNCATION     .75 /*default value Huber truncation point */

#define   CURRENTRSC      GlmGlobals[0] /*current robust scale factor*/
#define   RobTruncation   GlmGlobals[1] /* Huber truncation point */
/* RobBeta = cumchi(c*c,3) + 2*c*c*(1-cumnor(c)), c = RobTruncation*/
#define   RobBeta         GlmGlobals[2]
#define   ROBUSTVARK      GlmGlobals[3] /*K from page 40 of Huber 1977 */
#define   REGDIM          GlmGlobals[4] /* degrees of freedom in model */
#define   GLMSCALE        GlmGlobals[0]
#define   GLMPOWER        GlmGlobals[1]

/* constant related to logistic() and probit()*/
#define   BINOMPLIMIT      1e-8
#define   BinomPlimit      GlmGlobals[2]
#define   HitPlimit        GlmGlobals[3]

EXTERN char            STRMODELNAME[] INIT("STRMODEL");

EXTERN char           *RUNNINGGLM INIT(0); /*set != 0 at start of glm*/

#define STRMODEL       (CurrentModelInfo->strmodel)

#if (0) /*superseded declaration; access through macro*/
EXTERN char          **STRMODEL INIT(0);	/* string form of inputted model */
#endif /*0*/

/* ERRORTERMS is initialized in glmInit() */
#define ERRORTERMS     (MODELINFO->errorterms)

#if (0) /*superseded declaration; access through macro*/
EXTERN modelPointer    ERRORTERMS INIT((modelPointer) 0);
#endif /*0*/

#if (0)
EXTERN modelPointer    ERRORTERMS INIT((modelPointer) MODELINFO->errorterms);
#endif /*0*/

		/* terms used as errors */
#define NERRORTERMS    MODELINFO->nerrorterms
#define DEPVNAME       (CurrentModelInfo->depvName)
#define TERMNAMES      (CurrentModelInfo->termNames)
#define TERMNAMES_AT   (CurrentModelInfo->termNames_at)
#define REGX           (CurrentModelInfo->regx)
#define XTXINV         (CurrentModelInfo->xtxinv)
#define REGCOEF        (CurrentModelInfo->regcoef)
#define REGSS          (CurrentModelInfo->regss)

#if (0) /*superseded declarations; access through macros */
EXTERN double          NERRORTERMS;	/* number of error terms */
EXTERN char          **DEPVNAME INIT(0);    /* name of dependent variable */
EXTERN char          **TERMNAMES INIT(0);	/* names of terms */
EXTERN char          **TERMNAMES_AT INIT(0); /* full names of terms */
EXTERN double        **REGX INIT(0);	/* x matrix for regression */
#endif /*0*/

EXTERN double        **REGX2 INIT(0);	/* x matrix used only for scratch*/

#if (0) /*superseded declarations */
EXTERN double        **XTXINV INIT(0);	/* xtxinv matrix for regression */
EXTERN double        **REGCOEF INIT(0);	/* regression coefficients */
EXTERN double        **REGSS INIT(0);	/* sequential ss for regression */
#endif /*0*/

/* constants related to glm's */

enum variableNameCodes
{
	fullVariableName = 0x00,
	noBrackets       = 0x01,
	noTempChar       = 0x02
};

#define MAXGLMEXPRESSION  100

/*
  Codes associated with side effect variables
  The only value currently used is nSideEffectVars
*/
enum glmSideEffectCodes
{
	sideEffectRESIDUALS     = 0,
	sideEffectWTDRESIDUALS,
	sideEffectHII,
	sideEffectSS,
	sideEffectDF,
	sideEffectDEPVNAME,
	sideEffectTERMNAMES,
	sideEffectXTXINV,
	sideEffectCOEF,
	nSideEffectVars /*number of possible side effect variables*/
};

/* GLM side effect variable names*/
EXTERN mvSymbolName    SideEffectNames[INITDIM(nSideEffectVars)]
	INITARRAY(
		"RESIDUALS"    COMMA
		"WTDRESIDUALS" COMMA
		"HII"          COMMA
		"SS"           COMMA
		"DF"           COMMA
		"DEPVNAME"     COMMA
		"TERMNAMES"    COMMA
		"XTXINV"       COMMA
		"COEF"
	);


/*
  No variables are declared after this point
*/

/* GLM Codes used in MODELTYPE and PREVMODELTYPE*/
/* special cases, different algorithms */
#define NOMODEL      0x0000UL  /* 0 no current model */
#define LEAPS        0x0001UL  /* 1 screening regression*/
#define ANOVA        0x0002UL  /* 2 anova, balanced is special case */
#define FASTANOVA    0x0004UL  /* 4 iterative fast anova via EM */

/* single pass gram schmidt models */
#define MANOVA       0x0008UL  /* 8 manova */
#define OLSREG       0x0010UL  /* 16 standard regression */

/* multipass iterative gram schmidt models */
#define ROBUSTREG    0x0020UL  /* 32 robust regression */
#define GLMREG       0x0040UL  /* 64 general glm regression */
#define LOGITREG     0x0080UL  /* 128 logistic regression */
#define POISSONREG   0x0100UL  /* 256 poisson regression */
#define PROBITREG    0x0200UL  /* 512 probit regression */
#define ITERGLMS     0x03e0UL  /* ROBUSTREG|GLMREG|LOGITREG|POISSONREG|PROBITREG */
#define BINOMREG     0x0280UL  /* LOGITREG|PROBITREG*/

/* weight indicators */
#define CASEWEIGHTS  0x0400UL  /* 1024 user specified case weights */
#define MISSWEIGHTS  0x0800UL  /* 2048 missing values, so weights needed */
#define USEWEIGHTS   0x0fe0UL  /* ITERGLMS | CASEWEIGHTS | MISSWEIGHTS */

#define IPF          0x1000UL  /* 4096 iterative proportional fitting */

#define FSTATSOK     (ANOVA | FASTANOVA | MANOVA)
#define ALLGLMS      (FSTATSOK | OLSREG | ITERGLMS | LEAPS | IPF)

/* GLM Codes used in GLMCONTROL and PREVGLMCONTROL*/
 /* response distribution codes */
#define NORMALDIST    0x000001UL /* normal distribution */
#define BINOMDIST     0x000002UL /* binomial distribution */
#define POISSONDIST   0x000004UL /* poisson distribution */
#define GAMMADIST     0x000008UL /* gamma distribution */
#define NONNORMALDIST 0x00000eUL /* BINOMDIST | POISSONDIST | GAMMADIST */
#define ANYDIST       0x00000fUL /* NONNORMALDIST | NORMALDIST */
 /* link specification codes */
#define IDENTLINK     0x000100UL /* identity link */
#define LOGITLINK     0x000200UL /* logit link */
#define LOGLINK       0x000400UL /* natural log link */
#define PROBITLINK    0x000800UL /* invnor link */
#define SQRTLINK      0x001000UL /* square root link */
#define RECIPLINK     0x002000UL /* reciprocal link */
#define POWERLINK     0x004000UL /* power link */
#define CLOGLOGLINK   0x008000UL /* complementary log-log link */
#define EXPLINK       0x010000UL /* exponential link */
#define NONIDENTLINK  0x01fe00UL /* LOGITLINK | LOGLINK | ... | EXPLINK */
#define ANYLINK       0x01ff00UL /* NONIDENTLINK | IDENTLINK */
 /* other codes for GLMCONTROL */
#define UNBALANCED    0x100000UL /* use gramschmidt */
#define MULTIVAR      0x200000UL /* multivariate response */
#define MARGINALSS    0x400000UL /* marginal (type III) SS */

/* op codes for cellstat() */
enum cellStatCodes
{
	CELLSTATS  = 1,
	CELLSUMS,
	CELLCOUNTS,
	CELLMEANSANDSWEEP
};

/* codes used when accessing coefficients */
enum getAlphaCodes
{
	ALIASED     = -1, /* column is intrinsically aliased */
	PIVOTED     = -2, /* column is aliased, but not intrinsically */
	UNTRANSPOSED = 0, /* don't transpose effect matrix */
	TRANSPOSED   = 1  /* transpose effect matrix */
};

/* opcodes for gramschmidt() and/or itergs() */
#define DONTNOTIFY   0x000000
#define DONTDOSWEEP  0x000000
#define DONTDOCOEFS  0x000000
#define DONTDORINV   0x000000
#define DONTDOSS     0x000000

#define DONOTIFY     0x010000  /* warn if singular column found */
#define DOSWEEP      0x020000  /* actually do orthogonalization */
#define DOCOEFS      0x040000  /* compute coefficients */
#define DORINV       0x080000  /* compute R^(-1) in QR decomposition */
#define DOSS         0x100000  /* compute SS or SSCP */

/* opcodes for iterfns.c functions */
#define INITIAL      1L
#define STEP         2L
#define FINAL        4L


#endif /*GLMH_*/
