/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2003 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Iswhat
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "globals.h"

#include "mvIface.h"
#include "mvGraphs.h"

/*
  Operations isreal(), ischar(), islogic(), isvector(), ismatrix(),
  isscalar(), isfactor(), isnull(), ismacro(), isname()

  Each returns a logical vector of length NARGS(list)

  950821 isscalar(), isvector(), and ismatrix() all recognize keyword phrases
  real:T, char:T, and logic:T.

  980317 added isarray() which also recognizes real, char and logic

  990113 OK to repeat properties (e.g., vector("real","real")) on argvalue()
         and keyvalue()

  990226 Replaced putOUTSTR() by putErrorOUTSTR()

  990325 Added implemntation of isname()

  990522 Added implementation of isfunction()
  
  000814 Added keywords 'positive', 'negative', 'nonnegative', 'integer' to
         isscalar(), isvector(), ismatrix() and isreal()
  010210 Fixed bug in keyvalue()
         keyvalue(,name,"macro",default:macroArg) and
         keyvalue(,name,"graph",default:graphArg)
         gave error that the default was not the right type, even when it was
  010503 Add implementation of function isnumber()
  010726 Add implementation of function islocked()
*/

/*
  isVariableName(name) returns 1 if and only if name is a legal, possibly
  temporary, MacAnova name.  "NULL", "F", and "T" are not legal

  At some point this may go into utils.c
*/
static int isVariableName(char * name)
{
	int           nameOK = 0;

	if (strlen(name) <= NAMELENGTH &&
		(strcmp(name, NULLNAME) != 0 &&
		 (name[1] != '\0' || name[0] != 'T' && name[0] != 'F')))
	{
		if (istempname(name))
		{
			name++;
		}
		nameOK = isnamestart(*name);

		while (nameOK && *(++name) != '\0')
		{
			nameOK = isnamechar(*name);
		}
	}
	return (nameOK);
} /*isVariableName()*/

/*
  isNumberName(name) returns 1 if and only if name is a string that
  represents a single non-MISSING number, possibly padded before and
  after with white space.

  At some point this may go into utils.c
*/
static int isNumberName(char * name)
{
	char      *eptr;
	double     temp = mystrtod(name, &eptr);
	int        value;

	if (eptr == name || temp == TOOBIGVALUE)
	{
		value = 0;
	}
	else
	{
		while (*eptr != '\0' && isspace(*eptr))
		{
			eptr++;
		}
		value = (*eptr == '\0') ? 1 : 0;
	}
	return (value);
} /*isNumberName()*/

/*
  doIsname() implements MacAnova function isname()

  isname(arg1, arg2, ..., argk) where each argument is either missing
  or a CHARACTER scalar, returns a LOGICAL vector of length nargs.  Element
  k of the result is T is and only if argument k is a CHARACTER scalar whose
  value is a legal MacAnova name.  Values NULL, T and F are explicitly
  checked for 

  990235
*/
static Symbolhandle doIsname(Symbolhandle list)
{
	Symbolhandle        arg, result = SymbolZero;
	long                nargs = NARGS(list);
	long                nargs1 = 0;
	long                iarg, iarg1;
	int                 file = 0, path = 0, variable;
	int                 number = (strcmp(FUNCNAME, "isnumber") == 0);
	char               *keyword;
	
	for (iarg = 0; iarg < nargs; iarg++)
	{
		arg = COMPVALUE(list, iarg);
		keyword = isKeyword(arg);
		if (keyword != (char *) 0)
		{
			int      logicValue;
			
			if (number ||
				strcmp(keyword, "file") != 0 && strcmp(keyword,"path") != 0)
			{
				badKeyword(FUNCNAME, keyword);
				goto errorExit;
			}
			if (!isTorF(arg))
			{
				notTorF(keyword);
				goto errorExit;
			}
			logicValue = (DATAVALUE(arg, 0) != 0.0);
			if (keyword[0] == 'f')
			{
				file = logicValue;
			}
			else
			{
				path = logicValue;
			}
		} /*if (keyword != (char *) 0)*/
		else if (arg != SymbolZero && !isCharOrString(arg))
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s() is not CHARACTER scalar",
					iarg + 1, FUNCNAME);
			goto errorExit;
		}
		else
		{
			nargs1++;
		}
	} /*for (iarg = 0; iarg < nargs; iarg++)*/

	if (!number)
	{
		if (file && path)
		{
			sprintf(OUTSTR,
					"ERROR: illegal to use both 'file:T' and 'path:T' in %s()",
					FUNCNAME);
			goto errorExit;
		} /*if (file && path)*/
	} /*if (nameCheck)*/	
	variable = !number && !file && !path;

	result = RInstall(SCRATCH, nargs1);
	if (result == SymbolError)
	{
		goto errorExit;
	}
	setTYPE(result, LOGIC);
	for (iarg = 0, iarg1 = 0; iarg < nargs; iarg++)
	{
		
		arg = COMPVALUE(list, iarg);
		if (!isKeyword(arg))
		{
			double     value;
			
			if (arg != SymbolZero)
			{
				char      *name = STRINGPTR(arg);
				
				if (number && !isNumberName(name) ||
					variable && !isVariableName(name) ||
					file && !okfilename(name) ||
					path && !okpathname(name))
				{
					value = 0.0;
				}
				else
				{
					value = 1.0;
				}
			} /*if (arg != SymbolZero)*/
			else
			{
				value = 0.0;
			}
			DATAVALUE(result, iarg1) = value;
			iarg1++;
		} /*if (!isKeyword(arg))*/
	} /*for (iarg = 0, iarg1 = 0; iarg < nargs; iarg++)*/
	
	return (result);
	
  errorExit:
	putErrorOUTSTR();
	return (0);
} /*doIsname()*/

enum typeKeyCodes
{
	IREAL = 0,
	ILOGIC,
	ICHAR,
	NTYPEKEYS
};

enum valueKeyCodes
{
	IPOSITIVE,
	INONNEGATIVE,
	INEGATIVE,
	IINTEGER,
	NVALUEKEYS
};

static char     *TypeKeys[NTYPEKEYS] =
{
	"real", "logic", "char"
};
static char     *ValueKeys[NVALUEKEYS] =
{
	"positive", "nonneg", "negative", "integer"
};

#define CheckReal         okTypes[IREAL]
#define CheckLogic        okTypes[ILOGIC]
#define CheckChar         okTypes[ICHAR]

#define CheckPositive     okValues[IPOSITIVE]
#define CheckNonnegative  okValues[INONNEGATIVE]
#define CheckNegative     okValues[INEGATIVE]
#define CheckInteger      okValues[IINTEGER]

Symbolhandle iswhat(Symbolhandle list)
{
	long               nargs = NARGS(list);
	long               ntoCheck = nargs;
	long               i, jvalue, iarg;
	Symbolhandle       symh, result = SymbolZero;
	Symbolhandle       symhKey;
	char              *what = FUNCNAME + 2;
	long               keyIndex, checkType = 0, checkValue = 0;
	short              okTypes[NTYPEKEYS], okValues[NVALUEKEYS];
	int                shapeCheck, typeCheck;
	WHERE("iswhat");

	OUTSTR[0] = '\0';

	if (strcmp(what, "name") == 0 || strcmp(what, "number") == 0)
	{
		return (doIsname(list));
	}
	
	shapeCheck = strcmp(what, "scalar") == 0 || strcmp(what, "vector") == 0 ||
						strcmp(what, "matrix") == 0 ||
						strcmp(what, "array") == 0;

	typeCheck = strcmp(what, TypeKeys[IREAL]) == 0 ||
	  strcmp(what, TypeKeys[ICHAR]) == 0 ||
	  strcmp(what, TypeKeys[ILOGIC]) == 0;
	
	for (i = 0; i < NTYPEKEYS; i++)
	{
		okTypes[i] = -1;
	}
	for (i = 0; i < NVALUEKEYS; i++)
	{
		okValues[i] = 0;
	}

	if (shapeCheck)
	{
		/*
			isscalar(), isvector(), ismatrix() or isarray()
			All types, REAL, LOGIC, or CHARACTER are ok unless any of the
			real:T, logic:T, or char:T in which case only types
			that are explicitly indicated by, e.g., char:T, are ok.

			Check for type keywords
		*/

		for (i = 0; i < NTYPEKEYS; i++)
		{
			char         *keyword = TypeKeys[i];

			if ((keyIndex = findKeyword(list, keyword, 1)) >= 0)
			{
				symhKey = COMPVALUE(list,keyIndex);
				if (!isTorF(symhKey))
				{
					notTorF(keyword);
					goto errorExit;
				}
				okTypes[i] = (DATAVALUE(symhKey,0) != 0.0);
				setNAME(symhKey, USEDKEYWORD);
				ntoCheck--;
				checkType = 1;
			} /*if ((keyIndex = findKeyword(list, keyword, 1)) >= 0)*/
		} /*for (i = 0; i < NTYPEKEYS; i++)*/
		
		if (checkType)
		{/* some TypeKey keyword found */
			for (i = 0; i < NTYPEKEYS; i++)
			{ /* set unset types to false */
				okTypes[i] = (okTypes[i] < 0) ? 0 : okTypes[i];
			} /*for (i = 0; i < NTYPEKEYS; i++)*/

			if (okTypes[IREAL] + okTypes[ILOGIC] + okTypes[ICHAR] == 0)
			{
				sprintf(OUTSTR,
						"ERROR: no acceptable types specified for %s()",
						FUNCNAME);
				goto errorExit;
			} /*if (okTypes[IREAL] + okTypes[ILOGIC] + okTypes[ICHAR] == 0)*/
		} /*if (checkType)*/
	} /*if (shapeCheck)*/
	else if (typeCheck)
	{
		/*
			isreal(), islogic(), is(char()
		*/
		checkType = 1;
		for (i = 0; i < NTYPEKEYS; i++)
		{
			okTypes[i] = strcmp(what, TypeKeys[i]) == 0;
		}
	}

	if (shapeCheck || checkType)
	{
		/* Check for positive:T, nonneg:T, negative:T, integer:T */
		for (i = 0; i < NVALUEKEYS; i++)
		{
			char          *keyword = ValueKeys[i];

			if ((keyIndex = findKeyword(list, keyword, 1)) >= 0)
			{
				symhKey = COMPVALUE(list, keyIndex);
				if (!isTorF(symhKey))
				{
					notTorF(keyword);
					goto errorExit;
				}
				okValues[i] = (DATAVALUE(symhKey,0) != 0.0);
				if (!okValues[i])
				{
					sprintf(OUTSTR, "ERROR: %s:F is illegal on %s()",
							keyword, FUNCNAME);
					goto errorExit;
				}
				setNAME(symhKey, USEDKEYWORD);
				ntoCheck--;
				checkValue = 1;
			} /*if ((keyIndex = findKeyword(list, keyword, 1)) >= 0)*/
		} /*for (i = 0; i < NTYPEKEYS; i++)*/

		if (checkValue)
		{
			if (okTypes[ICHAR] > 0 || okTypes[ILOGIC] > 0 || !okTypes[IREAL])
			{
				char    *illegalCheck = "ERROR: illegal check of values with";
				
				if (!typeCheck)
				{
					sprintf(OUTSTR,
							"%s '%s:T', '%s:T' or '%s:F'", illegalCheck,
							TypeKeys[ICHAR],TypeKeys[ILOGIC],TypeKeys[IREAL]);
				}
				else
				{
					sprintf(OUTSTR, "%s %s()", illegalCheck, FUNCNAME);
				}
				goto errorExit;
			} /*if (okTypes[ICHAR] > 0||okTypes[ILOGIC] > 0||!okTypes[IREAL])*/
			checkType = okTypes[IREAL] = 1;
			okTypes[ICHAR] = okTypes[ILOGIC] = 0;
	
			if (okValues[IPOSITIVE] + okValues[INEGATIVE] +
					okValues[INONNEGATIVE] > 1)
			{
				sprintf(OUTSTR,
						"ERROR: only one of %s:T, %s:T and %s:T may be specified",
						ValueKeys[IPOSITIVE], ValueKeys[INEGATIVE], ValueKeys[INONNEGATIVE]);
				goto errorExit;
			}
		} /*if (checkValue)*/
	} /*if (shapeCheck || checkType)*/

	if (ntoCheck <= 0)
	{
		sprintf(OUTSTR,
				"ERROR: no arguments for %s() to check", FUNCNAME);
		goto errorExit;
	}

	result = RInstall(SCRATCH, ntoCheck);
	if (result == SymbolError)
	{
		goto errorExit;
	}
	setTYPE(result, LOGIC);

	jvalue = 0;
	for (iarg = 0;iarg < nargs;iarg++)
	{
		symh = COMPVALUE(list, iarg);
		if (symh == EmptyArgument || strcmp(NAME(symh), USEDKEYWORD) != 0)
		{
			int         value = symh != EmptyArgument;
			
			if (value)
			{
				long         type = TYPE(symh);
	
				if (typeCheck)
				{
					if (okTypes[IREAL])
					{
						value = (type == REAL);
					}
					else if (okTypes[ICHAR])
					{
						value = (type == CHAR);
					}
					else
					{
						value = (type == LOGIC);
					}
				}
				else if (shapeCheck)
				{
					if (strcmp(what, "scalar") == 0)
					{
						value = (isScalar(symh) != 0);
					}
					else if (strcmp(what, "vector") == 0)
					{
						value = (isVector(symh) != 0);
					}
					else if (strcmp(what, "matrix") == 0)
					{
						long               dims[2];

						value = (isMatrix(symh, dims) != 0);
					}
					else
					{
						value = (type == CHAR || type == REAL || type == LOGIC ||
								 type == LONG);
					}
				}
				else if (strcmp(what, "struc") == 0)
				{
					value = (type == STRUC);
				}
				else if (strcmp(what, "macro") == 0)
				{
					value = (type == MACRO);
				}
				else if (strcmp(what, "graph") == 0)
				{
					value = (type == PLOTINFO);
				}
				else if (strcmp(what, "function") == 0)
				{
					value = (type == BLTIN);
				}
				else if (strcmp(what, "factor") == 0)
				{
					value = (isFactor(symh) > 0);
				}
				else if (strcmp(what, "null") == 0)
				{
					value = (isDefined(symh) != 0 && TYPE(symh) == NULLSYM);
				}
				else if (strcmp(what, "defined") == 0)
				{
					value = (isDefined(symh) != 0);
				}
				else if (strcmp(what, "locked") == 0)
				{
					value = (isLocked(symh) != 0);
				}
				else
				{ /* should not happen*/
					sprintf(OUTSTR, "ERROR: ****programming error for operation %s", 
							FUNCNAME);
					Delete(result);
					goto errorExit;
				}
	
				if (value && checkType)
				{
					value = (CheckReal || type != REAL) &&
							(CheckLogic || type != LOGIC) &&
							(CheckChar || type != CHAR);
				} /*if (value && checkType)*/

				if (value && checkValue)
				{
					/* must be real */
					long      nvalues = symbolSize(symh);
					double   *x = DATAPTR(symh);
					
					for (i = 0; i < nvalues; i++)
					{
						double   v = x[i];

						if (isMissing(v) || CheckPositive && v <= 0 ||
							CheckNonnegative && v < 0 ||
							CheckNegative && v >= 0 ||
							CheckInteger && v != floor(v))
						{
							value = 0;
							break;
						}
					} /*for (i = 0; i < nvalues; i++)*/
				} /*if (value && checkValue)*/
			} /*if (value)*/
			DATAVALUE(result, jvalue++) = (double) value;
		} /*if (symh == EmptyArgument || strcmp(NAME(symh), USEDKEYWORD) != 0)*/
	} /*for (iarg = 0;iarg < nargs;iarg++)*/

	return (result);

  errorExit:
	putErrorOUTSTR();
	return (0);
} /*iswhat()*/

/*
   Function primarily for usage in macros to retrieve keyword values

   keyvalue(name1:arg1,name2:arg2, ..., targetName,type)
     where type is "real", "logical", "character", "structure", "macro",
     or "graph" and targetName is a quoted string or CHAR scalar
   Returns NULL, if no keyword name matches name.
   Returns 0 if there is a matching keyword name but value is wrong type
   Otherwise returns the value of the matching keyword.

   keyvalue(,targetName,type) returns NULL provided type is legal

   keyvalue(structure(name1:arg,name2:arg2,...),targetName,type) is equivalent
     to keyvalue(name1:arg1,name2:arg2,...,targetName,type)
     
   Typical usage in macro expecting optional keyword phrase 'weights:w'
      @w <- keyvalue($K,"weights","real")
	  if (isnull(@w)) {@w <- rep(1,@n)}

   960916 Installed
   980602 Enhanced keyvalue().  You can have several qualifiers, for example:
          keyvalue($K,"keyname",vector("positive","integer","scalar"))
   980620 Added property "notnull".  Use this to match keyword with non-
          null value of any type.  This was added since it is impossible
          to distinguish keyvalue(a:NULL,"a") from keyvalue(b:1,"a").
   990113 It's now OK to repeat a property, e.g., vector("real","real")
          You can include multiple properties in a single element of
          the last argument, e.g., "positive integer scalar" 
   990120 You can use the following abbreviated properties for scalars
          "number"  same as "nonmissing real scalar"
          "TF"      same as "nonmissing logical scalar"
          "string"  same as "character scalar"
   990121 New type property "square" implies and is not incompatible with
          "matrix"
   990414 Keyword phrase default:value as last argument to keyvalue() 
          provides a default value for the keyword
   000504 New abbreviated proprety for scalars
          "count"   same as "nonnegative integer scalar"
*/
enum
{
	TYPECODE      = 0,
	SHAPECODE,
	SIGNCODE,
	VALUECODE,
	MAXCODE,
	REALVAL       = 0,
	LOGICVAL,
	CHARVAL
};

#define PROPERTYLENGTH 20 /* must be longer than full name of any property*/

static char             *ScalarNames[] =
{
	"TF",
	"number",
	"count",
	"string",
	(char *) 0
};

static long             ScalarCodes[] =
{
	LOGICVALUE | SCALARTYPE | NONMISSINGTYPE,                    /*TF*/
	REALVALUE  | SCALARTYPE | NONMISSINGTYPE,                    /*number*/
	REALVALUE  | SCALARTYPE | INTEGERTYPE     | NONNEGATIVETYPE, /*count*/
	CHARVALUE  | SCALARTYPE                                      /*string*/
};

static char             *TypeNames[] =
{
	"real",
	"logical",
	"character",
	"macro",
	"graph",
	"notnull",
	(char *) 0
};
static long              TypeCodes[] =
{
	REALVALUE,          /*real*/
	LOGICVALUE,         /*logical*/
	CHARVALUE,          /*character*/
	-MACRO,             /*macro*/
	-PLOTINFO,          /*graph*/
	-NULLSYM            /*nonnull*/
};

static char             *ShapeNames[] =
{
	"scalar",
	"vector",
	"matrix",
	"array",
	"structure",
	"square",
	(char *) 0
};
static long              ShapeCodes[] =
{
	SCALARTYPE,                /*scalar*/
	VECTORTYPE,                /*vector*/
	MATRIXTYPE,                /*matrix*/
	ARRAYTYPE,                 /*array*/
	STRUCTURETYPE,             /*structure*/
	MATRIXTYPE | SQUARETYPE    /*square*/
};

static char             *SignNames[] = 
{
	"positive",
	"nonnegative",
	(char *) 0
};
static long              SignCodes[] =
{
	POSITIVETYPE | REALVALUE | NONMISSINGTYPE,     /*positive*/
	NONNEGATIVETYPE | REALVALUE | NONMISSINGTYPE   /*nonnegative*/
};

static char             *ValueNames[] = 
{
	"integer", "nonmissing", (char *) 0
};
static long              ValueCodes[] =
{
	INTEGERTYPE | REALVALUE| NONMISSINGTYPE,   /*integer*/
	NONMISSINGTYPE                             /*nonnegative*/
};

/*
  Match name to an element of one of TypeNames, ShapeNames, SignNames or
  ValueNames.  All of the characters name must match the first strlen(name)
  characters.
*/

static long keyDecode(char *name, long   codes[], char * words[])
{
	int          i, j, n = strlen(name), n1;
	int          minMatchLength;
	long         code;
	char        *word;

	if (mystrncmp(name, "non", 3) == 0 || mystrncmp(name, "str", 3) == 0)
	{
		minMatchLength = 4;
	}
	else
	{
		minMatchLength = 3;
	}

	for (i = 0; ScalarNames[i] != (char *) 0; i++)
	{
		word = ScalarNames[i];
		n1 = strlen(word);

		if (mystrncmp(name, word,
			(n1 < minMatchLength) ? n1 : minMatchLength) == 0)
		{
			long     typecode;

			code = ScalarCodes[i];
			
			if (codes[SHAPECODE] != 0 && codes[SHAPECODE] != SCALARTYPE)
			{
				return (-SHAPECODE - 1);
			}
			codes[SHAPECODE] = SCALARTYPE;
			
			typecode = code & (CHARVALUE | REALVALUE | LOGICVALUE);
			
			if (codes[TYPECODE] != 0 && codes[TYPECODE] != typecode)
			{
				return (-TYPECODE - 1);
			}
			codes[TYPECODE] = typecode;

			if (typecode != CHARVALUE)
			{
				codes[VALUECODE] |= NONMISSINGTYPE;
			}
			words[TYPECODE] = words[SHAPECODE] = words[VALUECODE] = word;
			
			return (code);
		} /*if (mystrncmp(name, ScalarNames[i], n) == 0)*/
	} /*for (i = 0; ScalarNames[i] != (char *) 0; i++)*/
	
	for (i = 0; TypeNames[i] != (char *) 0; i++)
	{
		word = TypeNames[i];
		n1 = strlen(word);

		if (mystrncmp(name, word,
			(n1 < minMatchLength) ? n1 : minMatchLength) == 0)
		{
			code = TypeCodes[i];
			j = TYPECODE;
			goto done;
		}
	} /*for (i = 0; TypeNames[i] != (char *) 0; i++)*/

	for (i = 0; ShapeNames[i] != (char *) 0; i++)
	{
		word = ShapeNames[i];
		n1 = strlen(word);

		if (mystrncmp(name, word,
			(n1 < minMatchLength) ? n1 : minMatchLength) == 0)
		{
			code = ShapeCodes[i];
			j = SHAPECODE;
			goto done;
		}
	} /*for (i = 0; ShapeNames[i] != (char *) 0; i++)*/

	for (i = 0; SignNames[i] != (char *) 0; i++)
	{
		word = SignNames[i];
		n1 = strlen(word);

		if (mystrncmp(name, word,
			(n1 < minMatchLength) ? n1 : minMatchLength) == 0)
		{
			code = SignCodes[i];
			j = SIGNCODE;
			goto done;
		}
	} /*for (i = 0; SignNames[i] != (char *) 0; i++)*/

	for (i = 0; ValueNames[i] != (char *) 0; i++)
	{
		word = ValueNames[i];
		n1 = strlen(word);

		if (mystrncmp(name, word,
			(n1 < minMatchLength) ? n1 : minMatchLength) == 0)
		{
			code = ValueCodes[i];
			j = VALUECODE;
			goto done;
		}
	} /*for (i = 0;  ValueNames[i] != (char *) 0; i++)*/

	return (0);

  done:
	if (codes[j] && (code & ~SQUARETYPE) != (codes[j]  & ~SQUARETYPE))
	{
		return (-j - 1);
	}
	codes[j] = code;
	words[j] = word;
	return (code);

} /*keyDecode()*/


/*
   980730 added new argument to reuseArg() so that notes will be kept.

   980801 added new function argvalue() with usage
          x <- argvalue(y, msg, properties)
   
   000817 corrected an error message
   010516 Use atLeastNArgs constant on badNargs() call

*/
#define PATTERNLENGTH    20
#define KEYNAMELENGTH   100
void myAlert(char * /*msg*/);

Symbolhandle keyvalue(Symbolhandle list)
{
	Symbolhandle         result = (Symbolhandle) NULLSYMBOL;
	Symbolhandle         symhStr, component, symhName;
	Symbolhandle         symhProperties = SymbolZero;
	Symbolhandle         symhArg = SymbolZero;
	Symbolhandle         symhDefault = SymbolZero;
	long                 nargs = NARGS(list), type;
	long                 ikey, nkeys = -1, i;
	long                 namePlace, propertyPlace;
	long                 code = 0;
	long                 propertyCodes[4];
	long                 argType, matchType;
	int                  strComp, keyCheck = strcmp(FUNCNAME, "keyvalue") == 0;
	int                  nProperties;
	char                 pattern[PATTERNLENGTH+1];
	char                *keyword = (char *) 0;
	char                 keyName[KEYNAMELENGTH + 1];
	char                *properties;
	char                *propertyNames[4];
	char                *cantUse =
		"ERROR: can't use property \"%s\" with property \"%s\"";
	WHERE("keyvalue");

	keyName[KEYNAMELENGTH] = '\0';


	if (keyCheck && nargs > 2)
	{
		symhDefault = COMPVALUE(list, nargs - 1);
		if ((keyword = isKeyword(symhDefault)) &&
			strcmp(keyword, "default") == 0)
		{
			nargs--;
		}
		else
		{
			symhDefault = SymbolZero;
		}
	} /*if (nargs > ((keyCheck) ? 2 : 3))*/
	
	if (nargs < 2)
	{
		badNargs(FUNCNAME, -2 + atLeastNArgs);
		goto errorExit;
	}
	else if (!keyCheck && nargs > 3)
	{
		badNargs(FUNCNAME, -3);
	}

	propertyCodes[TYPECODE] = propertyCodes[SHAPECODE] =
	  propertyCodes[SIGNCODE] = propertyCodes[VALUECODE] = 0;

	propertyNames[TYPECODE] = propertyNames[SHAPECODE] =
	  propertyNames[SIGNCODE] = propertyNames[VALUECODE] = NullString;

	symhStr = COMPVALUE(list, 0);

	if (!keyCheck)
	{ /*argvalue()*/
		Symbolhandle   symhMsg = COMPVALUE(list, 1);
		char          *symptom = (char *) 0;

		if (!isCharOrString(symhMsg))
		{
			char        outstr[40];

			sprintf(outstr, "argument 2 to %s()", FUNCNAME);
			notCharOrString(outstr);
			goto errorExit;
		}
		strncpy(keyName, STRINGPTR(symhMsg), KEYNAMELENGTH);

		symhArg = symhStr;
		/* Mimics code of argOK()*/
		if (symhArg == EmptyArgument)
		{
			symptom = "missing";
		}
		else if ((argType = TYPE(symhArg), argType == BLTIN))
		{
			symptom = "a built-in function";
		}
		else if (!myvalidhandle((char **) symhArg))
		{
			symptom = "damaged (deleted ?)";
		}
		else if (!isDefined(symhArg) || isNull(symhArg) && argType != NULLSYM)
		{
			symptom = (argType == UNDEF || argType == ASSIGNED) ?
			  "not defined" : "NULL";
		}
		
		if (symptom != (char *) 0)
		{
			sprintf(OUTSTR, "ERROR: %s is %s", keyName, symptom);
			goto errorExit;
		}
		namePlace = 1;
		strComp = 0;
	} /*if (!keyCheck)*/
	else
	{ /* keyvalue()*/
		strComp = !isKeyword(symhStr) && isStruc(symhStr);
		if (strComp)
		{
			nkeys = NCOMPS(symhStr);
			namePlace = 1;
		} /*if (!isKeyword(symhStr) && isStruc(symhStr))*/
		else /*if (strComp)*/
		{ /* first argument should be keyword phrase or empty argument */
			if (symhStr == EmptyArgument)
			{
				nkeys = 0;
				namePlace = 1;
			} /*if (symhStr == EmptyArgument)*/
			else
			{
				for (ikey = 0; ikey < nargs && isKeyword(COMPVALUE(list,ikey));
					 ikey++)
				{
					;
				}
				if (ikey == 0)
				{
					sprintf(OUTSTR,
							"ERROR: no keywords found by %s()", FUNCNAME);
					goto errorExit;
				}
				namePlace = nkeys = ikey;
			} /*if (symhStr == EmptyArgument){}else{}*/ 
			symhStr = list;
		} /*if (strComp){}else{}*/

		if (nargs <= namePlace)
		{
			sprintf(OUTSTR,
					"ERROR: no keyword name provided to %s()",
					FUNCNAME);
		}
		else if (nargs > namePlace + 2)
		{
			sprintf(OUTSTR, "ERROR: too many arguments to %s()", FUNCNAME);
		}

		if (*OUTSTR)
		{
			goto errorExit;
		}

		symhName = COMPVALUE(list, namePlace);
		if (!isCharOrString(symhName))
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s() is not quoted string or CHARACTER scalar",
					namePlace + 1, FUNCNAME);
			goto errorExit;
		}
		strncpy(keyName, STRINGPTR(symhName), KEYNAMELENGTH);

		if (!scanPat(keyName,&matchType,pattern,topicNameCheck,PATTERNLENGTH) ||
			matchType == EXACTMATCH && strlen(keyName) > NAMELENGTH - 2)
		{
			sprintf(OUTSTR,
					"ERROR: %s is not a legal keyword name or pattern for %s()",
					keyName, FUNCNAME);
			goto errorExit;
		}
	} /*if (!keyCheck){}else{}*/

	propertyPlace = namePlace + 1;

	if (propertyPlace < nargs)
	{
		symhProperties = COMPVALUE(list, propertyPlace);

		if (!argOK(symhProperties, 0, propertyPlace + 1))
		{
			goto errorExit;
		}
		
		if (TYPE(symhProperties) != CHAR || !isVector(symhProperties))
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s() is not a CHARACTER scalar or vector",
					propertyPlace + 1, FUNCNAME);
			goto errorExit;
		}
		properties = STRINGPTR(symhProperties);
		nProperties = symbolSize(symhProperties);

		for (i = 0; i < nProperties; i++)
		{
			long        thisCode;
			long        length = strlen(properties);
			char        thisProperty[PROPERTYLENGTH];
			char       *place = properties;
			
			/*
			  Each element may encode several white space separated properties
			*/
			while (place - properties < length)
			{
				if (copyField(place, (char *) 0) > PROPERTYLENGTH)
				{ /* be careful */
					sprintf(OUTSTR,
							"ERROR: property name for %s() is too long",
							FUNCNAME);
					goto errorExit;
				}
				place += copyField(place, thisProperty);
				thisCode = keyDecode(thisProperty, propertyCodes,
									 propertyNames);
				if (thisCode > 0)
				{
					code |= thisCode;
				}
				else if (thisCode == 0)
				{
					sprintf(OUTSTR,
							"ERROR: property \"%s\" not recognized by %s()",
							thisProperty, FUNCNAME);
				}
				else if (thisCode < 0 && -thisCode <= MAXCODE)
				{
					sprintf(OUTSTR,
							cantUse, thisProperty,
							propertyNames[-thisCode - 1]);
				}
				else if ((code = thisCode, nProperties > 1))
				{
					sprintf(OUTSTR,
							"ERROR: you can't use other properties for %s() with \"%s\"",
							FUNCNAME,
							(-thisCode == MACRO) ? TypeNames[3] : TypeNames[4]);
				}
				if (*OUTSTR)
				{
					goto errorExit;
				}
			} /*while (place - properties < length)*/
			
			properties = skipStrings(properties, 1);
		} /*for (i = 0; i < nProperties; i++)*/

		if (code > 0)
		{
			if ((propertyCodes[TYPECODE] & (LOGICVALUE | CHARVALUE | STRUCTURETYPE)) && 
				(propertyCodes[SIGNCODE] & (POSITIVETYPE | NONNEGATIVETYPE)))
			{
				sprintf(OUTSTR, cantUse,
						propertyNames[TYPECODE], propertyNames[SIGNCODE]);
				goto errorExit;
			}
			if ((propertyCodes[TYPECODE] & (LOGICVALUE | CHARVALUE | STRUCTURETYPE)) && 
				(propertyCodes[VALUECODE] & INTEGERTYPE))
			{
				sprintf(OUTSTR, cantUse,
						propertyNames[TYPECODE], propertyNames[VALUECODE]);
				goto errorExit;
			}
			if (code & NONMISSINGTYPE)
			{
				if (propertyCodes[TYPECODE] & CHARVALUE)
				{
					sprintf(OUTSTR, cantUse,
							propertyNames[TYPECODE],
							(propertyCodes[VALUECODE]) ?
							propertyNames[VALUECODE] : propertyNames[SIGNCODE]);
					goto errorExit;
				}
				if (propertyCodes[SHAPECODE] & STRUCTURETYPE)
				{
					sprintf(OUTSTR, cantUse,
							propertyNames[SHAPECODE],
							(propertyCodes[VALUECODE]) ?
							propertyNames[VALUECODE] : propertyNames[SIGNCODE]);
					goto errorExit;
				}
			} /*if (code & NONMISSINGTYPE)*/
		} /*if (code > 0)*/
	} /*if (propertyPlace < nargs)*/

	if (keyCheck)
	{
		for (ikey = 0; ikey < nkeys; ikey++)
		{
			component = COMPVALUE(symhStr, ikey);
			keyword = (symhStr == list) ? isKeyword(component) : NAME(component);
			if (matchName(keyword, matchType, pattern))
			{
				strcpy(keyName, keyword);
				break;
			}
		} /*for (ikey = 0; ikey < nkeys; ikey++)*/
		if (ikey < nkeys)
		{
			symhArg = component;
			argType = TYPE(symhArg);
			symhDefault = SymbolZero;
		}
		else if (symhDefault != SymbolZero)
		{
			symhArg = symhDefault;
			argType = TYPE(symhArg); /*added 010210 */			
			sprintf(keyName, "default for %s", STRINGPTR(symhName));
		}
	} /*if (keyCheck)*/
	else
	{
		ikey = 0;
	}

	if (symhArg != EmptyArgument)
	{
		if (symhProperties != SymbolZero)
		{
			type = propertyCodes[TYPECODE];

			if (type < 0)
			{
				/* macro, graph or notnull */
				char     *preposition = (keyCheck) ? "for" : "of";
				type = -type;

				if (type == NULLSYM && isNull(symhArg))
				{/* type == NULLSYM means NULL value illegal */
					sprintf(OUTSTR, "ERROR: value %s %s is NULL",
							preposition, keyName);
				}
				else if (type != NULLSYM && argType != type)
				{
					sprintf(OUTSTR,
							"ERROR: value %s %s is not %s",
							preposition, keyName, typeName(type));
				}
				if (*OUTSTR)
				{
					goto errorExit;
				}
			} /*if (type < 0)*/
			else
			{
				/* real, logic, char */
				char        msg[KEYNAMELENGTH + 2];

				if (keyCheck && symhDefault == SymbolZero)
				{
					sprintf(msg, "%s '%s'",
							(strComp) ? "structure component" : "keyword",
							keyName);
				}
				else
				{
					msg[0] = '#';
					strncpy(msg + 1, keyName, KEYNAMELENGTH);
					msg[KEYNAMELENGTH+1] = '\0';
				}

				if ((code & NONMISSINGTYPE) && isStruc(symhArg))
				{
					sprintf(OUTSTR, "ERROR: %s is not %s",
							(msg[0] != '#') ? msg : msg+1,
							(!(code & REALVALUE) && !(code & LOGICVALUE)) ?
							"REAL or LOGICAL" :
							((code & REALVALUE) ? "REAL" : "LOGICAL"));
					goto errorExit;
				}

				if (!checkArgType(symhArg, msg, code))
				{
					goto errorExit;
				}
			} /*if (type < 0){}else{}*/
		} /*if (symhProperties != SymbolZero)*/

		if (strComp)
		{ /* must be structure arg */
			if (isscratch(NAME(symhStr)))
			{ /* reuse component */
				result = reuseArg(symhStr, ikey, 1, 1);
				COMPVALUE(list, 0) = SymbolZero;
				Removesymbol(symhStr);
			}
			else
			{ /* make copy of component */
				result = Install(SCRATCH, type);
				if (result == SymbolError)
				{
					goto errorExit;
				}
				if (!Copy(symhArg, result))
				{
					Removesymbol(result);
					goto errorExit;
				}
			}
		} /*if (strComp)*/
		else
		{
			if (keyCheck || isscratch(NAME(symhArg)))
			{
				result = reuseArg(list,
								  (symhDefault == SymbolZero) ?
								  ikey : nargs,
								  1, 1);
			}
			else
			{
				result = Install(SCRATCH, argType);
				if (result == SymbolError)
				{
					goto errorExit;
				}
				if (!Copy(symhArg, result))
				{
					Removesymbol(result);
					goto errorExit;
				}
			}
		} /*if (symhStr != list){}else{}*/

		setNAME(result,SCRATCH);
	} /*if (symhArg != SymbolZero)*/

	return (result);

  errorExit:
	putErrorOUTSTR();
	return (0);
} /*keyvalue()*/

static void getTypeName(Symbolhandle symh, char name [])
{
	long          type = TYPE(symh);

	if (type == LOGIC)
	{
		strcpy(name, "LOGICAL");
	}
	else if (type == CHAR)
	{
		strcpy(name, "CHARACTER");
	}
	else if (type == BLTIN)
	{
		strcpy(name, "FUNCTION");
	}
	else if (type == STRUC)
	{
		strcpy(name, "STRUCTURE");
	}
	else if (type == UNDEF)
	{
		strcpy(name, "UNDEFINED");
	}
	else
	{
		strcpy(name, typeName(TYPE(symh)));
	}
} /*getTypeName()*/

static void getShapeName(Symbolhandle symh, char name [], int strict)
{
	long         type = TYPE(symh);
	
	if (type != REAL && type != LOGIC && type != CHAR && type != LONG)
	{
		getTypeName(symh, name);
	}
	else
	{
		char     *scalarname = "SCALAR";
		char     *vectorname = "VECTOR";
		char     *matrixname = "MATRIX";
		char     *arrayname = "ARRAY";
		char     *shapename;
		
		if (strict)
		{
			if (NDIMS(symh) == 1)
			{
				shapename = (DIMVAL(symh, 1) == 1) ? scalarname : vectorname;
			}
			else
			{
				shapename = (NDIMS(symh) == 2) ? matrixname : arrayname;
			}
		} /*if (strict)*/
		else
		{
			if (isScalar(symh))
			{
				shapename = scalarname;
			}
			else if (isVector(symh))
			{
				shapename = vectorname;
			}
			else if (isMatrix(symh, (long *) 0))
			{
				shapename = matrixname;
			}
			else
			{
				shapename = arrayname;
			}
		} /*if (strict){}else{}*/
		strcpy(name, shapename);
	}
} /*getShapeName()*/

/*
  vec <- nameof(arg_1, arg_2, ..., arg_k) returns vector of argument names
  vec <- getkeywords(arg_1, arg_2, ..., arg_k) returns CHARACTER vector.
    if arg_i is a keyword phrase, vec[i] is the keyword name; otherwise
    vec[i] is ""
  vec <- typeof(arg_1, arg_2, ..., arg_k) returns CHARACTER vector of
    argument types
  vec <- shapeof(arg_1, arg_2, ..., arg_k [,strict:T]) returns CHARACTER
    vector of argument shapes ("SCALAR","VECTOR","MATRIX","ARRAY",
    "STRUCTURE", "GRAPH", "NULL", "MACRO", or "FUNCTION")
  011026 nameof() returns keyword name for keyword phrase argument
         nameof() returns "" for empty argument; previously it was an error
         New function getkeywords() returns keyword names
  030514 New function typeof()
  030515 New function shapeof()
*/

Symbolhandle nameOf(Symbolhandle list)
{
	Symbolhandle      arg, result = SymbolZero;
	long              nargs = NARGS(list), i;
	long              needed = 0;
	char             *name, *place;
	char              typename[20];
	int               getkeywords = strcmp(FUNCNAME, "getkeywords") == 0;
	int               gettype = strcmp(FUNCNAME, "typeof") == 0;
	int               getshape = strcmp(FUNCNAME, "shapeof") == 0;
	int               strict = 0;
	WHERE("nameOf");

	if (getshape)
	{
		Symbolhandle       symhKey = COMPVALUE(list, nargs-1);
		char              *keyword = isKeyword(symhKey);
		
		if (keyword != (char *) 0 && strcmp(keyword, "strict") == 0)
		{
			if (!isTorF(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			strict = (DATAVALUE(symhKey, 0) != 0.0);
			nargs--;
			if (nargs == 0)
			{
				sprintf(OUTSTR,
						"ERROR: no variables in argument list to %s()",
						FUNCNAME);
				goto errorExit;
			}
		}
	} /*if (getshape)*/

	for (i = 0;i < nargs;i++)
	{
		arg = COMPVALUE(list, i);
		if (arg == EmptyArgument)
		{
			name = NullString;
		}
		else if (gettype)
		{
			getTypeName(arg, typename);
			name = typename;
		}
		else if (getshape)
		{
			getShapeName(arg, typename, strict);
			name = typename;
		}
		else
		{
			name = isKeyword(arg);
			if (name == (char *) 0)
			{
				if (getkeywords)
				{
					name = NullString;
				}
				else
				{
					name = NAME(arg);
					if (isscratch(name))
					{
						name += 2;
					}
				}
			} /*if (name == (char *) 0)*/
		}		
		needed += strlen(name) + 1;
	} /*for (i = 0;i < nargs;i++)*/

	result = CInstall(SCRATCH, needed);
	if (result == SymbolError)
	{
		goto errorExit;
	}
	TMPHANDLE = STRING(result);

	place = *TMPHANDLE;
	for (i = 0;i < nargs;i++)
	{
		arg = COMPVALUE(list, i);
		if (arg == SymbolZero)
		{
			name = NullString;
		}
		else if (gettype)
		{
			getTypeName(arg, typename);
			name = typename;
		}
		else if (getshape)
		{
			getShapeName(arg, typename, strict);
			name = typename;
		}
		else
		{
			name = isKeyword(arg);
			if (name == (char *) 0)
			{
				if (getkeywords)
				{
					name = NullString;
				}
				else
				{
					name = NAME(arg);
					if (isscratch(name))
					{
						name += 2;
					}
				}
			} /*if (name == (char *) 0)*/
		}
		place = copyStrings(name, place, 1);
	} /*for (i = 0;i < nargs;i++)*/

	setNDIMS(result,1);
	setDIM(result,1,nargs);

	return (result);

  errorExit:
	putErrorOUTSTR();
	Removesymbol(result);
	return (0);

} /*nameOf()*/


/*
   usage: rename(var, Name) where Name is CHARACTER scalar or an
   undefined variable.   The name of var is changed to Name or to NAME(Name).
      or: rename(expr, Name)
   where expr is a scratch variable (function result, expression, constant)
   is equivalent either to Name <- expr or <<Name>> <- expr.

   Written by C. Bingham (kb@stat.umn.edu) 950524

   The C-function name chosen to avoid conflict with system function rename()

   990325 Fixed some minor bugs.
*/
Symbolhandle renamer(Symbolhandle list)
{
	Symbolhandle          symh, symhName;
	char                 *name, newName[NAMELENGTH + 1];
	char                 *cannotChange =
	  "ERROR: %s() cannot change name of %s";
	long                  isexpr;
	long                  nargs = NARGS(list), type;
	WHERE("rename");

	*OUTSTR = '\0';
	if (nargs != 2)
	{
		badNargs(FUNCNAME, 2);
		goto errorExit;
	}


	symh = COMPVALUE(list, 0);
	symhName = COMPVALUE(list, 1);

	if (!argOK(symh, 0, 1))
	{
		goto errorExit;
	}
	if (symhName == EmptyArgument)
	{
		sprintf(OUTSTR,
				"ERROR: argument 2 to %s() is missing", FUNCNAME);
		goto errorExit;
	}
	
	if (symh == symhName)
	{
		return (NULLSYMBOL);
	}
	
	if (TYPE(symh) == BLTIN)
	{
		sprintf(OUTSTR, cannotChange, FUNCNAME,"built-in functions");
	}
	else if (isSpecial(symh))
	{
		sprintf(OUTSTR, cannotChange, FUNCNAME, NAME(symh));
	}
	else if (isAssigned(symh))
	{
		sprintf(OUTSTR, cannotChange, FUNCNAME, "variable being assigned to");
	}
	else if (isLocked(symh))
	{
		sprintf(OUTSTR, cannotChange, FUNCNAME, "locked variable");
	}
	if (*OUTSTR)
	{
		goto errorExit;
	}
	isexpr = isscratch(NAME(symh));

	type = TYPE(symhName);
	if (type == UNDEF)
	{
		strcpy(newName, NAME(symhName));
		Removesymbol(symhName);
		COMPVALUE(list, 1) = SymbolZero;
	} /*if (type == UNDEF)*/
	else
	{
		if (!isCharOrString(symhName))
		{
			name = NAME(symhName);
			if (isscratch(name))
			{
				sprintf(OUTSTR,
						"ERROR: Name must be undefined variable or CHARACTER scalar");
			}
			else if (type == BLTIN)
			{
				sprintf(OUTSTR,
						"ERROR: %s is the name of a built-in function", name);
			}
			else
			{
				sprintf(OUTSTR,
						"ERROR: a %s with name %s already exists",
						(type == MACRO) ? "macro" : "variable", name);
			}
		} /*if (type != CHAR || !isScalar(symhName))*/
		else if (strlen(name = STRINGPTR(symhName)) > NAMELENGTH)
		{
			sprintf(OUTSTR,
					"ERROR: name to be assigned is too long; length > %ld",
					(long) NAMELENGTH);
		}
		if (*OUTSTR)
		{
			goto errorExit;
		}

		if (!isVariableName(name))
		{
			sprintf(OUTSTR,
					"ERROR: '%s' is not a legal MacAnova name",
					STRINGPTR(symhName));
			goto errorExit;
		} /*if (!isVariableName(name))*/

		strcpy(newName, name);

		if (strcmp(newName, NAME(symh)) != 0 &&
			Lookup(newName) != SymbolZero)
		{
			sprintf(OUTSTR,
					"ERROR: a variable with name %s already exists", newName);
			goto errorExit;
		}
	} /*if (type == UNDEF){}else{}*/

	if (isexpr)
	{
		COMPVALUE(list, 0) = SymbolZero;
	}
	Cutsymbol(symh);
	setNAME(symh, newName);
	Addsymbol(symh);

	return (NULLSYMBOL);

  errorExit:
	putErrorOUTSTR();
	return (0);
} /*rename()*/

/*
  Function to implement lockvars() and unlockvars()
  usage:
	lockvars(a [,b ...] )
	unlockvars(a [,b ...] [, silent:T])

	a, b, ... variables of any type
	on lockvars() it is an error if any argument is special or scratch or
	missing;  a warning is given if an argument is a builtin function since
	they are effectively locked any way or if an argument is already locked.

	On unlockvars(), it is an error if any argument is missing.  A warning
	message is given if any argument is a scratch variable, a builtin function
	or is not locked.

	Warnings can be suppressed by 'silent:T'
*/
Symbolhandle lockvars(Symbolhandle list)
{
	Symbolhandle             result = SymbolZero;
	long                     nargs = NARGS(list);
	long                     i;
	int                      lockit = (FUNCNAME[0] == 'l');
	int                      silent = 0;

	nargs = (nargs == 1 && COMPVALUE(list, 0) == EmptyArgument) ? 0 : nargs;
	if (nargs > 0)
	{
		Symbolhandle      symhKey = COMPVALUE(list, nargs-1);
		char              *keyword = isKeyword(symhKey);
		
		if (keyword != (char *) 0 && strcmp(keyword, "silent") == 0)
		{
			if (!isTorF(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			silent = (DATAVALUE(symhKey, 0) != 0);
			nargs--;
		}
	} /*if (nargs > 0)*/
	
	if (nargs == 0)
	{
		sprintf(OUTSTR,
				"ERROR: no variables to %s", (lockit) ? "lock" : "unlock");
		goto errorExit;
	}
	for (i = 0; i < nargs; i++)
	{
		Symbolhandle       arg = COMPVALUE(list,i);
		char              *keyword = isKeyword(arg);

		if (arg == EmptyArgument)
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s() is missing", i+1, FUNCNAME);
		}
		else if (keyword != (char *) 0)
		{
			if (strcmp(keyword, "silent") == 0)
			{
				sprintf(OUTSTR,
						"ERROR: %s:T must be last argument to %s()",
						keyword, FUNCNAME);
			}
			else
			{
				badKeyword(FUNCNAME, keyword);
			}
			goto errorExit;
		}
		else if (!isDefined(arg))
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld (%s) to %s() is not defined",
					i+1, NAME(arg), FUNCNAME);
		}
		else if (lockit)
		{
			char       *cantlock = "ERROR: can't lock %s%s";
			char       *name = NAME(arg);
			
			if (isscratch(name) || arg == NULLSYMBOL)
			{
				sprintf(OUTSTR, cantlock,
						"expression or function result", NullString);
			}
			else if (istempname(name))
			{
				sprintf(OUTSTR, cantlock, "temporary variable", NullString);
			}
			else if (isSpecial(arg))
			{
				sprintf(OUTSTR, cantlock, "\"special\" variable ", name);
			}
			else if (strcmp(name, LASTGRAPH) == 0)
			{
				sprintf(OUTSTR, cantlock, "variable ", name);
			}
			else if (strcmp(name, "LASTLINE") == 0)
			{
				sprintf(OUTSTR, cantlock, "variable ", name);
			}
		}
		if (*OUTSTR)
		{
			goto errorExit;
		}
	} /*for (i = 0; i < nargs; i++)*/

	for (i = 0; i < nargs; i++)
	{
		Symbolhandle          arg = COMPVALUE(list, i);
		long                  type = TYPE(arg);
		char                 *name = NAME(arg);
		
		OUTSTR[0] = '\0';
		if (lockit)
		{
			if (type == BLTIN)
			{
				sprintf(OUTSTR, "WARNING: can't lock built in function %s",
						name);
			}
			else if (isLocked(arg))
			{
				sprintf(OUTSTR, "WARNING: %s is already locked", name);
			}
			else
			{
				setLocked(arg);
			}
		} /*if (lockit)*/
		else
		{
			char                 *cantunlock = "WARNING: can't unlock %s%s";
			
			if (isscratch(NAME(arg)) || arg == NULLSYMBOL)
			{
				sprintf(OUTSTR, cantunlock, "expression or function result",
						NullString);
			}
			else if (TYPE(arg) == BLTIN)
			{
				sprintf(OUTSTR, cantunlock, "built in function ", name);
			}
			else if (!isLocked(arg))
			{
				sprintf(OUTSTR, "WARNING: %s is not locked", name);
			}
			if (*OUTSTR == '\0')
			{
				clearLocked(arg);
			}
		} /*if (!lockit)*/
		if (*OUTSTR && !silent)
		{
			putErrorOUTSTR();
		}
	} /*for (i = 0; i < nargs; i++)*/
	
	return (NULLSYMBOL);
	
  errorExit:
	putErrorOUTSTR();
	return (SymbolError);
} /*lockvars()*/

/*
  function to process commands alltrue() and anytrue(), making use of
  the new capability of some functions having their arguments "quoted"
  like macro arguments, rather than evaluated.

  alltrue() and anytrue() process arguments one by one
  000504 New function
*/

Symbolhandle testLogicals(Symbolhandle list)
{
	Symbolhandle       result = SymbolZero;
	Symbolhandle       arg = COMPVALUE(list, 0);
	long               nargs = NARGS(list);
	long               i;
	int                all = (strncmp(FUNCNAME, "all", 3) == 0);
	int                logValue;
	WHERE("testLogicals");
	
#if (0) /*debugging code to test evalQuotedArg() */
	PRINT("FUNCNAME = %s, nargs = %ld, arg = %08x", FUNCNAME, nargs, arg, 0);
	if (nargs > 1 || arg != EmptyArgument)
	{
		for (i = 0; i < nargs; i++)
		{
			arg = COMPVALUE(list, i);
			if (arg == EmptyArgument)
			{
				PRINT("argument %ld is nil", i+1,0,0,0);
			}
			else if (TYPE(arg) != CHAR)
			{
				PRINT("argument %ld has type %s", i+1, typeName(TYPE(arg)),0,0);
			}
			else if (!isScalar(arg))
			{
				PRINT("argument %ld is not a scalar", i + 1, 0,0,0);
			}
			else
			{
				PRINT("argument %ld is '%s'",i+1,STRINGPTR(arg),0,0);
			}
			arg = evalQuotedArg(list, i);
			if (arg != COMPVALUE(list, i))
			{
				PRINT("arg = %08x, COMPVALUE(list,i) = %08x",arg,COMPVALUE(list,i),0,0);
			}
			
			if (arg == SymbolZero)
			{
				PRINT("evaluated argument %ld is nil", i+1,0,0,0);
			}
			else
			{
				PRINT("%s:",NAME(arg),0,0,0);
				prexpr(arg);
			}
		} /*for (i = 0; i < nargs; i++)*/
	}
	return (NULLSYMBOL);
#else
	if (nargs == 1 && arg == EmptyArgument)
	{
		badNargs(FUNCNAME, 0);
		goto errorExit;
	}
	
	for (i = 0; i < nargs; i++)
	{
		Symbolhandle      symhEval;
		
		arg = COMPVALUE(list, i);
		if (arg == EmptyArgument || TYPE(arg) == NULLSYM)
		{
			symhEval = NULLSYMBOL;
		}
		else if (!isCharOrString(arg))
		{ /* should not happen, but play it safe*/
			sprintf(OUTSTR, "ERROR: (internal) argument to %s() not quoted",
					FUNCNAME);
			goto errorExit;
		}
		else
		{
			symhEval = evalQuotedArg(list, i);
		}
		
		if (symhEval == SymbolError)
		{
			/* presumably someone else has issued error message*/
			goto errorExit;
		}

		if (!isTorF(symhEval))
		{
			char        msg[30];
			
			sprintf(msg, "argument %ld to %s()", i+1, FUNCNAME);
			notTorF(msg);
			goto errorExit;
		} /*if (!isTorF(symhEval))*/
		logValue = (DATAVALUE(symhEval, 0) != 0.0);
		
		if (all && !logValue || !all && logValue)
		{
			break;
		}
	} /*for (i = 0; i < nargs; i++)*/

	result = RInstall(SCRATCH, 1);
	if (result == SymbolError)
	{
		goto errorExit;
	}
	setTYPE(result, LOGIC);
	DATAVALUE(result, 0) = (logValue) ? 1.0 : 0.0;
	return (result);

  errorExit:
	putErrorOUTSTR();
	return (0);
#endif
} /*testLogicals()*/

		
				
