/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2003 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment MakeFact
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "globals.h"

enum makefactorOpCodes
{
	IMATCH = 1,
	IUNIQUE
};

/*
	factor(arg)
		If arg legal for factor (all posInts) return factor equivalent
		If arg already a factor, the number of levels is recomputed from
		the actual values in arg, not taken from NCLASS(arg).
	match(arg,vals [,noMatch] [, exact:F  or  fuzz:d [,relative:T]])
		result[i] = (arg[i] == vals[j]) ? j+1 : noMatch
		arg and vals must be both REAL or both CHAR
		vals must contain no missing values
		The default value for noMatch is length(vals) + 1
        exact:F is legal with CHARACTER arg and allows wild
        card characters in arg.
        fuzz:d is legal with REAL arg.  A match occurs if
		abs(val-arg) <= d
	unique(arg [,fuzz:d [,relative:T]])
		Produces vector of unique values in arg. It is an error for arg to
		be all MISSING.  arg must be REAL or CHAR
        With fuzz:d, values are the same if they differ by not more
        than d >= 0.  The result is order dependent.
	J <- unique(arg,index:T [,fuzz:d [,relative:T]]) produces vector J of
	    integers such that arg[J] is the same as 
	    unique(arg [,fuzz:d [,relative:T]])
*/

/*
	961013 Fixed bug affecting case of 1 logical scratch argument

	980420 added keyword 'index' to unique().  index:T means the
    indices of the unique elements are returned instead of the values
    so that x[unique(x,index:T)] and unique(x) are identical.

	980730 added new argument to reuseArg() so that notes will be kept.

    990226 match() and unique() now work with logical arguments
           changed most putOUTSTR() to putErrorOUTSTR()

    000526 implemented keyword phrase fuzz:d on unique() and match()

    010619 When match() argument 2 is not a vector, argument 1 must be
	       a vector and argument 2 is treated as an array of vectors

    011016 When an argument to factor() is already a factor, its
           nominal number of levels is ignored; the nominal number
           of levels is max(arg1,arg2,...)
*/

static Symbolhandle  doFactor(Symbolhandle list)
{
	Symbolhandle       result = (Symbolhandle) 0;
	Symbolhandle       symh;
	double            *in, *out;
	double             val;
	long               nargs = NARGS(list);
	long               type, type0;
	long               i, j, k, n, needed = 0,levels, nclasses = -1;
	WHERE("doFactor");
	
	*OUTSTR = '\0';

	for (i = 0; i < nargs; i++)
	{
		symh = COMPVALUE(list,i);
		if (!argOK(symh, 0, (nargs > 1) ? i + 1 : 0))
		{
			goto errorExit;
		}
		type = TYPE(symh);
		if (type != REAL && type != LOGIC)
		{
			badType(FUNCNAME,-type, (nargs > 1) ? i+1 : 0);
			goto errorExit;
		}
		if (!isVector(symh))
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s() is not a vector %s",i+1,
					FUNCNAME, (nargs == 1) ? "" : " or scalar");
		}
		else if (i == 0)
		{
			type0 = type;
		}
		else if (type != type0)
		{
			sprintf(OUTSTR,
					"ERROR: type of argument %ld to %s() not the same as argument 1",
					i + 1, FUNCNAME);
		}
		if (*OUTSTR)
		{
			goto errorExit;
		}
	} /*for (i = 0; i < nargs; i++)*/

	
	for (i = 0; i < nargs; i++)
	{
		symh = COMPVALUE(list, i);

		n = symbolSize(symh);
		in = DATAPTR(symh);
		for (j = 0; j < n; j++)
		{
			if (!isMissing(in[j]))
			{
				break;
			}
		} /*for (j = 0; j < n; j++)*/

		if (j < n)
		{ /* found at least 1 non-missing value */
			levels = isFactor(symh);
			if (levels == 0)
			{
				sprintf(OUTSTR,
				"ERROR: argument for %s() with non-integers or nonpositive integers",
					FUNCNAME);
				goto errorExit;
			}
			if (levels < 0)
			{
				levels = -levels;
			}
			else
			{
				double       extremes[2];
				
				symExtremes(symh, extremes);
				levels = (long) extremes[1];
			}
			
			if (levels > nclasses)
			{
				nclasses = levels;
			}
		} /*if (j < n)*/
		needed += n;
	} /*for (i = 0; i < nargs; i++)*/

	if (nclasses < 0)
	{
		sprintf(OUTSTR,
				"ERROR: no non-MISSING values in arguments to %s()",
				FUNCNAME);
	}
	else if (nclasses > MAXFACTORLEVELS)
	{
		sprintf(OUTSTR,
				"ERROR: maximum factor level > %ld is too big",
				(long) MAXFACTORLEVELS);
	}
	if (*OUTSTR)
	{
		goto errorExit;
	}
	
	if (nargs == 1)
	{
		if (isscratch(NAME(symh)))
		{/* reuse argument */
			result = reuseArg(list, 0, 1, 1);
		}
		else
		{
			result = Install(SCRATCH, REAL);
			if (result == (Symbolhandle) 0 || !Copy(symh, result))
			{
				goto errorExit;
			}
		}				
		setNAME(result, SCRATCH);
	} /*if (nargs == 1)*/
	else
	{
		result = RInstall(SCRATCH, needed);
		if (result == (Symbolhandle) 0)
		{
			goto errorExit;
		}
		setNDIMS(result, 1);
		setDIM(result,1,needed);
	} /*if (nargs == 1){}else{}*/

	setTYPE(result, REAL);
	setNCLASS(result, nclasses);

	if (type == LOGIC || nargs > 1)
	{
		out = DATAPTR(result);
		k = 0;
		for (i = 0; i < nargs; i++)
		{
			symh = (i > 0 || COMPVALUE(list, 0) != (Symbolhandle) 0) ?
					 COMPVALUE(list, i) : result;
			n = symbolSize(symh);
			in = DATAPTR(symh);

			for (j = 0; j < n; j++)
			{
				val = in[j];
				if (isMissing(val))
				{
					setMissing(out[k]);
				}
				else
				{
					if (type == LOGIC)
					{
						val = (val != 0.0) ? 2.0 : 1.0;
					}
					out[k] = val;
				}
				k++;
			} /*for (j = 0; j < n; j++)*/
		} /*for (i = 0;i < n; i++)*/
	} /*if (type == LOGIC || nargs > 1 || !isscratch(NAME(symH)))*/

	return (result);

  errorExit:
	putErrorOUTSTR();
	return (0);
	
} /*doFactor()*/

#define MAXPATLENGTH 500

/*
  010619 modified so it matches argument 2 a column at a time
  010703 modified to allow case to be ignored
  030522 renamed fuzzEqual() to fuzzyEqual(), rewrote it to allow
         for relative error and then moved it to mathutil.c
         Added argument relative to doMatch() and doUnique()
*/
static void doMatch(Symbolhandle args [], Symbolhandle result,
				   double noMatch, int exact, double fuzz, int relative)
{
	double     hugedbl = HUGEDBL;
	double    *values, val, *in, *out;
	long       n = symbolSize(args[0]);
	long       m = symbolSize(args[1]);
	long       type = TYPE(args[0]);
	long       i, j, notMatched = 0;
	long       patternTooLong = 0;
	long       icol, ncols = 1;
	long       dims[MAXDIMS + 1];
	int        quiet = noMatch != hugedbl;
	char      *chIn;
	char      *chValues, *chVal;

	if (!isVector(args[1]))
	{
		getDims(dims, args[1]);
		ncols = m/dims[1];
		m = dims[1];
		dims[1] = n;
	}
	else
	{
		getDims(dims, args[0]);
	}
	Setdims(result, dims[0], dims+1);

	if (noMatch == hugedbl)
	{
		noMatch = (double) (m+1);
	}
		
	if (type != CHAR)
	{
		in = DATAPTR(args[0]);
		values = DATAPTR(args[1]);
	}
	else
	{
		chIn = STRINGPTR(args[0]);
		chValues = STRINGPTR(args[1]);
	}
	out = DATAPTR(result);
		
	for (icol = 0; icol < ncols; icol++)
	{
		if (type == CHAR)
		{
			chVal = chIn;
		}
		

		for (i = 0; i < n;i++)
		{
			if (type != CHAR)
			{
				val = in[i];
				if (!isMissing(val))
				{
					for (j = 0;j < m;j++)
					{
						if (type == REAL &&
							fuzzyEqual(val, values[j], fuzz, relative) ||
							type == LOGIC &&
							(val != 0.0 && values[j] != 0.0 ||
							 val == 0.0 && values[j] == 0.0))
						{
							break;
						}
					} /*for (j = 0;j < m;j++)*/
				} /*if (!isMissing(val))*/
				else
				{
					j = -1;
				}
			} /*if (type != CHAR)*/
			else
			{ /*CHAR*/
				long      matchType = EXACTMATCH;
				char      pattern[MAXPATLENGTH + 1];
				char     *chVal1 = chValues; 
	
				if (!(exact & EXACTMATCH) &&
					!scanPat(chVal, &matchType, pattern,
							 noNameCheck, MAXPATLENGTH) &&
					matchType & PATTERNTOOLONG)
				{
					patternTooLong++;
					j = m;
				}
				else
				{
					long         chValLength = strlen(chVal);

					if ((exact & MATCHIGNORECASE) != 0)
					{
						matchType |= MATCHIGNORECASE;
					}
					
					for (j = 0;j < m;j++)
					{
						if (matchType == EXACTMATCH)
						{
							if (strcmp(chVal1,chVal) == 0)
							{
								break;
							}
						}
						else if (matchType == (EXACTMATCH | MATCHIGNORECASE))
						{
							if (mystrncmp(chVal1, chVal, 0) == 0)
							{
								break;
							}
						}
						else if (matchName(chVal1, matchType, pattern))
						{
							break;
						}
						chVal1 = skipStrings(chVal1, 1);
					} /*for (j = 0;j < m;j++)*/
				}
				chVal = skipStrings(chVal, 1);
			} /*if (type != CHAR){}else{}*/
	
			if (j == m)
			{ /* not matched */
				notMatched++;
				out[i] = noMatch;
			}
			else if (j >= 0)
			{ /* matched non missing*/
				out[i] = (double) (j+1);
			}
			else
			{
				setMissing(out[i]);
			}
		} /*for (i = 0; i< n;i++)*/
		out += n;
		if (type != CHAR)
		{
			values += m;
		}
		else if (icol < ncols - 1)
		{
			chValues = skipStrings(chValues, m);
		}
	} /*for (icol = 0; icol < ncols; icol++)*/

	if (notMatched != 0 && !quiet)
	{
		if (patternTooLong)
		{
			sprintf(OUTSTR,
					"WARNING: %ld patterns were too long to be matched",
					patternTooLong);
			putErrorOUTSTR();
		}
		sprintf(OUTSTR,"WARNING: %ld values not matched coded as %ld",
				notMatched, m+1);
		putErrorOUTSTR();
	}
} /*doMatch()*/

static int doUnique(Symbolhandle args [], Symbolhandle result,
			  int compIndex, double fuzz, int relative)
{
	double            val, *in, *out, **outH;
	long              m = 0, n = symbolSize(args[0]);
	long              i, j, type = TYPE(args[0]);
	long              cplace;
	char            **chInH, **chOutH;

		
	if (type != CHAR)
	{
		in = DATAPTR(args[0]);
	}
	else
	{
		cplace = 0;
		chInH = STRING(args[0]);
	}
	if (compIndex || type != CHAR)
	{
		outH = DATA(result);
		out = *outH;
	}
	
	for (i = 0; i < n;i++)
	{
		if (type != CHAR)
		{
			val = in[i];
			if (!isMissing(val))
			{
				if (!compIndex)
				{
					for (j = 0; j < m; j++)
					{
						if (fuzzyEqual(val, out[j], fuzz, relative))
						{
							break;
						}
					} /*for (j = 0; j < m; j++)*/
				} /*if (!compIndex)*/
				else
				{
					for (j = 0; j < m;j++)
					{
						if (fuzzyEqual(val, in[(long) out[j] - 1],
									  fuzz, relative))
						{
							break;
						}
					} /*for (j=0;j<m;j++)*/
					val = (double) (i + 1);
				}
				if (j == m)
				{
					/* not matched; must be one of unique values */
					out[m++] = val;
				}
			} /*if (!isMissing(val))*/
		} /*if (type != CHAR)*/
		else
		{ /* type == CHAR */
			char       *chVal = *chInH + cplace, *chVal1;
			long        chOutSize, chLength = strlen(chVal) + 1;

			if (i == 0)
			{
				/* first value is always one of unique values */
				if (!compIndex)
				{
					chOutSize = chLength;
					chOutH = mygethandle(chOutSize);
					setSTRING(result, chOutH);
					if (chOutH == (char **) 0)
					{
						goto errorExit;
					}
					strcpy(*chOutH, *chInH);
				} /*if (!compIndex)*/
				else
				{
					out[0] = 1.0;
				}
				m = 1;
			} /*if (i == 0)*/
			else
			{
				if (!compIndex)
				{
					chVal1 = *chOutH;
					for (j = 0; j < m;j++)
					{
						if (strcmp(chVal,chVal1) == 0)
						{
							break;
						}
						chVal1 = skipStrings(chVal1, 1);
					} /*for (j = 0; j < m;j++)*/

					if (j == m)
					{/* not matched */
						chOutH = mygrowhandle(chOutH,chOutSize + chLength);
						setSTRING(result,chOutH);
						if (chOutH == (char **) 0)
						{
							goto errorExit;
						}
						strcpy(*chOutH + chOutSize, *chInH + cplace);
						chOutSize += chLength;
						m++;
					} /*if (j == m)*/
				} /*if (!compIndex)*/
				else
				{
					char    *cval = *chInH;

					for (j = 0; j < m; j++)
					{
						if (strcmp(chVal, cval) == 0)
						{
							break;
						}
						if (j < m - 1)
						{
							cval = skipStrings(cval,
											   (long) (out[j+1] - out[j]));
						}
					} /*for (j = 0; j < m; j++)*/
					if (j == m)
					{
						out[m++] = (double) (i + 1);
					}
				} /*if (!compIndex){}else{}*/
			} /*if (i == 0){}else{}*/
			cplace += chLength;
		} /*if (type != CHAR){}else{}*/
	} /*for (i = 0;i < n;i++)*/

	if (m == 0)
	{
		sprintf(OUTSTR,
				"ERROR: all values in argument to %s() are MISSING",
				FUNCNAME);
		goto errorExit;
	}

	if (TYPE(result) != CHAR && m < n)
	{ /* resize outH */
		outH = (double **) mygrowhandle((char **) outH, m * sizeof(double));
		setDATA(result,outH);
		if (outH == (double **) 0)
		{
			goto errorExit;
		}
	} /*if (TYPE(result) != CHAR && m < n)*/
	setNDIMS(result, 1);
	setDIM(result, 1, m);

	return (1);
	
  errorExit:
	return (0);
} /*doUnique()*/


/*
  usages:
    factor(l1, l2, ...) where l1, l2, ... are either all REAL vectors 
	of positive integers or all LOGICAL vectors.

	match(x,vec [,nomatch] [,exact:F or fuzz:d [,relative:T]]),
	x REAL or CHARACTER, vec avector of the same type as x and
	nomatch a REAL scalar, d >= 0 a scalar

	unique(x [,index:T] [,fuzz:d [,relative:T]]), x REAL or CHARACTER,
	scalar d >= 0

	980730 added new argument to reuseArg() so that notes will not be kept.

    000525 added keyword phrase fuzz:d to match() and unique()

    010516 Use atLeastNArgs constant on badNargs() call

    010619 match() now accepts matrix argument 2

    010703 match() now recognizes keyword phrase ignorecase:T

    030522 match() and unique() now recognize 'relative:T' with 'fuzz:d'
           allowing for relative rather than absolute differences
*/
Symbolhandle    makefactor(Symbolhandle list)
{
	Symbolhandle    args[3], result = (Symbolhandle) 0;
	double          hugedbl = HUGEDBL;
	double          noMatch = hugedbl;
	double          fuzz = -1.0;
	long            i, m, n, op, size;
	long            nargs = NARGS(list);
	long            type;
	int             maxNargs, minNargs;
	int             compIndex = 0, exact = -1, ignoreCase = 0, relative = 0;
	char           *indexKey = "index";
	char           *exactKey = "exact";
	char           *ignoreKey = "ignorecase";
	char           *fuzzKey = "fuzz", *relativeKey = "rel";
	char           *keyword;
	WHERE("makefactor");

	*OUTSTR = '\0';
	if (strcmp(FUNCNAME,"factor") == 0)
	{
		return (doFactor(list));
	}

	if (strcmp(FUNCNAME,"match") == 0)
	{
		op = IMATCH;
		maxNargs = 3;
		minNargs = 2;
	}
	else
	{
		op = IUNIQUE;
		maxNargs = minNargs = 1;
	}
		
	if (nargs < minNargs)
	{
		badNargs(FUNCNAME, -minNargs + atLeastNArgs);
		goto errorExit;
	}
	
	for (i = 0; i < minNargs; i++)
	{
		if (isKeyword(COMPVALUE(list, i)))
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s() must not be keyword phrase",
					i + 1, FUNCNAME);
			goto errorExit;
		}
	} /*for (i = 0; i < minNargs; i++)*/
	
	/* scan for keywords from the end of the argument list */
	for (i = nargs - 1; i >= minNargs; i--)
	{
		Symbolhandle     symhKey = COMPVALUE(list, i);

		keyword = isKeyword(symhKey);
		if (keyword == (char *) 0)
		{
			break;
		}

		if (strcmp(keyword, fuzzKey) == 0)
		{
			if (!isNumber(symhKey, NONNEGATIVEVALUE))
			{
				notNonNegativeReal(keyword);
				goto errorExit;
			}
			fuzz = DATAVALUE(symhKey, 0);
		}
		else if (strncmp(keyword, relativeKey, 3) == 0)
		{
			if (!isTorF(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			relative = (DATAVALUE(symhKey, 0) != 0.0);
		}
		else if (op == IMATCH && strcmp(keyword, exactKey) == 0 ||
				 op == IUNIQUE && strcmp(keyword, indexKey) == 0 ||
				 op == IMATCH && strncmp(keyword, ignoreKey, 6) == 0)
		{
			if (!isTorF(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			if (strcmp(keyword, exactKey) == 0)
			{
				exact = (DATAVALUE(symhKey,0) != 0.0) ?
				  EXACTMATCH : 0;
			}
			else if (strcmp(keyword, indexKey) == 0)
			{
				compIndex = (DATAVALUE(symhKey,0) != 0.0);
			}
			else
			{
				ignoreCase = (DATAVALUE(symhKey, 0) != 0.0) ?
				  MATCHIGNORECASE : 0;
			}
		}
		else
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		nargs--;
	} /*for (i = nargs - 1; i >= minNargs; i--)*/
	
	if (nargs > maxNargs)
	{
		sprintf(OUTSTR,
				"ERROR: too many non-keyword arguments for %s()",FUNCNAME);
		goto errorExit;
	}
	
	for (i = 0; i < nargs;i++)
	{
		args[i] = COMPVALUE(list,i);
		if (!argOK(args[i],0,(nargs > 1) ? i+1 : 0))
		{
			goto errorExit;
		}
	} /*for (i = 0; i < nargs;i++)*/

	type = TYPE(args[0]);

	if (type != REAL && type != CHAR && type != LOGIC)
	{
		badType(FUNCNAME, -type, (op == IMATCH) ? 1 : 0);
		goto errorExit;
	}

	if (relative && fuzz <= 0)
	{
		sprintf(OUTSTR,
				"WARNING: %s() argument 'relative:T' ignored without 'fuzz:eps'",
				FUNCNAME);
		putErrorOUTSTR();
	}
	
	if (fuzz >= 0 && type == CHAR)
	{
		sprintf(OUTSTR,
				"ERROR: '%s:d' illegal with CHARACTER argument to %s()",
				fuzzKey, FUNCNAME);
		goto errorExit;
	}

	if ((exact > 0 || ignoreCase) && type != CHAR)
	{
		sprintf(OUTSTR,
				"ERROR: '%s:T' and '%s:T' illegal with %s argument to %s()",
				exactKey, ignoreKey, typeName(type), FUNCNAME);
		goto errorExit;
	}

	exact = (exact < 0) ? EXACTMATCH : exact;
	exact |= ignoreCase;

	fuzz = (fuzz < 0.0) ? 0.0 : fuzz;

	if (op == IMATCH)
	{
		if (TYPE(args[1]) != type)
		{
			sprintf(OUTSTR,
					"ERROR: %s arguments to %s() must have same type",
					(nargs == 2) ? "both" : "first two", FUNCNAME);
		}			
		else if (nargs == 3)
		{ /* value to return on no match */
			if (TYPE(args[2]) != REAL || !isScalar(args[2]))
			{
				char          msg[30];
				
				sprintf(msg, "%s() argument 3", FUNCNAME);
				notNumberOrReal(msg);
				goto errorExit;
			}
			noMatch = DATAVALUE(args[2], 0);
		}
		else if (type != CHAR && anyMissing(args[1]))
		{
			sprintf(OUTSTR,
					"ERROR: argument 2 to %s() contains MISSING values",
					FUNCNAME);
		}
		else if (!isVector(args[1]) && !isVector(args[0]))
		{
			sprintf(OUTSTR,
					"ERROR: illegal when first two %s() args both not vectors",
					FUNCNAME);
		}
		if (*OUTSTR)
		{
			goto errorExit;
		}
		m = (isVector(args[1])) ? symbolSize(args[1]) : DIMVAL(args[1], 1);
	} /*if (op == IMATCH)*/

	n = symbolSize(args[0]);
	size = (op != IMATCH) ? n : (n*symbolSize(args[1]))/m;

	if (op == IMATCH && type == REAL && n == size &&
		isscratch(NAME(args[0])))
	{ /* reuse argument */
		result = reuseArg(list, 0, 0, 0);
		setNCLASS(result, -1);
	}
	else if (type != CHAR || compIndex || op != IUNIQUE)
	{
		result = RInstall(SCRATCH, size);
	}
	else
	{
		result = Install(SCRATCH, CHAR);
	}
	
	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	else if (type == LOGIC && op == IUNIQUE && !compIndex)
	{
		setTYPE(result, LOGIC);
	}
	
	if (op == IMATCH)
	{
		doMatch(args, result, noMatch, exact, fuzz, relative);
	}
	else
	{
		if (!doUnique(args, result, compIndex, fuzz, relative))
		{
			goto errorExit;
		}
	}

	return (result);

  errorExit:
	putErrorOUTSTR();
	Removesymbol(result);

	return (0);
} /*makefactor()*/

