/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.12 or later
*(C)
*(C)* Copyright (c) 2001 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Makesymb
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "globals.h"

/*
  Function to return character codes
  usage:
    symbolCodes <- makesymbols(ascii), where ascii is a scalar, vector,
	  matrix or array of integers between 1 and 255
    symbolCodes is a CHARACTER array of the same size and shape as ascii,
	  whose elements have ASCII codes given by the corresponding elements
      of ascii

    symbolCodes <- plotsymbols(symbolNames [,medium:T or small:T]),
	  where symbolNames is a CHARACTER scalar, vector, matrix or array

    symbolCodes will be a CHARACTER array of the same size and shape
	as symbolNames.

	If the first three letters of symbolNames[i] match the first three
    letters of "diamond", "plus", "square", "cross", "triangle", "star",
	"dot", or "circle", then symbolCodes[i] will be
    "\001", "\002","\003","\004","\005","\006","\007","\010" ("\011",
	"\012","\013","\014","\015","\016","\017","\020" with medium:T and
	"\021", "\022","\023","\024","\025","\026","\027","\030" with small:T)

     Otherwise, symbolCodes[i] = symbolNames[i].

  010926 installed
*/

enum shapeNameCodes
{
	diamondShape = 1,
	plusShape,
	squareShape,
	crossShape,
	triangleShape,
	starShape,
	dotShape,
	circleShape,
	nShapes = circleShape
};

enum symbolSizeCode
{
	bigSymbol = 1,
	mediumSymbol = bigSymbol + nShapes,
	smallSymbol = mediumSymbol + nShapes
};

static char         ShapeNames[nShapes][9] =
{
	"diamond",
	"plus",
	"square",
	"cross",
	"triangle",
	"star",
	"dot",
	"circle",
};


Symbolhandle makesymbols(Symbolhandle list)
{
	Symbolhandle             symhCodes;
	Symbolhandle             result = SymbolZero;
	Symbolhandle             symhKey = SymbolZero;
	long                     type;
	long                     nargs = countArgs(list, (long *) 0);
	long                     length;
	long                     i;
	long                     dims[MAXDIMS+1];
	char                    *keyword;
	char                     message[100];
	unsigned char           *charCodes;
	WHERE("makesymbols");

	if (nargs == 0)
	{
		badNargs(FUNCNAME, atLeast1Args);
		goto errorExit;
	}
	if (nargs > 2)
	{
		badNargs(FUNCNAME, atMost2Args);
		goto errorExit;
	}

	symhCodes = COMPVALUE(list, 0);
	if (!argOK(symhCodes, 0, (nargs > 1) ? 1 : 0))
	{
		goto errorExit;
	}
	type = TYPE(symhCodes);

	if (type != REAL && type != CHAR)
	{
		sprintf(OUTSTR,
				"ERROR: argument 1 to %s() not REAL or CHARACTER", FUNCNAME);
	}

	length = symbolSize(symhCodes);

	sprintf(message, "argument 1 to %s()", FUNCNAME);

	if (type == REAL)
	{
		double         *asciiCodes = DATAPTR(symhCodes);

		if (!checkArgType(symhCodes, message, NONNEGATIVEINTARRAY))
		{
			goto errorExit;
		}
		if (doubleMax(asciiCodes, length) > 255.0)
		{
			sprintf(OUTSTR,
					"ERROR: %s has an element > 255", message);
			goto errorExit;
		}
		result = CInstall(SCRATCH, 2*length);
		if (result == SymbolError)
		{
			goto errorExit;
		}
		asciiCodes = DATAPTR(symhCodes); /*play it safe*/
		charCodes = (unsigned char *) STRINGPTR(result);

		for (i = 0; i < length; i++)
		{
			*charCodes++ = (unsigned char) asciiCodes[i];
			*charCodes++ = '\0';
		}
	} /*if (type == REAL)*/
	else
	{
		enum symbolSizeCode      size = bigSymbol;
		long                     needed;
		char                    *codeNames;
		char                    *newCodeNames;

		if (nargs > 1)
		{
			symhKey = COMPVALUE(list, 1);
			keyword = isKeyword(symhKey);
			if (keyword == (char *) 0)
			{
				sprintf(OUTSTR,
						"ERROR: argument 2 to %s() must be ",
						FUNCNAME);
				goto errorExit;
			} /*if (keyword == (char *) 0)*/

			if (mystrncmp(keyword, "small", 5) != 0 &&
				mystrncmp(keyword, "medium", 3) != 0 &&
				mystrncmp(keyword, "large", 5) != 0 &&
				mystrncmp(keyword, "big", 3) != 0)
			{
				badKeyword(FUNCNAME, keyword);
				goto errorExit;
			}
			if (!isTorF(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			if (DATAVALUE(symhKey, 0) == 0.0)
			{
				sprintf(OUTSTR,
						"ERROR: value of keyword '%s' must be T",
						keyword);
				goto errorExit;
			} /*if (DATAVALUE(symhKey, 0) == 0.0)*/

			if (tolower(keyword[0]) == 's')
			{
				size = smallSymbol;
			}
			else if (tolower(keyword[0]) == 'm')
			{
				size = mediumSymbol;
			}
		} /*if (nargs > 1)*/

		codeNames = STRINGPTR(symhCodes);

		needed = length;
		for (i = 0; i < length; i++)
		{
			int       j;

			for (j = 0; j < nShapes; j++)
			{
				if (mystrncmp(codeNames, ShapeNames[j], 3) == 0)
				{
					needed++;
					break;
				}
			} /*for (j = 0; j < nShapes; j++)*/
			if (j == nShapes)
			{
				needed += strlen(codeNames);
			}
			codeNames = skipStrings(codeNames, 1);
		} /*for (i = 0; i < length; i++)*/

		result = CInstall(SCRATCH, needed);
		if (result == SymbolError)
		{
			goto errorExit;
		}

		codeNames = STRINGPTR(symhCodes);
		newCodeNames = STRINGPTR(result);

		for (i = 0; i < length; i++)
		{
			int       j;

			for (j = 0; j < nShapes; j++)
			{
				if (mystrncmp(codeNames, ShapeNames[j], 3) == 0)
				{
					*newCodeNames++ = j + size;
					*newCodeNames++ = '\0';
					break;
				}
			} /*for (j = 0; j < nShapes; j++)*/

			if (j == nShapes)
			{
				newCodeNames = copyStrings(codeNames, newCodeNames, 1);
			}
			codeNames = skipStrings(codeNames, 1);
		} /*for (i = 0; i < length; i++)*/
	} /*if (type == REAL){}else{}*/

	getDims(dims, symhCodes);
	Setdims(result, dims[0], dims+1);

	return (result);

  errorExit:
	putErrorOUTSTR();
	return (SymbolError);
} /*makesymbols()*/

