
/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2003 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Utils
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/


#include "globals.h"
#include "plot.h"
#include <ctype.h>

/*
   Contains some purely mathematical routines, possibly of use in
   more than one MacAnova function.
     doSwp(), fsolve(), factorize()
	 epslon()
   Also contains substitutes for various functions that may not be
   available on all systems.
     mystrtod() (synonym for strtod()), pythag() (synonym for hypot())
	 intpow() (replacement for pow(x,y) when y is an integer)
	 mylgamma() (synonym for lgamma())
	 index(), rindex()
  970217 added fsolve()
  000525 added mylog2()
  001123 added factorize()
  020528 replaced mylgamma() by a more accurate algorithm
  030522 added fuzzyEqual() to test for approximate equality
*/

/*
   Do single non-symmetric Beaton SWP on matrix.
	This version of swp does a non-symmetric sweep of the	m by n  matrix cp.
	The pivot row/column is specified by k.  The original k-th diagonal
	element is cpdiag

	If full != 0, a complete swp of all of cp is done.
	If full == 0, only columns > k are swept;  use this if you are
       sweeping on an increasing sequence of columns and are primarily
	   interested in the lower right hand corner of cp.
	In case of singularity, 0 is returned; otherwise 1
	
   011217 added argument tol for testing whether pivot is too small;
          previously used global REGTOL; 0 value means use REGTOL
*/
#define SINGULAR    0
#define NONSINGULAR 1

long doSwp(double * cp,long m,long n,long k,double cpdiag, double tol,
		   long full)
{
	long             i,j, istart, jstart;
	register double  beta;
	double           pivot;
	double          *cpj, *cpk;
	WHERE("doSwp");

	tol = (tol > 0) ? tol : REGTOL;

	istart = jstart = (full) ? 0 : k + 1;
	cpk = cp + m*k;		/* column k of cp */
	cpj = cp + m*jstart; /* column j of cp */

	pivot = cpk[k]; /* cp[k,k]*/
	if (fabs(pivot) <= tol*fabs(cpdiag))
	{
		return (SINGULAR);
	}

	/* sweep out of all columns except k */
	for (j = jstart;j < n;j++)
	{
		if (j != k)
		{
			beta =  cpj[k]/pivot; /* cp[k,j]/c[k,k] */
			for (i = istart;i < m;i++)
			{ /* no test for i == k => cost 1 redundant operation */
				cpj[i] -= beta*cpk[i]; /* cp[i,j]-=cp[i,k]*beta */
			} /*for (i = istart;i < m;i++)*/
			if (full)
			{
				cpj[k] = -beta; /* cp[k,j] = -beta*/
			}
		} /*if (j != k)*/
		cpj += m;
	} /*for (j = jstart;j < n;j++)*/

	/* Now fix up column k*/
	if (full)
	{
		for (i = 0;i < m;i++)
		{
			if (i != k)
			{
				cpk[i] /= pivot; /* cp[i,k] /= cp[k,k];*/
			} /*if (i != k)*/
			else
			{
				cpk[i] = 1.0/pivot;	/* cp[k,k] = 1/cp[k,k]; */
			}
		} /*for (i = 0;i < m;i++)*/
	} /*if (full)*/
	
	return (NONSINGULAR);

} /*doSwp()*/

#include "handlers.h"

enum primeConsts
{
	nPrimes    = 54, /* number of primes < 256 */
	primeChunk = 1000 /* check for interrupt about every primerChunk products*/
};

static unsigned char    Primes[] = /*primes < 256*/
{
	  2,   3,   5,   7,  11,  13,  17,  19,  23,  29,  31,
	 37,  41,  43,  47,  53,  59,  61,  67,  71,  73,  79,
	 83,  89,  97, 101, 103, 107, 109, 113, 127, 131, 137,
	139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193,
	197, 199, 211, 223, 227, 229, 233, 239, 241, 251
};

/*
	Find prime factors of x, returning them in array factors;
	The number of factors is returned as value.

    The first 54 primes are explicitly tested; then brute force takes over,
	every odd number > 251 and <= sqrt(unfactored part of x) is tested

	001123 added to mathutil.c
*/

int factorize(double x, double factors [])
{
	int               jprime = 0;

	clearNproducts();
	if (x <= 3.0)
	{
		factors[jprime++] = x;
	} /*if (x <= 3.0)*/
	else
	{
		int         i = 0;
		int         nOps = 0;
		double      thisPrime = Primes[i++];

		while (thisPrime * thisPrime <= x)
		{
			double      y = floor(x / thisPrime);

			if (y * thisPrime < x)
			{
				thisPrime = (i < nPrimes) ? Primes[i++] : thisPrime + 2;
			}
			else
			{
				factors[jprime++] = thisPrime;
				x = y;
			}
			nOps += 2;
			if (nOps > primeChunk)
			{
				incNproducts(nOps);
				testNproducts(interrupted);
				nOps = 0;
			}
		} /*while (thisPrime * thisPrime <= x)*/
		if (x > 1)
		{
			factors[jprime++] = x;
		}
	} /*if (x <= 3.0){}else{}*/
	return (jprime);

  interrupted:
	return (-1);
} /*factorize()*/


#undef MAX
#undef MIN
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))

#undef INT
#define INT       long

/*
  Routine to find the solution z of f(z)=a where the function f(.)
  is strictly monotonic either increasing or decreasing.

  Arguments:
  All real arguments and func are in double precision.
  arg    the value for which the solution is to be found
         in other words we want z to solve f(z) = arg
  func   an external function that evaluates f(z) and returns
         as its value f(z)
  param, nparam
         additional arguments that func may require.  They are passed
         straight through when func is called.  func
         is called as  y = func(z,param,nparam).  param is a vector
         of additional parameters, and nparam is the number of
         such additional parameters.
  ntab  the size of the working arrays x,y, and pr
  xmax, xmin
        given limits within which the solution
        is known a priori to lie.  If you don't know the range
        of possible values, set large positive and negative values
        for xmax and xmin respectively.
  eps   the value |f(z)-arg| must attain for the procedure
        to be considered to have converged.
  x and y
        work arrays for intermediate trial z values and their associated
        f(z)-arg. On input, the first two elements of x must be set by the
        user to initial guesses at the solution.  Ideally
        these guesses should straddle the true solution.
  nxy is returned
        The number of x,y, pairs studied.
  pr    a work array.
  soln  the solution z to the equation - that is
        func(soln,param,nparam) = arg
  iter is returned
        the number of iterations
  ifault is an error code.  it should be zero.

   Code received from Douglas Hawkins, Applied Statistics,
   University of Minnesota, 1/23/96

   Modified by C. Bingham
   970217 added argument itmax  also nxy, iter and ifault are returned
   in argument vecto report

   971016 func() now has a 4th long * argument to report errors
   or interruption
*/
#define Nxy    report[0]
#define Iter   report[1]
#define Ifault report[2]


double fsolve(double arg, double (*func) (double ,double [], INT, INT *),
			  double param [], INT nparam, double xmax, double xmin,
			  double eps, INT itmax, INT ntab,
			  double x [], double y [], double pr [], INT report[])
/*	double x[ntab], y[ntab], param[nparam], pr[ntab];*/
{
	double     aminim = 0.8, big = 1e20, small = 1.e-20;
	double     step, cerr, up, xlo, xhi, ydif, prnew, guess;
	double     off;
	INT        nt, i, iii, low, between;
	WHERE("fsolve");
	
	Iter = 2;
	Ifault = 0;
	nt = 1;
	iii = 1;
	low = -1;
	step = 1.0;
	if (ntab < 4 || xmax <= xmin || eps <= 0.0)
	{
		Ifault = FSOLVEBADARGUMENTS;
		goto errorExit;
	} /*if (ntab < 4 || xmax <= xmin || eps <= 0.0)*/

	y[0] = func(x[0], param, nparam, &Ifault) - arg;
	if (Ifault)
	{
		goto errorExit;
	}
	
	y[1] = func(x[1], param, nparam, &Ifault) - arg;
	if (Ifault)
	{
		goto errorExit;
	}

	if (fabs(y[0] - y[1]) < small)
	{
		Ifault = FSOLVEBADGUESSES;
		goto errorExit;
	}

	cerr = MIN(fabs(y[0]), fabs(y[1]));

	up = ((x[0] - x[1])*(y[0] - y[1]) > 0.0) ? 1.0 : -1.0;

	if (y[0]*y[1] <= 0.0)
	{
		/* solution between x[0] and x[1] */
		xlo = MIN(x[0], x[1]);
		xhi = MAX(x[0], x[1]);
	} /*if (y[0]*y[1] <= 0.0)*/
	else if (y[0] * up > 0.0)
	{
		/* both x[0] and x[1] too high*/
		xlo = xmin;
		xhi = MIN(x[0], x[1]);
	}
	else
	{
		/* both x[0] and x[1] too low */
		xlo = MAX(x[0], x[1]);
		xhi = xmax;
	}
	
	while (1)
	{
		guess = 0.0;
		if (low < 2)
		{
			low++;
		}
		
		if (nt - low < 2)
		{
			pr[low] = x[low];
		}
		prnew = x[nt];
		for (i = low;i < nt ;i++)
		{
			ydif = y[i] - y[nt];
			prnew *= y[i]/ydif;
			pr[i] *= -y[nt]/ydif;
			guess += pr[i];
		} /*for (i = low;i < nt ;i++)*/

		pr[nt] = prnew;
		guess += prnew;
		between = ((xhi - guess)*(xlo - guess) <= 0.0);
		while (2)
		{
			if (between)
			{
				double          oldcerr;
				
/*70*/			if (cerr < eps)
				{
					goto normalExit;
				}

				if (++Iter > itmax)
				{
					Ifault = FSOLVENOTCONVERGED;
					goto errorExit;
				}
				off = func(guess, param, nparam, &Ifault) - arg;
				if (Ifault)
				{
					goto errorExit;
				}
				for (i = 0;i <= nt ;i++)
				{
					if ((y[i] - off)*up*(x[i] - guess) < 0.0)
					{
						Ifault = FSOLVENONMONOTONIC;
						goto errorExit;
					}
/*80*/			} /*for (i = 0;i <= nt ;i++)*/
				if (up*off < 0.0)
				{
					xlo = guess;
				}
				if (up*off > 0.0)
				{
					xhi = guess;
				}
				oldcerr = cerr;
				cerr = fabs(off);
				if (cerr <= aminim*oldcerr)
				{
					break;
				}
			} /*if (between)*/
			else
			{
				between = 1;
			}
			
			if (xlo > (-big) && xhi < big) /*60*/
			{
				guess = 0.5*(xlo + xhi);
			} /*if (xlo > (-big) && xhi < big)*/
			else if (cerr > eps) /*40*/
			{
				if (xlo <= (-big))
				{
					guess = xhi - step;
					guess = MAX(guess, xmin);
				}
				else
				{
					guess = xlo + step;
					guess = MIN(guess, xmax);
				} 
				step += step;
			} /*if (xlo > (-big) && xhi < big){}else if (cerr > eps){}*/
		} /*while (2)*/

		if (nt < ntab - 1)
		{
			nt++;
		} /*if (nt < ntab - 1)*/
		else
		{
			if (++iii >= ntab)
			{
				iii = 2;
			}
			for (i=low;i < ntab ;i++)
			{
				pr[i] *= -(y[i] - y[iii])/y[iii];
			}
			x[iii] = x[nt];
			y[iii] = y[nt];
			pr[iii] = pr[nt];
		} /*if (nt < ntab - 1){}else{}*/
		x[nt] = guess;
		y[nt] = off;
	} /*while (1)*/
	
  normalExit:
	/*fall through*/
	
  errorExit:
	Nxy = nt;
	return (guess);
	
} /*fsolve()*/

/*NOSTRTOD defined implies no working strtod in standard library*/
#if defined(NOSTRTOD)
/*
  version of strtod not using sscanf

  Note:  This is non-conforming since it always expects '.' as "radix
  character", regardless of locale

  C. Bingham, 911012
  960403 Corrections made
*/

#define MAXPOW 20
static double powers[MAXPOW+1] =
{
	1e0,1e1,1e2,1e3,1e4,1e5,1e6,1e7,1e8,1e9,1e10,
	1e11,1e12,1e13,1e14,1e15,1e16,1e17,1e18,1e19,1e20
};

static double mystrtod1(char *nptr, char **eptr)
{
	long       ssign = 1, esign = 1, decPlaces = 0;
	double     mantissa = 0.0, expon = 0.0;
	char      *endptr;
	long       place;
	short      c;
	WHERE("mystrtod");
	
	if (eptr != (char **) 0)
	{
		*eptr = nptr;
	}
	if (*nptr == '\0')
	{
		return (0.0);
	}
	/* expected pattern is [-+]?[0-9]*\.?[0-9]*\([eE][ -+]?[0-9][0-9]*\)? */

	place = 0;
	while (isspace( nptr[place] ))
	{
		place++;					/* skip leading nonnumeric */
	}

	c = nptr[place];
	if (c == '+' || c == '-')
	{
		if (c == '-')
		{
			ssign = -1;
		}
		c = nptr[++place];
	}

	if (isdigit(c))
	{							/* decode integral part of mantissa */
		do
		{
			mantissa = 10.0*mantissa + (c - '0');
			c = nptr[++place];
		} while (isdigit(c));
	}
	else if (c != '.' || !isdigit(nptr[place+1]))
	{							/* not a number */
		return (0.0);
	}

	if (c == '.')
	{							/* decode fractional part of mantissa */
		c = nptr[++place];
		if (isdigit(c))
		{
			do
			{
				mantissa = 10.*mantissa + (c - '0');
				decPlaces++;
				c = nptr[++place];
			}while (isdigit(c));
		} /*if (isdigit(c))*/
	} /*if (c == '.')*/

	endptr = nptr + place;
	place++;

	if (c == 'e' || c == 'E')
	{ /* decode exponent */
		if ((c = nptr[place]) == '-' || c == '+' || c == ' ')
		{
			if (c == '-')
			{
				esign = -1;
			}
			c = nptr[++place];
		}
		if (isdigit(c))
		{
			do
			{ /* decode values of exponent */
				expon = 10.0*expon + (c - '0');
				c = nptr[++place];
			} while (isdigit(c));
			endptr = nptr + place;
			if (esign < 0)
			{
				expon = -expon;
			}
		}
	}
	expon -= (double) decPlaces;

	if (fabs(expon) > (double) MAXPOW)
	{
#ifdef DBL_MAX_10_EXP
		if (expon > DBL_MAX_10_EXP)
		{
			mantissa = TOOBIGVALUE;
		}
		else
#endif /*DBL_MAX_10_EXP*/
		{
			mantissa *= pow(10.,expon);
		}
	} /*if (fabs(expon) > (double) MAXPOW)*/
	else
	{
		if (expon >= 0.0)
		{
			mantissa *= powers[(long) expon];
		}
		else
		{
			mantissa /= powers[-(long) expon];
		}
#ifdef HASINFINITY
		if (isInfinite(mantissa))
		{
			mantissa = HUGE_VAL;
		}
#endif /*HASINFINITY*/
#ifdef HASNAN
		if (isNaN(mantissa))
		{
			mantissa = HUGE_VAL;
		}
#endif /*HASNAN*/
	} /*if (fabs(expon) > (double) MAXPOW){}else{}*/

	if (eptr != (char **) 0)
	{
		*eptr = endptr;
	}

	return ((ssign < 0) ? -mantissa : mantissa);

} /*mystrtod()*/
#else /*NOSTRTOD*/
/* system library has strtod() */
static double mystrtod1(char *nptr, char **eptr)
{
#ifndef MW_CW
	return (strtod(nptr, eptr));
#else /*MW_CW*/
/*
	This code is intended to adjust for a bug in Codewarrior strtod():
	After strtod("12e 1, nptr, &eptr), say, the value returned
	is 12 but eptr = nptr.
	This resulted in an infinite loop in vecread()
*/
	double      value = strtod(nptr, eptr);
	char       *numberChars = "+-0123456789.";

	if (*eptr == nptr)
	{
		while (strchr(numberChars, *nptr) != (char *) 0)
		{
			nptr++;
		}
		*eptr = nptr;
	}
	return (value);
#endif /*MW_CW*/
} /*mystrtod()*/

#endif /*NOSTRTOD*/
#undef NL
#define NL    10
/*
  Summer 2002
  Now ignores '_' in matissa of numbers
  Treats 'd' and 'D' as if they were 'e' so Fortran style double precision
  literals are correctly translated
*/
double mystrtod(char *nptr, char **eptr)
{
	double         value;
	int            place;
	char          *numberChars = "+-0123456789.";
	char           oldNl, *nlplace;

	for (place = 0; !isNewline(nptr[place]) && isspace(nptr[place]); place++)
	{
		;
	}
	if (strchr(numberChars, nptr[place]) == (char *) 0)
	{
		/*can't be a number*/
		*eptr = nptr;
		return (0.0);
	}
	
	nlplace = strchr(nptr, '\n');
#ifdef MACINTOSH
	if (nlplace == (char *) 0)
	{
		nlplace = strchr(nptr, NL);
	}
#endif /*MACINTOSH*/
	if (nlplace == (char *) 0)
	{
		nlplace = nptr + strlen(nptr);
	}
	/*
		This is a little risque; temporarily replace newline by '\0'
	*/
	oldNl = *nlplace;
	*nlplace = '\0';
	if (strchr(nptr + place, '_') == (char *) 0 &&
		strchr(nptr + place, 'd') == (char *) 0 &&
		strchr(nptr + place, 'D') == (char *) 0)
	{
		/*no '_' or 'd' or 'D' in line*/
		*nlplace = oldNl;
		value = mystrtod1(nptr, eptr);
	}
	else
	{
		/* 
		   Copy number to tempInput, skipping any '_' in mantissa
		   and changing 'd' or 'D' to 'e'
		*/
		char        c = nptr[place];
		char        tempInput[1011], *eptr1;
		int         j = 0, foundPoint = 0;
		
		*nlplace = oldNl;
		if (c == '+' || c == '-')
		{
			tempInput[j++] = c;
			place++;
		}

		while ((c = nptr[place], j < 1000) &&
			   (isdigit(c) || (!foundPoint && c == '.') || c == '_'))
		{
			if (c != '_')
			{
				if (c == '.')
				{
					foundPoint = 1;
				}
				tempInput[j++] = c;
			}
			place++;
		}

		if (c == 'd' || c == 'D' || c == 'E')
		{
			c = 'e';
		}
		
		if (c == 'e')
		{
			int       place1 = place+1;
			
			if (nptr[place1] == '-' || nptr[place1] == '+')
			{
				place1++;
			}
			if (isdigit(nptr[place1]))
			{
				/* Copy exponent*/
				tempInput[j++] = c;
				if (nptr[++place] == '-' || nptr[place] == '+')
				{
					tempInput[j++] = nptr[place++];
				}
				while (nptr[place] == '0' && isdigit(nptr[place+1]))
				{
					/* skip leading 0's of exponent */
					place++;
				}
				while (isdigit(nptr[place]) && j < 1010)
				{
					/* copy rest of exponent */
					tempInput[j++] = nptr[place++];
				}
			}
		}
		tempInput[j] = '\0';
		value = mystrtod1(tempInput, &eptr1);
		*eptr = nptr + place;
	}
	return (value);
} /*mystrtod()*/
#ifdef NOHYPOT
#undef MAX
#undef MIN
#define MAX(A,B) (((A) > (B)) ? (A) : (B))
#undef MIN
#define MIN(A,B) (((A) < (B)) ? (A) : (B))

/*
   Finds sqrt(a**2+b**2) without overflow or destructive underflow

   980821 added check for overflow
*/
#if (1)
/*
  Version using sqrt()
  On HPGCC version, gives same result as library hypot(), more accurate
  than iterative version below
*/
double pythag(double a, double b)
{
	double        p;
#ifdef DBL_MAX
	double        dbl_max = DBL_MAX;
#else /*DBL_MAX*/
	double        dbl_max = TOOBIGVALUE;
#endif /*DBL_MAX*/

	a = fabs(a);
	b = fabs(b);
	p = MAX(a, b);

	if (p != 0.0)
	{
		double        r = MIN(a, b)/p;
		double        p1 = sqrt(1.0 + r*r);

		p = (p > dbl_max/p1) ? TOOBIGVALUE : (p1*p);
	} /*if (p != 0.0)*/
	return (p);
} /*pythag()*/
#else /*1*/
/*
  Version adapted from function PYTHAG in Eispack that computes
  square root iteratively
*/
double pythag(double a, double b)
{
	double        p;

	a = fabs(a);
	b = fabs(b);
	p = MAX(a, b);

	if (p != 0.0)
	{
		double        r = MIN(a, b)/p;
		double        p1 = 1.0;
		double        s, t, u, temp;
		
		r *= r;
		while (1)
		{
			t = 4.0 + r;
			if (t == 4.0)
			{
				break ;
			}
			s = r/t;
			u = 1.0 + 2.0*s;
			p1 *= u;
			temp = s/u;
			r *= temp*temp;
		} /*while (1)*/
		p = (p > TOOBIGVALUE/p1) ? TOOBIGVALUE : (p1*p);
	} /*if (p != 0.0)*/
	return (p);
} /*pythag()*/
#endif /*1*/
#else /*NOHYPOT*/
#ifndef MPW /*following is inconsistant with MPW*/
extern double hypot(double, double);
#endif /*MPW*/

double pythag(double a, double b)
{
	return hypot(a,b);
} /*pythag()*/
#endif /*NOHYPOT*/

#ifndef USEPOW
/*
  To be used when system pow() is inefficient with integer exponents
  as with MPW C, or is inaccurate as with gcc in HPUX
*/

double intpow(register double x, double p)
{
	register double         y;
	register long           iexp = labs((long) p);
	
	y = (iexp & 1) ? x : 1.0;
	
	for (iexp >>= 1 ;iexp != 0;iexp >>= 1)
	{
		x *= x;
		if (iexp & 1)
		{
			y *= x;
		}
	} /*for (iexp >>= 1 ;iexp != 0;iexp >>= 1)*/
	
	return ((p < 0.0) ? 1.0/y : y);
} /*intpow()*/
#else /*USEPOW*/
double intpow(double x, double p)
{
	double       y = pow(fabs(x), p);

	return ((x < 0 && ((long) p) % 2 != 0) ? -y : y);
} /*intpow()*/
#endif /*USEPOW*/

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Trans
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#if defined(MW_CW) && !defined(powerpc) && !defined(__ppc__) && !defined(ppc) && !defined(__POWERPC__)
/*
	Avoid problems with using both math.h and fp.h on 68K Mac compile
*/
#define NOLGAMMA
#endif

#ifndef NOLGAMMA
#ifdef MW_CW
#include <fp.h>
#else /*MW_CW*/
extern double lgamma(double);
#endif /*MW_CW*/

double mylgamma(double x)
{
	return (lgamma(x));
} /*mylgamma()*/
#else /*NOLGAMMA*/

/* No system lgamma available */
#if (1)
/*
  020528 New version of mylgamma() based on a translation of
  gamln() by D. E. Amos
*/
#define gamln  mylgamma
/*
  Written by D. E. Amos,  September,  1977.

  References
      Sand - 77 - 1518

      Computer Approximations by J.F.Hart,  et.al.,  SIAM Series in
      Applied Mathematics,  Wiley,  1968,  p.135 - 136.

      NBS Handbook of Mathematical Functions,  AMS 55,  by
      M. Abramowitz and I.A. Stegun,  December. 1955,  p.257.

  Abstract
      gamln computes the natural log of the gamma function for
      x.gt.0. A rational Chebyshev approximation is used on
      8.lt.x.lt.1000.,  the asymptotic expansion for x.ge.1000. and
      a rational Chebyshev approximation on 2.lt.x.lt.3. For
      0.lt.x.lt.8. and x non - integral,  forward or backward
      recursion fills in the intervals  0.lt.x.lt.2 and
      3.lt.x.lt.8. For x=1., 2., ..., 100.,  gamln is set to
      natural logs of factorials.

  Description of arguments

      Input
        x - x.gt.0

      Output
        gamln - natural log of the gamma function at x

  Error conditions
      improper input argument - a fatal error
*/

/*
  In this C version, the exact values of log(gamma(j)) were recomputed
  to more significant digits using Linux bc.

  Constant MaxExactLgamma determines the largest value of j for which
  the log(gamma(j)) will use the exact values.  As currently coded,
  it should be either 16 or 100
*/
enum exactSize
{
	MaxExactLgamma = 16
};

static double    P[5] =
{
	7.66345188000000e-04, -5.94095610520000e-04, 7.93643110484500e-04,
	-2.77777775657725e-03, 8.33333333333169e-02
};

static double    Q[2] =
{
	-2.77777777777778e-03, 8.33333333333333e-02
};

static double    Pcoe[9] =
{
	2.97378664481017e-03, 9.23819455902760e-03,	1.09311595671044e-01,
	3.98067131020357e-01, 2.15994312846059e+00, 6.33806799938727e+00,
	2.07824725317921e+01, 3.60367725300248e+01, 6.20038380071273e+01
};

static double    Qcoe[4] =
{
	1.00000000000000e+00, -8.90601665949746e+00, 9.82252110471399e+00,
	6.20038380071270e+01
};

#if (1)
/* The following computed using Linux bc */
static double            Gln[MaxExactLgamma] =
{
	                        0,                        0
	,   .69314718055994530941,   1.79175946922805500081
	,  3.17805383034794561964,   4.78749174278204599424
	,  6.57925121201010099506,   8.52516136106541430016
	, 10.60460290274525022841,  12.80182748008146961120
	, 15.10441257307551529522,  17.50230784587388583928
	, 19.98721449566188614951,  22.55216385312342288557
	, 25.19122118273868150009,  27.89927138384089156608
#if MaxExactLgamma > 16
	, 30.67186010608067280375,  33.50507345013688888400
	, 36.39544520803305357621,  39.33988418719949403622
	, 42.33561646075348502965,  45.38013889847690802616
	, 48.47118135183522387963,  51.60667556776437357044
	, 54.78472939811231919009,  58.00360522298051993929
	, 61.26170176100200198476,  64.55753862700633105895
	, 67.88974313718153498289,  71.25703896716800901007
	, 74.65823634883016438548,  78.09222355331531063141
	, 81.55795945611503717850,  85.05446701758151741396
	, 88.58082754219767880362,  92.13617560368709248333
	, 95.71969454214320248495,  99.33061245478742692932
	,102.96819861451381269875, 106.63176026064345912620
	,110.32063971475739542905, 114.03421178146170323292
	,117.77188139974507153883, 121.53308151543863396231
	,125.31727114935689512520, 129.12393363912721488259
	,132.95257503561630988282, 136.80272263732636846964
	,140.67392364823425939870, 144.56574394634488600891
	,148.47776695177303206753, 152.40959258449735783918
	,156.36083630307878519406, 160.33112821663090702821
	,164.32011226319518141181, 168.32744544842765233048
	,172.35279713916280156383, 176.39584840699735171524
	,180.45629141754377105184, 184.53382886144949050245
	,188.62817342367159118728, 192.73904728784490243603
	,196.86618167288999399138, 201.00931639928152667928
	,205.16819948264119853578, 209.34258675253683564643
	,213.53224149456326119131, 217.73693411395422725098
	,221.95644181913033395006, 226.19054832372759333227
	,230.43904356577695232139, 234.70172344281826774272
	,238.97838956183432305376, 243.26884900298271418285
	,247.57291409618688393664, 251.89040220972319437723
	,256.22113555000952545608, 260.56494097186320930525
	,264.92164979855280104211, 269.29109765101982253628
	,273.67312428569370414855, 278.06757344036614291413
	,282.47429268763039602742, 286.89313329542699395089
	,291.32395009427030756623, 295.76660135076062402108
	,300.22094864701413175397, 304.68685676566871547255
	,309.16419358014692194486, 313.65282994987906178318
	,318.15263962020932684999, 322.66349912672617689115
	,327.18528770377521720079, 331.71788719692847313811
	,336.26118197919847703435, 340.81505887079901786896
	,345.37940706226685410744, 349.95411804077023692956
	,354.53908551944080884919, 359.13420536957539877604
#endif
};
#else
/* The following were the original constants */
static double            Gln[MaxExactLgamma] =
{
	0.0,                  0.0,                  6.93147180559945e-01,
	1.79175946922806e+00, 3.17805383034795e+00, 4.78749174278205e+00,
	6.57925121201010e+00, 8.52516136106541e+00, 1.06046029027453e+01,
	1.28018274800815e+01, 1.51044125730755e+01, 1.75023078458739e+01,
	1.99872144956619e+01, 2.25521638531234e+01, 2.51912211827387e+01,
	2.78992713838409e+01, 3.06718601060807e+01, 3.35050734501369e+01,
	3.63954452080331e+01, 3.93398841871995e+01, 4.23356164607535e+01,
	4.53801388984769e+01, 4.84711813518352e+01, 5.16066755677644e+01,
	5.47847293981123e+01, 5.80036052229805e+01, 6.12617017610020e+01,
	6.45575386270063e+01, 6.78897431371815e+01, 7.12570389671680e+01,
	7.46582363488302e+01, 7.80922235533153e+01, 8.15579594561150e+01,
	8.50544670175815e+01, 8.85808275421977e+01, 9.21361756036871e+01,
	9.57196945421432e+01, 9.93306124547874e+01, 1.02968198614514e+02,
	1.06631760260643e+02, 1.10320639714757e+02, 1.14034211781462e+02,
	1.17771881399745e+02, 1.21533081515439e+02, 1.25317271149357e+02,
	1.29123933639127e+02, 1.32952575035616e+02, 1.36802722637326e+02,
	1.40673923648234e+02, 1.44565743946345e+02, 1.48477766951773e+02,
	1.52409592584497e+02, 1.56360836303079e+02, 1.60331128216631e+02,
	1.64320112263195e+02, 1.68327445448428e+02, 1.72352797139163e+02,
	1.76395848406997e+02, 1.80456291417544e+02, 1.84533828861449e+02,
	1.88628173423672e+02, 1.92739047287845e+02, 1.96866181672890e+02,
	2.01009316399282e+02, 2.05168199482641e+02, 2.09342586752537e+02,
	2.13532241494563e+02, 2.17736934113954e+02, 2.21956441819130e+02,
	2.26190548323728e+02, 2.30439043565777e+02, 2.34701723442818e+02,
	2.38978389561834e+02, 2.43268849002983e+02, 2.47572914096187e+02,
	2.51890402209723e+02, 2.56221135550010e+02, 2.60564940971863e+02,
	2.64921649798553e+02, 2.69291097651020e+02, 2.73673124285694e+02,
	2.78067573440366e+02, 2.82474292687630e+02, 2.86893133295427e+02,
	2.91323950094270e+02, 2.95766601350761e+02, 3.00220948647014e+02,
	3.04686856765669e+02, 3.09164193580147e+02, 3.13652829949879e+02,
	3.18152639620209e+02, 3.22663499126726e+02, 3.27185287703775e+02,
	3.31717887196928e+02, 3.36261181979198e+02, 3.40815058870799e+02,
	3.45379407062267e+02, 3.49954118040770e+02, 3.54539085519441e+02,
	3.59134205369575e+02
};
#endif

double gamln(double x)
{
	double          xlim1 = 8.0, xlim2 = 1000.0;
	double          rtwpil = 9.18938533204673e-01;/*log(2*PI)/2*/
	double          t, px, qx, rx, rxx, xx, dgam;
	int             i, ndx, nxm;
	
	if (x <= 0)
	{
#if (0)
		xerror("gamln argument is less than or equal to zero ", 46, 1,2);
#endif /*0*/
		/* Minimum possible value for lgamma(x) is -0.1214862905358496 */
		return (-1.0);
	}

	ndx = x;
	t = x - (double) ndx;

	if (t == 0 && ndx <= MaxExactLgamma)
	{
		/*
			Table look up for integer arguments less than or equal
			MacExactLgamma.
		*/
		return (Gln[ndx - 1]);
	}

	if (x <= xlim1)
	{
		/* Small Non-integral x */
		/*
		  Rational chebyshev approximation on 2.lt.x.lt.3 for gamma(x)
		*/

		nxm = ndx - 2;
		px = Pcoe[0];
		for (i = 1;i < 9 ;i++)
		{
			px = t*px + Pcoe[i];
		}
		qx = Qcoe[0];
		for (i = 1;i < 4 ;i++)
		{
			qx = t*qx + Qcoe[i];
		}
		dgam = px/qx;
		if (nxm > 0)
		{
			/* Forward recursion for 3.lt.x.lt.8 */
			xx = 2.0 + t;
			for (i = 0;i < nxm ;i++)
			{
				dgam *= xx;
				xx += 1.0;
			}
		} /*if (nxm > 0) */
		else if (nxm != 0)
		{
			/* Backward recursion for 0.lt.x.lt.2 */
			dgam /= 1.0 + t;
			if (nxm != -1)
			{
				dgam /= t;
			}
		}
		return (log(dgam));
	} /*if (x <= xlim1) */

	/*      x.gt.xlim1 */

	rx = 1.0/x;
	rxx = rx*rx;
	if (x >= xlim2)
	{
		px = Q[0]*rxx + Q[1];
	}
	else
	{
		/* x.lt.xlim2 */

		px = P[0];
		for (i = 1;i < 5 ;i++)
		{
			px = px*rxx + P[i];
		}
	}
	return (px*rx + (x - .5)*log(x) - x + rtwpil);
} /*gamln()*/

#else /*1*/
static double   ln2pio2 = 0.9189385332046727417803297;

static double   clocal[] =
{
	-0.577191652,
	0.988205891,
	-0.897056937,
	0.918206857,
	-0.756704078,
	0.482199394,
	-0.193527818,
	0.035868343,
};
static double   casymp[] =
{
	12.,
	-360.,
	1260.,
	-1680.,
};


double          mylgamma(double x)
{
	long            i;
	double          out, z, z2, value;

	if (x > 8.)
	{			/* asymptotic, 6.1.41 of Abramowitz and Stegun */
		z = 1 / x;
		z2 = z * z;
		out = 0.0;
		for (i = 0; i < 4; i++)
		{
			out += z / casymp[i];
			z *= z2;
		}
		return ((x - .5) * log(x) - x + ln2pio2 + out);
	}

	/* small x, 6.1.36 of Abramowitz and Stegun */
	value = 0.0;
	while (x < 1.0)
	{
		value -= log(x++);
	}
	while (x > 2.0)
	{
		value += log(--x);
	}
	
	z = x - 1.;
	z2 = z;
	out = 1.;
	for (i = 0; i < 8; i++)
	{
		out += clocal[i] * z2;
		z2 *= z;
	}
	return (log(out) + value);
} /*mylgamma()*/
#endif /*1*/
#endif /*NOLGAMMA*/

#undef MAX
#undef MIN
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))


/*
 ***Begin prologue  dpsifn
 ***Date written   820601   (yymmdd)
 ***Revision date  851111   (yymmdd)
 ***Category no.  c7c
 ***Keywords  derivatives of the gamma function, double precision, 
              gamma function, polygamma function, psi function
 ***Author  Amos,  Donald E.,  Sandia National Laboratories
 ***Purpose  Compute derivatives of the psi function.
             (18 or fewer digits)
 ***Description

        **** double precision version of psifn ****
          The following definitions are used in dpsifn:

       Definition 1
          psi(x) = d/dx (ln(gamma(x)),  the first derivative of
                   the log gamma function.
       Definition 2
                      k   k
          psi(k, x) = d /dx (psi(x)),  the k - th derivative of psi(x).
    ___________________________________________________________________
       dpsifn computes a sequence of scaled derivatives of
       the psi function;i.e. for fixed x and m it computes
       the m - member sequence

                     ((-1)**(k + 1)/gamma(k + 1))*psi(k, x)
                        for k = n, ..., n + m-1

       where psi(k, x) is as defined above.   For kode=1,  dpsifn returns
       the scaled derivatives as described.  kode=2 is operative only
       when k=0 and in that case dpsifn returns - psi(x) + ln(x).  That
       is,  the logarithmic behavior for large x is removed when kode=2
       and k=0.  When sums or differences of psi functions are computed
       the logarithmic terms can be combined analytically and computed
       separately to help retain significant digits.

          Note that call dpsifn(x, 0,1, 1,ans) results in
                    ans = -psi(x)

      Input      x is double precision
            x - argument,  x .gt. 0.0e0
            n - first member of the sequence,  0 .le. n .le. 100
                     n=0 gives ans(1) = -psi(x)       for kode=1
                                        -psi(x) + ln(x) for kode=2
            kode - selection parameter
                     kode=1 returns scaled derivatives of the psi
                     function.
                     kode=2 returns scaled derivatives of the psi
                     function except when n=0. in this case, 
                     ans(1) = -psi(x) + ln(x) is returned.
            m - number of members of the sequence,  m.ge.1

     Output     ans is double precision
            ans - a vector of length at least m whose first m
                     components contain the sequence of derivatives
                     scaled according to kode.
            nz - underflow flag
                     nz.eq.0,  a normal return
                     nz.ne.0,  underflow,  last nz components of ans are
                              set to zero,  ans(m - k+1)=0.0,  k=1, ..., nz
            ierr - error flag
                     ierr=0,  a normal return,  computation completed
                     ierr=1,  input error,      no computation
                     ierr=2,  overflow,         x too small or n + m-1 too
                             large or both
                     ierr=3,  error,            n too large. dimensioned
                             array trmr(nmax) is not large enough for n

          The nominal computational accuracy is the maximum of unit
          roundoff (=d1mach(4)) and 1.0e-18 since critical constants
          are given to only 18 digits.

          psifn is the single precision version of dpsifn.
 ***long description

          The basic method of evaluation is the asymptotic expansion
          for large x.ge.xmin followed by backward recursion on a two
          term recursion relation

                   w(x + 1) + x**(-n - 1) = w(x).

          This is supplemented by a series

                   sum( (x + k)**(-n - 1) , k=0, 1,2, ... )

          which converges rapidly for large n. Both xmin and the
          number of terms of the series are calculated from the unit
          roundoff of the machine environment.
 ***References  Handbook of Mathematical Functions,  AMS 55,  National
                  Bureau of Standards by M. Abramowitz and I. A.
                  Stegun,  1964,  pp.258 - 260,  eqtns. 6.3.5,  6.3.18,  6.4.6, 
                  6.4.9,  6.4.10
                A portable fortran subroutine for derivatives of the
                  psi function by D. E. Amos,  ACM Trans. Math Software, 
                  1983,  Algorithm 610
 ***Routines called  d1mach, i1mach
 ***End prologue  dpsifn
*/
#include <float.h>
#if (1)
#ifndef BCPP
#define i1mach14  DBL_MANT_DIG
#define i1mach15  DBL_MIN_EXP
#define i1mach16  DBL_MAX_EXP
#define d1mach4   DBL_EPSILON
#else /*BCPP*/
#define i1mach14  53
#define i1mach15  -1021
#define i1mach16  1024
#define d1mach4   2.2204460492503131e-16
#endif /*BCPP*/
#define d1mach5   0.30102999566398119521 /*log10(2)*/
#else
int                 i1mach(int);
double              d1mach(int);
#define i1mach14  i1mach(14)
#define i1mach15  i1mach(15)
#define i1mach16  i1mach(16)
#define d1mach4   d1mach(4)
#define d1mach5   d1mach(5)
#endif /*1*/

/*
              Bernoulli numbers
*/

static double       Bern[] =
{
	1.00000000000000000e+00, -5.00000000000000000e-01,
	1.66666666666666667e-01, -3.33333333333333333e-02,
	2.38095238095238095e-02, -3.33333333333333333e-02,
	7.57575757575757576e-02, -2.53113553113553114e-01,
	1.16666666666666667e+00, -7.09215686274509804e+00,
	5.49711779448621554e+01, -5.29124242424242424e+02,
	6.19212318840579710e+03, -8.65802531135531136e+04,
	1.42551716666666667e+06, -2.72982310678160920e+07,
	6.01580873900642368e+08, -1.51163157670921569e+10,
	4.29614643061166667e+11, -1.37116552050883328e+13,
	4.88332318973593167e+14, -1.92965793419400681e+16
};

#define NMAX  100 /* maximum number of different orders */
#ifndef INT
#define INT long
#endif /*INT*/

#define dpsifn mypolygamma

void dpsifn(double x, int n, int kode, int m, double ans [], int * nz,
			int * ierr)
{
	INT                 i, j, k , mm, mx, nn, np, nx;
	double              trm[22], trmr[NMAX];
	double              arg, den, elim, eps;
	double              fln, fn, fnp, fns, fx, rln, rxsq;
	double              r1m4, r1m5;
	double              s, slope;
	double              t, ta, tk, tol, tols, tss, tst, tt, t1, t2;
	double              wdtol, xdmln, xdmy, xinc, xln, xm, xmin, yint;
	int                 count = 0;


	/* ***first executable statement  dpsifn */
	*ierr = 0;
	*nz = 0;
	if (x <= 0.0 || n < 0 || kode < 1 || kode > 2 || m < 1)
	{
		*ierr = 1;
		return;
	}
	
	/*
	  n: n_th derivative of psi(x) = (d/dx) log(gamma(x))
	  m: number of such derivatives from n to n + m - 1
	*/
	mm = m;
	/*
	  i1mach14 = T = DBL_MANT_DIG = number of binary digits in mantissa (53)
	  i1mach15 = DBL_MIN_EXP = minimum exponent (-1021)
	  i1mach16 = DBL_MAX_EXP = maximum exponent (1024)
	  d1mach4  = DBL_EPSILON = 2^(1 - T) = 2^-52 = 2.2204e-16)
	  d1mach5  = log10(2)
	*/
	if (-i1mach15 < i1mach16)
	{
		nx = -i1mach15;
	}
	else
	{
		nx = i1mach16;
	}
	//nx = MIN(-i1mach15, i1mach16);
	r1m5 = d1mach5;
	r1m4 = d1mach4*0.5;
	wdtol = MAX(r1m4, 0.5e-18);
	/*
	  elim = approximate exponential over and underflow limit
	*/
	elim = 2.302*((double) nx*r1m5 - 3.0);
	xln = log(x);

	while (1) 
	{
#if (0)
		if (++count > 200)
		{
			printf("count > 200, x = %g\n", x);
			exit (1);
		}
#endif /*0*/
		nn = n + mm - 1; /* highest order derivative wanted*/
		fn = (double) nn;
		fnp = fn + 1.0;
		t = fnp*xln;
		/*
		  Overflow and underflow test for small and large x
		*/

		if (fabs(t) > elim) 
		{
			if (t <= 0.0)
			{
				*nz = 0;
				*ierr = 2;
				return;
			}
		} /*if (fabs(t) > elim) */
		else 
		{
			if (x < wdtol)
			{
				ans[0] = pow(x, (double) -(n + 1));
				for (k = 0, i = 2;i <= mm ;i++, k++)
				{
					ans[k + 1] = ans[k]/x;
				}
				if (n == 0 && kode == 2)
				{
					ans[0] += xln;
				}
				return;
			} /*if (x < wdtol)*/
			/*
			  Compute xmin and the number of terms of the series,  fln + 1
			*/
			/* i1mach14 is number of base_b digits (IEEE 53) */
			rln = r1m5 * (double) i1mach14;

			rln = MIN(rln, 18.06);
			fln = MAX(rln, 3.0) - 3.0;
			yint = 3.50 + 0.40*fln;
			slope = 0.21 + fln*(0.0006038*fln + 0.008677);
			xm = yint + slope*fn;
			mx = (INT) xm + 1;
			xmin = (double) mx;

			if (n != 0) 
			{
				xm = -2.302*rln - MIN(0.0, xln);
				fns = (double) n;
				arg = xm/fns;
				arg = MIN(0.0, arg);
				eps = exp(arg);
				xm = 1.0 - eps;
				if (fabs(arg) < 1.0e-3)
				{
					xm = -arg;
				}
				fln = x*xm/eps;
				xm = xmin - x;
				if (xm > 7.0 && fln < 15.0)
				{
					nn = (INT) fln + 1;
					np = n + 1;
					t1 = (fns + 1.0)*xln;
					t = exp(-t1);
					s = t;
					den = x;
					for (i = 0;i < nn ;i++)
					{
						den++;
						trm[i] = pow(den, (double) -np);
						s += trm[i];
					} /*for (i = 0;i < nn ;i++)*/

					ans[0] = s;
					if (n == 0 && kode == 2)
					{
						ans[0] += xln;
					} /*if (n == 0 && kode == 2)*/

					/*
					  Generate higher derivatives,  j.gt.n
					*/

					tol = wdtol/5.0;
					for (j = 1;j < mm ;j++)
					{
						t /= x;
						s = t;
						tols = t*tol;
						den = x;
						for (i = 0;i < nn ;i++)
						{
							den++;
							trm[i] /= den;
							s += trm[i];
							if (trm[i] < tols)
							{
								break;
							}
						} /*for (i = 0;i < nn ;i++)*/
						ans[j] = s;
					} /*for (j = 1;j < mm ;j++)*/
					return;
				} /*if (xm > 7.0 && fln < 15.0)*/
			} /*if (n != 0) */
			if (x < xmin) 
			{
				xinc = xmin - (double) (INT) x;
				xdmy = x + xinc;
				xdmln = log(xdmy);
			}
			else
			{
				xinc = 0.0;
				xdmy = x;
				xdmln = xln;
			}

			/*
			  Generate w(n + mm - 1, x) by the asymptotic expansion
			*/

			t = fn*xdmln;
			t1 = xdmln + xdmln;
			t2 = t + xdmln;
			if (fabs(t) <= elim && fabs(t1) <= elim && fabs(t2) <= elim)
			{
				break; /*from while*/
			}
		} /*if (fabs(t) > elim) {}else{}*/
		(*nz)++;
		ans[--mm] = 0.0;
		if (mm == 0)
		{
			return;
		}
	} /*while (1) */

	tss = exp(-t);
	tt = 0.5/xdmy;
	t1 = tt;
	tst = wdtol*tt;
	if (nn != 0)
	{
		t1 = tt + 1.0/fn;
	}
	rxsq = 1.0/(xdmy*xdmy);
	ta = 0.5*rxsq;
	t = fnp*ta;
	s = t*Bern[2];

	if (fabs(s) >= tst) 
	{
		tk = 2.0;
		for (k = 3;k < 22 ;k++)
		{
			t = t*((tk + fn + 1.0)/(tk + 1.0))*((tk + fn)/(tk + 2.0))*rxsq;
			trm[k] = t*Bern[k];
			if (fabs(trm[k]) < tst)
			{
				break;
			}
			s += trm[k];
			tk += 2.0;
		} /*for (k = 3;k < 22 ;k++)*/
	} /*if (fabs(s) >= tst) */
	s = (s + t1)*tss;

	if (xinc != 0.0) 
	{
		/*      
			Backward recur from xdmy to x
		*/

		nx = (INT) xinc;
		np = nn + 1;
		if (nx > NMAX) 
		{
			*nz = 0;
			*ierr = 3;
			return;
		}

		if (nn == 0)
		{
			goto recurPsi0;
		}

		xm = xinc;
		/*      
			This loop should not be changed. fx is accurate when x is small
		*/

		for (i = 0;i < nx ;i++)
		{
			xm -= 1.0;
			fx = x + xm;
			trmr[i] = pow(fx, (double) -np);
			s += trmr[i];
		} /*for (i = 0;i < nx ;i++)*/
	} /*if (xinc != 0.0) */

	ans[mm - 1] = s;
	if (fn == 0.0)
	{
		goto adjustPsi0;
	}

	/*
      generate lower derivatives,  j.lt.n + mm - 1
	*/

	for (j = 2;j <= mm ;j++)
	{
		fnp = fn;
		fn -= 1.0;
		tss *= xdmy;
		t1 = tt;
		if (fn != 0.0)
		{
			t1 = tt + 1.0/fn;
		}
		t = fnp*ta;
		s = t*Bern[2];
		if (fabs(s) >= tst) 
		{
			tk = 3.0 + fnp;
			for (k = 3;k < 22 ;k++)
			{
				trm[k] *= fnp/tk;
				if (fabs(trm[k]) < tst)
				{
					break;
				}
				s += trm[k];
				tk += 2.0;
			} /*for (k = 3;k < 22 ;k++)*/
		} /*if (fabs(s) >= tst) */
		s = (s + t1)*tss;
		if (xinc != 0.0) 
		{
			if (fn == 0.0)
			{
				goto recurPsi0;
			}
			xm = xinc;
			for (i = 0;i < nx ;i++)
			{
				xm -= 1.0;
				fx = x + xm;
				trmr[i] *= fx;
				s += trmr[i];
			} /*for (i = 0;i < nx ;i++)*/
		} /*if (xinc != 0.0) */
		ans[mm - j] = s;
		if (fn == 0.0)
		{
			goto adjustPsi0;
		}
	} /*for (j = 1;j <= mm ;j++)*/
	return;

  recurPsi0:
	for (i = 1; i <= nx; i++)
	{
		s += 1.0/(x + (double) (nx - i));
	}
	
  adjustPsi0:
	if (kode != 2)
	{
		ans[0] = s - xdmln;
	}
	else if (xdmy != x) 
	{
		ans[0] = s - log(xdmy/x);
	}
} /*dpsifn()*/

#define NOLOG2
#ifndef NOLOG2
double log2(double);
#endif /*NOLOG2*/

double             mylog2(double x)
{
#ifndef NOLOG2
	return (log2(x));
#else /*NOLOG2*/
	return (MV_LOG2E*log(x));
#endif /*NOLOG2*/
} /*mylog2()*/

/*
  Test for approximate equality of val1, val2
  When fuzz <= 0, non-zero is returned only when val1 == val2 exactly
  When fuzz > 0 and relative != 0, non-zero is returned only when
   |val1 - val2| <= fuzz*(|val1| + |val2|)
  When fuzz > 0 and relative == 0, non-zero is returned only when
   |val1 - val2| <= fuzz
*/

int fuzzyEqual(double val1, double val2, double fuzz, int relative)
{
	return ((fuzz == 0.0) ? val1 == val2 :
			((relative) ?
			 fabs(val1 - val2) <= fuzz*(fabs(val1) + fabs(val2)) :
			 fabs(val1 - val2) <= fuzz));
} /*fuzzyEqual()*/

/* From Eispack.  Moved here from eigutils.c */
/*      estimate unit roundoff in quantities of size x. */


/*
	This program should function properly on all systems
      satisfying the following two assumptions,
         1.  The base used in representing floating point
             numbers is not a power of three.
         2.  The quantity  a  in statement 10 is represented to
             the accuracy used in floating point variables
             that are stored in memory.
      The label LAB10 and the goto LAB10 are intended to
      force optimizing compilers to generate code satisfying
      assumption 2.
      Under these assumptions, it should be true that,
             a  is not exactly equal to four-thirds,
             b  has a zero for its last bit or digit,
             c  is not exactly equal to one,
             eps  measures the separation of 1.0 from
                  the next larger floating point number.
      The developers of eispack would appreciate being informed
      about any systems where these assumptions do not hold.

      This version dated 4/6/83.
*/

double          epslon(double x)
{
	double          a, b, c, eps;

	a = 4.0 / 3.0;
/* use of label designed to out-fox optimizing compiler; see above */
  LAB10:
	b = a - 1.0;
	c = b + b + b;
	eps = fabs(c - 1.0);
	if (eps == 0)
	{
		goto LAB10;
	}
	return (eps * fabs(x));
} /*epslon()*/

#ifdef PROVIDEINDEX
/*
   These are needed under gcc on Epx since one of the libraries
   refers to them.  It seemed easier to provide the code
   than to try to find which, if any, additional library to link
   If these are needed for an installation, PROVIDEINDEX should
   be defined in platform.h
*/
char *index(char *s, char c)
{
	return (strchr(s, c));
}

char *rindex(char *s, char c)
{
	return (strrchr(s, c));
}

#endif /*PROVIDEINDEX*/
