/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.11 or later
*(C)*
*(C)* Copyright (c) 2001 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Odometer
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/


#include "globals.h"

#define MAXSTEP   MaxLong

/*
  An odometer is structure(digits:D,lower:L,upper:H, place:N)
    D is a vector of length M
    L and U are vectors of length M or scalars
      (interpreted as rep(L,M) or rep(U,M)),
      with L[i] <= U[i],
    N is an integer
*/
enum odometerPartNumbers
{
	odometerDigits,
	odometerLower,
	odometerUpper,
	odometerPlace,
	odometerNparts
};

static char         *OdometerParts[odometerNparts] =
{
	"digits", "lower", "upper", "place"
};

static int checkOdometer(Symbolhandle odometer)
{
	Symbolhandle     symh;
	long             length;
	int              ipart;
	
	if (TYPE(odometer) != STRUC || NCOMPS(odometer) != odometerNparts)
	{
		sprintf(OUTSTR,
				"ERROR: odometer variable not structure with %d components",
				odometerNparts);
		goto errorExit;
	}

	for (ipart = 0; ipart < odometerNparts; ipart++)
	{
		long      thisLength;
		
		symh = COMPVALUE(odometer, ipart);
		if (!isVector(symh) || TYPE(symh) != REAL)
		{
			sprintf(OUTSTR,
					"ERROR: odometer component %d not REAL %s",
					ipart+1, (ipart == odometerPlace) ?
						"scalar" : "vector or scalar");
			goto errorExit;
		} /*if (!isVector(symh) || TYPE(symh) != REAL)*/

		if (strcmp(NAME(symh), OdometerParts[ipart]) != 0)
		{
			sprintf(OUTSTR,
					"ERROR: argument not an odometer; name of component %d is not %s",
					ipart + 1, OdometerParts[ipart]);
			goto errorExit;
		}
		thisLength = DIMVAL(symh, 1);
		if (ipart == odometerDigits)
		{
			length = thisLength;
		}
		else if (ipart == odometerPlace)
		{
			if (thisLength > 1)
			{
				sprintf(OUTSTR,
						"ERROR: odometer component %s not a scalar",
						OdometerParts[ipart]);
				goto errorExit;
			}
		}
		else if (thisLength > 1 && thisLength != length)
		{
			sprintf(OUTSTR,
					"ERROR: vector odometer component %s not same length as %s",
					OdometerParts[ipart], OdometerParts[odometerDigits]);
			goto errorExit;
		}
	} /*for (ipart = 0; iparti < odometerNparts; iparti++)*/

	return (1);
	
  errorExit:
	putErrorOUTSTR();
	return (0);
} /*checkOdometer()*/

/*
  Function to set the odometer to place.
  Component 'place' is set to place
  Component 'digits' is set appropriately
*/
static void doSetOdometer(Symbolhandle symhOdometer, double place)
{
	Symbolhandle        symhDigits = COMPVALUE(symhOdometer, odometerDigits);
	Symbolhandle        symhLower = COMPVALUE(symhOdometer, odometerLower);
	Symbolhandle        symhUpper = COMPVALUE(symhOdometer, odometerUpper);
	Symbolhandle        symhPlace = COMPVALUE(symhOdometer, odometerPlace);
	double             *digitsVals, *lowerVals, *upperVals;
	long                inc1 = (isScalar(symhLower)) ? 0 : 1;
	long                inc2 = (isScalar(symhUpper)) ? 0 : 1;
	int                 ndigits = symbolSize(symhDigits);
	
	/* reset odometer*/

	digitsVals = DATAPTR(symhDigits);
	lowerVals = DATAPTR(symhLower);
	upperVals = DATAPTR(symhUpper);

	/*
	  first reset 'digits' to place 0
	*/
	if (isScalar(symhLower))
	{
		doubleFill(digitsVals, lowerVals[0], ndigits);
	}
	else
	{
		doubleCopy(lowerVals, digitsVals, ndigits);
	}

	DATAVALUE(symhPlace, 0) = place;
	if (place <= 1)
	{
		digitsVals[0] += place;
	}
	else 
	{
		long         jLower = 0;
		long         jUpper = 0;
		int          digit;
		int          ndigits = symbolSize(symhDigits);

		for (digit = 0; digit < ndigits; digit++)
		{
			double      range, remainder;
				
			range = (upperVals[jUpper] - lowerVals[jLower]) + 1;
			remainder = fmod(place, range);
				
			digitsVals[digit] = lowerVals[jLower] + remainder;
			place = floor(place / range);
			if (place == 0)
			{
				break;
			}
			jLower += inc1;
			jUpper += inc2;
		} /*for (digit = 0; digit < ndigits; digit++)*/			
	}
} /*doSetOdometer()*/

static void doStepOdometer(Symbolhandle symhOdometer, long nsteps)
{
	Symbolhandle        symhDigits = COMPVALUE(symhOdometer, odometerDigits);
	Symbolhandle        symhLower = COMPVALUE(symhOdometer, odometerLower);
	Symbolhandle        symhUpper = COMPVALUE(symhOdometer, odometerUpper);
	Symbolhandle        symhPlace = COMPVALUE(symhOdometer, odometerPlace);
	double             *digitsVals, *lowerVals, *upperVals;
	long                jstep, jLower, jUpper;
	long                inc1 = (isScalar(symhLower)) ? 0 : 1;
	long                inc2 = (isScalar(symhUpper)) ? 0 : 1;
	int                 digit;
	int                 ndigits = symbolSize(symhDigits);
	/* reset or step odometer*/
		
	digitsVals = DATAPTR(symhDigits);
	lowerVals = DATAPTR(symhLower);
	upperVals = DATAPTR(symhUpper);

	/* step the odometer by nsteps*/
	DATAVALUE(symhPlace, 0) += (double) nsteps;

	if (nsteps > 0)
	{
		for (jstep = 0; jstep < nsteps; jstep++)
		{
			jLower =  jUpper = 0;
			for (digit = 0; digit < ndigits; digit++)
			{
				digitsVals[digit]++;
				if (digitsVals[digit] <= upperVals[jUpper])
				{
					break;
				}
				digitsVals[digit] = lowerVals[jLower];
				jLower += inc1;
				jUpper += inc2;
			} /*for (digit = 0; digit < ndigits; digit++)*/
			if (digit == ndigits)
			{
				jLower = 0;
				for (digit = 0; digit < ndigits; digit++)
				{
					digitsVals[digit] = lowerVals[jLower];
					jLower += inc1;
				}
			}	
		} /*for (jstep = 0; jstep < nsteps; jstep++)*/
	} /*if (nsteps > 0)*/
	else
	{
		nsteps = -nsteps;
		for (jstep = 0; jstep < nsteps; jstep++)
		{
			jLower = jUpper = 0;
			for (digit = 0; digit < ndigits; digit++)
			{
				digitsVals[digit]--;
				if (digitsVals[digit] >= lowerVals[jLower])
				{
					break;
				}
					
				digitsVals[digit] = upperVals[jUpper];
				jLower += inc1;
				jUpper += inc2;
			} /*for (digit = 0; digit < ndigits; digit++)*/
			if (digit == ndigits)
			{
				jUpper = 0;
				for (digit = 0; digit < ndigits; digit++)
				{
					digitsVals[digit] = upperVals[jUpper];
					jUpper += inc2;
				}
			}	
		} /*for (jstep = 0; jstep < nsteps; jstep++)*/
	} /*if (nsteps > 0){}else{}*/
} /*doStepOdometer()*/

/*
  Code to implement function setodometer()

  An odometer is a structure of the form
    structure(digits:digits, lower:lower, upper:upper, place:count)
	  digits is an integer vector of length p
      lower and upper are integer scalars or integer vectors of length p
      count is an integer scalar
  usage:
    odom <- setodometer([lower:L,] upper:H [,ndigits:p])  
	odom <- setodometer(odom [,step:intScalar])
  	odom <- setodometer(odom [,set:place])
  000212 New function
  010516 Use atLeastNArgs constant on badNargs() call
*/

Symbolhandle  odometer(Symbolhandle list)
{
	Symbolhandle      result = (Symbolhandle) 0;
	Symbolhandle      symhOdometer = (Symbolhandle) 0;      
	Symbolhandle      symhDigits = (Symbolhandle) 0;
	Symbolhandle      symhLower = (Symbolhandle) 0;
	Symbolhandle      symhUpper = (Symbolhandle) 0;
	Symbolhandle      symhPlace = (Symbolhandle) 0;
	Symbolhandle      symh;
	double           *lowerVals, *upperVals;
	double            maxPlace;
	double            place = UNSETREAL;
	long              digit, jLower, jUpper;
	long              iarg, i, inc1, inc2, ndigits = 0, nsteps = UNSETLONG;
	long              nargs = NARGS(list);
	long              argnoLower, argnoUpper;
	int               setOdometer = 0, createOdometer = 0;
	char             *illegalKeyUse =
	  "ERROR: keyword '%s' on %s() illegal %s structure argument 1";
	char             *keyword;
	WHERE("odometer");
	
	maxPlace = MaxInteger - 1.0;
	
	if (nargs > 4)
	{
		badNargs(FUNCNAME, -4);
		goto errorExit;
	}
	if (nargs == 1 && COMPVALUE(list, 0) == (Symbolhandle) 0)
	{
		badNargs(FUNCNAME, -1 + atLeastNArgs);
		goto errorExit;
	}
	
	/* check arguments */
	for (iarg = 0; iarg < nargs; iarg++)
	{
		symh = COMPVALUE(list, iarg);
		if (!argOK(symh, 0, (nargs > 1) ? iarg + 1 : 0))
		{
			goto errorExit;
		}
		keyword = isKeyword(symh);
		if (keyword == (char *) 0)
		{
			/* not a keyword */
			if (iarg > 0)
			{
				sprintf(OUTSTR,
						"ERROR: argument %ld to %s() must be a keyword phrase",
						iarg + 1, FUNCNAME);
				goto errorExit;
			}
			if (!checkOdometer(symh))
			{
				goto errorExit;
			}
			symhOdometer = symh;
			ndigits = symbolSize(COMPVALUE(symhOdometer, odometerDigits));
		} /*if (keyword == (char *) 0)*/
		else if (strncmp(keyword, "step", 4) == 0)
		{
			if (!isInteger(symh, ANYVALUE))
			{
				notInteger(keyword);
				goto errorExit;
			}
			if (symhOdometer == (Symbolhandle) 0)
			{
				sprintf(OUTSTR, illegalKeyUse, keyword, FUNCNAME, "without");
				goto errorExit;
			}
			if (fabs(DATAVALUE(symh, 0)) > MAXSTEP)
			{
				sprintf(OUTSTR, "ERROR: value for '%s' %s%ld", keyword,
					(DATAVALUE(symh, 0) < 0) ? "< -" : "> ", (long) MAXSTEP);
				goto errorExit;
			}
			nsteps = (long) DATAVALUE(symh, 0);
		}
		else if (strcmp(keyword, OdometerParts[odometerPlace]) == 0)
		{
			if (!isInteger(symh, NONNEGATIVEVALUE))
			{
				notPositiveInteger(keyword);
				goto errorExit;
			}
			place = DATAVALUE(symh, 0);
			if (place > maxPlace)
			{
				sprintf(OUTSTR, "ERROR: value for '%s' %s%.17g", keyword,
					(DATAVALUE(symh, 0) < 0) ? "< -" : "> ", maxPlace);
				goto errorExit;
			}
			setOdometer = 1;
		}
		else if (strcmp(keyword, "ndigits") == 0)
		{
			long       thisLength;
			
			if (!isInteger(symh, POSITIVEVALUE))
			{
				notPositiveInteger(keyword);
				goto errorExit;
			}
			if (symhOdometer != (Symbolhandle) 0)
			{
				sprintf(OUTSTR, illegalKeyUse, keyword, FUNCNAME, "with");
				goto errorExit;
			}
			
			thisLength = DATAVALUE(symh, 0);
			if (ndigits == 0)
			{ /* not previously set */
				ndigits = thisLength;
			}
			else if (ndigits > 0)
			{
				/* ndigits previously set by 'ndigits' */
				if (thisLength != ndigits)
				{
					sprintf(OUTSTR,
							"ERROR: repeat inconsistent uses of keyword '%s' in %s()",
							keyword, FUNCNAME);
					goto errorExit;
				}				
			}
			else if (ndigits != -1 && thisLength != -ndigits)
			{
				/* ndigits length of 'lower' or 'upper' */
				sprintf(OUTSTR,
						"ERROR: value of '%s' does not match length of '%s' and/or '%s'",
						keyword, OdometerParts[odometerLower],
						OdometerParts[odometerUpper]);
				goto errorExit;
			}
			else
			{
				ndigits = thisLength; /* value > 0 means from 'ndigits'*/
			}
		}
		else if (strcmp(keyword, OdometerParts[odometerLower]) == 0 ||
				 strcmp(keyword, OdometerParts[odometerUpper]) == 0)
		{
			long        thisLength = symbolSize(symh);
			int         partNo = (strcmp(keyword,
										 OdometerParts[odometerLower]) == 0) ?
			  odometerLower : odometerUpper;

			if (!checkArgType(symh, keyword, INTVECTOR))
			{
				sprintf(OUTSTR,
						"ERROR: value for %s not integer scalar or vector in %s()",
						keyword, FUNCNAME);
				goto errorExit;
			}
			if (symhOdometer != (Symbolhandle) 0)
			{
				sprintf(OUTSTR, illegalKeyUse, keyword, FUNCNAME, "with");
				goto errorExit;
			}
						
			if (partNo == odometerLower && symhLower != (Symbolhandle) 0 ||
				partNo == odometerUpper && symhUpper != (Symbolhandle) 0)
			{
				sprintf(OUTSTR,
						"ERROR: repeat use of keyword '%s' illegal in %s()",
						keyword, FUNCNAME);
				goto errorExit;
			}
			
			if (ndigits == 0)
			{ /* not previously set */
				ndigits = -thisLength; /* ndigits < 0 means length of min or max*/
			}
			else if (ndigits > 0)
			{
				if (thisLength > 1 && thisLength != ndigits)
				{
					sprintf(OUTSTR,
							"ERROR: length of '%s' value != value of '%s'",
							keyword, "ndigits");
					goto errorExit;
				}
			}
			else if (ndigits == -1)
			{ /* other one of lower or upper was scalar */
				ndigits = -thisLength;
			}
			else if (thisLength != 1 && thisLength != -ndigits)
			{
				sprintf(OUTSTR,
						"ERROR: inconsistant lengths of '%s' and '%s' values",
						OdometerParts[odometerLower],
						OdometerParts[odometerUpper]);
				goto errorExit;
			}
			
			if (partNo == odometerLower)
			{
				argnoLower = iarg;
				symhLower = symh;
			}
			else
			{
				argnoUpper = iarg;
				symhUpper = symh;
			}
		}
		else
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
	} /*for (iarg = 0; iarg < nargs; i++)*/
		
	if (realIsSet(place) && longIsSet(nsteps))
	{
		sprintf(OUTSTR,
				"ERROR: illegal use of both '%s' and '%s' in %s()",
				"step", OdometerParts[odometerPlace], FUNCNAME);
		goto errorExit;
	}
	if (!realIsSet(place))
	{
		place = 0.0;
	}
	if (!longIsSet(nsteps))
	{
		nsteps = 1;
	}
	
	createOdometer = (symhOdometer == (Symbolhandle) 0);
	
	if (createOdometer || place != 0.0)
	{
		double        odometerSize = 1.0;

		if (!createOdometer)
		{
			symhLower = COMPVALUE(symhOdometer, odometerLower);
			symhUpper = COMPVALUE(symhOdometer, odometerUpper);
			ndigits = symbolSize(COMPVALUE(symhOdometer, odometerDigits));
		} /*if (!createOdometer)*/
		else
		{
			nsteps = 0;
			if (symhUpper == (Symbolhandle) 0)
			{
				sprintf(OUTSTR,
						"ERROR: you must specify '%s' when creating an odometer",
						OdometerParts[odometerUpper]);
				goto errorExit;
			}

			if (symhLower == (Symbolhandle) 0)
			{
				/* default for 'lower' is scalar 0 */
				symhLower = RInstall(SCRATCH, 1);
				if (symhLower == (Symbolhandle) 0)
				{
					goto errorExit;
				}
				DATAVALUE(symhLower, 0) = 0.0;
				argnoLower = -1;
			} /*if (symhLower == (Symbolhandle) 0)*/
		
			ndigits = labs(ndigits);
		} /*if (!createOdometer){}else{}*/
		
		inc1 = (isScalar(symhLower)) ? 0 : 1;
		inc2 = (isScalar(symhUpper)) ? 0 : 1;

		/* check values of lower and upper */
		lowerVals = DATAPTR(symhLower);
		upperVals = DATAPTR(symhUpper);
		jLower = jUpper = 0;
		for (digit = 0; digit < ndigits; digit++)
		{
			double     range = upperVals[jUpper] - lowerVals[jLower] + 1;
			
			odometerSize *= range;

			if (createOdometer)
			{
				if (range <= 0.0)
				{
					sprintf(OUTSTR, "ERROR: %s >= %s in %s() at digit %ld",
							OdometerParts[odometerLower],
							OdometerParts[odometerUpper],
							FUNCNAME, digit + 1);
				}
				else if (range > (double) MaxLong)
				{
					sprintf(OUTSTR,
							"ERROR: max(%s - %s) > %ld at digit %ld",
							OdometerParts[odometerUpper],
							OdometerParts[odometerLower],
							(long) MaxLong, digit + 1);
				}
				else if (odometerSize > maxPlace + 1.0)
				{
					sprintf(OUTSTR,
							"ERROR: prod(%s - %s + 1) > %.17g",
							OdometerParts[odometerUpper],
							OdometerParts[odometerLower], maxPlace + 1.0);
				}
				if (*OUTSTR)
				{
					goto errorExit;
				}
			} /*if (createOdometer)*/
			
			jLower += inc1;
			jUpper += inc2;
		} /*for (digit = 0; digit < ndigits; digit++)*/

		if (!createOdometer)
		{
			if (place >= odometerSize)
			{
				sprintf(OUTSTR,
						"ERROR: value for '%s' outside range of odometer in %s()",
						OdometerParts[odometerPlace], FUNCNAME);
				goto errorExit;
			}
		}
		else
		{
			result = StrucInstall(SCRATCH, odometerNparts);
			if (result == (Symbolhandle) 0)
			{
				goto errorExit;
			}

			/* component digits */
			COMPVALUE(result, odometerDigits) = symh = Makereal(ndigits);
			if (symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}

			if (isScalar(symhLower))
			{
				doubleFill(DATAPTR(symh), DATAVALUE(symhLower, 0), ndigits);
			}
			else
			{
				doubleCopy(DATAPTR(symhLower), DATAPTR(symh), ndigits);
			}

			/* component lower */
			if (isscratch(NAME(symhLower)))
			{
				/* reuse argument or previously constructed scalar*/
				COMPVALUE(result, odometerLower) = symh =
				  (argnoLower >= 0) ?
				  reuseArg(list, argnoLower, 0, 0) : symhLower;
				Cutsymbol(symh); /* remove from symbol table */
			}
			else
			{
				long        sizeLower = symbolSize(symhLower);

				COMPVALUE(result, odometerLower) = symh = Makereal(sizeLower);
				if (symh == (Symbolhandle) 0)
				{
					goto errorExit;
				}
				doubleCopy(DATAPTR(symhLower), DATAPTR(symh), sizeLower);
			}

			/* component upper */
			if (isscratch(NAME(symhUpper)))
			{
				/* reuse argument */
				COMPVALUE(result, odometerUpper) = symh =
				  reuseArg(list, argnoUpper, 0, 0);
				Cutsymbol(symh); /* remove from symbol table */
			}
			else
			{
				long        sizeUpper = symbolSize(symhUpper);

				COMPVALUE(result, odometerUpper) = symh = Makereal(sizeUpper);
				if (symh == (Symbolhandle) 0)
				{
					goto errorExit;
				}
				doubleCopy(DATAPTR(symhUpper), DATAPTR(symh), sizeUpper);
			}

			/* component place */
			COMPVALUE(result, odometerPlace) = symh = Makereal(1);
			if (symh == (Symbolhandle) 0)
			{
				goto errorExit;
			}
			DATAVALUE(symh, 0) = 0; /* place runs from 0 to sum of steps taken*/

			lowerVals = DATAPTR(COMPVALUE(result, odometerLower));
			upperVals = DATAPTR(COMPVALUE(result, odometerUpper));
			jLower = jUpper = 0;

			for (i = 0; i < odometerNparts; i++)
			{
				setNAME(COMPVALUE(result, i), OdometerParts[i]);
			}
		}
	} /*if (createOdometer || place != 0.0)*/
	
	if (!createOdometer)
	{
		if (isscratch(NAME(symhOdometer)))
		{
			result = reuseArg(list, 0, 0, 0);
		}
		else
		{
			result = Install(SCRATCH, STRUC);
			if (result == (Symbolhandle) 0 ||
				!Copy(symhOdometer, result))
			{
				goto errorExit;
			}
			setNAME(result, SCRATCH);
		}
	} /*if (!createOdometer)*/
	
	if (setOdometer)
	{
		doSetOdometer(result, place);
	}
	else if (nsteps != 0)
	{
		doStepOdometer(result, nsteps);
	}
	
	return (result);

  errorExit:
	putErrorOUTSTR();
	Removesymbol(result);
	return (0);
} /*odometer()*/

	
