/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 5.00 or later
*(C)*
*(C)* Copyright (c) 2004 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Power
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/


#include "globals.h"

enum designCodes
{
	CRD = 1,
	RBD
};

enum argListCodes
{
	noncenArg = 0,
	df1Arg,
	ngrpArg = df1Arg,
	df2Arg,
	nrepArg = df2Arg,
	alpha1Arg = df2Arg,
	powerArg,
	alpha2Arg = powerArg,
	nonKeyArgs
};


/*
  samplesize(noncen,ngrp,alpha,pwr [,design:"crb"] [,maxn:n] [,silent:T])
*/

#ifndef MAXN
#define         MAXN 256.0
#endif /*MAXN*/

#ifndef BETANCERRMAX
#define BETANCERRMAX   1e-10
#endif /*BETANCERRMAX*/

#ifndef BETANCITRMAX
#if (0)
#define BETANCITRMAX   200
#else
#define BETANCITRMAX  10000
#endif
#endif /*BETANCITRMAX*/

/*
  970730 Added keyword maxn, changed default from 127 to 256 and
         made other cosmetic modifications
  980303 Fixed bug in error message
  980616 Fixed more bugs in error message
  990226 Replaced putOUTSTR() by putErrorOUTSTR()
  020805 Added argument upper = 1 to calls to Qsne() and noncentF()
         and modified code accordingly
  021007 Argument list should now be samplesize(noncen,g, n, alpha)
  041006 Added errmax, itrmax to noncentF() call
*/
Symbolhandle    samplesize(Symbolhandle list)
{
	Symbolhandle    result, args[nonKeyArgs];
	Symbolhandle    symhDesign = SymbolZero;
	double          tmp, left, right, noncen, alpha, powr,trialPower;
	double          numDf, denomDf;
	double          ngrp;
	double          critval;
	double          maxn = MAXN;
	double          errmax = BETANCERRMAX;
	long            itrmax = BETANCITRMAX;
	long            i;
	long            minNgrp;
	long            nargs = NARGS(list);
	int             design = CRD;
	int             silent = 0;
	char           *keyword;
	char           *usage = "usage: samplesize(noncen,ngrp,alpha,pwr [,design:\"rbd\"])";
	char           *badargsFmt = "%s (arg. %d to %s())";
	char           *argWhat[nonKeyArgs];
	char            errormsg[80];
	WHERE("samplesize");

	argWhat[noncenArg] = "non-centrality";
	argWhat[ngrpArg] = "number of groups";
	argWhat[alpha1Arg] = "significance level";
	argWhat[powerArg] = "desired power";

	if (nargs < nonKeyArgs)
	{
		badNargs(FUNCNAME, -nonKeyArgs + atLeastNArgs);
		goto errorExit;
	}
	else if (nargs > 6)
	{
		badNargs(FUNCNAME,-6);
		goto errorExit;
	}
	for (i = 0;i < nonKeyArgs;i++)
	{
		args[i] = COMPVALUE(list,i);
	}

	/* non-centrality*/
	if (!isNumber(args[noncenArg], NONNEGATIVEVALUE))
	{
		sprintf(errormsg, badargsFmt, argWhat[noncenArg], noncenArg + 1,
				FUNCNAME);
		notNonNegativeReal(errormsg);
		goto errorExit;
	}

	/* number of groups */
	if (!isInteger(args[ngrpArg], POSITIVEVALUE))
	{
		sprintf(errormsg, badargsFmt, argWhat[ngrpArg], ngrpArg + 1, FUNCNAME);
		notPositiveInteger(errormsg);
		goto errorExit;
	}
	/* significance level */
	if (!isNumber(args[alpha1Arg], POSITIVEVALUE))
	{
		sprintf(errormsg, badargsFmt, argWhat[alpha1Arg], alpha1Arg + 1,
				FUNCNAME);
		notPositiveReal(errormsg);
		goto errorExit;
	}
	if (!isNumber(args[powerArg], POSITIVEVALUE))
	{
		sprintf(errormsg, badargsFmt, argWhat[powerArg], powerArg + 1,
				FUNCNAME);
		notPositiveReal(errormsg);
		goto errorExit;
	}
	
	for (i = nonKeyArgs; i < nargs; i++)
	{
		Symbolhandle         symhKey = COMPVALUE(list, i);

		keyword = isKeyword(symhKey);
		if (keyword)
		{
			if (strcmp(keyword, "design") == 0)
			{
				symhDesign = symhKey;
			}
			else if (strcmp(keyword, "maxn") == 0)
			{
				if (!isInteger(symhKey, POSITIVEVALUE))
				{
					notPositiveInteger(isKeyword(symhKey));
					goto errorExit;
				}
				maxn = DATAVALUE(symhKey, 0);
			}
			else if (strcmp(keyword, "silent") == 0)
			{
				if (!isTorF(symhKey))
				{
					notTorF(isKeyword(symhKey));
					goto errorExit;
				}
				silent = (DATAVALUE(symhKey, 0) != 0.0);
			}
			else
			{
				badKeyword(FUNCNAME, keyword);
				goto errorExit;
			}
		} /*if (keyword)*/
		else
		{
			sprintf(OUTSTR,
					"ERROR: argument %ld to %s() must be keyword phrase",
					i+1, FUNCNAME);
			goto errorExit;
		}
	} /*for (i = nonKeyArgs; i < nargs; i++)*/
	
	if (symhDesign != SymbolZero)
	{
		char         *designName;
		
		if (!isCharOrString(symhDesign))
		{
			notCharOrString(isKeyword(symhDesign));
			goto errorExit;
		}
		designName = STRINGPTR(symhDesign);
		if (strcmp(designName,"crd") == 0 || strcmp(designName,"cr") == 0)
		{
			design = CRD;
		}
		else if (strcmp(designName,"rbd") == 0 || strcmp(designName,"rb") == 0 ||
				 strcmp(designName,"rcb") == 0)
		{
			design = RBD;
		}
		else
		{
			sprintf(OUTSTR,"ERROR: design name %s not recognized by %s()",
					designName,FUNCNAME);
			goto errorExit;
		}
	} /*if (symhDesign != SymbolZero)*/
	
	minNgrp = (design == CRD) ? 1 : 2;
	if (DATAVALUE(args[ngrpArg], 0) < minNgrp)
	{
		sprintf(OUTSTR, "ERROR: %s for %s() must be integer >= %ld",
				argWhat[ngrpArg], FUNCNAME, minNgrp);
		goto errorExit;
	}
	
	for (i = 2; i < 3; i++)
	{
		if (DATAVALUE(args[i], 0) >= 1.0)
		{
			sprintf(OUTSTR,
					"ERROR: %s for %s() must be between 0 and 1",
					argWhat[i], FUNCNAME);
			goto errorExit;
		}
	}

	noncen = DATAVALUE(args[noncenArg], 0);
	ngrp = DATAVALUE(args[ngrpArg], 0);
	alpha = DATAVALUE(args[alpha1Arg], 0);	
	powr = DATAVALUE(args[powerArg], 0);

	right = 1.0;

	/* for CRD, treat problem as one sample t-test of H0: mu = 0 */
	numDf = (ngrp > 1.0) ? ngrp - 1.0 : 1.0;

	do /*while (trialPower < powr)*/
	{
		if (interrupted(DEFAULTTICKS) != INTNOTSET)
		{
			goto interruptExit;
		}
		left = right;
		if (right >= maxn)
		{
			break;
		}
		right *= 2.0;
		if (right > maxn)
		{
			right = maxn;
		}
		
		denomDf = (right - 1.0) * ((design == CRD) ? ngrp : (ngrp - 1.0));
		critval = Qsne(alpha, numDf, denomDf, 1);
		trialPower = noncentF(critval, right * noncen, numDf, denomDf, 
							  errmax, itrmax, 1);
	} while (trialPower < powr);


	if (trialPower > powr)
	{
		left = floor(left) + .1;

		while (right - left > 1.1)
		{
			if (interrupted(DEFAULTTICKS) != INTNOTSET)
			{
				goto interruptExit;
			}
			tmp = floor((right + left) / 2.0 + .1);
			denomDf = (tmp - 1.0) * ((design == CRD) ? ngrp : (ngrp - 1.0));
			critval = Qsne(alpha, numDf, denomDf, 1);
			trialPower = noncentF(critval,tmp * noncen, numDf, denomDf, 
								  errmax, itrmax, 1);
			if (trialPower > 1.0)
			{ /* not fully converged, ignore */
				trialPower -= floor(trialPower);
			}
			if (trialPower < powr)
			{
				left = tmp;
			}
			else
			{
				right = tmp;
			}
		} /*while (right - left > 1.1)*/
	} /*if (trialPower > powr)*/
	else if (!silent)
	{
		sprintf(OUTSTR,"WARNING: %s() truncated at %4.0f", FUNCNAME, right);
		putErrorOUTSTR();
	}

	result = RInstall(SCRATCH, 1L);

	if (result != SymbolError)
	{
		DATAVALUE(result,0) = right;
	}
	
	return (result);

  errorExit:
	putErrorOUTSTR();

	putOutMsg(usage);

  interruptExit:
	
	return (0);

} /*samplesize()*/

enum powerOpCode
{
	IPOWER = 1,
	IPOWER2
};

/* 
  power(noncen,ngrp,groupsize,alpha [,design="rcb"])
  power2(noncen2,ndf,ddf,alpha)
  also, but deprecated,
   power(noncen,ngrp,alpha,groupsize [,design="rcb"])
   power2(noncen2,ndf,alpha,ddf)

  010516 Use atLeastNArgs constant on badNargs() call
  021008 Changed order of argument list argument list
  041006 Added errmax, itrmax to call to noncentF()
*/
/* name changed for use in MPW where pow is equivalenced to power*/
Symbolhandle    power1(Symbolhandle list)
{
	Symbolhandle    args[nonKeyArgs], result = SymbolZero;
	double          noncen, ngrp, alpha, groupSize, powr;
	double          numDf, denomDf, critval;
	double          errmax = BETANCERRMAX;
	long            itrmax = BETANCITRMAX;
	long            nargs = NARGS(list);
	long            nvals = 1;
	long            i, j, length;
	char           *usage;
	char           *argWhat[nonKeyArgs];
	int             inc[nonKeyArgs], ii[nonKeyArgs];
	int             op, design = CRD;
	int             silent = 0;
	WHERE("power1");
	
	*OUTSTR = '\0';
	if (strcmp(FUNCNAME,"power") == 0)
	{
		op = IPOWER;
		usage = "power(noncen,ngroups,groupsize,alpha [,design:\"rbd\"])";
		if (nargs < nonKeyArgs || nargs > nonKeyArgs + 2)
		{
			badNargs(FUNCNAME, (nargs < nonKeyArgs) ?
					 -nonKeyArgs + atLeastNArgs : -nonKeyArgs + 2);
			goto errorExit;
		}
	}
	else
	{
		op = IPOWER2;
		usage = "power2(noncen2,numDf,denomDf,alpha [,silent:T])";
	
		if (nargs < nonKeyArgs || nargs > nonKeyArgs + 1)
		{
			badNargs(FUNCNAME, (nargs < nonKeyArgs) ?
					 -nonKeyArgs + atLeastNArgs : -nonKeyArgs + 1);
		}
	}

	argWhat[noncenArg] = "non-centrality";
	argWhat[df1Arg] = (op == IPOWER) ? "number of groups" : "numerator d.f.";
	argWhat[df2Arg] = (op == IPOWER) ? "group size" : "denominator d.f.";
	argWhat[alpha2Arg] = "significance level";
	
	for (j = 0;j < nonKeyArgs;j++)
	{
		args[j] = COMPVALUE(list,j);
		if (!argOK(args[j], REAL, j+1))
		{
			goto errorExit;
		}

		if (!isVector(args[j]))
		{
			sprintf(OUTSTR,
					"ERROR: %s for %s() must be REAL scalar or vector",
					argWhat[j], FUNCNAME);
			goto errorExit;
		}

		if (anyMissing(args[j]))
		{
			sprintf(OUTSTR,
					"ERROR: %s for %s() contains MISSING values",
					argWhat[j], FUNCNAME);
			goto errorExit;
		} /*if (anyMissing(args[j]))*/
		
		ii[j] = 0;
		inc[j] = 0;
		length = DIMVAL(args[j], 1);
		if (j == 0)
		{
			nvals = length;
		}
		else if (length > 1)
		{
			if (nvals > 1 && length != nvals)
			{
				sprintf(OUTSTR,
						"ERROR: %s for %s() is vector with non-matching length",
						argWhat[j], FUNCNAME);
				goto errorExit;
			}
			nvals = length;
		}
		inc[j] = (length == 1) ? 0 : 1;
	} /*for (j = 0;j < nonKeyArgs;j++)*/

	for (i = nonKeyArgs; i < nargs; i++)
	{
		Symbolhandle       symhKey = COMPVALUE(list, i);
		char              *keyword = isKeyword(symhKey);
		
		if (keyword == (char *) 0)
		{
			sprintf(OUTSTR,
					"ERROR: argument %d to %s() must be keyword phrase",
					nonKeyArgs, FUNCNAME);
			goto errorExit;
		}
		
		if (strcmp(keyword, "design") == 0)
		{
			char         *designName;

			if (!isCharOrString(symhKey))
			{
				notCharOrString(keyword);
				goto errorExit;
			}
			if (op == IPOWER2)
			{
				badKeyword(FUNCNAME, keyword);
				goto errorExit;
			}
			
			designName = STRINGPTR(symhKey);
			if (strcmp(designName,"crd") == 0 || strcmp(designName,"cr") == 0)
			{
				design = CRD;
			}
			else if (strcmp(designName,"rbd") == 0 || strcmp(designName,"rb") == 0 ||
					 strcmp(designName,"rcb") == 0)
			{
				design = RBD;
			}
			else
			{
				sprintf(OUTSTR,"ERROR: illegal value \"%s\" for keyword %s on %s()",
						designName, keyword, FUNCNAME);
				goto errorExit;
			}
		}
		else if (strcmp(keyword, "silent") == 0)
		{
			if (!isTorF(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			silent = (DATAVALUE(symhKey, 0) != 0.0);
		}
		else
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
	} /*for (i = nonKeyArgs; i < nargs; i++)*/

	if (doubleMax(DATAPTR(args[df2Arg]), symbolSize(args[df2Arg])) < 1.0 &&
		doubleMax(DATAPTR(args[alpha2Arg]), symbolSize(args[alpha2Arg])) > 1.0)
	{
		Symbolhandle      symhTemp = args[df2Arg];
		int               incTemp = inc[df2Arg];
		
		args[df2Arg] = args[alpha2Arg];
		args[alpha2Arg] = symhTemp;
		inc[df2Arg] = inc[alpha2Arg];
		inc[alpha2Arg] = incTemp;
	}
	
	length = (inc[noncenArg] > 0) ? nvals : 1;
	if (doubleMin(DATAPTR(args[noncenArg]), length) < 0.0)
	{
		sprintf(OUTSTR,
				"ERROR: %s for %s() not all non-negative",
				argWhat[0], FUNCNAME);
		goto errorExit;
	}
	
	length = (inc[df1Arg] > 0) ? nvals : 1;
	for (i = 0;i < length;i++)
	{
		numDf = DATAVALUE(args[df1Arg],i);
		if (numDf < 1.0 || numDf != floor(numDf))
		{
			sprintf(OUTSTR,
					"ERROR: %s for %s() not all positive integers",
					argWhat[df1Arg], FUNCNAME);
			goto errorExit;
		}
	} /*for (i = 0;i < length;i++)*/

	length = (inc[df2Arg] > 0) ? nvals : 1;
	for (i = 0;i < length;i++)
	{
		denomDf = DATAVALUE(args[df2Arg],i);
		if (denomDf < 1.0 || denomDf != floor(denomDf) ||
			(op == IPOWER && denomDf < 2.0))
		{
			sprintf(OUTSTR,
					"ERROR: %s for %s() not all integers > %ld",
					argWhat[df2Arg], FUNCNAME, (op == IPOWER) ? 1L : 0L);
			goto errorExit;
		}
	} /*for (i = 0;i < length;i++)*/

	length = (inc[alpha2Arg] > 0) ? nvals : 1;
	if (doubleMin(DATAPTR(args[alpha2Arg]), length) <= 0.0 ||
		doubleMax(DATAPTR(args[alpha2Arg]), length) >= 1.0)
	{
		sprintf(OUTSTR,
				"ERROR: %s for %s() not all between 0 and 1",
				argWhat[alpha2Arg], FUNCNAME);
		goto errorExit;
	}

	result = RInstall(SCRATCH, nvals);

	if (result == SymbolError)
	{
		goto errorExit;
	}
	
	for (i = 0;i < nvals;i++)
	{
		noncen = DATAVALUE(args[noncenArg],ii[noncenArg]);
		if (noncen == 0.0)
		{
			powr = alpha;
		}
		else
		{
			numDf = DATAVALUE(args[df1Arg], ii[df1Arg]);
			alpha = DATAVALUE(args[alpha2Arg], ii[alpha2Arg]);
			denomDf = DATAVALUE(args[df2Arg], ii[df2Arg]);
			if (op == IPOWER)
			{
				ngrp = numDf;
				groupSize = denomDf;
				numDf = (ngrp > 1.0) ? ngrp - 1.0 : 1.0;
				denomDf = (groupSize - 1.0) * 
					((design == CRD || ngrp == 1) ? ngrp : (ngrp - 1.0));
				noncen *= groupSize;
			} /*if (op == IPOWER)*/

			critval = Qsne(alpha, numDf, denomDf, 1);
			powr = noncentF(critval, noncen, numDf, denomDf, 
							errmax, itrmax, 1);

			if (powr > 1.0)
			{
				if (!silent)
				{
					sprintf(OUTSTR,
							"WARNING: non-central F not fully converged in %s() after %g terms",
							FUNCNAME, floor(powr));
					putErrorOUTSTR();
				}
				powr -= floor(powr);
			}
		}
		DATAVALUE(result,i) = powr;
		for (j = 0; j < nonKeyArgs; j++)
		{
			ii[j] += inc[j];
		}
	} /*for (i = 0;i < nvals;i++)*/
	
	return (result);

  errorExit:
	putErrorOUTSTR();
	
	putOutMsg(usage);
	return (0);
} /*power1()*/

#if (0)
/* 
  This has been obsoleted by Qsne in pvalsub.c which calls ppbeta in
  ppbeta() in betabase.c
*/

double          Fquant(double p, double n1, double n2)
{
	/* find F quantile by interval bisection */

	/*  first get starting values via Cornish-Fisher approx
	    eqn 17 p. 82 of Johnson and Kotz */

	double          left, right, center, tmp, delta, sigma;
	long            fault;

	tmp = ppnd(p, &fault);
	sigma = sqrt(0.5 * (1.0 / n1 + 1.0 / n2));
	delta = 0.5 * (1. / n2 - 1. / n1);
	center = delta * (1.0 + (tmp * tmp - 1.) / 3.) +
		tmp * sigma / sqrt(1.0 - sigma * sigma);

	if (Csne(center, n1, n2) > p)
	{
		right = center;
		left = right * 0.9;
		while (Csne(left, n1, n2) > p)
		{
			left *= 0.9;
		}
	}
	else
	{
		left = center;
		right = left * 1.1;
		while (Csne(right, n1, n2) < p)
		{
			right *= 1.1;
		}
	}

	while ((right - left) * 2 / (right + left) > 0.001)
	{
		center = (left + right) / 2.;
		if (Csne(center, n1, n2) < p)
		{
			left = center;
		}
		else
		{
			right = center;
		}
	}

	return (center);
} /*Fquant()*/

/* This has been replaced by a new version in pvalsub.c calling betanc */
double          noncentF(double x, double lam, double n1, double n2)
{
	/* approximate cumulative for noncentral F using Tiku's approx
	   or Severo and Zelen approx if Tiku fails
	   p 195 of Johnson and kotz */

	double          H, K, n1p, c, b, n1pluslam, tmp;
	extern double   Csne(), Cnor();
	WHERE("noncentF");
	
	n1pluslam = n1 + lam;
	if (n2 > 2.0)
	{
		H = 2.*n1pluslam*n1pluslam*n1pluslam +
			3.*n1pluslam*(n1 + 2.*lam)*(n2 - 2.) +
			(n1 + 3.*lam)*(n2 - 2.)*(n2 - 2.);
		K = n1pluslam*n1pluslam + (n2 - 2.)*(n1 + 2*lam);

		n1p = 0.5*(n2 - 2.)*(sqrt(H*H/(H*H - 4.*K*K*K)) - 1.);
		c = (n1p/n1)*H/K/(2*n1p + n2 - 2.);
		b = -n2*(c - 1. - lam/n1)/(n2 - 2.);

		return (Csne((x - b)/c, n1p, n2));
	}

	tmp = pow(n1*x/(n1 + lam), 1./3.);
	b = sqrt((2./9.)*((n1 + 2.*lam)/(n1pluslam*n1pluslam) + tmp*tmp/n2));

	c = (1. - (2./9.)/n2)*tmp - 1. + 
		(2./9.)*(n1 + 2.*lam)/(n1pluslam*n1pluslam);
	return (Cnor(c/b));
} /*noncentF()*/
#endif /*0*/
