/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 5.03 or later
*(C)*
*(C)* Copyright (c) 2007 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Print
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "globals.h"

#include "mvIface.h"

#include "mvIo.h"

#ifdef CONSOLENAME
#define ISCONSOLE(NAME) (strcmp(NAME,CONSOLENAME) == 0)
#else /*CONSOLENAME*/
#define ISCONSOLE(NAME)  0
#endif /*CONSOLENAME*/

#define KEYNOTSET -1
/*
  990212 replaced putOUTSTR() by putErrorOUTSTR()
  071012 added keyword phrase multiline:T to putascii()
*/
enum printOpCodes
{
	IPRINT      = 0x01,
	IWRITE      = 0x02,
	IMACRO      = 0x04,
	IMAT        = 0x08,
	IFILE       = 0x10,
	IERR        = 0x20,
	IFPRINT     = IPRINT | IFILE,
	IFWRITE     = IWRITE | IFILE,
	IMATPRINT   = IPRINT | IMAT | IFILE,
	IMATWRITE   = IWRITE | IMAT | IFILE,
	IMACROWRITE = IMACRO | IFILE,
	IERROR      = IERR | IPRINT
};


	/* formerly a routine to print a single macro */
/*
  Operations print(), write(), fprint(), fwrite(), matprint(), matwrite(),
  macrowrite(), error()

  The first argument for fprint(), fwrite(), matprint(), matwrite(),
  and macrowrite(), must be a file name.

  If the last argument is the CHAR string "new" then the file is overwritten;
  otherwise, stuff is written at the end.

  For print(), write(), fprint(), and fwrite() the remaining arguments may
  be any printable object (types REAL, LOGIC, CHAR, MACRO, STRUC, PLOTINFO).

  For matprint() and matwrite(), arguments must be REAL OR LOGIC.  LOGIC
  values T and F are translated to 1.0 and 0.0 and a comment of the form ')
  LOGICAL' is prepended.  A warning message is printed for any argument
  that is not a structure, or a REAL, LOGIC or CHARACTER variable.

  For macrowrite(), a warning message is printed for any argument that is not
  MACRO

  write(), fwrite(), and matwrite() differ from print(), fprint(), and
  matprint(), only in printing more significant digits.

  If keyword phrase file:fileName appears on print() or write(), it
  operates like fprint() or fwrite() with that fileName.  In that case,
  new:T is acceptable.

  All operations accept keywords nsig (e.g., nsig:8) and format (e.g.,
  format:"10'4f"), although these are meaningless for macrowrite.  They
  control the printing of any REAL or LOGIC objects later in the argument
  list.  These keywords may be repeated to change the significance.

  nsig:m is is equivalent to a format of %(m+7).mg  .  The last nsig or
  format after all non-keywords or unrecognized keywords is taken as a
  starting default.  Thus print(a, b, c, nsig:9) is equivalent to
  print(nsig:9, a, b c) or write(a,b,c)

  Key word new (e.g., new:T or new:F) can appear immediately after the file
  name.

  error() is identical to print() except it will cause immediate termination
  of the statement in which it occurs.  It is designed for printing fatal
  error messages in macros.  As of 990207 it also adds " in macro XXXX"
  to the macro when called from a macro, unless keyword phrase macroname:F
  is an argument
*/

typedef struct opEntry
{
	char *name;
	short iop1;
	short iop2;
} opEntry;

static opEntry Ops[] =
{ /* name          iop1       iop2 */
	{"write",      IWRITE,      IPRINT},
	{"fwrite",     IFWRITE,     IPRINT},
	{"matwrite",   IMATWRITE,   IMAT},
	{"print",      IPRINT,      IPRINT},
	{"fprint",     IFPRINT,     IPRINT},
	{"matprint",   IMATPRINT,   IMAT},
	{"macrowrite", IMACROWRITE, IMACRO},
	{"error",      IERROR,      IERR},
	{(char *) 0,   0}
};

typedef struct keylist
{
	char        *name;
	short        matchlength; /* 0 means exact match */
	short        type;
	short        legalOps;
} keylist;

static keylist LegalKeys[] =
{
/*  name         matchlength  type     legalOps*/
	{"new",           0,      LOGIC,   IPRINT | IMAT | IMACRO},
	{"file",          0,      CHAR,    IPRINT | IMAT | IMACRO},
	{"sep",           0,      CHAR,    IMAT},
	{"header",        4,      LOGIC,   IPRINT | IMAT | IMACRO},
	{"missing",       4,      -1,      IPRINT | IMAT | IERR},
	{"nsig",          0,      REAL,    IPRINT | IMAT | IERR},
	{"format",        0,      CHAR,    IPRINT | IMAT | IERR},
	{"zero",          4,      CHAR,    IPRINT | IERR},
	{"width",         0,      REAL,    IPRINT | IMAT | IERR},
	{"height",        0,      REAL,    IPRINT | IERR},
	{"labels",        5,      LOGIC,   IPRINT | IMAT | IERR},
	{"notes",         4,      LOGIC,   IPRINT | IMAT | IMACRO},
	{"name",          0,      CHAR,    IPRINT | IMAT | IMACRO},
	{"quoted",        5,      LOGIC,   IMAT},
	{"bylines",       6,      LOGIC,   IMAT},
	{"comments",      7,      CHAR,    IMAT | IMACRO},
	{"oldstyle",      0,      LOGIC,   IMAT | IMACRO},
	{"stripdols",     8,      LOGIC,   IMAT | IMACRO},
	{"macroname",     0,      LOGIC,   IERR | IPRINT | IWRITE},
	{(char *) 0,      0,      0,       0},
};

/* This order must match the order in LegalKeys */
enum printKeyCodes
{
	KNEW = 0,
	KFILE,
	KSEP,
	KHEADER,
	KMISSING,
	KNSIG,
	KFORMAT,
	KZERO,
	KWIDTH,
	KHEIGHT,
	KLABELS,
	KNOTES,
	KNAME,
	KQUOTED,
	KBYLINES,
	KCOMMENTS,
	KOLDSTYLE,
	KSTRIPDOLS,
	KMACRONAME,
	KNOKEY
};

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Print
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

/*
  030921 new function created by moving code from print()
*/

static void printSingleString(char * thisString, FILE * outFile, int macroname,
							  int isError)
{
	if (PRINTWARNINGS || strncmp(thisString, WarningStart, 8) != 0)
	{
		if (isError && strncmp(thisString, ErrorStart,6) != 0 &&
			strncmp(thisString, WarningStart, 8) != 0)
		{
			myprint(ErrorStart);
		}
		if (thisString[0] == '\0')
		{
			fmyeol(outFile);
		}
		else
		{
			/* last arg is addNL */
			mymultiprint(thisString, outFile, !macroname);
		}
		if (macroname || isError)
		{
			if (thisString[0])
			{
				int    iMacro;
				if (macroname)
				{
					myprint(inWhichMacro(0));
					fmyeol(outFile);
				}

				for (iMacro = 1; TRACEBACK && iMacro < ACTIVEMACROS;
					 iMacro++)
				{
					myprint(inWhichMacro(iMacro));
					fmyeol(outFile);
				}
			} /*if (thisString[0])*/
		} /*if (macroname || isError)*/
	} /*if (PRINTWARNINGS || strncmp(thisString, WarningStart, 8) != 0)*/
} /*printSingleString()*/


/*
   General entry for print(), write(), fprint(), fwrite(), error(),
   matprint(), matwrite(), macrowrite()

   960502 changed fopen() to fmyopen()
   960503 changed macOpen() to macFindFile()
   960827 Illegal values for keywords now cause an error before any
          output, instead of just a warning.  Also keywords header,
		  labels, and sep can appear several times, applying to
		  later objects being printed.  Also name at end is same
		  as name at beginning, like other keywords
   960906 Added keyword comments on matwrite(), matprint() and macrowrite()
   970109 Added keyword width on print(), write(), matprint(), matwrite(),
          and error(), and keyword height on print(), write() and error()
   970302 Allow STRUC, NULLSYM, and PLOTINFO types as arguments to
          mat{print,write}() and macrowrite().  PLOTINFO variables are
		  written as NULL variables for now
   970506 Use macros TEXTWRITEMODE and TEXTAPPENDMODE for fmyopen() mode
   971101 Added keyword oldstyle to macrowrite()
   980430 Cosmetic changes; also, for example, error("N < 0") now prints 
          "ERROR: N < 0"
   980723 New keyword notes for matprint(), matwrite() and macwrite() and
          made keyword labels legal for matprint() and matwrite().
          Added notes and labels to argument list for matWrite()
          On print("CHARACTER SCALAR", header:T and/or labels:T), the
          string is enclosed in quotes and a header printed unless
          header:F
   990207 error() now adds information about which macro was active and
          recognizes new keyword phrase 'macroname:F' which suppresses the
          new information.
   990330 Single test for MISSING as keyword value
   990924 Added keyword stripdols for matwrite(), matprint() and macrowrite()
   990926 Rewrote much of the code parsing keywords, replacing
          if (strcmp() == 0){}else if () ... by switch statements
          Also, abbreviations are allowed on some keywords
   991116 'macroname' now legal on print() and write(), default F
   000830 Fixed minor bug which appeared on print("WARNING:...") after
          setoptions(warning:F)
   001214 Fixed bug that caused crash when printing single LOGICAL variable
          whose first element was T
   010522 Changed call to macSetInfo()
   010523 Replaced macSetInfo() by macSetFInfo()
   010524 Changed calling sequence of macFindFile()
   010525 Changed macFindFile() to mvFindFile() and dropped wxFindFile()
   010601 Include mvIface.h instead of macIface.h or wxIface.h
   010614 Add keyword phrase zero:charZero to print(). charZero, padded
          on the left with spaces, will be printed for REAL 0's
          Replaced instances of (Symbolhandle) 0 by SymbolZero, SymbolError
          or EmptyArgument
   010810 Added argument to inWhichMacro() and code to give traceback
   010811 Traceback given only when option 'traceback' is True
   030921 Moved code out of print() into new function printSingleString()
   030923 notes:T now affects print(); if a variable has notes, they are
          printed
   060922 Introduced KEYNOTSET macro to clarify code
*/
Symbolhandle    print(Symbolhandle list)
{
	Symbolhandle    arg1 = COMPVALUE(list,0);
	Symbolhandle    argi, filearg = SymbolZero;
	Symbolhandle    symhComment = SymbolZero;
	Symbolhandle    result = SymbolZero;
	FILE           *outFile = STDOUT;
	char           *fileMode = (char *) 0;
	long            nargs = NARGS(list), i, type, targetType, ifile;
	long            unusedArgs = nargs;
	long            op, keyOp;
	long            toFile = 0, useConsole = 0;
	long            fmt[3];
	int             header = KEYNOTSET, labels = KEYNOTSET, notes = KEYNOTSET;
	long            width = KEYNOTSET, height = KEYNOTSET;
	int             charFormat = 0;
	int             oldStyle = 0;
	int             stripDols = 0;
	int             macroname;
	int             notSingleString = 0;
	struct opEntry *who;
	struct keylist *key;
	char            separator[2];
	char          **missingCode = (char **) 0;
	char          **zeroCode = (char **) 0;
	char           *name, **nameH = (char **) 0;
	char           *keyword;
	char           *fileName = (char *) 0;
	double          missValue;
	long            nsig = KEYNOTSET;
	WHERE("print");
	
/* save current format values in case they are changed */
	saveFormat();
	setMissing(missValue);
	SETUPINT(cleanup);
	
	*OUTSTR = '\0';
	*separator = '\0';

	for (who = Ops;who->iop1 != 0;who++)
	{
		if (strcmp(who->name,FUNCNAME) == 0)
		{
			op = who->iop1;
			keyOp = who->iop2;
			break;
		}
	} /*for (who = Ops;who->iop != 0;who++)*/

	macroname = op & IERR;

	if (!(op & IERR))
	{ /* Look for file:"filename" */
		for (i = 0; i < nargs;i++)
		{
			argi = COMPVALUE(list,i);
		
			if ((keyword = isKeyword(argi)) && strcmp(keyword,"file") == 0)
			{
				if (filearg != SymbolZero)
				{
					goto repeatedKey;
				}
				
				unusedArgs--;
				op |= IFILE;
				ifile = i + 1;
				filearg = argi;
				if (!ISCONSOLE(NAME(filearg)) && !isCharOrString(filearg))
				{
					notCharOrString(keyword);
					goto errorExit;
				}
				setNAME(argi,USEDKEYWORD);
			} /*if ((keyword = isKeyword(argi)) && strcmp(keyword,"file") == 0)*/
		} /*for (i=0;i<nargs;i++)*/
	} /*if (!(op & IERR))*/
	
	if (op & IFILE)
	{ /* operation writing to a file */
		if (filearg == SymbolZero)
		{ /* no file:fileName ; file must be first argument */
			ifile = toFile = 1;
			filearg = arg1;
			if (!ISCONSOLE(NAME(filearg)) && !isCharOrString(filearg))
			{
				notCharOrString("file name");
				goto errorExit;
			}
			unusedArgs--;
		}
		else
		{ /* file:"fileName" was present" */
			toFile = 2;
		}
		
		useConsole = ISCONSOLE(NAME(filearg)) || ISCONSOLE(STRINGPTR(filearg));

		if (useConsole)
		{
			fileName = CONSOLENAME;
		}
	} /*if (op & IFILE)*/
	
	if (toFile)
	{ /* writing to a file */
		fileName = (useConsole) ? CONSOLENAME : STRINGPTR(filearg);
	} /*if (toFile)*/

	for (i=(toFile == 1) ? 1 : 0;i<nargs;i++)
	{ /* check arguments and keywords, saving trailing formatting info */
		int        restoreDefault = 0;

		argi = COMPVALUE(list,i);
		if (!argOK(argi, (!(op & (IMAT|IMACRO))) ? NULLSYM : 0,i+1))
		{
			goto errorExit;
		}
		type = TYPE(argi);

		if ((keyword = isKeyword(argi)))
		{
			for (key = LegalKeys;key->type != 0;key++)
			{
				if (((key->matchlength) ?
					strncmp(keyword, key->name, key->matchlength) :
					strcmp(keyword, key->name)) == 0)
				{
					break;
				}
			} /*for (key = LegalKeys;key->type != 0;key++)*/

			targetType = key->type;
			
			if (key->legalOps & keyOp)
			{
				int       whichKey = key - LegalKeys;

				if (whichKey == KMISSING)
				{
					targetType = (op & IMAT) ? REAL : CHAR;
				}
				
				if (whichKey == KCOMMENTS)
				{
					if (type != targetType || !isVector(argi))
					{
						sprintf(OUTSTR,
								"ERROR: value for %s not CHARACTER vector or scalar",
								keyword);
						goto errorExit;
					}
				} /*if (whichKey == KCOMMENTS)*/
				else if (!isScalar(argi) || type != targetType)
				{
					if (targetType == CHAR)
					{
						notCharOrString(keyword);
					}
					else if (targetType == LOGIC)
					{
						notTorF(keyword);
					}
					else
					{
						notNumberOrReal(keyword);
					}
					goto errorExit;
				} /*if (!isScalar(argi) || type != targetType)*/

				
				if (targetType != CHAR && isMissing(DATAVALUE(argi, 0)))
				{
					if (targetType == LOGIC)
					{
						notTorF(keyword);
					}
					else
					{
						sprintf(OUTSTR,
								"ERROR: MISSING value not legal for keyword %s",
								keyword);
					}
					goto errorExit;
				} /*if (targetType != CHAR && isMissing(DATAVALUE(argi, 0)))*/
				
				switch (whichKey)
				{
				  case KNEW:
					if (!toFile)
					{
						sprintf(OUTSTR,
								"ERROR: keyword %s illegal unless a file is specified",
								keyword);
						goto errorExit;
					}
					if (fileMode)
					{
						goto repeatedKey;
					}
					fileMode = (DATAVALUE(argi,0) == 0.0) ?
						TEXTAPPENDMODE : TEXTWRITEMODE;
					setNAME(argi,USEDKEYWORD);
					break;
					
				  case KSEP:
					if (strlen(STRINGPTR(argi)) > 1)
					{
						sprintf(OUTSTR,
								"ERROR: value for '%s' more than 1 character",
								keyword);
						goto errorExit;
					}
					strcpy(separator, STRINGPTR(argi));
					break;
					
				  case KHEADER:
					header = (DATAVALUE(argi,0) == 0.0) ? 0 : 1;
					notSingleString = 1;
					break;

				  case KLABELS:
					labels = (DATAVALUE(argi,0) == 0.0) ? 0 : 1;
					notSingleString = 1;
					break;

				  case KNOTES:
					notes = (DATAVALUE(argi,0) == 0.0) ? 0 : 1;
					notSingleString = 1;
					break;

				  case KMACRONAME:
					macroname = (DATAVALUE(argi,0) == 0.0) ? 0 : 1;
					break;
					
				  case KNSIG:
					{
						double          dsig  = DATAVALUE(argi,0);

						if (dsig <= 0.0 || dsig != floor(dsig))
						{
							notPositiveInteger(keyword);
							goto errorExit;
						}
						nsig = (long) dsig;
						fmt[0] = nsig+7;
						fmt[1] = nsig;
						fmt[2] = (long) 'g';
					}
					break;
					
				  case KFORMAT:
					if (!setFormat(STRINGPTR(argi),fmt))
					{
						sprintf(OUTSTR,
								"ERROR: '%s' is not a legal value for '%s'",
								STRINGPTR(argi), keyword);
						goto errorExit;
					}
					nsig = fmt[1];
					break;
					
				  case KZERO:
					zeroCode = STRING(argi);
					break;

				  case KMISSING:
					if (op & IMAT)
					{
						missValue = DATAVALUE(argi,0);
					}
					else
					{
						missingCode = STRING(argi);
					}
					break;
					
				  case KNAME:
					nameH = STRING(argi);
					notSingleString = 1;
					break;
					
				  case KQUOTED:
					charFormat = (DATAVALUE(argi,0) != 0.0) ?
						BYQUOTEDFIELDS : 0;
					break;

				  case KBYLINES:
					charFormat = (DATAVALUE(argi,0) != 0.0) ?
						BYLINES : 0;
					break;

				  case KCOMMENTS:
					symhComment = argi;
					break;
					
				  case KHEIGHT:
				  case KWIDTH:
					{
						double         value = DATAVALUE(argi,0);
						long           ivalue;
						int            isheight = (whichKey == KHEIGHT);
						long           minvalue =
							((isheight) ? MINSCREENHEIGHT : MINSCREENWIDTH);

						if (value <= 0.0 || value != floor(value))
						{
							notPositiveInteger(keyword);
							goto errorExit;
						}
						ivalue = (long) value;
						if (ivalue < minvalue)
						{
							sprintf(OUTSTR,
									"ERROR: value for %s is %ld < %ld; illegal",
									keyword, ivalue, minvalue);
							goto errorExit;
						}
						if (isheight)
						{
							height = ivalue;
						} /*if (isheight)*/
						else
						{
							width = ivalue;
						}
					}
					break;
					
				  case KOLDSTYLE:
					oldStyle = (DATAVALUE(argi,0) != 0.0) ? 1 : 0;
					break;
					
				  case KSTRIPDOLS:
					stripDols = (DATAVALUE(argi,0) != 0.0) ? 1 : 0;
					break;
					
				} /*switch (whichKey)*/
				unusedArgs--;
			} /*if (key->legalOps & keyOp)*/
			else if (strcmp(keyword, USEDKEYWORD+2) != 0)
			{  /* not recognized as a legal keyword for this command */
				restoreDefault = 1;
			}
		} /*if ((keyword = isKeyword(argi)))*/
		else /* not a keyword */
		{
			restoreDefault = 1;
		}
		if (restoreDefault)
		{
			nameH = (char **) 0;
			setMissing(missValue); /* wipes out previous missValue*/
			missingCode = (char **) 0;
			zeroCode = (char **) 0;
			nsig = KEYNOTSET; /*wipes out previous format or nsig keywords*/
			width = height = KEYNOTSET;
			charFormat = 0;
			header = labels = notes = KEYNOTSET; /* restore default */
			separator[0] = '\0';
			symhComment = SymbolZero;
			oldStyle = 0;
			stripDols = 0;
		} /*if (restoreDefault)*/
	} /*for (i=(toFile == 1) ? 1 : 0;i<nargs;i++)*/

	if (unusedArgs == 0)
	{
		sprintf(OUTSTR,"ERROR: nothing specified to print or write");
		goto errorExit;
	} /*if (unusedArgs == 0)*/
	
	if (toFile)
	{
		if (fileMode == (char *) 0)
		{ /* default mode is append (new:T) */
			fileMode = TEXTAPPENDMODE;
		}
		
#ifdef HASFINDFILE
		fileName = mvFindFile(fileName, "File for output:", NullString,
							  mvWriteit | mvTextFile,
							  getMvDirInfo(printVol));

		if (fileName == (char *) 0)
		{
			goto cleanup;
		}

#endif /*HASFINDFILE*/

		fileName = expandFilename(fileName);
		if (fileName == (char *) 0 || !isfilename(fileName))
		{
			goto cleanup;
		}
	} /* if (toFile) */

	if (toFile && !useConsole)
	{
		outFile = fmyopen(fileName,fileMode);
		if (outFile == (FILE *) 0)
		{
			sprintf(OUTSTR,"ERROR: %s() cannot open %s for writing",
					FUNCNAME,STRINGPTR(arg1));
			putErrorOUTSTR();
			goto cleanup;
		}
#ifdef MACINTOSH
		{
			mvFileInfo     fileInfo;

			setMvFileVolume(&fileInfo, getMvVolID(printVol));
			setMvFileDirectory(&fileInfo, getMvDirID(printVol));
			setMvFileType(&fileInfo, 'TEXT');
			c2pstrcpy(getMvFileName(&fileInfo), fileName);
			macSetFInfo(&fileInfo, TEXTCREATOR);
		}
#endif /*MACINTOSH*/
	} /*if (toFile && !useConsole)*/

	if (nsig < 0)
	{ /* no format has been set */
		for (i = 0; i < 3;i++)
		{
			fmt[i] = (op & IWRITE) ? WRITEFORMAT[i] : PRINTFORMAT[i];
		}
	} /*if (nsig < 0)*/

	installFormat(fmt[0], fmt[1], fmt[2]);

	/* Note: SCREENWIDTH and SCREENHEIGHT are restored by restoreFormat() */
	installScreenDims((width > 0) ? width : SCREENWIDTH,
					  (height > 0) ? height : SCREENHEIGHT);

	if ((op & (IWRITE|IPRINT)) != 0 && (op & IMAT) == 0 && unusedArgs == 1
		&& !notSingleString)
	{
		/*
		   only one thing to write or print
		*/
		for (i = (toFile == 1) ? 1 : 0;i < nargs;i++)
		{
			argi = COMPVALUE(list,i);
			if (!argOK(argi, NULLSYM,i+1))
			{
				goto cleanup;
			}
			
			if (!isKeyword(argi) && isCharOrString(argi))
			{/* if single character variable, print as is, no header */
				printSingleString(STRINGPTR(argi), outFile, macroname,
								  op & IERR);
				goto okExit;
			} /*if (!isKeyword(argi) && isCharOrString(argi))*/
		} /*for (i = (toFile == 1) ? 1 : 0;i < nargs;i++)*/
	} /*if ((op&(IWRITE|IPRINT)) != 0 && (op&IMAT) == 0 && unusedArgs == 1 ...)*/

	for (i = (toFile == 1) ? 1 : 0; i < nargs && unusedArgs > 0; i++)
	{
		argi = COMPVALUE(list, i);
		if (!argOK(argi, NULLSYM,i+1))
		{
			goto cleanup;
		}
		type = TYPE(argi);

		/*
		   960827 Keywords should all have legal values and need not be checked
		*/
		if ((keyword = isKeyword(argi)))
		{
			for (key = LegalKeys; key->type != 0; key++)
			{
				if (((key->matchlength) ?
					strncmp(keyword, key->name, key->matchlength) :
					strcmp(keyword, key->name)) == 0)
				{
					break;
				}
			} /*for (key = LegalKeys;key->type != 0;key++)*/

			if (key->legalOps & keyOp)
			{
				int      whichKey = key - LegalKeys;

				switch (whichKey)
				{
				  case KFORMAT:
					if (setFormat(STRINGPTR(argi), fmt))
					{
						installFormat(fmt[0], fmt[1], fmt[2]);
					}
					break;
					
				  case KNSIG:
					nsig = (long) DATAVALUE(argi, 0);
					installFormat(nsig+7, nsig, (long)'g');
					break;
					
				  case KMISSING:
					if (op & IMAT)
					{
						missValue = DATAVALUE(argi, 0);
					}
					else
					{
						missingCode = STRING(argi);
					}
					break;

				  case KZERO:
					zeroCode = STRING(argi);
					break;

				  case KSEP:
					strcpy(separator, STRINGPTR(argi));
					break;
					
				  case KHEADER:
					header = (DATAVALUE(argi,0) == 0.0) ? 0 : 1;
					break;
					
				  case KLABELS:
					labels = (DATAVALUE(argi,0) == 0.0) ? 0 : 1;
					break;
					
				  case KNOTES:
					notes = (DATAVALUE(argi,0) == 0.0) ? 0 : 1;
					break;
					
				  case KNAME:
					nameH = STRING(argi);
					break;
					
				  case KQUOTED:
					charFormat = (DATAVALUE(argi,0) != 0.0) ?
						BYQUOTEDFIELDS : 0;
					break;
					
				  case KBYLINES:
					charFormat = (DATAVALUE(argi,0) != 0.0) ?
						BYLINES : 0;
					break;
					
				  case KCOMMENTS:
					symhComment = argi;
					break;
					
				  case KHEIGHT:
				  case KWIDTH:
					{
						long      ivalue = (long) DATAVALUE(argi, 0);
						int       isheight = (whichKey == KHEIGHT);

	/* Note: SCREENWIDTH and SCREENHEIGHT are restored by restoreFormat() */
						installScreenDims((!isheight) ? ivalue : SCREENWIDTH,
										  (isheight) ? ivalue : SCREENHEIGHT);
					}
					break;
					
				  case KOLDSTYLE:
					oldStyle = (DATAVALUE(argi,0) != 0.0) ? 1 : 0;
					break;
					
				  case KSTRIPDOLS:
					stripDols = (DATAVALUE(argi,0) != 0.0) ? 1 : 0;
					break;
					
				} /*switch (whichKey)*/
				setNAME(argi,USEDKEYWORD);
			} /*if (key->legalOps & keyOp)*/			
			if (strcmp(NAME(argi), USEDKEYWORD) == 0)
			{
				argi = SymbolZero;
			}
		} /*if ((keyword = isKeyword(argi)))*/

		if (argi != SymbolZero)
		{
			unusedArgs--;
		
			if ((op & (IMAT + IMACRO)) && type == STRUC)
			{
				if (strucAnyType(argi, PLOTINFO))
				{
					sprintf(OUTSTR,
							"WARNING: %s() will write GRAPH components of structure as %s component",
							FUNCNAME, NULLNAME);
					putErrorOUTSTR();
				}
			} /*if ((op & (IMAT + IMACRO)) && type == STRUC)*/

			name = (nameH != (char **) 0) ? *nameH : NAME(argi);
			nameH = (char **) 0; /* use name only once */
			if (op & (IMAT + IMACRO))
			{
				if (type == REAL || type == LOGIC || type == CHAR ||
					type == STRUC || type == NULLSYM || type == PLOTINFO)
				{
					if (op & IMACRO)
					{
						sprintf(OUTSTR,
								"WARNING: %s variable written by %s()",
								typeName(type), FUNCNAME);
						putErrorOUTSTR();
					}
				}
				else if (type == MACRO)
				{
					if (op & IMAT)
					{
						sprintf(OUTSTR,
								"WARNING: %s written by %s()",
								typeName(type), FUNCNAME);
						putErrorOUTSTR();
					}
				}
				else
				{ /* not REAL, LOGIC, CHAR, MACRO, STRUC or NULLSYM*/
					sprintf(OUTSTR,
							"ERROR: %s() cannot write variable of type %s",
							FUNCNAME, typeName(type));
					putErrorOUTSTR();
					goto cleanup;
				}
				/* 971103 all calls routed through matWrite()*/
				matWrite(argi, outFile, missValue,
						 (header < 0) ? (separator[0] == '\0') : header,
						 labels, notes, separator, charFormat, oldStyle,
						 stripDols, name, symhComment);
						 
				symhComment = SymbolZero;
			} /*if (op & (IMAT + IMACRO))*/
			else
			{/* not matwrite, matprint, or macrowrite */
				if (isKeyword(argi) || isscratch(name))
				{
					name += 2;
				}
				else if (argi == NULLSYMBOL)
				{
					name = "NULL";
				}
				if (header)
				{
					sprintf(OUTSTR,"%s:", name);
					fmyprint(OUTSTR,outFile);
					*OUTSTR = '\0';
					fmyeol(outFile);
				}
#ifndef NOTESARECHAR
				if (notes > 0 && HASNOTES(argi))
				{
					NotesType  **symhNotes = NOTES(argi);
					char        *theseNotes = STRINGPTR(symhNotes);
					long         nLines = DIMVAL(symhNotes, 1);
					long         iLine;
					
					for (iLine = 0; iLine < nLines; iLine++)
					{
						mymultiprint(theseNotes, outFile, 1);
						theseNotes = skipStrings(theseNotes, 1);
					} /*for (iLine = 0; iLine < nLines; iLine++)*/
				} /*if (notes > 0 && HASNOTES(argi))*/
#endif /*NOTESARECHAR*/
 				fprexpr(argi,outFile,labels, missingCode, zeroCode);
			} /*if (op & (IMAT + IMACRO)){...}else{...}*/
		} /*if (argi != SymbolZero)*/

		if (INTERRUPT != INTNOTSET)
		{
			if (INTERRUPT != PRINTABORT)
			{
				goto cleanup;
			}
			break;
		}
	} /*for (i = (toFile == 1) ? 1 : 0; i < nargs; i++)*/

  okExit:
	result = (op != IERROR) ? NULLSYMBOL : SymbolZero;

	goto cleanup;

  repeatedKey:
	sprintf(OUTSTR,
			"ERROR: repeated use of keyword '%s' as argument to %s()",
			keyword, FUNCNAME);
/* fall through*/
  errorExit:
	putErrorOUTSTR();

  cleanup:
	if (outFile != (FILE *) 0 && outFile != STDOUT)
	{
		fclose(outFile);
		outFile = (FILE *) 0;
	}
	restoreFormat();

	*OUTSTR = '\0';
	
	return (result);
} /*print()*/

/*
  putascii(vec [, keep:T] [,multiline:T] [,file:filename]),
  where vec[i] > 0 && vec[i] < 256
  routine to output raw ascii codes
  putascii(vec1, vec2 [, vec3] ...) same as putascii(vector(vec1, vec2 [, vec3] ...), ...)

  getascii(charVec1 [,charVec2 ...]) translates character variables to
  a vector of integer ascii codes 

  960502 changed fopen() to fmyopen()
  980310 now checks for MISSING in arguments
  980903 added code for getascii()
  071012 added keyword phrase multiline:T; without keep:T, output is vector
         of separate characters; with keep:T, separate characters are written
         on separate lines of file
*/
#define GBUFFER  0

Symbolhandle    putascii(Symbolhandle list)
{

	Symbolhandle    arg1, symh;
	Symbolhandle    result = SymbolZero;
	long            i, j, n = 0, length;
	long            nargs = NARGS(list);
	int             new = 0, toFile = 0, keep = 0, multiline = 0;
	int             put = strcmp(FUNCNAME, "putascii") == 0;
	double          val;
	char           *keyword, *fileName = (char *) 0;
	char           *fileMode = TEXTAPPENDMODE;
	unsigned char **buffer = (unsigned char **) 0;
	FILE           *fp = STDOUT;
	WHERE("putascii");
	TRASH(1,errorExit);
	
	SETUPINT(errorExit);

	if ((arg1 = COMPVALUE(list,0)) == EmptyArgument && nargs == 1)
	{
		nargs = 0;
	}
	
	for (i=nargs-1;i>=0;i--)
	{
		symh = COMPVALUE(list,i);
		if (!(keyword = isKeyword(symh)))
		{
			break;
		}
		if (!put)
		{ /* no keywords allowed on getascii()*/
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		if (strcmp(keyword,"file") == 0)
		{
			if (!isCharOrString(symh))
			{
				notCharOrString("'file'");
				goto errorExit;
			}
			toFile = 1;
			fileName = STRINGPTR(symh);
		}
		else if (strcmp(keyword,"new") == 0 || strcmp(keyword,"keep") == 0)
		{
			int      logicalValue;
			
			if (!isTorF(symh))
			{
				notTorF(keyword);
				goto errorExit;
			}
			logicalValue = (DATAVALUE(symh, 0) != 0.0);
			
			if (keyword[0] == 'n')
			{
				new = logicalValue;
			}
			else
			{
				keep = logicalValue;
			}
		}
		else if (strcmp(keyword,"multiline") == 0)
		{
			if (!isTorF(symh))
			{
				notTorF(keyword);
				goto errorExit;
			}
			multiline = DATAVALUE(symh, 0) != 0.0;
		}
		else
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
		nargs--;
	} /*for (i=nargs-1;i>=0;i--)*/

	if (new && !toFile)
	{
		sprintf(OUTSTR,
				"ERROR: 'new:T' illegal without 'file:fileName'");
	}
	else if (keep && toFile)
	{
		sprintf(OUTSTR,
				"ERROR: 'keep:T' is illegal with 'file:fileName'");
	}
	else if (nargs == 0)
	{
		sprintf(OUTSTR,
				"ERROR: no data supplied for %s()", FUNCNAME);
	}

	if (*OUTSTR)
	{
		goto errorExit;
	}
	
	for (i = 0;i < nargs;i++)
	{
		symh = COMPVALUE(list,i);
		if (!argOK(arg1,0,i+1))
		{
			goto errorExit;
		}
		if (!isVector(symh) || TYPE(symh) != ((put) ? REAL : CHAR) ||
			put && anyMissing(symh))
		{
			sprintf(OUTSTR,
					"ERROR: data arguments to %s() must be %s scalars or vectors",
					FUNCNAME, (put) ? "non-MISSING REAL" : "CHARACTER");
			goto errorExit;
		}
		n += symbolSize(symh);
	} /*for (i = 0;i < nargs;i++)*/
	
	if (toFile)
	{
		if (new)
		{ /* default mode is "a" (new:F) */
			fileMode = TEXTWRITEMODE;
		}

#ifdef HASFINDFILE
		fileName = mvFindFile(fileName, "File for putascii():", NullString,
							  mvWriteit | mvTextFile,
							  getMvDirInfo(printVol));

		if (fileName == (char *) 0)
		{
			goto errorExit;
		}
#endif /*HASFINDFILE*/
		fileName = expandFilename(fileName);
		if (fileName == (char *) 0 || !isfilename(fileName))
		{
			goto errorExit;
		}
		fp = fmyopen(fileName,fileMode);
		if (fp == (FILE *) 0)
		{
			sprintf(OUTSTR,"ERROR: %s() cannot open file %s",
					FUNCNAME, fileName);
			goto errorExit;
		}
	} /* if (toFile) */
		
	if (put)
	{ /*putascii()*/
		long       needed = (multiline) ? 2*n : n + 1;

		if (!getScratch(buffer, GBUFFER, needed, unsigned char))
		{
			goto errorExit;
		}

		n = 0;
		for (i = 0;i < nargs;i++)
		{
			symh = COMPVALUE(list,i);
			length = symbolSize(symh);
			for (j = 0;j < length;j++)
			{
				val = DATAVALUE(symh,j);
				if (floor(val) != val || val <= 0.0 || val > 255.0)
				{
					sprintf(OUTSTR,
							"ERROR: all values for %s() must be integers >= 1 and <= 255",
							FUNCNAME);
					goto errorExit;
				}
				(*buffer)[n++] = (unsigned char) val;
				if (multiline) {
					(*buffer)[n++] = (!keep) ? '\n' : '\0';
				}
			} /*for (j = 0;j < length;j++)*/
		} /*for (i = 0;i < nargs;i++)*/
		if (!multiline || !keep)
		{
			(*buffer)[n] = '\0';
		}
		
		if (!keep)
		{
			fmyprint((char *) *buffer, fp);
			fmyflush(fp);
			result = NULLSYMBOL;
		}
		else
		{
			if ((result = Install(SCRATCH,CHAR)) == SymbolError)
			{
				goto errorExit;
			}
			unTrash(GBUFFER);
			setSTRING(result,(char **) buffer);
			setNDIMS(result,1);
			setDIM(result,1,(multiline) ? n/2 : 1);
		}
	} /*if (put)*/
	else
	{ /*getascii()*/
		double    *data;
		long       length = 0, k;
		char      *place;
		
		for (i = 0; i < nargs; i++)
		{
			long       size;
			
			symh = COMPVALUE(list, i);
			
			place = STRINGPTR(symh);
			size = DIMVAL(symh, 1);
			for (j = 0; j < size; j++)
			{
				length += strlen(place);
				place = skipStrings(place, 1);
			}
		} /*for (i = 0; i < nargs; i++)*/

		if (length > 0)
		{
			result = RInstall(SCRATCH, length);
			if (result == SymbolError)
			{
				goto errorExit;
			}

			k = 0;
			data = DATAPTR(result);
			for (i = 0; i < nargs; i++)
			{
				long       size;
			
				symh = COMPVALUE(list, i);
			
				place = STRINGPTR(symh);
				size = DIMVAL(symh, 1);
				for (j = 0; j < size; j++)
				{
					unsigned char     *pc = (unsigned char *) place;
				
					while (*pc)
					{
						data[k++] = (double) *pc++;
					}
					place = skipStrings(place, 1);
				} /*for (j = 0; j < size; j++)*/
			} /*for (i = 0; i < nargs; i++)*/
		} /*if (length > 0)*/
		else
		{
			result = NULLSYMBOL;
		}
	} /*if (put){}else{}*/
	
	/* fall through */

  errorExit:
	putErrorOUTSTR();

	if (toFile && fp != STDOUT && fp != (FILE *) 0)
	{
		fclose(fp);
	}
	emptyTrash();
	
	return (result);
} /*putascii()*/
