/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2004 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/

#include "globals.h"
#include "keywords.h"

#define READDATA__
#include "mvIo.h"
#include "mvIface.h"

#include "mvhelp.h"

#ifdef CONSOLENAME
#define ISCONSOLE(NAME) (strcmp(NAME, CONSOLENAME) == 0)
#define CONSOLEOK
#else /*CONSOLENAME*/
#define ISCONSOLE(NAME)  0
#define CONSOLENAME ""
#endif /*CONSOLENAME*/

/*
  Usage of functions
	  matread(fileName, matName) [matName may be "" or missing]
      macroread(fileName, matName)
      inforead(fileName, matName)
	  vecread(fileName) and vecread(fileName, byfields:T) read REAL data
	  vecread(fileName, bywords:T), vecread(fileName, bylines:T)
	  and vecread(fileName,bychars:T) read CHARACTER data

	  fileName may be of form string:CharVec or file:CharScalar, in which
	  case it need not be first argument.
	  If matName is omitted on matread(), macroread(), or inforead(),
	  the first non-blank line in the file is assumed to be a header for
	  a macro or data set in the right format.

  In windowed versions, if fileName is "", the user selects the file using the
    standard file dialog box.

    it assumes the file starts with a matrix or macro in the right format.

  Possible other keyword phrases
   skip:"P"    P punctuation, not'+', '-', ',', '.' or '*' (e.g. '#'),
               skips any lines beginning with P (vecread() only)
   stop:"S"    S punctuation, not'+', '-', ',', '.', or '*' (default '!');
               read will be terminated on first occurence of S (vecread()
               only)
   go:"G"      G any character; read will be terminated by first line that
               does not have G in column 1 (vecread() only); incompatible
               with stop
   skipthru:"C" Skip upto and including the first line starting with "C",
               vecread() only
   n:N         N > 0 integer on vecread() limits number of items read
   badvalue:V  V will replace unreadable items (vecread() only)
   quiet:T     Do not echo header lines ({mat,macro,info}read())
   quiet:F     Echo "skipped" lines (vecread())
   echo:T      Echo any lines after the header lines ({mat,macro,info}read())
               Echo any data lines (vecread())
   printname:F Print the file name ({vec,mat,mac,info,}read())
   silent:T    Echo nothing and suppress warning messages; incompatible with
               echo:T or quiet:F
   notfoundok:T No message and NULL returned if data set or macro not found
               ({mat,macro,info}read())
   nofileok:T  No message and NULL returned if file can't be opened or is empty
   badkeyok:T  Unrecognized keywords are not errors

   980924 changed opcode ICOMPREAD to fix bug
   981208 fixed bug that kept files from being closed
   990212 Changed most uses of putOUTSTR() to putErrorOUTSTR()
          Replaced myerrorout() with putOutErrorMsg() and putOutMsg()
   990329 Added keyword skipthru to vecread(). 
          vecread(filename, skipthru:c) starts reading the line
          after the first line starting with c.  If c == '\n', then
		  the read starts the first line after the first completely
		  empty line.  The line is echoed the same way as skipped lines
		  are echoed.  The skip, stop and go characters are not operative
          until after the skipthru character has been found.
   991222 Add keyword phrase 'n:N' to vecread(); vecread() will read at most
          N items.
   000323 vecread(filename, bywords:T) now returns leading comma as "";
          previously a leading comma was ignored.
   000328 With vecread(filename, bywords:T), the stop character is
          recognized at the start of a word;  the is the same behavior as
          with 'byfields:T'
   000518 New vecread keyword phrase startline:n, starts reading on line
          n of the file, that is the first n-1 lines are skipped
   000523 Implemented new function getfilename()
   000601 Added vecread() keyword phrase bypass:n and made separated
          code that just skips initial lines into skipLines() and
          modified calling sequences to findLine(), charVecRead(),
          realVecRead() and doVecRead()
   000614 Treat 'character' and 'bywords' as complete synonyms, with the
          latter as the normal usage ('character' is kept for backward
          compatibility)
   001025 Use allocateLINE() and freeLINE() for allocating and disposing of
          LINE.
   010130 Modified askForFileName() so that it returns complete path name
          on Macintosh; it already did on Windows and Motif; also implemented
          keyword phrases nameonly:T and pathonly:T to strip off path and name
   010305 The name of the file read is printed by default except with 'silent:T'
          and 'string:charVar'
          Added keyword 'printname'.  With printname:F, the name of the
          file being read is not printed.  With printname:T, the name is
		  printed, even with silent:T
   010319 The file name printed is the complete path from new global
          LastPathOpened
   010324 getfilename(last:T) returns the name of the last file
          successfully opened, that is the contents of LastPathOpened
   010524 Changed calling sequence of macFindFile()
   010525 Changed macFindFile() to mvFindFile() and dropped wxFindFile()
   010529 Moved splitFileName() to unxIo.c, wxIo.c, macIo.c
   010601 Include mvIface.h instead of wxIface.h or macIface.h
   010604 Changed askForFileName()
   010616 Added keyword phrase help:T to getfilename()
*/

enum readOpCodes
{
	IVECREAD   = 0x01,
	IMATREAD   = 0x02,
	IMACROREAD = 0x04,
	IREAD      = IMATREAD | IMACROREAD,
	IINFOREAD  = 0x08,
	ICOMPREAD  = 0x10 | IREAD,
	IMACROSEARCH = 0x20 | IMACROREAD,
	IALLREADS  = IVECREAD | IMATREAD | IMACROREAD | IINFOREAD
};


enum readdataKeyCodes
{
	KFILE = 0,
	KSTRING,
	KECHO,
	KQUIET,
	KSILENT,
	KPRINTNAME,
	KPROMPT,
	KLABELS,
	KSTOP,
	KGO,
	KSKIP,
	KSKIPTHRU,
	KN,
	KSTARTLINE,
	KBYPASS,
	KCHAR,
	KBYWORDS,
	KBYCHARS,
	KBYLINES,
	KBADVAL,
	KBYFIELDS,
	KNOTFOUNDOK,
	KNOFILEOK,
	KBADKEYOK
};

static keywordList ReaddataKeys[] = 
{
	InitKeyEntry("file",       0, IALLREADS,               CHARSCALAR),    
	InitKeyEntry("string",     0, IALLREADS,               CHARVECTOR),    
	InitKeyEntry("echo",       0, IALLREADS,               LOGICSCALAR),   
	InitKeyEntry("quiet",      0, IALLREADS,               LOGICSCALAR),   
	InitKeyEntry("silent",     4, IALLREADS,               LOGICSCALAR),   
	InitKeyEntry("printname",  0, IALLREADS,               LOGICSCALAR),   
	InitKeyEntry("prompt",     0, IALLREADS,               LOGICSCALAR),   
	InitKeyEntry("labels",     5, IMATREAD,                SYMHVALUE),     
	InitKeyEntry("stop",       0, IVECREAD,                CHARSCALAR),    
	InitKeyEntry("go",         0, IVECREAD,                CHARSCALAR),    
	InitKeyEntry("skip",       0, IVECREAD,                CHARSCALAR),    
	InitKeyEntry("skipthru",   7, IVECREAD,                CHARSCALAR),    
	InitKeyEntry("n",          0, IVECREAD,                POSITIVEINT),   
	InitKeyEntry("startline",  0, IVECREAD,                POSITIVEINT),   
	InitKeyEntry("bypass",     0, IVECREAD,                NONNEGATIVEINT),
	InitKeyEntry("character",  4, IVECREAD,                LOGICSCALAR),   
	InitKeyEntry("bywords",    6, IVECREAD,                LOGICSCALAR),   
	InitKeyEntry("bychars",    6, IVECREAD,                LOGICSCALAR),   
	InitKeyEntry("bylines",    6, IVECREAD,                LOGICSCALAR),   
	InitKeyEntry("badvalue",   4, IVECREAD,                REALSCALAR),    
	InitKeyEntry("byfields",   5, IVECREAD,                LOGICSCALAR),   
	InitKeyEntry("notfoundok", 0, IMATREAD|IMACROREAD|IINFOREAD,
                                                           LOGICSCALAR),   
	InitKeyEntry("nofileok",   0, IALLREADS,               LOGICSCALAR),   
	InitKeyEntry("badkeyok",   4, IALLREADS,               LOGICSCALAR)    
};

#define Keys       ReaddataKeys

#define File       (KeyCharValue(ReaddataKeys,KFILE))
#define String     (KeySymhValue(ReaddataKeys,KSTRING))
#define Echo       (KeyLogValue(ReaddataKeys,KECHO))
#define Quiet      (KeyLogValue(ReaddataKeys,KQUIET))
#define Labels     (KeySymhValue(ReaddataKeys,KLABELS))
#define Silent     (KeyLogValue(ReaddataKeys,KSILENT))
#define Printname  (KeyLogValue(ReaddataKeys,KPRINTNAME))
#define Prompt     (KeyLogValue(ReaddataKeys,KPROMPT))
#define Stop       (KeyCharValue(ReaddataKeys,KSTOP))
#define Go         (KeyCharValue(ReaddataKeys,KGO))
#define Skip       (KeyCharValue(ReaddataKeys,KSKIP))
#define Skipthru   (KeyCharValue(ReaddataKeys,KSKIPTHRU))
#define N          (KeyIntValue(ReaddataKeys,KN))
#define Startline  (KeyIntValue(ReaddataKeys,KSTARTLINE))
#define Bypass     (KeyIntValue(ReaddataKeys,KBYPASS))
#define CharRead   (KeyLogValue(ReaddataKeys,KCHAR))
#define ByWords    (KeyLogValue(ReaddataKeys,KBYWORDS))
#define ByChars    (KeyLogValue(ReaddataKeys,KBYCHARS))
#define ByLines    (KeyLogValue(ReaddataKeys,KBYLINES))
#undef  BadValue /*defined in X.h*/
#define BadValue   (KeyRealValue(ReaddataKeys,KBADVAL))
#define ByFields   (KeyLogValue(ReaddataKeys,KBYFIELDS))
#define NotFoundOK (KeyLogValue(ReaddataKeys,KNOTFOUNDOK))
#define NoFileOK   (KeyLogValue(ReaddataKeys,KNOFILEOK))
#define BadKeyOK   (KeyLogValue(ReaddataKeys,KBADKEYOK))

static char     *Opnames[] = 
{
	"vecread", "matread", "macroread", "read", "inforead", "macrosearch",
	(char *) 0
};

static short     Opcodes[] =
{
	IVECREAD, IMATREAD, IMACROREAD, IREAD, IINFOREAD, IMACROSEARCH
};

#define FromString (StringToRead != (unsigned char **) 0)

/*
   Globals for communicating with datagetc() (in unxio.c or macIo.c)

   StringToRead == (unsigned char **) 0 signals read by getc() from a file;
     otherwise mygetc() returns (*StringToRead)[PlaceInString], or EOF if
     PlaceInString >= Stringlength ; PlaceInString is incremented.
     If (*StringToRead)[PlaceInString] == '\0', mygetc() returns '\n'

   LastDataWasCr is 1 if last character was CR; otherwise 0.  This allows
     correctly reading text files with CR, LF or CR/LF line separators
     files should be opened in binary mode.

   980805 Changed type of StringToRead to unsigned char **
          Changed because an octal 254 in s stopped vecread(string:s,bychars)
*/
#ifndef DATAPROMPT
#define DATAPROMPT "Input> "
#endif /*DATAPROMPT*/

unsigned char **StringToRead;
long            StringLength;
long            PlaceInString;
char            LastDataWasCr;
/*
  980501 DataPrompt is prompt for next line of data read interactively
  from CONSOLE
*/
char          *DataPrompt;

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Readdata1
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

/*
	Read next data value from *LINE, starting at *LINE + place
	*next is set to position of next character to read
	*badchar is set to '\0' if a number is successfully decoded or if
	? if found;  otherwise it is set to the first non-white bad character
	found or EOF if the scan goes off the end.

	If a non-numeric item is found, scanning continues to the next ',',
	'-', '+' or digit (or end of line) and *next is set to the offset of
	the place stopped at.  When 2 successive commas are found, *next is set	
	to the offset of the second 1;  when a successful decoding is done or '?'
	found, next is set to the offset of the first position after what was
	decoded.

	000115 Started switch to new default behavior which expects comma
	or white space separated fields.  Any field that is not a well formed
    number or a token for missing ('?', '??', '???', ..., '.', '*') is a
	bad item, even if it contains numerals.  Two successive

   000204 'NA' followed by end of line, comma or whitespace is now read as
   MISSING	

   030813 added work around for strtod() bug in CodeWarrior
*/
static int         LastWasComma = 0;
#define isCommaOrSpace(C) (isspace(C) || (C) == ',')
#define isEOL(C) ((C) == '\0')

static double nextRealItem(long place, long *next, int *badchar, int byFields)
{
	char      *cplace = *LINE + place, c, *cnext;
	double     value;
	int        ncommas = 0;
	WHERE("nextRealItem");

	*badchar = '\0';
	while (ncommas < 2 && LastWasComma < 2 &&
		   (c = *cplace , isCommaOrSpace(c)))
	{
		if (c == ',')
		{
			if (byFields)
			{
				LastWasComma++;
			}
			else 
			{
				ncommas++;
			}
		}
		cplace++;
	} /*while (ncommas < 2 && LastWasComma < 2 && (c = *cplace ,isCommaOrSpace(c)))*/

	if (!byFields && ncommas > 1)
	{
		*badchar = ',';
		cplace--;
	}
	else if (LastWasComma > 1)
	{
		LastWasComma--;
		setMissing(value);
	}
	else
	{
		if (isEOL(c))
		{
			if (LastWasComma)
			{
				setMissing(value);
			}
			else
			{
				*badchar = EOF;
			}
		}
		else if (c == '?')
		{
			setMissing(value);
			while ((c = *(++cplace)) && c == '?')
			{/* multiple '?' the same as 1 */
				;
			}
		} /*if (*cplace == '?')*/
		else if ((c == '.' || c == '*') &&
				 (isEOL(cplace[1]) || isCommaOrSpace(cplace[1])))
		{/* isolated '.' is MISSING (like SAS) */
			setMissing(value);
			cplace++;
		}
		else if (c == 'N' && cplace[1] == 'A' &&
				 (isEOL(cplace[2]) || isCommaOrSpace(cplace[2])))
		{
			setMissing(value);
			cplace += 2;
		}
		else
		{
			value = mystrtod(cplace, &cnext);
#ifdef MW_CW
			/*
			  code to work around bug in strtod() that reads things
			  like ".", "-", "+", "..","--", "-+", ... as valid numbers
			  with value 0.  If it starts with '+', '-' or '.', '.' must be
			  followed by a digit and '+' and '-' must be followed by
			  a digit or '.' followed by a digit.
			*/
			if (value == 0.0 && cnext > cplace && 
				strchr("+-.", cplace[0]) != (char *) 0)
			{
				if (cnext == cplace + 1)
				{ /* solitary +, - or .*/
					cnext = cplace;
				}
				else if (!isdigit(cplace[1]))
				{
					if (cplace[0] == '.' || cplace[1] != '.' ||
						cnext == cplace + 2 || !isdigit(cplace[2]))
					{
						cnext = cplace;
					}
				}
			}
#endif /*MW_CW*/
			if (cnext != cplace &&
				(!byFields || isEOL(*cnext) || isCommaOrSpace(*cnext)))
			{ /*decoded number */
				cplace = cnext;
			} /*if (cnext != cplace)*/
			else
			{ /* not a number, skip past bad stuff */
				if (isEOL(*cplace))
				{
					*badchar = EOF;
				} /*if (isEOL(*cplace))*/
				else
				{
					c = *badchar = *cplace;
					if (!byFields)
					{
						if (c == '-' || c == '+' || c == '.')
						{
							cplace++;
						}
				
						while (!isEOL(c = *cplace))
						{
							if (isspace(c) || isdigit(c) || c == '-' || c == '+' ||
								c == '.' || c == ',')
							{
								break;
							}
							cplace++;
						} /*while (!isEOL(c = *cplace))*/
					} /*if (!byFields)*/
					else
					{
						while(!isEOL(c = *cplace) && !isCommaOrSpace(c))
						{
							cplace++;
						}
					} /*if (!byFields){}else{}*/ 
				} /*if (isEOL(*cplace)){}else{}*/
			} /*if (cnext != cplace){}else{}*/
		} /*if (*cplace == '?'){}else{}*/
		LastWasComma = 0;
	}
	
	*next = cplace - *LINE;
	return (value);
	
} /*nextRealItem()*/

/*
   Find the next comma- or whitespace-separated field
   If the field is null (successive commas or trailing comma)
     return a pointer to NullString and set Next so that
     (*LINE)[next] is the next character to be read
   If nothing is found, return (char *) 0;
   Otherwise, return a pointer to the start of the field and set Next
   so that the field starts at *LINE + next 

   000323 added argument atStart which is non-zero only at the start of
          a line.  This is to make it possible to treat a leading
          comma as ""
*/
static char * nextCharItem(long place, long *next, int atStart)
{
	char      *cplace = *LINE + place, c;
	char      *cvalue;
	int        ncommas = (atStart) ? 1 : 0;
	WHERE("nextCharItem");
	
	while (ncommas < 2 && !isEOL(c = *cplace) && isCommaOrSpace(c))
	{
		cplace++;
		if (c == ',')
		{
			ncommas++;
		}
	}

	if (ncommas == 2 || ncommas == 1 && isEOL(c))
	{ /*
		A leading comma, 2 commas with nothing non-white between them,
		or a trailing comma
	  */
		*next = cplace - *LINE;
		if (ncommas == 2)
		{
			(*next)--;
		}
		cvalue = NullString;
	}
	else if (isEOL(c))
	{ /* nothing found */
		cvalue = (char *) 0;
	}
	else
	{
		*next = cplace - *LINE;
		cvalue = cplace;
	}
	return (cvalue);
} /*nextCharItem()*/

static void headerError(long errorNumber, char *matName, char *fileName,
						int warningOnly)
{
	long        printLine = 0;
	char        prefix[10];

	strcpy(prefix, (warningOnly) ? "WARNING" : "ERROR");
	switch(errorNumber)
	{
	  case NOTFOUND:
		sprintf(OUTSTR, "%s: dataset or macro %s not found", prefix, matName);
		break;

	  case COMPONENTNOTFOUND:
		  {
			  int         place;
			  char        c;

			  place = strlen(matName) - 1;
			  c = matName[place];
			  matName[place] = '\0';
				   
			  sprintf(OUTSTR, "%s: component for structure %s not found",
					  prefix, matName);
			  matName[place] = c;
			  break;
		  }
	  
	  case BADHEADER:
		sprintf(OUTSTR, "%s: header for %s in wrong format", prefix, matName);
		printLine = 1;
		break;
	  case TWOFORMATS:
		sprintf(OUTSTR,
				"%s: two scanf formats found in header lines for %s",
				prefix, matName);
		printLine = 1;
		break;
	  case BADFORMAT:
	  case BADCFORMAT:
		sprintf(OUTSTR,
				"%s: something wrong with format associated with %s",
				prefix, matName);
		printLine = 1;
		break;
	  case NODATA:
		sprintf(OUTSTR, "%s: endfile hit before data or macro reached", prefix);
		break;
	  case NOFIELDS:
		sprintf(OUTSTR,
				"%s: header line for %s specifies zero cases", prefix, matName);
		printLine = 1;
		break;
	  case NOLINES:
		sprintf(OUTSTR, "%s: header line specifies 0 lines in macro %s()",
				prefix, matName);
		printLine = 1;
		break;
	  default:
		sprintf(OUTSTR,
				"%s: headerError %ld for %s", prefix, errorNumber, matName);
		printLine = 1;
	} /*switch(errorNumber)*/
	putPieces(OUTSTR, " on file ", fileName, (char *) 0);
	if (printLine)
	{
		putOutMsg("Last line read was:");
		echoLINE();
	}
	*OUTSTR = '\0';
} /*headerError()*/

static void readError(int errorNumber, char *matName, char *fileName,
					  int warningOnly)
{
	char        prefix[10];
	
	strcpy(prefix, (warningOnly) ? "WARNING" : "ERROR");
	
	switch (errorNumber)
	{
	  case NODATA:
		sprintf(OUTSTR, "%s: no data read from dataset %s on file ",
				prefix, matName);
		break;
	  case NOMACRO:
		sprintf(OUTSTR, "%s: no body found for macro %s() on file ",
				prefix, matName);
		break;

	  case CANTALLOC:
		sprintf(OUTSTR, "%s: memory allocation problem reading file ", prefix);
		break;
		
	  default:
		sprintf(OUTSTR,
				"%s: problem %d reading dataset or macro %s from file ",
				prefix, errorNumber, matName);
	} /*switch (errorNumber)*/
	putPieces(OUTSTR, fileName, (char *) 0, (char *) 0);
	putOutMsg("Last line read was:");
	echoLINE();
	*OUTSTR = '\0';
} /*readError()*/

/*
  000601 new function
  If bypass > 0 lines are skipped until bypass lines starting with
  stopCh are read.

  Then, if startline > 1, startline - 1 lines are skipped
  Then if skipthruCh != '\0', additional lines are read, and possibly
  echoed untile a line starting with skipthruCh is read
*/
static int skipLines(FILE *fn, unsigned char stopCh, unsigned char skipthruCh,
					 long bypass, long startline, int echo)
{
	while (bypass > 0 || --startline > 0)
	{
		if (fillLINE(fn) == EOF)
		{
			return (0);
		}
		if (bypass > 0 && (*LINE)[0] == stopCh)
		{
			bypass--;
		}
	}
	while (skipthruCh != '\0')
	{
		if (fillLINE(fn) == EOF)
		{
			return (0);
		}
		if (skipthruCh == '\n' && isAllWhite(*LINE) ||
			(*LINE)[0] == skipthruCh)
		{
			skipthruCh = '\0';
		}
		if (echo)
		{
			echoLINE();
		}
	}
	
	return (1);
} /*skipLines()*/
	
/*
  000601 removed arguments startline and skipthruCh and code
         processing them;  it's now in skipLines()
*/

static int findLine(FILE *fn, long echo, long verbose, unsigned char skipCh,
					unsigned char goCh, int charOp, long * status)
{
	long                 length;
	int                  skipEmptyLines = (charOp == KCHAR);
	unsigned char        c;
	WHERE("findLine");
	
	do
	{
		*status = fillLINE(fn);
		length = strlen(*LINE);
		
		if (length > 0)
		{
			c = (unsigned char) (*LINE)[0];

			if (charOp != KBYCHARS && (*LINE)[length-1] == '\n')
			{
				(*LINE)[--length] = '\0';
			}
			if (echo || verbose && c == skipCh)
			{
				echoLINE();
			}
		} /*if (length > 0)*/
	} while (*status != EOF && ((skipCh != '\0' && c == skipCh) ||
								skipEmptyLines && length == 0));
	if (goCh != '\0' && (*LINE)[0] != goCh)
	{
		*status = EOF;
		length = 0;
	}

	return (*status != EOF || length > 0);
} /*findLine()*/

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Readdata2
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

/*
  970530 modified storage allocation.  The amount allocated each time
         is approximately 1/CHUNKFRACTION of what has already been allocated
         This was to fix very slow reading of very large files
*/

enum readChunks
{
	CHARCHUNK = 400, /*initial increment for allocating space for CHAR data */
	REALCHUNK =  50 ,/*initial increment for allocating space for REAL data */
	CHUNKFRACTION = 20
};

/*
  991223 moved code for doReadVec() to new function realVecRead()
  000121 added argument byFields and modified code to handle new
         behavior when byFields != 0
  000601 removed arguments startline and skipthruCh; lines
         to be skipped should already be passed
*/

static long realVecRead(FILE *fn, char **fnameH, Symbolhandle result,
						long verbose, long silent, long echo, long byFields,
						double badValue,
						unsigned char stopCh, unsigned char goCh,
						unsigned char skipCh, long maxNitems)
{
	long            status, nItems = 0;
	long            allocated = symbolSize(result);
	double        **dataH = DATA(result);
	double          value, tmp;
	long            place;
	long            chunk = REALCHUNK;
	int             warnflag = silent, foundInfinite = 0;
	int             gotsome, foundBad = 0;
	int             badchar;
	WHERE("realVecRead");

	/* read until an acceptable line is found */
	gotsome = findLine(fn, echo, verbose, skipCh, goCh, 0, &status);

	if (gotsome && byFields)
	{
		char        *s = *LINE;
		
		/* make sure leading comma interpreted as MISSING */
		while (isspace(*s))
		{
			s++;
		}
		LastWasComma = (*s == ',') ? 1 : 0;
	}
	
	place = 0;

	while ((maxNitems <= 0 || nItems < maxNitems) &&
		   (gotsome || status != EOF))
	{
		if (allocated <= nItems)
		{/* make more space */
			chunk = (nItems < CHUNKFRACTION*REALCHUNK) ?
				chunk : nItems/CHUNKFRACTION; 
			allocated = nItems + chunk;
			dataH = (double **) mygrowhandle((char **) dataH,
											 allocated * sizeof(double));
			setDATA(result, dataH);
			if (dataH == (double **) 0)
			{
				goto errorExit;
			}
		} /*if (allocated <= nItems)*/

		while (allocated > nItems && gotsome)
		{
			value = nextRealItem(place, &place, &badchar, byFields);
			if (badchar != EOF)
			{
				if (badchar != 0)
				{
					if (badchar == stopCh)
					{
						status = EOF;
						gotsome = 0;
					}
					else if(realIsSet(badValue))
					{
						(*dataH)[nItems++] = badValue;
					}
					else
					{
						foundBad = 1;
						if (byFields)
						{
							setMissing((*dataH)[nItems++]);
						}
					}
				} /*if (badchar != 0)*/
				else
				{
					if (!isMissing(value) &&
						(tmp = fabs(value), doubleEqualBdouble(tmp, TooBigValue)))
					{
						foundInfinite = 1;
						setMissing(value);
					}
					(*dataH)[nItems++] = value;
					if (maxNitems > 0 && nItems >= maxNitems)
					{
						break;
					}

				} /*if (badchar != 0){}else{}*/
			} /*if (badchar != EOF)*/				

			if (maxNitems > 0 && nItems >= maxNitems)
			{
				break;
			}

			if (isEOL((*LINE)[place]) && (!byFields || !LastWasComma))
			{ /* nothing left in line */
				if (status != EOF)
				{ /* try to refill it */
					gotsome = findLine(fn, echo, verbose, skipCh, goCh, 0, &status);
					place = 0;
					if (gotsome && byFields)
					{
						char        *s = *LINE;

						/* make sure leading comma interpreted as MISSING */
						while (isspace(*s))
						{
							s++;
						}
						LastWasComma = (*s == ',') ? 1 : 0;
					}
				}
				else
				{
					gotsome = 0;
				}
			} /*if (isEOL((*LINE)[place]) && (!byFields || !LastWasComma))*/
		} /*while (allocated > nItems && gotsome)*/
	} /*while (gotsome || status != EOF)*/

	if (foundBad && !warnflag)
	{
		putPieces("WARNING: nonnumeric character(s) in ", *fnameH,
				  (byFields) ? " returned as MISSING" : " ignored",
				  (char *) 0);
	}

	if (nItems > 0)
	{/* some data was read */
		if (foundInfinite && !silent)
		{
			putOutErrorMsg("WARNING: numbers outside representable range replaced by MISSING");
		}
		/* shrink data to its actual size */
		if (allocated > nItems)
		{
			dataH = (double **) mygrowhandle((char **) dataH,
											 nItems * sizeof(double));
			setDATA(result, dataH);
			if (dataH == (double **) 0)
			{
				goto errorExit;
			}
		} /*if (allocated > nItems)*/			
	} /*if (nItems > 0)*/
	else
	{
		mydisphandle((char **) dataH);
		setDATA(result, (double **) 0);
	}
	return (nItems);

  errorExit:
	return (-1);
} /*realVecRead()*/

/*
  991222 added argument maxNitems and code to process it.  If maxNitems > 0,
         no more than maxNitems will be read
  000601 removed arguments startline and skipthruCh; lines
         to be skipped should already be passed
*/

static long charVecRead(FILE *fn, Symbolhandle result, long verbose,
						int charOp, long echo,
						unsigned char stopCh, unsigned char goChar,
						unsigned char skipCh, long maxNitems)
{
	char         **chdata = (char **) 0;
	long           needed, allocated = 0, used = 0, place, nItems = 0;
	char          *field, *out, c;
	long           next;
	long           status = 0;
	long           chunk = CHARCHUNK;
	long           length;
	int            foundStop = 0;
	int            atStart;
	int            byLines = (charOp == KBYLINES);
	int            byChars = (charOp == KBYCHARS);
	WHERE("charVecRead");
	
	while (status != EOF && !foundStop &&
		   (maxNitems <= 0 || nItems < maxNitems) &&
		   (atStart = 1,
			findLine(fn, echo, verbose, skipCh, goChar, charOp, &status)))
	{
		if ((byLines  || byChars) && (*LINE)[0] == stopCh)
		{
			break;
		}
		length = strlen(*LINE);

		needed = used + length + 1;

		if (byChars)
		{
			if (used == 0 && maxNitems > 0)
			{
				needed = 2 * maxNitems;
			}
			else
			{
				needed += length;
			}
			
			if (length != 0 && isNewline((*LINE)[length-1]))
			{
				needed--;
			}
		} /*if (byChars)*/
		
		if (needed > allocated)
		{
			chunk = (needed < CHUNKFRACTION * CHARCHUNK) ?
				chunk : needed/CHUNKFRACTION;
			
			if (needed - used < chunk)
			{
				needed = used + chunk;
			}
			
			chdata = (allocated == 0) ?
				mygethandle(needed) : mygrowhandle(chdata, needed);
			setSTRING(result, chdata);
			if (chdata == (char **) 0)
			{
				goto errorExit;
			}
			allocated = needed;
		} /*if (needed > allocated)*/
		
		if (byLines)
		{
			used = copyStrings(*LINE, *chdata + used, 1) - *chdata;
			nItems++;
		} /*if (byLines)*/
		else if (byChars)
		{
			char      *line = *LINE;
			
			do
			{
				c = *line++;
				
				(*chdata)[used++] = c;

				if (!isEOL(c))
				{
					(*chdata)[used++] = '\0';
				}
				nItems++;
				if (maxNitems > 0 && nItems >= maxNitems)
				{
					break;
				}
			} while (!isEOL(c) && !isNewline(c));
		} /*if (byLines) {}else if(byChars){}*/
		else
		{ /*bywords*/
			place = 0;

			while (!foundStop && (maxNitems <= 0 || nItems < maxNitems))
			{
				out = *chdata;
				/* there should be room to copy whatever we find to chdata*/
				field = nextCharItem(place, &next, atStart);
				atStart = 0;
				if (field == (char *) 0)
				{ /* off the end of the line */
					break;
				}

				if (isEOL(*field))
				{
					/* null field */
					out[used++] = '\0';
					nItems++;
					place = next;
					if (isEOL((*LINE)[next]))
					{
						break;
					}
				}
				else if (*field == stopCh)
				{ /* field starting with '!' or other stop character */
					foundStop = 1;
				}
				else
				{ /* anything else; copy up to whitespace, comma, or '!' */
					/* 000328 '&& c != stopCh' removed from test */
					while (!isEOL(c = *field) && !isCommaOrSpace(c))
					{
						out[used++] = c;
						field++;
					}
					out[used++] = '\0';
					nItems++;
					place = field - *LINE;
				}
			} /*while (!foundStop)*/
		} /*if (byLines) {}else if(byChars){}else{}*/
	}

	if (nItems == 0)
	{ /* nothing found */
		mydisphandle(chdata);
		setSTRING(result, (char **) 0);
	} /*if (nItems == 0)*/
	else
	{
		if (byChars && used > 1 &&
			isEOL((*chdata)[used-1]) && isEOL((*chdata)[used-2]))
		{
			nItems--;
			used--;
		}
		
		chdata = mygrowhandle(chdata, used);
		setSTRING(result, chdata);
		if (chdata == (char **) 0)
		{
			goto errorExit;
		}
	} /*if (nItems == 0){}else{}*/
	
	return (nItems);

  errorExit:
	return (-1);
	
} /*charVecRead()*/


/*
  991222 added argument maxNitems and code to process it.  If maxNitems > 0,
         no more than maxNitems will be read
  991223 moved REAL vector code to new function realVecRead()
  000115 added argument oldStyle
  000121 changed argument oldStyle to byFields
  000601 new argument bypass and skipping of bypass-1 stop chars
         and then startline - 1 lines is done here.  Removed
         startline from arguments to charVecRead() and realVecRead()
*/
static Symbolhandle doReadVec(FILE *fn, char **fnameH, long verbose, long echo,
							  long silent, long byFields, int charOp,
							  double badValue, unsigned char stopCh,
							  unsigned char goCh, unsigned char skipthruCh,
							  unsigned char skipCh, long maxNitems,
							  long bypass, long startline)
{
	long            nItems;
	Symbolhandle    result;
	long            warnflag = silent;
	long            chunk = REALCHUNK;
	WHERE("doReadVec");

	if (charOp)
	{
		result = CInstall(SCRATCH, 0);
	}
	else
	{
		result = RInstall(SCRATCH, (maxNitems <= 0) ? chunk : maxNitems);
	}
	if (result == SymbolError)
	{
		goto errorExit;
	}

	/*
	  Skip startline - 1 lines after passing bypass lines
	  starting with stopCh and then, when skipthruCh != '\0'
	  skip to just past the first line starting with skipthruCh
	*/
	if (!skipLines(fn, stopCh, skipthruCh, bypass,
				   startline, echo || verbose))
	{
		sprintf(OUTSTR, "%sno data found in %s", ErrorStart, *fnameH);
		goto errorExit;
	}
  
			
	if (!charOp)
	{
		nItems = realVecRead(fn, fnameH, result, verbose, silent, echo,
							 byFields,
							 badValue, stopCh, goCh, skipCh, maxNitems);
	} /*if (!charOp)*/
	else
	{
		nItems = charVecRead(fn, result, verbose, charOp,
							 echo, stopCh, goCh, skipCh, maxNitems);
	} /*if (!charOp){}else{}*/

	if (nItems < 0)
	{
		goto errorExit;
	}
	
	if (nItems > 0)
	{ /* some data was read */
		setNDIMS(result, 1);
		setDIM(result, 1, nItems);
		setNCLASS(result, -1);
		if (maxNitems > 0 && nItems < maxNitems && !warnflag)
		{
			sprintf(OUTSTR,
					"WARNING: fewer than %ld items read by %s()",
					maxNitems, FUNCNAME);
			putErrorOUTSTR();
		}
	} /*if (nItems > 0)*/
	else
	{
		if (!silent)
		{
			sprintf(OUTSTR,
					"WARNING: no data found by %s()", FUNCNAME);
			putErrorOUTSTR();
		}

		setTYPE(result, NULLSYM);
		setNAME(result, NULLSCRATCH);
		setNDIMS(result, 1);
		setDIM(result, 1, 0);
	} /*if (nItems > 0){}else{}*/
	
	return (result);

  errorExit:
	putErrorOUTSTR();
  	Removesymbol(result);
	return (SymbolError);

} /*doReadVec()*/

/*
  991123 New function to attach labels read from a file to the
         symbol with which they are associated.  As of now, it
         does no expansion of labels so the number of labels must
         match the sum of the dimensions of result.  At some point
         it will probably recognize labels like
            vector("@","[","X1","X2","X3")
         and expand them appropriately.
  991216 when there are fewer labels than
         sum(DIMVAL(result,i),i=1,NDIMS(result)), labels starting with
         @, or of the form (, [, {, <, /, or \ are expanded similarly to
         the expansion done by setlabels().  If labels for any dimension
         are to be expanded, any dimension label starting with '@' is
         assumed to be a "singleton" to be expanded.
*/

static int attachLabelsFromFile(Symbolhandle result,
								Symbolhandle labelsFromFile)
{
	long         nlabels, mlabels =	DIMVAL(labelsFromFile, 1);
	int          ndims, i, j;
	char       **labelsH = STRING(labelsFromFile);
	char       **labelsH1 = labelsH;
	WHERE("attachLabelsFromFile");

	if (TYPE(result) == STRUC)
	{
		nlabels = NCOMPS(result);
		ndims = 1;
	}
	else
	{
		int        i;

		ndims = NDIMS(result);

		nlabels = 0; /* total number of label elements required */
		for (i = 1; i <= ndims; i++)
		{
			nlabels += DIMVAL(result, i);
		}
	}

	if (mlabels > nlabels)
	{
		return (0);
	}
	
	if (mlabels < nlabels)
	{
		long      mlabels1 = mlabels, needed = 0;
		char     *place = *labelsH;
		
		for (i = 1; i <= ndims; i++)
		{
			long        dimi = DIMVAL(result, i);

			if (place[0] == NUMERICLABEL ||
				isEOL(place[1]) && strchr(Lparens, place[0]) != (char *) 0)
			{
				mlabels1 += dimi - 1;
				if (mlabels1 > nlabels)
				{
					return (0);
				}
				needed += expandLabels(place, dimi, (char *) 0);
				place = skipStrings(place,1);
			}
			else
			{
				for (j = 0; j < dimi; j++)
				{
					needed += strlen(place) + 1;
					place = skipStrings(place, 1);
				}
			}
		} /*for (i = 1; i <= NDIMS(result))*/
		if (mlabels1 != nlabels)
		{
			return (0);
		}
		
		labelsH1 = mygethandle(needed);
		if (labelsH1 == (char **) 0)
		{
			return (-1);
		}
	} /*if (mlabels < nlabels)*/
	if (!setLabels(result, labelsH1))
	{
		return (-1);
	}

	/*
	  At this point, if expansion is to occur, labelsH is still attached
	  to labelsFromFile and labelsH1 is attached to result; otherwise
	  labelsH and labelsH1 are the same and are attached to both
	  labelsFromFile and result
	*/
	if (mlabels < nlabels)
	{
		char      *place = *labelsH;
		char      *place1 = *labelsH1;
		
		for (i = 1; i <= ndims; i++)
		{
			long         dimi = DIMVAL(result, i);
			
			if (place[0] == NUMERICLABEL ||
				isEOL(place[1]) && strchr(Lparens, place[0]) != (char *) 0)
			{
				(void) expandLabels(place, dimi, place1);
				place = skipStrings(place, 1);
				place1 = skipStrings(place1, dimi);
			}
			else
			{
				for (j = 0; j < dimi; j++)
				{
					place1 = copyStrings(place, place1, 1);
					place = skipStrings(place, 1);
				}
			}
		} /*for (i = 1; i <= ndims; i++)*/
		mydisphandle(labelsH);
	} /*if (mlabels < nlabels)*/		
	setSTRING(labelsFromFile, (char **) 0); /* detach from labelsFromFile */
	
	return (1);
} /*attachLabelsFromFile()*/


/*
  970227 added new argument endMacro to calls to readHeader() and readMacro()
  970302 added capability to read structures and NULL symbols.
  970619 added capability to distinguish between macros to be expanded
         in-line and out-of-line
  990220 Added argument dollars to readHeader to pass the presence or
         absense of keyword DOLLARS on a macro
  Note: matNameH is either a real MacAnova handle, if user provided
        a name, or is an pointer to a pointer to an array of length
		MAXNAME + 1, if the user did not provide a name, or if
		doReads() is called from itself to read a structure component
		in which case *matNameH should end with '$'
  991029 Default for 'quiet' is now F on inforead().
  991123 Labels read from file are now attached by new function
         attachLabelsFromFile()
  991216 attachLabelsFromFile() now expands labels starting with @ or of
         the form (, [, {, /, <, or \
  010325 Fixed minor bug, thus re-enabling error message when macroread()
         finds non-macro or matread() finds a macro
*/

static Symbolhandle doReads(FILE *fp, char **matNameH, char **fnameH,
							long op, long verbose, long echo, long silent,
							long notFoundOK,
							Symbolhandle symhLabels)
{
	Symbolhandle     result = SymbolZero, result1;
	Symbolhandle     labelsOrNotesFromFile = SymbolZero;
	char             format[MAXFMT + 1];
	char             endMacro[ENDMACROLENGTH + 1];
	char            *matName = *matNameH;
	char            *fmt = format, **string = (char **) 0;
	double          *y;
	double           yi, missValue;
	long             nFields, nSkip;
	int              byRows, quoted, inLine = -1, locked;
	int              labels, notes, dollars;
	long             i, size, length = 0;
	long             ndims, dims[MAXDIMS];
	long             type;
	long             reply, labelError;
	long             printHeader = verbose;
	long             dataType;
	int              nameLength = strlen(matName);
	short            badLogical = 0, foundInfinite = 0, checkMissing;
	int              nTries;
	long             status = 1;
	unsigned long    report;
	WHERE("doReads");
	
	*OUTSTR = '\0';
	setMissing(missValue);
	endMacro[0] = '\0';

	/*
		When the user did not specify a name or when reading a structure
		component, the actual name of the object read is copied to *matNameH
		matNameH in that case points to a buffer of length at least MAXNAME + 1
	*/

	reply = readHeader(fp, dims, &nFields, &fmt, &nSkip, &byRows, &type,
					   &quoted, &inLine, &locked, &labels, &notes, &dollars,
					   &missValue, *matNameH, endMacro,
					   (!(op & IINFOREAD)) ? printHeader : -1);

	if (notFoundOK && reply == NOTFOUND)
	{
		goto returnNull;
	}
	
	if (reply != READNOERROR)
	{
		if (isEOL(**matNameH))
		{
			matName = "first data set";
			matNameH = &matName;
		}
		if (op != ICOMPREAD ||
			!silent && reply != NOTFOUND && reply != COMPONENTNOTFOUND)
		{
			headerError(reply, *matNameH, *fnameH, op == ICOMPREAD);
		}
		goto errorExit;
	} /*if (reply != READNOERROR)*/

	dataType = type == REAL || type == LOGIC || type == CHAR;

	if (symhLabels != SymbolZero)
	{		
		if (!dataType)
		{
			symhLabels = SymbolZero;
			if (!silent)
			{
				sprintf(OUTSTR,
						"WARNING: labels ignored when reading %s",
						(type == MACRO) ? "macro" :
						((type == STRUC) ? "structure" : "NULL symbol"));
				putErrorOUTSTR();
			}
		}
		labels = 0; /* ignore any labels on symbol */
	} /*if (symhLabels != SymbolZero)*/

	if (!(op & IINFOREAD))
	{
		checkMissing = !isMissing(missValue);
		
		if (type != CHAR && nFields > MAXN)
		{
			sprintf(OUTSTR,
					"%snumber of items specified by header is %ld > %ld",
					ErrorStart, nFields, (long) MAXN);
			goto errorExit;
		} /*if (nFields > MAXN)*/

		if (!silent && !(op & IMATREAD && op & IMACROREAD) &&
			(op & IMATREAD && type == MACRO ||
			 op & IMACROREAD && type != MACRO))
		{
			char        *what1 = "macro", *what2 = "data set";

			sprintf(OUTSTR, "WARNING: found %s instead of %s",
					(op & IMATREAD) ? what1 : what2,
					(op & IMATREAD) ? what2 : what1);
			putErrorOUTSTR();
		}
	
		ndims = dims[0];
		if (dataType)
		{
			if (ndims == 1)
			{
				dims[2] = 1;
			}
		
			if (dims[1] == 0)
			{
				if (!silent)
				{
					putOutErrorMsg("WARNING: 0 lines of data in data set");
				}
				goto returnNull;
			} /*if (dims[1] == 0)*/
		
			if (isEOL(*fmt) && type != CHAR)
			{
				fmt = DefaultFormat;
			}
		
			if (type == CHAR)
			{
				result = CInstall(SCRATCH, 0);
			}
			else
			{
				result = RInstall(SCRATCH, 0);
			}
			
			if (result == SymbolError)
			{
				goto errorExit;
			}
			setTYPE(result, type);
			setNDIMS(result, ndims);
			size = 1;
			for (i = 1; i <= ndims; i++)
			{
				size *= dims[i];
				setDIM(result, i, dims[i]);
			} /*for (i = 1; i <= ndims; i++)*/
			
			if (symhLabels != SymbolZero)
			{
				labelError = checkLabels(symhLabels, ndims, dims + 1);

				if (labelError != LABELSOK &&
					(!silent || labelError & LABELSERROR))
				{
					badLabels(labelError);
				}
				if (labelError & LABELSERROR ||
					!(labelError & WRONGSIZELABELS) &&
					!installLabels(symhLabels, result))
				{
					goto errorExit;
				}
			} /*if (symhLabels != SymbolZero)*/
			
	/* on non-error, returns number of items actually read, if <= expected */
			reply = readFile(fp, nFields, fmt, nSkip, byRows, quoted,
							 result, echo, !silent);
			if (reply < 0)
			{
				readError(-reply, *matNameH, *fnameH, op == ICOMPREAD);
				goto errorExit;
			}

			if (!silent && reply != size)
			{
				sprintf(OUTSTR,
						"WARNING: %s data set incomplete; %ld items set to %s",
						typeName(type), size - reply,
						(type != CHAR) ? "MISSING" : "\"\"");
				putErrorOUTSTR();
			} /*if (!silent && reply != size)*/

			if (type != CHAR)
			{
				y = DATAPTR(result);
				for (i = 0;i < reply;i++)
				{
					if (checkMissing && y[i] == missValue)
					{
						setMissing(y[i]);
					}
					yi = y[i];
					if (!isMissing(yi))
					{
						if (type == LOGIC)
						{
							if (yi != 0.0 && yi != 1.0)
							{
								badLogical = 1;
							}
							y[i] = (double) (yi != 0.0);
						} /*if (type == LOGIC)*/
						else if ((yi = fabs(yi),
								  doubleEqualBdouble(yi, TooBigValue)))
						{ /* outside legal range */
							setMissing(y[i]);
							foundInfinite = 1;
						}
					} /*if (!isMissing(yi))*/				
				} /*for (i = 0;i < reply;i++)*/
			} /*if (type != CHAR)*/
			else
			{
				if (byRows && dims[0] > 1)
				{ /* Fix dimensions and Transpose */
					for (i = 1; i <= dims[0]; i++)
					{
						setDIM(result, i, dims[dims[0] - i + 1]);
					} /*for (i = 1; i <= dims[0]; i++)*/					

					result1 = doTranspose(result, (double *) 0, 0, &report);

					if (result1 == SymbolError)
					{
						goto errorExit;
					}
					else if (result1 != result)
					{
						Removesymbol(result);
						result = result1;
					}
				} /*if (byRows)*/
			} /*if (type != CHAR){}else{}*/
		} /*if (dataType)*/
		else if (type == MACRO)
		{
			char       *zeroLines = "WARNING: 0 lines in macro on file";

			labels = 0;
			if (dims[1] == 0)
			{
				if (!silent)
				{
					putOutErrorMsg(zeroLines);
				}
				goto returnNull;
			} /*if (dims[1] == 0)*/

			result = Install(SCRATCH, MACRO);
			if (result == SymbolError)
			{
				goto errorExit;
			}

			setScratchMacroName(result, *matNameH);
			
			if ((reply = readMacro(fp, dims[1], endMacro, result,
								   echo, !silent)) < 0)
			{
				goto errorExit;
			}
			if (reply == 0)
			{
				if (!silent)
				{
					putOutErrorMsg(zeroLines);
				}
				Removesymbol(result);
				goto returnNull;
			}
			setInline(result, inLine);
			setNDIMS(result, 1);
			setDIM(result, 1, 1);
			length = prepMacro(STRINGPTR(result), (char *) 0, dollars);
			if (length != strlen(STRINGPTR(result)) + 1)
			{
				TMPHANDLE = mygethandle(length);
				if (TMPHANDLE == (char **) 0)
				{
					goto errorExit;
				}
				
				(void) prepMacro(STRINGPTR(result), *TMPHANDLE, dollars);
				mydisphandle(STRING(result));
				setSTRING(result, TMPHANDLE);					
			} /*if (length != strlen(STRINGPTR(result)) + 1)*/			
		}
		else if (type == STRUC)
		{
			long           ncomps = dims[1], icomp;
			char           compName[MAXNAME + 1];
			char           targetName[MAXNAME + 1];
			char          *compNameP = compName;
			char         **compNameH = &compNameP;
			Symbolhandle   symhComp;
			
			result = StrucInstall(SCRATCH, ncomps);
			
			if (result == SymbolError)
			{
				goto errorExit;
			}

			setScratchName(result);
			matName = *matNameH;
			nameLength = strlen(matName);
			strcpy(targetName, matName);
			targetName[nameLength++] = COMPNAMESTART;
			targetName[nameLength] = '\0';

			for (icomp = 0; icomp < ncomps; icomp++)
			{
				strcpy(compName, targetName);
				symhComp = doReads(fp, compNameH, fnameH, ICOMPREAD,
								   verbose, echo, silent, 0, SymbolZero);
				if (symhComp == SymbolError)
				{
					sprintf(OUTSTR,
							"%s%s() can't find component %ld of structure %s",
							ErrorStart, FUNCNAME, icomp + 1, *matNameH);
					putPieces(OUTSTR, " on file ", *fnameH, (char *) 0);
					*OUTSTR = '\0';
					goto errorExit;
				}
				Cutsymbol(symhComp);
				COMPVALUE(result, icomp) = symhComp;
				strcpy(compName, strrchr(compName, COMPNAMESTART) + 1);
				compName[NAMELENGTH] = '\0';
				setNAME(symhComp, compName);
			} /*for (icomp = 0; icomp < ncomps; icomp++)*/
		}
		else if (type == NULLSYM)
		{
			goto returnNull;
		}
		else
		{
			sprintf(OUTSTR,
					"%s%s() does not know how to read type %s object",
					ErrorStart, FUNCNAME, typeName(type));
			goto errorExit;
		}

		nTries = labels + notes;
		for (i = 0; i < nTries && (labels || notes); i++)
		{
			char            labelsOrNotesName[MAXNAME + 1];
			char           *labelsOrNotesNameP = labelsOrNotesName;
			char          **labelsOrNotesNameH = &labelsOrNotesNameP;
			
			matName = *matNameH;
			nameLength = strlen(matName);

			strcpy(labelsOrNotesName, matName);
			labelsOrNotesName[nameLength++] = COMPNAMESTART;
			labelsOrNotesName[nameLength] = '\0';

			labelsOrNotesFromFile = doReads(fp, labelsOrNotesNameH,
											fnameH, ICOMPREAD,
											echo, echo, silent, 0, SymbolZero);

			if (labelsOrNotesFromFile == SymbolZero ||
				TYPE(labelsOrNotesFromFile) != CHAR ||
				NDIMS(labelsOrNotesFromFile) != 1)
			{
				break;
			}
				
			
			if (labels && mystrncmp(labelsOrNotesName + nameLength,
									LABELSSCRATCH+2,
									strlen(LABELSSCRATCH+2)) == 0)
			{
				int      reply = attachLabelsFromFile	(result,
													  labelsOrNotesFromFile);
				setSTRING(labelsOrNotesFromFile, (char **) 0);
				labels = 0;
				if (reply < 0)
				{
					goto errorExit;
				}
				if (reply == 0)
				{
					break;
				}
			}
			else if (notes && mystrncmp(labelsOrNotesName + nameLength,
										NOTESSCRATCH+2,
							  			strlen(NOTESSCRATCH+2)) == 0)
			{
				if (!setNotes(result, STRING(labelsOrNotesFromFile)))
				{
					goto errorExit;
				}
				setSTRING(labelsOrNotesFromFile, (char **) 0);
				notes = 0;
			}
			else 
			{
				break;
			}
			Removesymbol(labelsOrNotesFromFile);
			labelsOrNotesFromFile = SymbolZero;
		} /*for (i = 0; i < nTries && (labels || notes); i++)*/
		
		if (!silent)
		{
			char      *onFile = " on file ";
			
			if (labels)
			{
				sprintf(OUTSTR,
						"WARNING: %s() can't find labels for '%s'",
						FUNCNAME, *matNameH);
				putPieces(OUTSTR, onFile, *fnameH, (char *) 0);
			}
			if (notes)
			{
				sprintf(OUTSTR,
						"WARNING: %s() can't find notes for '%s'",
						FUNCNAME, *matNameH);
				putPieces(OUTSTR, onFile, *fnameH, (char *) 0);
			}
			if (badLogical)
			{
				putOutErrorMsg("WARNING: non 0 or 1 value(s) found in LOGICAL variable read; set to T");
			}
			else if (foundInfinite)
			{
				putOutErrorMsg("WARNING: numbers outside representable range replaced by MISSING");
			}
			*OUTSTR = '\0';
		} /*if (!silent)*/		
		if (locked)
		{
			Setname(result, LOCKEDSCRATCH);
		}
	} /*if (!(op & IINFOREAD))*/
	else
	{ /* inforead()*/
		if (verbose)
		{ /* print 1st line (name and dimensions) */
			(*LINE)[strlen(*LINE)-1] = '\0'; /* trim off '\n' */
			/* use putOutMsg to make sure line is counted */
			putOutMsg(*LINE);
		} /*if (verbose)*/

		result = CInstall(SCRATCH, 0);
		if (result == SymbolError)
		{
			goto errorExit;
		}
		setDIM(result, 1, 1);
		while (status != EOF &&
			  (status = fillLINE(fp), (*LINE)[0] == COMMENTSTART))
		{
			if ((*LINE)[1] != '"')
			{ /* ignore lines starting ')"' */
				size = strlen(*LINE + 1);
				if (string == (char **) 0)
				{
					setSTRING(result, string = mygethandle(size + 1));
				}
				else
				{
					setSTRING(result,
							  string = mygrowhandle(STRING(result),
													length + size + 1));
				}
				if (string == (char **) 0)
				{
					goto errorExit;
				}
				strcpy(*string + length, *LINE + 1);
				length += size;
				if (verbose > 1 || verbose == 1 && (*LINE)[1] != COMMENTSTART)
				{
					(*LINE)[size] = '\0'; /* trim off '\n' */
					putOutMsg(*LINE);
				}
			} /*if ((*LINE)[1] != '"')*/
		} /*while (...)*/
		if (length > 0 && (*string)[length-1] == '\n')
		{
			length--;
			(*string)[length] = '\0';
			setSTRING(result, string = mygrowhandle(string, length + 1));
			if (string == (char **) 0)
			{
				goto errorExit;
			}
		} /*if (length > 0 && (*string)[length-1] == '\n')*/

		if (string == (char **) 0)
		{ /* no comment lines found */
			setSTRING(result, string = mygethandle(1));
			if (string == (char **) 0)
			{
				goto errorExit;
			}
			(*string)[0] = '\0';
		} /*if (string == (char **) 0)*/
	} /*if (!(op & IINFOREAD)){}else{}*/
	
	return (result);

  returnNull:
	result = Install(NULLSCRATCH, NULLSYM);
	return (result);
	
  errorExit:
	putErrorOUTSTR();

	Removesymbol(labelsOrNotesFromFile);
	Removesymbol(result);

	return (SymbolError);
} /*doReads()*/


/*
   Functions vecread(), matread(), macroread(), inforead()

  Routines to read data and macros in from file.
  'Read' has been replaced by synonym 'vecread'

  Usage:
	vecread(fileName)
	matread(fileName [,setName], [,quiet:T])
	macroread(fileName [,setName], [,quiet:T])
	inforead(fileName [,setName], [,quiet:T])

  where fileName and setName must be CHARACTER scalars.

  inforead() returns as a CHARACTER scalar the comment lines describing
  dataset or macro name, stripping off the leading ')' (COMMENTSTART)

  As usual on the Macintosh, fileName may be "".

  NOTE:  As of 931111, it is no longer legal on the Macintosh for fileName
  to be missing.

  950119:  If fileName is of the form string:charVec, a read will
  be simulated from the character vector, with each element starting on
  a new line.  Thus, for example, vecread(string:"1 2 3 4") is equivalent
  to run(4).

  960207 vecread(fileName, character:T) returns CHARACTER vector containing
  whitespace- or comma-separated fields read.  vecread(fileName,badvalue:val)
  returns numbers read interspersed with val for every unreadable item
  found.

  960304 On vecread() of reals, single commas are ignored, a string of k > 1
  commas will result in k - 1 bad character possibilities.

  960304 "?", "??", "???", ... are all equivalent; successive commas found
  by vecread(fileName,character:T) are interpreted as delimiting empty
  strings (""), as does a trailing comma

  960502 changed fopen() to fmyopen()
*/

#ifdef MACINTOSH
#define printableChar(C) (isgraph(C) || (C) > DEL)
#else /*MACINTOSH*/
#define printableChar(C) isgraph(C)
#endif /*MACINTOSH*/

#undef isascii
#define isascii(C) ((unsigned char) (C) <= 0x7f)

#define ISPUNCT(C) (ispunct(C) && (C) != '+' && (C) != '-' &&\
					(C) != '.' && (C) != '?' && (C) != ',')

#define okSkip(C) ((charData) ? printableChar(C) : ISPUNCT(C))

#define okStop(C) ((charData) ? (ispunct(C) && (C) != ',') : ISPUNCT(C))


enum consoleCodes
{
	nonConsole = 0,
	interactiveConsole,
	batchConsole
};

/*
  New 000523 to implement getfilename()
    usage:
      fullpath <- getfilename([type:"text" or type:"workspace"])
      filename <- getfilename(nameonly:T [type:value])
      pathname <- getfilename(pathonly:T [type:value])
      helpfullpath <- getfilename(help:T)
      helpfilename <- getfilename(help:T,nameonly:T)
      helppathname <- getfilename(help:T,pathonly:T)
	  lastfullpath <- getfilename(last:T)
	  lastfilename <- getefilename(last:T, nameonly:T)
	  lastpathname <- getfilename(last:T,pathonly:T)

   010525 Changed macFindFile() to mvFindFile() and dropped wxFindFile()
   010604 changed call to splitPathName() to provide buffer length
          increased size of buffer
   010616 Added keyword phrase help:T
   021213 Added keyword phrase cancelOK:T, meaning NULL is returned
          on cancellation to allow testing in a macro
   021218 Keyword processing using keyword-related macros
*/

#ifndef FULLPATHSIZE
#define FULLPATHSIZE 1024
#endif /*FULLPATHSIZE*/

enum getFileNameKeyCodes
{
	KTYPE     = 0,
	KNAMEONLY,
	KPATHONLY,
	KHELP,
	KLAST,
	KCANCELOK
};

/*
  Definition and typing of legal keywords for getfilename()
  Order must match getFileNameKeyCodes
*/
static keywordList    GetFileNameKeys[] =
{
	InitKeyEntry("type",     0, 1, CHARSCALAR), 
	InitKeyEntry("nameonly", 0, 1, LOGICSCALAR),
	InitKeyEntry("pathonly", 0, 1, LOGICSCALAR),
	InitKeyEntry("help",     0, 1, LOGICSCALAR),
	InitKeyEntry("last",     0, 1, LOGICSCALAR),
	InitKeyEntry("cancelok", 0, 1, LOGICSCALAR) 
};

/* Values in array GetFileNameKeys can be referred to by these macros*/
#define         FileType  KeyCharValue(GetFileNameKeys,KTYPE)
#define         NameOnly  KeyLogValue(GetFileNameKeys,KNAMEONLY)
#define         PathOnly  KeyLogValue(GetFileNameKeys,KPATHONLY)
#define         HelpFile  KeyLogValue(GetFileNameKeys,KHELP)
#define         LastFile  KeyLogValue(GetFileNameKeys,KLAST)
#define         CancelOK  KeyLogValue(GetFileNameKeys,KCANCELOK)

static Symbolhandle askForFileName(Symbolhandle list)
{
	Symbolhandle        result;
	char               *fileName = NullString;
	char                pathName[FULLPATHSIZE + 1];
	int                 workspaceFile = -1;
	int                 keyStatus;
	long                nargs = NARGS(list);
	char               *notTogether =
		"%s'%s:T' and '%s:T' can't be used together on %s()";
	WHERE("askForFileName");

	NameOnly = PathOnly = CancelOK = LastFile = HelpFile = 0;
	FileType = (char **) 0;

	if (nargs == 1 && COMPVALUE(list,0) == EmptyArgument)
	{
		nargs = 0;
	}
	if (nargs > 3)
	{
		badNargs(FUNCNAME, -3);
		goto errorExit;
	}

	/* pick off keyword values */
	keyStatus = getAllKeyValues(list, 0, 1, GetFileNameKeys, 
								NKeys(GetFileNameKeys));
	if (keyStatus < 0)
	{
		goto errorExit;
	}
	nargs -= keyStatus;
	if (PathOnly && NameOnly)
	{
		sprintf(OUTSTR, notTogether, ErrorStart, "nameonly", "pathonly",
				FUNCNAME);
		goto errorExit;
	}

	if (LastFile && HelpFile)
	{
		sprintf(OUTSTR, notTogether, ErrorStart, "last", "help", FUNCNAME);
		goto errorExit;
	}

	if (FileType != (char **) 0)
	{
		if (mystrncmp(*FileType, "text",0) != 0 &&
			mystrncmp(*FileType, "workspace", 4) != 0)
		{
			sprintf(OUTSTR,
					"%svalue for %s must be \"text\" or \"workspace\" in %s()",
					ErrorStart, KeyName(GetFileNameKeys, KTYPE), FUNCNAME);
			goto errorExit;
		}
		workspaceFile = tolower((*FileType)[0]) == 'w';
	} /*if (FileType != (char **) 0)*/
	
	if (nargs > 0)
	{
		/* getAllKeyValues() didn't use all arguments */
		char               *keyword = isKeyword(COMPVALUE(list, 0));
		
		if (keyword == (char *) 0)
		{
			sprintf(OUTSTR,
					"%sall arguments to %s() must be keyword phrases",
					ErrorStart, FUNCNAME);
		}
		else
		{
			badKeyword(FUNCNAME, keyword);
		}
		goto errorExit;
	} /*if (nargs > 0)*/

	if (!LastFile && !HelpFile)
	{
#ifndef HASFINDFILE
		sprintf(OUTSTR,
				"%s%s() without 'last:T' or 'help:T' illegal in non-windowed versions",
				ErrorStart, FUNCNAME);
		goto errorExit;
#else
		workspaceFile = (workspaceFile < 0) ? 1 : workspaceFile;
	
		if (!workspaceFile)
		{
			fileName = mvFindFile(fileName, "Specify the data file", NullString,
								  mvReadit | mvTextFile | mvROTextFile,
								  getMvDirInfo(readVol));
		} /*if (!workspaceFile)*/
		else
		{
			fileName = mvFindFile(fileName, "Specify the restore file",
								  NullString, mvReadit | mvTextFile |
								  mvROTextFile | mvAsciiSaveFile | mvSaveFile,
								  getMvDirInfo(restoreVol));
		} /*if (!workspaceFile){}else{}*/

		if (fileName == (char *) 0)
		{
			if (!CancelOK)
			{
				sprintf(OUTSTR, "%s%s() cancelled", ErrorStart, FUNCNAME);
				goto errorExit;
			}
			return (NULLSYMBOL);
		}

#ifdef MACINTOSH
		/*
		  fileName is already just the file name;  if needed, the path name
		  is retrieved separately
		*/
		if (!NameOnly)
		{
			int      errorFlag = getMacPathName(pathName);

			if (errorFlag == 0)
			{
				if (!PathOnly)
				{
					if (strlen(pathName) + strlen(fileName) > FULLPATHSIZE)
					{
						strncpy(pathName, fileName, FULLPATHSIZE);
					}
					else
					{
						strcat(pathName, fileName);
					}
				}
			}
			else
			{
				sprintf(OUTSTR,
						"%serror in getting path name; error code = %d",
						ErrorStart, errorFlag);
				goto errorExit;
			}
		} /*if (!NameOnly)*/
		else
		{
			strncpy(pathName, fileName, FULLPATHSIZE);
		}
#else /*MACINTOSH*/
		/*
		  On Windows and in Motif, fileName is the complete path name
		*/
		if (NameOnly || PathOnly)
		{
			/* put name or directory name in pathName */
			splitPathName(NameOnly, fileName, pathName, FULLPATHSIZE + 1);
		} /*if (NameOnly || PathOnly)*/
		else
		{
			strncpy(pathName, fileName, FULLPATHSIZE);
		}
#endif /*MACINTOSH*/
#endif /*HASFINDFILE*/
	} /*if (!LastFile && !HelpFile)*/
	else if (LastFile)
	{
		if (LastPathOpened[0] == '\0')
		{
			sprintf(OUTSTR,
					"%slast file name read is unavailable", ErrorStart);
			goto errorExit;
		}
		if (NameOnly || PathOnly)
		{
			splitPathName(NameOnly, LastPathOpened, pathName, FULLPATHSIZE + 1);
		}
		else
		{
			strncpy(pathName, LastPathOpened, FULLPATHSIZE);
		}
	}
	else
	{
		char       *helpFileName;
		if (HelpFileName != (char **) 0)
		{
			helpFileName = *HelpFileName;
		}
		else if (HELPFILENAME != (char **) 0)
		{
			helpFileName = *HELPFILENAME;
		}
		else
		{
			sprintf(OUTSTR,
					"%shelp file name is unavailable", ErrorStart);
			goto errorExit;
		}
		if (NameOnly || PathOnly)
		{
			splitPathName(NameOnly, helpFileName, pathName, FULLPATHSIZE + 1);
		}
		else
		{
			strncpy(pathName, helpFileName, FULLPATHSIZE);
		}
	}
	pathName[FULLPATHSIZE] = '\0';

	result = CInstall(SCRATCH, strlen(pathName) + 1);
	if (result == SymbolError)
	{
		goto errorExit;
	}
	strcpy(STRINGPTR(result), pathName);
	
	return (result);

  errorExit:
	putErrorOUTSTR();
	return (SymbolError);
} /*askForFileName()*/


static Symbolhandle askForDirName(void)
{
	Symbolhandle        result;
	char               *dirName = NullString;
	WHERE("askForFileName");

#ifndef HASFINDFILE
	sprintf(OUTSTR,"getdirname() only implemented in Carapace version");
		goto errorExit;
#endif

	dirName = mvFindDir("Specify the directory",LastPathOpened);

	if (dirName == (char *) 0)
	{
		return (NULLSYMBOL);
	}
	result = CInstall(SCRATCH, strlen(dirName) + 1);
	if (result == SymbolError)
	{
		goto errorExit;
	}
	strcpy(STRINGPTR(result), dirName);
	
	return (result);

  errorExit:
	putErrorOUTSTR();
	return (SymbolError);
} /*askForFileName()*/


/*
   General entry for vecread(), matread(), macroread(), and inforead()

   960503 changed macOpen() to macFindFile()

   970210 allowed keywords file and string anywhere.  If they are used, then,
   for {mat,macro,info}read(), data set or macro name, if present, may also
   be in any position.

   971030 on commands other than vecread(), quiet:F forces printing of all
   headerlines;  otherwise header lines starting '))' are not printed

   971104 added check on value of myhandlelength()

   980501 added keyword phrase prompt:F to suppress prompts when reading
          from CONSOLE
   980507 added keyword phrase go:G; vecread() will stop reading lines
          on the first line that does not start with G; cannot be used with
          stop:S
   981116 Added new function name "macrosearch";  this is not in the symbol
          table but is used by searchForMacro() in yylex.c to use
          readdata to search and load undefined macros found in a
          command line
   981120 New keyword phrase nofileok:T causes a failure to open a
          file to not be an error, returning NULL.
   991222 New keyword phrase n:N on vecread() limits number of items read;
          a WARNING message is printed if fewer items are read
   000115 New keyword phrase oldstyle:T or oldstyle:F controls how
          vecread() treats commas and unreadable items when reading
          REAL data.  Initial default is T; will probably be F
   000316 New keyword phrase badkeyok:T on all commands, instructs command
          to ignore duplicate or unrecognized keywords.  This is can be
          useful in a macro allowing, say, vecread(string:@line,byword:T,$K)
          without having to worry about whether string or byword are
          keywords in the macro argument list.
   000323 vecread(filename, bywords:T) now returns leading comma as ""
   000614 Moved setting of ByWords, ByChars and ByLines to early in
          the program and made 'character' and 'bywords' explicit synonyms
          "#" is now the default for 'skip'.  skip:"" is now legal
          and means no skip character

   010305 Recognize TF keyword 'printname', default T
   010324 Modified parsing of argument list.  All keyword arguments are
          moved after nonkeyword arguments; uses new utilities countArgs()
		  and moveKeysToEnd()
   010516 Use atLeastNArgs constant on badNargs() call
   010524 Changed calling sequence of macFindFile()
   010525 Changed macFindFile() to mvFindFile() and dropped wxFindFile()
*/
Symbolhandle    readdata(Symbolhandle list)
{
	char           *fname, *whatRead;
	char           *quote = "\"";
	char            setname[MAXNAME + 1], *setnameP = setname;
	char          **fnameH, **setnameH;
	FILE           *fn = (FILE *) 0;
	Symbolhandle    arg1, arg2 = SymbolZero, result = SymbolZero;
	long            argCounts[nArgCountCodes];
	long            nargs;
	long            useConsole = nonConsole;
	long            keyStatus;
	long            nkeys = NKeys(ReaddataKeys);
	long            nonKeys, unusedKeys;
	keywordListPtr  keys = ReaddataKeys;
	long            i;
	int             setPlace;
	unsigned long   labelError;
	char           *keyword, *bywordsKeyname;
	/* 000614 default for 'skip' is now '#'*/
	unsigned char   skipChar = '#';
	unsigned char   stopChar = '\0', goChar = '\0';
	unsigned char   skipthruChar = '\0';
	char           *notSinglechar =
	  "ERROR: value for %s must be single character";
	char           *notPunctuation =
	  "ERROR: value for %s not punctuation character or is '+', '-', ',', or '.'";
	char           *badCharStop =
	  "ERROR: value for '%s' is \",\" or not a punctuation character";
	char           *badCharSkip =
	  "ERROR: value for '%s' is a non-printable character";
	long            op;
	long            lineChunk = (MAXLINE < LINECHUNK) ? MAXLINE : LINECHUNK;
	int             verbose, charData = 0;
	WHERE("readdata");
	
	if (strcmp(FUNCNAME, "getfilename") == 0)
	{
		return (askForFileName(list));
	}
	if (strcmp(FUNCNAME, "getdirname") == 0)
	{
		return (askForDirName());
	}
	setname[0] = OUTSTR[0] = '\0';
	
	freeLINE();
	StringToRead = (unsigned char **) 0;
	StringLength = PlaceInString = 0;
	nargs = countArgs(list, argCounts);

	if (nargs == 0)
	{
		badNargs(FUNCNAME, atLeast1Args);
		goto errorExit;
	}

	nonKeys = argCounts[nonKeywordArgCount];
	unusedKeys = argCounts[keywordArgCount];
	
	for (i = 0; Opnames[i] != (char *) 0;i++)
	{
		if (strcmp(FUNCNAME, Opnames[i]) == 0)
		{
			op = Opcodes[i];
			break;
		}
	} /*for (i = 0; Opnames[i] != (char *) 0;i++)*/

	unsetKeyValues(keys, nkeys);
	/*
	    010324
		Make sure any non-keyword arguments are at start; any data
		set or macro name should then be argument 1 (with keywords
		'file' or 'string') 2 (without 'file' or 'string')
	*/
	moveKeysToEnd(list);

	/*
	  970210 "file" and "string" now in keyword list and can appear
	  anywhere in keywords.

	  getAllKeyValues() moves used keywords to end so any unused
	  keyword arguments will immediately follow non-keyword arguments
	*/

	if (op != IMACROSEARCH)
	{
		keyStatus = getAllKeyValues(list, 0, op, keys, nkeys);
		if (keyStatus < 0)
		{
			goto errorExit;
		}
		unusedKeys -= keyStatus;
		if (charIsSet(File) && symhIsSet(String))
		{
			sprintf(OUTSTR,
					"%sillegal to use both '%s' and '%s' in %s()",
					ErrorStart,KeyName(keys,KFILE),KeyName(keys,KSTRING),
					FUNCNAME);
			goto errorExit;
		}

		/*
		  Check keywords that specify CHARACTER data will be read
		  'character' and 'bywords' are complete synonyms
		*/
		if (logicalIsSet(CharRead) && logicalIsSet(ByWords))
		{
			if (CharRead && !ByWords || !CharRead && ByWords)
			{
				sprintf(OUTSTR,
						"ERROR: %s and %s on %s() must both be True or both be False",
						KeyName(keys,KCHAR), KeyName(keys,KBYWORDS), FUNCNAME);
				goto errorExit;
			}
			bywordsKeyname = KeyName(keys,KBYWORDS);
		}
		else if (logicalIsSet(CharRead))
		{
			ByWords = CharRead;
			bywordsKeyname = KeyName(keys, KCHAR);
		}
		else
		{
			bywordsKeyname = KeyName(keys,KBYWORDS);
		}

		/* set the default values for 'bylines', 'bywords' and 'bychar'*/
		if (!logicalIsSet(ByLines))
		{
			setLogicalFalse(ByLines);
		}
		if (!logicalIsSet(ByChars))
		{
			setLogicalFalse(ByChars);
		}
		if (!logicalIsSet(ByWords))
		{
			setLogicalFalse(ByWords);
		}

		charData = ByLines || ByChars || ByWords;
		
		if (charData)
		{
			char       *illegal = "ERROR: %s:T is illegal with %s:T";
			
			if (ByWords && (ByChars || ByLines))
			{
				sprintf(OUTSTR, illegal, bywordsKeyname,
						(ByChars) ?
						KeyName(keys,KBYCHARS) : KeyName(keys,KBYLINES));
			}
			else if (ByChars && ByLines)
			{
				sprintf(OUTSTR, illegal, KeyName(keys, KBYLINES),
						KeyName(keys,KBYCHARS));
			}
			if (*OUTSTR)
			{
				goto errorExit;
			}
		}
	} /*if (op != IMACROSEARCH)*/
	else
	{
		/*
		  call from yylex implicitly has
            notfoundok:T, nofileok:T, echo:F, silent:T, badkeyok:T
		*/
		setLogicalTrue(NotFoundOK);
		setLogicalTrue(NoFileOK);
		setLogicalTrue(BadKeyOK);
		setLogicalFalse(Echo);
		setLogicalTrue(Silent);

		/*play it safe*/
		setLogicalFalse(ByLines);
		setLogicalFalse(ByChars);
		setLogicalFalse(ByWords);
		keyStatus = 0;
	} /*if (op != IMACROSEARCH){}else{}*/

	
	if (symhIsSet(String))
	{
		StringToRead = (unsigned char **) STRING(String);
		StringLength = myhandlelength((char **) StringToRead);
		if (StringLength < 0)
		{
			goto errorExit;
		}
		setPlace = (nonKeys > 0) ? 0 : -1;
	}
	else if (charIsSet(File))
	{
#ifdef CONSOLEOK
	/*
	   useConsole == 0 if read not from CONSOLE
	   useConsole == 1 if read from CONSOLE which is interactive input
	   useConsole == 2 if read from CONSOLE while reading batch file or
	                   noninteractive (redirected) input
	*/
		if (ISCONSOLE(*File))
		{
			useConsole = (!(ISATTY & ITTYIN) || BDEPTH > 0) ?
			  batchConsole : interactiveConsole;
			fname = CONSOLENAME;
			fnameH = &fname;
		}
		else
#endif /*CONSOLEOK*/
		{
			fnameH = File;
		}
		setPlace = (nonKeys > 0) ? 0 : -1;
	}
	else
	{ /* neither 'file' or 'string' used; 1st arg must be file name*/
		/* N.B. arg1 may not have originally been argument 1*/
		arg1 = COMPVALUE(list, 0);
		if (nonKeys == 0 || arg1 == EmptyArgument)
		{
			sprintf(OUTSTR,
					"%syou must provide a file name for %s() without '%s'",
					ErrorStart, FUNCNAME, KeyName(keys,KSTRING));
			goto errorExit;
		}
#ifdef CONSOLEOK
		if ((ISCONSOLE(NAME(arg1)) || 
			TYPE(arg1) == CHAR && isScalar(arg1) && ISCONSOLE(STRINGPTR(arg1))))
		{
			useConsole = (!(ISATTY & ITTYIN) || BDEPTH > 0) ? 
			  batchConsole : interactiveConsole;
			fname = CONSOLENAME;
			fnameH = &fname;
		}
		else
#endif /*CONSOLEOK*/
		{
			if (!isCharOrString(arg1))
			{
				notCharOrString("file name");
				goto errorExit;
			}
			fnameH = STRING(arg1);
		}
		nonKeys--;
		setPlace = (nonKeys > 0) ? 1 : -1;
	}
	/*
	  setPlace is now set to current position in list where any data set or
	  macro name is expected or to -1 if nonKeys == 0
	*/
	
	if (!logicalIsSet(NotFoundOK))
	{
		setLogicalFalse(NotFoundOK);
	} /*if (!logicalIsSet(NotFoundOK))*/
	
	if (!logicalIsSet(NoFileOK))
	{
		setLogicalFalse(NoFileOK);
	} /*if (!logicalIsSet(NoFileOK))*/
	
	if (!logicalIsSet(BadKeyOK))
	{
		setLogicalFalse(BadKeyOK);
	} /*if (!logicalIsSet(BadKeyOK))*/
	
	if (!logicalIsSet(Silent))
	{
		if (!NotFoundOK)
		{
			setLogicalFalse(Silent);
		}
		else
		{
			setLogicalTrue(Silent);
		}
	} /*if (!logicalIsSet(Silent))*/
	
	if (!logicalIsSet(Printname))
	{
		if (FromString || Silent)
		{
			setLogicalFalse(Printname);
		}
		else
		{
			setLogicalTrue(Printname);
		}
	} /*if (!logicalIsSet(Printname))*/
	
	if (!logicalIsSet(Prompt))
	{
		if (op != IMACROSEARCH)
		{
			setLogicalTrue(Prompt);
		}
		else
		{
			setLogicalFalse(Prompt);
		}
	} /*if (!logicalIsSet(Prompt))*/
	else if (useConsole == nonConsole)
	{
		if (!Silent)
		{
			sprintf(OUTSTR,
					"WARNING: keyword %s ignored when not reading from CONSOLE",
					KeyName(keys,KPROMPT));
			putErrorOUTSTR();
		}		
		setLogicalTrue(Prompt);
	}

	if (charIsSet(Stop))
	{
		keyword = KeyName(keys,KSTOP);
		if (strlen(*Stop) != 1)
		{
			sprintf(OUTSTR, notSinglechar, keyword);
			goto errorExit;
		}
		stopChar = (*Stop)[0];
	} /*if (charIsSet(Stop))*/

	if (charIsSet(Go))
	{
		keyword = KeyName(keys,KGO);
		if (strlen(*Go) != 1)
		{
			sprintf(OUTSTR, notSinglechar, keyword);
			goto errorExit;
		}
		goChar = (*Go)[0];
	} /*if (charIsSet(Go))*/

	if (charIsSet(Skip))
	{
		keyword = KeyName(keys,KSKIP);
		if (strlen(*Skip) > 1)
		{
			sprintf(OUTSTR, notSinglechar, keyword);
			goto errorExit;
		}
		skipChar = (*Skip)[0];
	} /*if (charIsSet(Skip))*/

	if (charIsSet(Skipthru))
	{
		keyword = KeyName(keys,KSKIPTHRU);
		if (strlen(*Skipthru) != 1)
		{
			sprintf(OUTSTR, notSinglechar, keyword);
			goto errorExit;
		}
		skipthruChar = (*Skipthru)[0];
	} /*if (charIsSet(Skipthru))*/

	if (symhIsSet(Labels))
	{
		labelError = checkLabels(Labels, 0, (long *) 0);
		if (labelError != LABELSOK)
		{
			badLabels(labelError);
			goto errorExit;
		}
	} /*if (symhIsSet(Labels))*/
	else
	{
		Labels = SymbolZero;
	}
	
	/*
	  At this point,
	    nonKeys = number of non processed non-keyword arguments
		unusedKeys = number of unused keyword arguments, which may
		   include unrecognized or duplicate keywords
		setPlace < 0 if nonKeys == 0
		setPlace = 0 if nonKeys > 0 and 'file' or 'string' used
		setPlace = 1 if nonKeys > 0 and 'file' and 'string' not used
	*/
	if (op & IVECREAD)
	{
		/* Check values of special vecread() keywords */
		if (stopChar != '\0')
		{
			if (goChar != '\0')
			{
				sprintf(OUTSTR,
						"ERROR: keywords '%s' and '%s' are incompatible on %s()",
						KeyName(keys,KSTOP), KeyName(keys,KGO), FUNCNAME);
				goto errorExit;
			}
			
			/*
			  When not reading from CONSOLE, stopping character can
			  be any non-ascii character, such as "\377", usually
			  selected to ensure the whole file will be read
			*/
			if (!okStop(stopChar) &&
				(useConsole != nonConsole || isascii(stopChar)))
			{
				sprintf(OUTSTR, (charData) ? badCharStop :notPunctuation,
						KeyName(keys, KSTOP));
				goto errorExit;
			}
		} /*if (stopChar != '\0')*/
		else if (goChar == '\0')
		{
			stopChar = '!';
		}
		
		if (skipChar != '\0')
		{
			if (skipChar == goChar)
			{
				sprintf(OUTSTR,
						"ERROR: both '%s' and '%s' have the same value on %s()",
						KeyName(keys, KSTOP), KeyName(keys, KGO), FUNCNAME);
				goto errorExit;
			}
			
			if (skipChar == stopChar)
			{
				if (ByLines)
				{
					sprintf(OUTSTR,
							"ERROR: illegal for '%s' and '%s' to be the same with %s:T",
							KeyName(keys,KSTOP), KeyName(keys,KSKIP),
							KeyName(keys,KBYLINES));
					goto errorExit;
				}
				if (!Silent)
				{
					sprintf(OUTSTR,
							"WARNING: '%s' and '%s' characters are the same",
							KeyName(keys,KSTOP),KeyName(keys,KSKIP));
					putErrorOUTSTR();
				} /*if (!Silent)*/
			} /*if (skipChar == stopChar)*/
			if (!okSkip(skipChar))
			{
				sprintf(OUTSTR, (charData) ? badCharSkip : notPunctuation,
						KeyName(keys, KSKIP));
				goto errorExit;
			}
		} /*if (skipChar != '\0')*/
		else if (skipthruChar == '\0' && logicalIsSet(Quiet) && !Quiet && !Silent)
		{
			sprintf(OUTSTR,
					"WARNING: '%s:F' without skip or skipthru character ignored by %s()",
					KeyName(keys, KQUIET), FUNCNAME);
			putErrorOUTSTR();
		}
		if (nonKeys > 0)
		{
			sprintf(OUTSTR,
					"ERROR: all non-filename arguments to %s() must be keyword phrases",
					FUNCNAME);
			goto errorExit;
		}
	} /*if (op & IVECREAD)*/
	else if (nonKeys > 1)
	{
		sprintf(OUTSTR,
				"ERROR: %s() has too many non-keyword arguments", FUNCNAME);
		goto errorExit;
	}
	
	if (unusedKeys > 0 && !BadKeyOK)
	{
		/*
			010324 no longer trying to give more informative messages about
			extra keywords
		*/
		sprintf(OUTSTR,
				"ERROR: %s() found unrecognized or duplicate keyword",
				FUNCNAME);
		goto errorExit;
	} /*if (unusedKeys > 0 && !BadKeyOK)*/
	
	if (op & IINFOREAD || op == IREAD)
	{
		whatRead = "data or macro name";
	}
	else if (op & IMACROREAD)
	{
		whatRead = "macro name";
	}
	else if (op & IMATREAD)
	{
		whatRead = "data set name";
	}
	
	if (!logicalIsSet(Echo))
	{
		if (!Silent)
		{
#ifdef HASCONSOLEDIALOG
	/* echo anything from CONSOLE dialog box*/
			Echo = useConsole;
#else  /*HASCONSOLEDIALOG*/
		/* echo anything from batch or non-interactive CONSOLE*/
			Echo = (useConsole == batchConsole);
#endif /*HASCONSOLEDIALOG*/
		} /*if (!Silent)*/
		else
		{
			Echo = 0;
		}
	} /*if (!logicalIsSet(Echo))*/
	else if (Echo && Silent)
	{
		sprintf(OUTSTR,
				"ERROR: %s:T and %s:T are incompatible on %s()",
				KeyName(keys, KECHO), KeyName(keys, KSILENT), FUNCNAME);
		goto errorExit;
	}
	
	
	/*
	   For vecread(), Quiet == 0 means to echo any skipped lines
	   for matread() and macroread(), Quiet == 0 means to echo header lines
	*/
	if (!logicalIsSet(Quiet))
	{
		if (Silent || op & (IVECREAD | IINFOREAD))
		{ /*
			don't echo skipped lines on vecread()
		    don't echo header on inforead()
		  */
			verbose = 0;
			Quiet = 1;
		}
		else
		{
#ifdef HASCONSOLEDIALOG
			/* echo all header lines not starting with '))'*/
			verbose = 1;
#else  /*HASCONSOLEDIALOG*/
			/*
			  echo all header lines not starting '))' except
			  interactively typed ones
			*/
			verbose = (useConsole == interactiveConsole) ? 0 : 1;
#endif /*HASCONSOLEDIALOG*/
		}
	} /*if (!logicalIsSet(Quiet))*/
	else if (!Quiet && Silent)
	{
		sprintf(OUTSTR,
				"ERROR: %s:F and %s:T are incompatible on %s()",
				KeyName(keys, KQUIET), KeyName(keys, KSILENT), FUNCNAME);
		goto errorExit;
	}
	else if (!(op & IVECREAD))
	{
		verbose = (Quiet) ? 0 : 2; /* all header lines printed on quiet:F */
	}
	else
	{
		verbose = !Quiet;
	}
	
	if (!longIsSet(N))
	{
		N = 0;
	}
	
	if (!longIsSet(Startline))
	{
		Startline = 1;
	}

	if (!longIsSet(Bypass))
	{
		Bypass = 0;
	}

#if (0) /* 000614 completely ignore badvalue with CHARACTER data */
	if (charData && realIsSet(BadValue) && !Silent)
	{
		char      *keyname;
		
		if (ByWords)
		{
			keyname = bywordsKeyname;
		}
		else if (ByLines)
		{
			keyname = KeyName(keys,KBYLINES);
		}
		else
		{
			keyname = KeyName(keys,KBYCHARS);
		}
		
		sprintf(OUTSTR,
				"WARNING: %s:value ignored with %s:T",
				KeyName(keys,KBADVAL), keyname);
		putErrorOUTSTR();
	}
#endif /*0*/
	
	if (!logicalIsSet(ByFields))
	{
		ByFields = VecreadByfields;
	}
	else if (charData)
	{
		setLogicalTrue(ByFields);
		if (!Silent)
		{
			sprintf(OUTSTR,
					"WARNING: keyword '%s' ignored when reading CHARACTER data",
					KeyName(keys, KBYFIELDS));
			putErrorOUTSTR();
		}
	}
	
	if (setPlace >= 0)
	{
		arg2 = COMPVALUE(list, setPlace);
		if (!isScalar(arg2) || TYPE(arg2) != CHAR)
		{
			if (op != IMACROSEARCH)
			{
				notCharOrString(whatRead);
			}
			goto errorExit;
		}
	} /*if (setPlace >= 0)*/
	else
	{
		arg2 = SymbolZero;
	}
	
	if (!(op & IVECREAD))
	{
		if (arg2 != SymbolZero && !isKeyword(arg2))
		{
			setnameH = STRING(arg2);
			if (strlen(*setnameH) > MAXNAME)
			{
				sprintf(OUTSTR,
						"ERROR: data set name longer than %ld characters",
						(long) MAXNAME);
				goto errorExit;
			}
			else if ((*setnameH)[0] == '\0' || useConsole != nonConsole)
			{ /* setName is ""; seek first name */
				if (useConsole != nonConsole && !Silent)
				{
					sprintf(OUTSTR,
							"WARNING: %s ignored by %s() when reading from CONSOLE",
							whatRead, FUNCNAME);
					putErrorOUTSTR();
				} /*if (useConsole != nonConsole && !Silent)*/
				/* setnameP points to a buffer of size MAXNAME + 1 */
				setnameH = &setnameP;
			}
		} /*if (arg2 != SymbolZero && !isKeyword(arg2))*/
		else
		{
			setnameH = &setnameP;
		}
	} /*if (!(op & IVECREAD))*/

	if (*OUTSTR)
	{
		goto errorExit;
	}

	if (!FromString && useConsole == nonConsole)
	{
		fname = *fnameH;
#ifdef HASFINDFILE
		fname = mvFindFile(fname, "Specify the data file", NullString,
						   mvReadit | mvTextFile | mvROTextFile,
						   getMvDirInfo(readVol));

		if (fname == (char *) 0)
		{ /* open cancelled */
			if (!Silent)
			{
				sprintf(OUTSTR, "WARNING: %s() cancelled", FUNCNAME);
			}
			result = NULLSYMBOL;
			goto errorExit;
		}
#endif /*HASFINDFILE*/

		fname = expandFilename(fname);
		if (fname == (char *) 0 || !isfilename(fname))
		{
			goto errorExit;
		}
		fnameH = &fname;
	} /*if (!FromString && useConsole == nonConsole)*/

	if (!allocateLINE(lineChunk + 1))
	{
		goto errorExit;
	}
	(*LINE)[0] = '\0';
	fn = (FILE *) 0;


	if (useConsole == nonConsole && !FromString)
	{
		fn = fmyopen(*fnameH, TEXTREADMODE);
		LastDataWasCr = 0;
		
		if (fn == (FILE *) 0)
		{
			sprintf(OUTSTR,
					"ERROR: %s() cannot open file %s", FUNCNAME, *fnameH);
		}
		else if ((fnameH = &LastPathOpened, myfeof(fn)))
		{/* no data */
			sprintf(OUTSTR, "ERROR: file %s is empty", *fnameH);
		}
		if (NoFileOK && *OUTSTR)
		{
			OUTSTR[0] = '\0';
			result = NULLSYMBOL;
			goto errorExit; /* not an error */
		} /*if (NoFileOK && *OUTSTR)*/
		whatRead = "file ";
	} /*if (useConsole == nonConsole && !FromString)*/ 
	else if (useConsole != nonConsole)
	{ /* read from CONSOLE */
		fn = INPUTFILE[BDEPTH];
		fnameH = INPUTFILENAMES[BDEPTH];
		whatRead = (BDEPTH == 0) ? "input file " : "batch file ";
	} /*if (useConsole == nonConsole && !FromString){}else if (useConsole != nonConsole){}*/
	else
	{
		fname = "string variable";
		fnameH = &fname;
		whatRead = quote = NullString;
		fn = (FILE *) 0;
	}

	if (*OUTSTR)
	{
		goto errorExit;
	}

	DataPrompt = (Prompt) ? DATAPROMPT : NullString;

#ifdef CONSOLEOK

	if (useConsole != nonConsole && Prompt)
	{
		if (op & IVECREAD)
		{
			char        termination[50];

			if (N > 0)
			{
				char       *what;
				
				if (ByLines)
				{
					what = "lines";
				}
				else if (ByChars)
				{
					what = "characters";
				}
				else if (ByWords)
				{
					what = "words";
				}
				else
				{
					what = "numbers or '?'";
				}
				
				sprintf(termination, "Provide %ld %s", N, what);
			}
			else if (goChar != '\0')
			{
				sprintf(termination,
						"End by line not starting with '%c'",
						goChar);
			}
			else if (ByLines)
			{
				sprintf(termination,
						"End by line starting with '%c'", stopChar);
			}
			else
			{
				sprintf(termination,
						"End by line containing%s '%c'",
						(stopChar == skipChar) ? " but not starting with" : "",
						stopChar);
			}
			
			if (!ByLines && !ByChars)
			{
				sprintf(OUTSTR,
						"Type %s separated by spaces. %s",
						(ByWords) ? "unquoted 'words'" : "numbers or '?'",
						termination);
			}
			else
			{
				sprintf(OUTSTR, "Type lines.  %s", termination);
			}
		}
		else if (op & (IMATREAD | IMACROREAD))
		{
			sprintf(OUTSTR,
					"Type header and comments starting with '%c', followed by %s",
					COMMENTSTART,
					(strcmp(FUNCNAME, "matread") == 0) ? "data" : "macro");
		}
		else
		{ /* inforead() */
			sprintf(OUTSTR,
					"Type header, followed by comments starting with '%c'",
					COMMENTSTART);
		}
		
#ifndef HASCONSOLEDIALOG
		putOUTSTR();
#else /*HASCONSOLEDIALOG*/
		if (useConsole == interactiveConsole)
		{
			myAlert(OUTSTR);
			*OUTSTR = '\0';
		}
		else
		{
			putOUTSTR();
		}
#endif /*HASCONSOLEDIALOG*/
	} /*if (useConsole != nonConsole && Prompt)*/
	
#endif /*CONSOLEOK*/
	if (op & IVECREAD)
	{
		int    charOp;

		if (ByLines)
		{
			charOp = KBYLINES;
		}
		else if (ByChars)
		{
			charOp = KBYCHARS;
		}
		else if (ByWords)
		{
			charOp = KCHAR;
		}
		else
		{
			charOp = 0;
		}

		result = doReadVec(fn, fnameH, verbose, Echo, Silent, ByFields, charOp,
						   BadValue, stopChar, goChar, skipthruChar,
						   skipChar, N, Bypass, Startline);
	} /*if (op & IVECREAD)*/
	else
	{
		result = doReads(fn, setnameH, fnameH, op, verbose, Echo, Silent,
						 NotFoundOK, Labels);
	} /*if (op & IVECREAD){}else{}*/
	if (result != SymbolError && Printname)
	{
		sprintf(OUTSTR, "Read from %s%s%s%s", whatRead, quote, *fnameH,	quote);
		putOUTSTR();
	}

#ifdef MACINTOSH
	if (UseWindows && useConsole != nonConsole && !FromString && BDEPTH == 0)
	{
		InvalRect(&(CmdWind->portRect));
		macUpdate((WindowPtr) 0); /* update all windows */
	} /*if (UseWindows && useConsole != nonConsole && !FromString && BDEPTH == 0)*/
#endif /*MACINTOSH*/
	*OUTSTR = '\0';
/* fall through */

  errorExit:
	if (fn != STDIN && fn != (FILE *) 0 &&
		useConsole == nonConsole && !FromString)
	{
		fclose(fn);
	}
	if (op != IMACROSEARCH)
	{
		putErrorOUTSTR();
	}
	
	freeLINE();
	UNLOADSEG(findLine);
	return (result);
} /*readdata()*/

/* 
   new function findfile(name) returns the first matche to name
   in the search path
*/
Symbolhandle    findfile(Symbolhandle list)
{
	Symbolhandle    symhString,result;
	long            nargs = NARGS(list);
	char           *filename;
	char           *pathname;
	WHERE("findfiles");
	
	*OUTSTR = '\0';
	
	if (nargs != 1)
	{
		badNargs(FUNCNAME, 1);
		goto errorExit;
	}

	symhString = COMPVALUE(list,0);

	filename = (char *) 0;
	if (symhString == EmptyArgument || !isCharOrString(symhString) ||
        	*(filename = STRINGPTR(symhString)) == 0)
	{
		notCharOrString("filename");
		goto errorExit;
	}

	pathname = findFileByName(filename,"r");
	if(pathname == (char *) 0)
	{
		return(NULLSYMBOL);
	}


	result = Install(SCRATCH,CHAR);

	if (result == (Symbolhandle) 0)
	{
		goto errorExit;
	}
	TMPHANDLE = mygethandle(strlen(pathname) + 5);  /* why worry? */
	setSTRING(result, TMPHANDLE);
	if (TMPHANDLE == (char **) 0)
	{
		goto errorExit;
	}
	
	strcpy(*TMPHANDLE,pathname);
	setNAME(result, SCRATCH);
	setTYPE(result, CHAR);
	setNDIMS(result,1);
	setDIM(result,1,1);
	return (result);

  errorExit:
	putErrorOUTSTR();
	Removesymbol(result);
	return (0);
}


