/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2003 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#include "globals.h"
#include "mvsave.h"

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Save
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "mvIface.h"

#include "mvIo.h" /*added 010212*/
#include "mvGraphs.h"

#include <ctype.h>

/*
  replaced putOUTSTR() by putErrorOUTSTR() in most places
  replaced myerrorout() by putOutErrorMsg() and putOutMsg()

  001105 modified so that when keyword phrase 'foreignok:T' is an
  argument to restore() and a binary file saved on a "foreign" machine
  is being restore, the file will be restored, with an attempt made to
  permute the byte order of doubles and replace '\n' and '\r' by the
  local definition of '\n' in CHAR and MACRO variables

  001114 add kludge to cope with the problem that the NT version represents
  SAMPLEDOUBLES with the least significant bit one greater than in
  Windows, Mac, HPUX or Linux

  010118 Fixed minor bug in keyword processing.
         replaced all occurrences (Symbolhandle) 0 by SymbolError or
         SymbolZero
  010212 When restoring from a foreign file, option 'matchdelay' is
         not changed
  010411 File/Folder name variables (HOME, DATAPATHS, MACROFILES, DATAFILE) are
         not restored, the original variables being preserved.  The
		 ones from the workspace file are bundled in structure
		 SAVEDNAMES
  010601 Include mvIface.h instead of macIface.h or wxIface.h
  010618 On Windowed systems, if no Window File directory information has
         been set, restore file information is transferred
  010727 Changed ISSHORT to isShort
  010729 Recognizes SYMNONTYPE and uses value to restore locked status
  011003 When restoring ascii workspace file, '\n' and '\r' in CHARACTER
         variables or macros are recognized
  030319 Added HELPFILES and HELPINDICES to list of folder variables that
         won't be overwritten but will be included in SAVEDNAMES
  030615 Added defines related to Carapace
  031211 restore() is more careful about closing files
*/

#define CHARCHUNK     10 /* guess at size of CHAR element */
#define MACROCHUNK  sizeof(double)  /* size that is likely to be in a cache */

#define TIMELENGTH    30 /* length of buffer for time information */

#define HEADERLENGTH  24 /* strlen("////MacAnovaSaveFile////") */

enum folderSymbolCodes
{
	homeSymbol,
	dataFileSymbol,
	nFolderScalars,
	dataPathsSymbol = nFolderScalars,
	macroFilesSymbol,
	helpFilesSymbols,
	helpIndicesSymbols,
	nFolderSymbols
};

static Symbolhandle    FolderSymbols[nFolderSymbols];
static Symbolhandle    SavedFolderSymbols[nFolderSymbols];

static mvSymbolName    FolderSymbolNames[nFolderSymbols+1] =
{
	"HOME",
	"DATAFILE",
	"DATAPATHS",
	"MACROFILES",
	"HELPFILES",
	"HELPINDICES",
	"SAVEDNAMES"	/*name for structure with symbols from file*/
};

static char           LastRestoreWasCr = 0;

/*
  Constants, typedefs and variables related to translation of foreign
  binary save files
*/

enum permutationConstants
{
	badDimensions = 1,
	differentDoubles,
	badDoubleHexLength,
	badDoubleHexString,
	duplicateDoubleByte,
	unmatchedDoubleByte,
	badDoublePermutation,
	differentLongs,
	badLongHexLength,
	badLongHexString,
	duplicateLongByte,
	unmatchedLongByte,
	badLongPermutation,
	noPermute,
	yesPermute,
	maybePermute
};
enum swapConstants
{
	noSwap,
	yesSwap
};


/*
  Control translation of "foreign" binary save files
  byte j of double comes from byte DoublePermutation[j] of double from file
  byte j of long comes from byte LongPermutation[j] of long from file

  Permutation is not done unless DoPermute == yesPermute

  Swapping of local '\n' for foreign '\n' in CHAR and MACRO variables
  is not done unless SwapNewlines == yesSwap

  These are operative only when
    1.  the file being restored is binary
    2.  the saving machine is not recognized as being compatible
    3.  keyword phrase foreignok:T is an argument
*/
static char           DoublePermutation[sizeof(double)];
static char           LongPermutation[sizeof(long)];
static char           DoPermute;
static char           SwapNewlines;

/*
  ForeignFile != 0 if and only if the saving machine is not recognized
  as compatible, regardless of whether the file is binary or whether
  foreignok:T is an argument
*/
static char           ForeignFile;


static int whichSpecialSymbol(char * name)
{
	int          i;
	
	for (i = 0; i < NSPECIALSYMBOLS; i++)
	{
		if (strcmp(name, SpecialSymbolNames[i]) == 0)
		{
			return (i);
		}
	} /*for (i = 0; i < NSPECIALSYMBOLS; i++)*/
	return (-1);
} /*whichSpecialSymbol()*/

/*
  000614 ^Z always recognized as EOF on ascii workspace file
         Formerly this happend only with MSDOS defined
*/
static int restoregetc(FILE * fp)
{
	int           c;

	c = fgetc(fp);
	if (!Binary)
	{
		c = (c == NL && LastRestoreWasCr) ? fgetc(fp) : c;
		LastRestoreWasCr = (c == CR);
#ifndef MACINTOSH
		c = (LastRestoreWasCr) ? '\n' : c;
#else
		c = (c == NL) ? '\n' : c;
#endif /*MACINTOSH*/

	/* 000614 following line no longer protected by #ifdef MSDOS*/
	c = (c == MSDOSEOF) ? EOF : c;

	} /*if (!Binary)*/
	
	return (c);
} /*restoregetc()*/

/*
  Test for a name match.  If a match is found, the data part of symh
  is copied to pointer dPointer and 1 is returned.  Otherwise 0 is returned.
  Used in restoring globals
*/
static long symbol2Double(double *dPointer,char *name,long size,
						  Symbolhandle symh)
{
	WHERE("symbol2Double");

	if (strcmp(NAME(symh), name) != 0)
	{
		return (0);
	}
	size = (size <= DIMVAL(symh, 1)) ? size : DIMVAL(symh, 1);
	doubleCopy(DATAPTR(symh), dPointer, size);
	return (1);
} /*symbol2Double()*/

/*
  Test for a name match.  If a match is found, the data part of symh
  is copied to pointer lPointer and 1 is returned.  Otherwise 0 is returned.
  Used in restoring globals
*/
static long symbol2Long(long *lPointer,char *name,long size, Symbolhandle symh)
{
	long                i;
	WHERE("symbol2Long");

	if(strcmp(NAME(symh),name) != 0)
	{
		return (0);
	}
	size = (size <= DIMVAL(symh,1)) ? size : DIMVAL(symh,1);
	for (i = 0;i < size;i++)
	{
		*lPointer++ = LONGDATAVALUE(symh,i);
	}
	return (1);
} /*symbol2Long()*/

/*
  Test for a name match.  If a match is found, the current handle is
  disposed of, it is reset to DATA(symh), DATA(symh) is cleared, and 1
  is returned.  Otherwise 0 is returned.
  Used in restoring globals
*/

static long symbol2Handle(double ***handle,char *name,Symbolhandle symh)
{
	if (strcmp(NAME(symh),name) != 0)
	{
		return (0);
	}

	if (*handle != (double **) 0)
	{
		mydisphandle((char **) *handle);
	}

	*handle = DATA(symh);
	setDATA(symh, (double **) 0);
	return (1);
} /*symbol2Handle()*/

/*
  980717 New function to search for SYMBOLMARKER (%#$SYMBOL$#%)
         put before each symbol name.  Someday this may facilitate
         recovery from restore errors.
*/
static int getSymbolName(char name [])
{
	if (!Pre407Save && (!restoreName(name) || strcmp(name, SYMBOLMARKER) != 0))
	{
		return (0);
	}
	
	if (!restoreName(name) || strlen(name) > NAMELENGTH)
	{
		return (0);
	}
	
	return (1);
} /*getSymbolName()*/

/*
  restoreSymbolLabels() used for restores from files in Pre407Save formats
*/
static int restoreSymbolLabels(Symbolhandle symh, long labelsLength)
{
	long          i, j, ndims = NDIMS(symh);
	long          labelsActual = 0;
	char         *labels;
	WHERE("restoreSymbolLabels");
	
	TMPHANDLE = mygethandle(labelsLength);
	if (TMPHANDLE == (char **) 0 || !setLabels(symh, TMPHANDLE))
	{
		mydisphandle(TMPHANDLE);
		return (0);
	}

	labels = *TMPHANDLE;
	for (i = 1; i <= ndims; i++)
	{
		for (j = 0; j < DIMVAL(symh, i); j++)
		{
			if (!restoreString(labels))
			{
				return (0);
			}
			labelsActual += strlen(labels) + 1;
			labels = skipStrings(labels,1);
		} /*for (j = 0; j < DIMVAL(symh, i); j++)*/				
	} /*for (i = 1; i <= ndims; i++)*/

	if (labelsActual != labelsLength)
	{
		return (0);
	}

	return (1);
} /*restoreSymbolLabels()*/

/*
  Read symbol from file.  It is immediately put in symbol table and returned
  as an installed symbol.  This means a structure component must be cut from
  the symbol table before being attached.

  980717 symbol name is now read by getSymbolName()
  980718 symbol labels are now read by restoreSymbolLabels()
  990502 moved code related to restoring MACRO, CHAR, REAL, LOGIC and LONG
         handles to new public function restoreHandle()
  990522 includes mvGraphs.h
  000209 restores GLM global TERMNAMES_AT
  010516 Use atLeastNArgs constant on badNargs() call
*/
static Symbolhandle restoresym(void)
{
	long            i, tot;
	long            type, dimi, nclass;
	long            ndims = -1, mdims;
	long            labelsLength = -1;
	Symbolhandle    symh = SymbolZero, component;
	char          **chardata = (char **) 0, name[NAMELENGTH+1];
	char            itemCvalue[MAXCHARITEMSIZE + 1];
	char           *typename = itemCvalue;
	whole_graph   **graph = (whole_graph **) 0;
	WHERE("restoresym");

	/* get symbol name */
	if (!getSymbolName(itemCvalue) || strlen(itemCvalue) > NAMELENGTH+1)
	{
		goto errorExit;
	}
	strcpy(name, itemCvalue);

	/* get symbol type */
	if (ASCII == SAVEBINARY1 || ASCII == SAVEASCII1)
	{ /* numeric type of non-ascii and old ascii format */
		if (!restoreLong(&type))
		{
			goto errorExit;
		}
		/* make sure codes are restore correctly even if future codes change */
		switch (type)
		{
		  case CHAR_V31:
			type = CHAR;
			break;
		  case REAL_V31:
			type = REAL;
			break;
		  case LOGIC_V31:
			type = LOGIC;
			break;
		  case STRUC_V31:
			type = STRUC;
			break;
		  case MACRO_V31:
			type = MACRO;
			break;
		  case PLOTINFO_V31:
			type = PLOTINFO;
			break;
		  case LONG_V31:
			type = LONG;
			break;
		  default:
			;
		} /*switch (type)*/
	} /*if(ASCII == SAVEBINARY1 || ASCII == SAVEASCII1)*/
	else
	{
		if (!restoreName(typename))
		{/* name-specified type for new ascii formats and new binary format */
			goto errorExit;
		}
		if (strncmp(typename, "SHORT", 5) == 0)
		{
			typename += 5;
			type = SHORTSYMBOL;
		}
		else
		{
			type = 0;
		}
		
		if(strcmp(typename,"REAL") == 0)
		{
			type |= REAL;
		}
		else if(strcmp(typename,"LOGIC") == 0)
		{
			type |= LOGIC;
		}
		else if(strcmp(typename,"LONG") == 0)
		{
			type |= LONG;
		}
		else if(strcmp(typename,"CHAR") == 0)
		{
			type |= CHAR;
		}
		else if(strcmp(typename,"STRUC") == 0)
		{
			type |= STRUC;
		}
		else if(strcmp(typename,"MACRO") == 0)
		{
			type |= MACRO;
		}
		else if(strcmp(typename,"GRAPH") == 0)
		{
			type |= PLOTINFO;
		}
		else if(strcmp(typename,"NULL") == 0)
		{
			type |= NULLSYM;
		}
		else
		{ /* not recognized */
			goto errorExit;
		}
	}
	
	symh = Install(name, type); /* immediately add to symbol table */
	
	if (symh == SymbolError)
	{
		goto errorExit;
	}

	if (Pre407Save)
	{
		/* get NCLASS */
		if (!restoreLong(&nclass))
		{
			goto errorExit;
		}
		setNCLASS(symh,nclass);

		/* get NDIMS */
		if (!restoreLong(&ndims))
		{
			goto errorExit;
		}
		setNDIMS(symh,ndims);
		
		/* get dimensions */
		mdims = (ASCII >= SAVEASCII3 || ASCII <= SAVEBINARY3) ? ndims : MAXDIMS;
		for (i = 1; i <= mdims; i++)
		{
			if (!restoreLong(&dimi))
			{
				goto errorExit;
			}
			if(i <= ndims)
			{
				setDIM(symh,i,dimi);
			}
		} /*for (i = 1; i <= mdims; i++)*/

		if (!Pre400Save)
		{
			if (!restoreLong(&labelsLength) ||
				labelsLength > 0 && !restoreSymbolLabels(symh, labelsLength))
			{
				goto errorExit;
			}
		}
	} /*if (Pre407Save)*/
	else
	{
		double        itemDvalue;
		long          itemLvalue, itemLength;
		long          nlabels = -1;
		int           itemNumber, itemType;
		char          itemCvalue[MAXCHARITEMSIZE + 1];
#ifndef LABELSARECHAR
		Symbolhandle  labelsSymh = SymbolZero;
#endif /*LABELSARECHAR*/
#ifndef NOTESARECHAR
		Symbolhandle  notesSymh = SymbolZero;
#endif /*NOTESARECHAR*/

		/* Look for Symbol items until itemNumber == EndItems */
		while (1)
		{
			if (!getNextRestoreItem(SymbolPrefix, &itemNumber, &itemType,
									&itemLength, &itemDvalue, &itemLvalue,
									itemCvalue))
			{
				goto errorExit;
			}

			if (itemNumber == EndItems)
			{
				break;
			}
			
			if (itemNumber < SYMENDSYM)
			{ /* known item */
				if (itemType != SYMITEMTYPE(itemNumber))
				{
					goto errorExit;
				}

				switch (itemNumber)
				{
				  case SYMNCLASS:
					setNCLASS(symh, itemLvalue);
					break;

				  case SYMDIMS:
					ndims = itemLength;
					if (isShort(symh) && ndims != 1)
					{
						goto errorExit;
					}
					setNDIMS(symh, ndims);
					nlabels = 0;
					
					for (i = 1; i <= ndims; i++)
					{
						long         dimi;

						if (!restoreLong(&dimi))
						{
							goto errorExit;
						}
						setDIM(symh, i, dimi);
						nlabels += dimi;
					} /*for (i = 1; i <= ndims; i++)*/
					break;

#ifdef LABELSARECHAR
				  case SYMLABLEN:
					labelsLength = itemLvalue;
					if (labelsLength < 0)
					{
						goto errorExit;
					}
					break;

				  case SYMLABELS:
					if (labelsLength <= 0 || nlabels != itemLength ||
						!restoreSymbolLabels(symh, labelsLength))
					{
						goto errorExit;
					}
					labelsLength = -1;
					break;
#else /*LABELSARECHAR*/
				  case SYMLABELS:
					labelsSymh = restoresym();

					if (labelsSymh == SymbolError ||
						NDIMS(labelsSymh) != 1 ||
						DIMVAL(labelsSymh, 1) != nlabels ||
						!setLabels(symh, STRING(labelsSymh)))
					{
						Removesymbol(labelsSymh);
						goto errorExit;
					}
					setSTRING(labelsSymh, (char **) 0);
					Removesymbol(labelsSymh);
					break;
					
#endif /*LABELSARECHAR*/

#ifndef NOTESARECHAR
				  case SYMNOTES:
					notesSymh = restoresym();

					if (notesSymh == SymbolError ||
						NDIMS(notesSymh) != 1 ||
						!setNotes(symh, STRING(notesSymh)))
					{
						Removesymbol(notesSymh);
						goto errorExit;
					}
					setSTRING(notesSymh, (char **) 0);
					Removesymbol(notesSymh);
					break;
					
#endif /*NOTESARECHAR*/
				  case SYMNONTYPE:
					if ((itemLvalue << 16) & LOCKEDMASK)
					{
						setLocked(symh);
					}
					break;

				} /*switch (itemNumber)*/
			} /*if (itemNumber < SYMENDSYM)*/
			else if (itemType & SAVEVECTOR)
			{ /*unrecognized vector type ; skip value*/
				if (!skipRestoreVector(itemType, itemLength))
				{
					goto errorExit;
				}
			} /*else if (itemType & SAVEVECTOR)*/
		} /*while (1)*/
	} /*if (Pre407Save){}else{}*/
	
	if (DIMVAL(symh, 1) == 0)
	{
		setTYPE(symh, NULLSYM | ((isShort(symh)) ? SHORTSYMBOL : 0));
	} /*if (DIMVAL(symh, 1) == 0)*/
	else
	{
		tot = symbolSize(symh);
		
		switch  (type & TYPEMASK)
		{
		  case CHAR:
		  case MACRO:
		  case REAL:
		  case LOGIC:
		  case LONG:
			if (!restoreHandle(symh, Binary))
			{
				goto errorExit;
			}
			break;

		  case STRUC:
			TMPHANDLE = mygethandle((long) tot * sizeof(Symbolhandle));
			setCOMP(symh,(Symbolhandle**) TMPHANDLE);
			if (COMP(symh) == (Symbolhandle**) 0)
			{
				goto errorExit;
			}
			for (i = 0; i < tot; i++)
			{
				COMPVALUE(symh, i) = SymbolZero;
			}
			for (i = 0; i < tot; i++)
			{
				component = restoresym();
				Cutsymbol(component);
				COMPVALUE(symh, i) = component;
				if (component == SymbolError)
				{
					goto errorExit;
				}
			}
			break;

		  case PLOTINFO:
#if (1) /* 980714 new calling sequence for restoreGraph()*/
			if (!restoreGraph(symh))
			{
				goto errorExit;
			}
#else /*1*/
			graph = restoreGraph(); /* restoreGraph() is in plotutil.c */
			if(graph == (whole_graph **) 0)
			{
				goto errorExit;
			}
			setGRAPH(symh, graph);
#endif /*1*/
			break;

		} /*switch  (type & TYPEMASK)*/
	} /*if (DIMVAL(symh, 1) == 0){}else{}*/

	return (symh);

  errorExit:
	Removesymbol(symh);

	return (SymbolError);

} /*restoresym()*/

/* bit definitions used in previous versions */
#define BALANOVA     0x002L  /* 2 balanced anova */
#define UNBALANO     0x010L  /* 16 unbalanced anova */
#define OLDOLSREG    0x020L  /* 32 standard regression */
#define OLDROBUSTREG 0x040L  /* 64 robust regression */
/*
  971104 added check on value of myhandlelength()
*/
static void restoreGlobals(Symbolhandle globals)
{
	long                icomp, i, j, n, k, nvars;
	Symbolhandle        symh, symh1;
	double              dTemp[MAXVARS];
	double            **model = (double **) 0;
	double              errorterms[MAXERRTRMS + 1];
	unsigned long       tempModelType = 0;
	unsigned long       tempGlmControl = 0;
	long                maxvars = MAXVARS;
	long                neweststyle =
	  (ASCII <= SAVEBINARY5 || ASCII >= SAVEASCII5);
	WHERE("restoreGlobals");

	nvars = -1;
	for(icomp=0;icomp<NCOMPS(globals);icomp++)
	{
		symh = COMPVALUE(globals,icomp);

		if(symbol2Double(dTemp,"NVARS",1,symh))
		{
			NVARS = (long) dTemp[0];
			nvars = MODELINFO->nvars = NVARS;
		}
		else if(symbol2Long(&maxvars,"MAXVARS",1,symh))
		{
			if(maxvars > MAXVARS)
			{
				sprintf(OUTSTR,
						"ERROR: Model information saved by version allowing %ld variables in models",
						maxvars);
				putOUTSTR();
				sprintf(OUTSTR,
						"       This version allows only %ld.  Model information not restored",
						(long) MAXVARS);
				putErrorOUTSTR();
				NVARS = MODELINFO->nvars = -1;
				break;
			}
		}
		else if(symbol2Double(dTemp, "NFACTORS",1,symh))
		{
			NFACTORS = (long) dTemp[0];
			MODELINFO->nfactors = NFACTORS;
		}
		else if(symbol2Double(dTemp,"NDATA",1,symh))
		{
			NDATA = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"NOTMISSING",1,symh))
		{
			NOTMISSING = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"NTERMS",1,symh))
		{
			NTERMS = (long) dTemp[0];
			MODELINFO->nterms = NTERMS;
		}
		else if(symbol2Double(dTemp,"NREGX",1,symh))
		{
			NREGX = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"NY",1,symh))
		{
			NY = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"NERRORTERMS",1,symh))
		{
			NERRORTERMS = (long) dTemp[0];
			MODELINFO->nerrorterms = NERRORTERMS;
		}
		else if(symbol2Double(&PREVGLMSCALE,"PREVGLMSCALE",1,symh))
			;
		else if(symbol2Double(dTemp,"PREVMODELTYP",1,symh))
		{
			tempModelType = (long) dTemp[0];
			if (!neweststyle)
			{
				if (tempModelType & OLDROBUSTREG)
				{
					putOutErrorMsg("WARNING: cannot restore information from robust() in previous version");
					NVARS = MODELINFO->nvars = -1;
					break;
				} /*if (tempModelType & OLDROBUSTREG)*/
				
				tempGlmControl = UNBALANCED;
				if (tempModelType & LEAPS)
				{
					tempGlmControl |= NORMALDIST | IDENTLINK;
				}
				else if (tempModelType & BALANOVA)
				{
					tempModelType &= ~BALANOVA;
					tempModelType |= ANOVA;
					tempGlmControl &= ~UNBALANCED;
					tempGlmControl |= (NORMALDIST|IDENTLINK);
				}
				else if (tempModelType & FASTANOVA)
				{
					tempGlmControl |= NORMALDIST | IDENTLINK;
				}
				else if (tempModelType & MANOVA)
				{
					tempGlmControl |= (NORMALDIST||IDENTLINK|MULTIVAR);
				}
				else if (tempModelType & UNBALANO)
				{
					tempModelType &= ~UNBALANO;
					tempModelType |= ANOVA;
					tempGlmControl |= (NORMALDIST|IDENTLINK);
				}
				else if (tempModelType & OLDOLSREG)
				{
					tempModelType &= ~OLDOLSREG;
					tempModelType |= OLSREG;
					tempGlmControl |= NORMALDIST | IDENTLINK;
				}
				else if (tempModelType & LOGITREG)
				{
					tempGlmControl |= BINOMDIST | LOGITLINK;
				}
				else if (tempModelType & POISSONREG)
				{
					tempGlmControl |= POISSONDIST | LOGLINK;
				}
			}
		}
		else if(symbol2Double(dTemp,"PREVGLMCONTR",1,symh))
		{
			tempGlmControl = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"INCREMENTAL",1,symh))
		{
			INCREMENTAL = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"USEGLMOFFSET",1,symh))
		{
			USEGLMOFFSET = (long) dTemp[0];
		}
		else if(symbol2Double(dTemp,"NCLASSES",MAXVARS,symh))
		{
			for(j = 0;j < maxvars;j++)
			{
				NCLASSES[j] = (j < nvars) ? (long) dTemp[j] : 0;
			}
		}
		else if(symbol2Double(errorterms,"ERRORTERMS",MAXERRTRMS+1,symh))
		{
			for(j = 0;j <= MAXERRTRMS;j++)
			{
				zeroTerm(MODELINFO->errorterms[j]);
				MODELINFO->errorterms[j][0] = (long) errorterms[j];
			}
		}
		else if(symbol2Long((long *) MODELINFO->errorterms,"ERRORTERMS1",
							(WORDSPERTERM)*(MAXERRTRMS+1),symh))
			;
		else if(symbol2Handle((double ***) &DEPVNAME,"DEPVNAME",symh))
			;
		else if(strcmp(NAME(symh),"VARNAMES") == 0)
		{
			n = (DIMVAL(symh,1) <= maxvars+1) ? DIMVAL(symh,1) : maxvars+1;
			k = 0;
			for(j=0;j<=maxvars;j++)
			{
				if(j < n)
				{
					strcpy(VARNAMES[j],STRINGPTR(symh) + k);
					k += strlen(VARNAMES[j]) + 1;
				}
				else
				{
					VARNAMES[j][0] = '\0';
				}
			}
		}
		else if(symbol2Handle(&model,"MODEL",symh))
		{
			long      length = myhandlelength((char **) model);

			if (length > 0)
			{
				length /= sizeof(double);
				TMPHANDLE = mygethandle(length * sizeof(modelType));
			}
			else
			{
				TMPHANDLE = (char **) 0;
			}
		
			if(TMPHANDLE == (char **) 0)
			{
				NVARS = MODELINFO->nvars = -1;
				putOutMsg("       Model information not restored");
				break;
			}
			MODEL = (modelHandle) TMPHANDLE;
			for(i=0;i<length;i++)
			{
				zeroTerm((*MODEL)[i]);
				(*MODEL)[i][0] = (termWord) (*model)[i];
			}
		}
		else if(symbol2Handle((double ***) &MODEL,"MODEL1",symh))
		{
			;
		}
		/*
		  Under current version, Y should always be the same as
		  DATA(MODELVARS[0]) so we no longer restore it;  however, we
		  do recognize it
		*/
		else if(strcmp(NAME(symh),"Y") == 0)
			;
		else if(symbol2Handle(&RESIDUALS,"RESIDUALS",symh))
			;
		else if(symbol2Handle(&WTDRESIDUALS,"WTDRESIDUALS",symh))
			;
		else if(symbol2Handle(&GLMOFFSET,"GLMOFFSET",symh))
			;
		else if(symbol2Handle(&HII,"HII",symh))
			;
		else if(symbol2Handle(&SS,"SS",symh))
			;
		else if(symbol2Handle(&DF,"DF",symh))
			;
		else if(symbol2Handle(&MISSWTS,"MISSWTS",symh))
			;
		else if(symbol2Handle(&CASEWTS,"CASEWTS",symh))
			;
		else if(symbol2Handle(&ITERWTS,"ITERWTS",symh))
			;
		else if(symbol2Handle(&ALLWTS,"ALLWTS",symh))
			;
		else if(symbol2Handle(&LOGITN,"LOGITN",symh))
			;
		else if(strcmp(NAME(symh),STRMODELNAME) == 0)
		{
			if(STRMODEL != (char **) 0)
			{
				mydisphandle(STRMODEL);
			}
			STRMODEL = STRING(symh);
			setSTRING(symh,(char **) 0);
		}
		else if(strcmp(NAME(symh),"TERMNAMES") == 0)
		{
			if(TERMNAMES != (char **) 0)
			{
				mydisphandle(TERMNAMES);
			}
			TERMNAMES = STRING(symh);
			setSTRING(symh,(char **) 0);
		}
		else if(strcmp(NAME(symh),"TERMNAMES_AT") == 0)
		{
			if(TERMNAMES_AT != (char **) 0)
			{
				mydisphandle(TERMNAMES_AT);
			}
			TERMNAMES_AT = STRING(symh);
			setSTRING(symh,(char **) 0);
		}
		else if(symbol2Handle(&REGX,"REGX",symh))
			;
		else if(symbol2Handle(&REGX2,"REGX2",symh))
			;
		else if(symbol2Handle(&XTXINV,"XTXINV",symh))
			;
		else if(symbol2Handle(&REGCOEF,"REGCOEF",symh))
			;
		else if(symbol2Handle(&REGSS,"REGSS",symh))
			;

		/*
		  Under current version, X[i] should always be the same as
		  DATA(MODELVARS[i+1]) so we no longer restore it.  We do
		  recognize it, however
		*/
		else if(strcmp(NAME(symh),"X") == 0)
			;
		else if(strcmp(NAME(symh),"MODELVARS") == 0)
		{
			n = NCOMPS(symh);
			for(j=0;j<=maxvars;j++)
			{
				symh1 = (j < n) ? COMPVALUE(symh,j) : SymbolZero;
				MODELVARS[j] = symh1;
				if(j < n)
				{
					COMPVALUE(symh,j) = SymbolZero;
					if(j==0)
					{
						Y = DATA(MODELVARS[0]);
					}
					else
					{
						X[j-1] = DATA(MODELVARS[j]);
					}
				} /*if(j < n)*/
			} /*for(j=0;j<=maxvars;j++)*/
		}
		else
		{
			putOutErrorMsg("WARNING: unrecognized global variable on file. Globals not restored");
			NVARS = MODELINFO->nvars = -1;
			break;
		}
	} /*for(icomp=0;icomp<NCOMPS(globals);icomp++)*/
	if (NVARS >= 0)
	{
		PREVMODELTYPE = tempModelType;
		PREVGLMCONTROL = tempGlmControl;
	}
} /*restoreGlobals()*/

#ifdef PLATFORM
/*
	980403 added checks for PLATFORM_ALT5 through PLATFORM_ALT8
*/
enum machineStatusConstants
{
	okPlatform,
	badPlatform,
	badGlm
};

/*
	Attempt to find permutation of bytes in doubles and longs that will work
	on this system
*/
static void setPermutation(Symbolhandle symh)
{
	SaveDouble     localDoubles[nSampleDoubles], foreignDoubles[nSampleDoubles];
	SaveLong       localLongs[nSampleLongs], foreignLongs[nSampleLongs];
	double         x;
	long           lx;
	char          *place = STRINGPTR(symh);
	char           c1, c2;
	unsigned char  c3;
	int            nrows = (nSampleDoubles > nSampleLongs) ?
	  nSampleDoubles : nSampleLongs;
	int            i, j; /*byte indices*/
	int            iNumber;
	WHERE("setPermutation");

	if (NDIMS(symh) != 2 || DIMVAL(symh, 1) != 2*nrows || DIMVAL(symh, 2) != 2)
	{
		DoPermute = badDimensions;
		goto errorExit;
	}

	for (i = 0; i < sizeof(double); i++)
	{
		DoublePermutation[i] = (char) sizeof(double);
	}
	for (i = 0; i < sizeof(long); i++)
	{
		LongPermutation[i] = (char) sizeof(long);
	}
	for (iNumber = 0; iNumber < nSampleDoubles; iNumber++)
	{
		if (sscanf(place, "%lf", &(localDoubles[iNumber].value)) != 1 ||
			sscanf((char *) SAMPLEDOUBLES[iNumber], "%lf", &x) != 1 ||
			localDoubles[iNumber].value != x)
		{
			DoPermute = differentDoubles;
			goto errorExit;
		}
		place = skipStrings(place, 1);
		if (strlen(place) != 2*sizeof(double))
		{
			DoPermute = badDoubleHexLength;
			goto errorExit;
		}
#if (0)
		PRINT("Double hex representation %d is '%s'",iNumber, place,0,0);
#endif /*0*/		
		/* reconstruct foreign double from hex representation*/
		for (i = 0; i < sizeof(double); i++)
		{
			c1 = place[2*i];
			c2 = place[2*i+1];
			if (!isalnum(c1) || !isalnum(c2))
			{
				DoPermute = badDoubleHexString;
				goto errorExit;
			}
			c1 = (isalpha(c1)) ? 10 + toupper(c1) - 'A' : c1 - '0';
			c2 = (isalpha(c2)) ? 10 + toupper(c2) - 'A' : c2 - '0';
			foreignDoubles[iNumber].bytes[i] = 16*c1 + c2;
		} /*for (i = 0; i < sizeof(double); i++)*/
		place = skipStrings(place, 1);
	} /*for (iNumber = 0; iNumber < nSampleDoubles; iNumber++)*/

	for (iNumber = 0; iNumber < nSampleLongs; iNumber++)
	{
		if (sscanf(place, "%ld", &(localLongs[iNumber].value)) != 1 ||
			sscanf((char *) SAMPLELONGS[iNumber], "%ld", &lx) != 1 ||
			localLongs[iNumber].value != lx)
		{
			DoPermute = differentLongs;
			goto errorExit;
		}
		place = skipStrings(place, 1);
		if (strlen(place) != 2*sizeof(long))
		{
			DoPermute = badLongHexLength;
			goto errorExit;
		}
		/* reconstruct foreign long from hex representation*/
#if (0)
		PRINT("Long hex representation %d is '%s'",iNumber, place,0,0);
#endif /*0*/		
		for (i = 0; i < sizeof(long); i++)
		{
			c1 = place[2*i];
			c2 = place[2*i+1];
			if (!isalnum(c1) || !isalnum(c2))
			{
				DoPermute = badLongHexString;
				goto errorExit;
			}
			c1 = (isalpha(c1)) ? 10 + toupper(c1) - 'A' : c1 - '0';
			c2 = (isalpha(c2)) ? 10 + toupper(c2) - 'A' : c2 - '0';
			foreignLongs[iNumber].bytes[i] = 16*c1 + c2;
		} /*for (i = 0; i < sizeof(long); i++)*/
		place = skipStrings(place, 1);
	} /*for (iNumber = 0; iNumber < nSampleLongs; iNumber++)*/

#ifndef BASE_MICROSOFT
	/*
		001114
		This is a kludge to correct for a one bit difference in the
		representation of SAMPLEDOUBLES in Windows NT
        The first byte (least significant) duplicates second and is 1 too
		large as compared with Windows, Macintosh, HPUX and Linux
	*/
	for (iNumber = 0; iNumber < nSampleDoubles; iNumber++)
	{
		if (foreignDoubles[iNumber].bytes[0] ==
			foreignDoubles[iNumber].bytes[1])
		{
			foreignDoubles[iNumber].bytes[0]--;
		}
	} /*for (iNumber = 0; iNumber < nSampleDoubles; iNumber++)*/
#else /* BASE_MICROSOFT*/
	/*
	  Might as well check and correct the locals
	  Presumably they will also be a bit off when running under NT
	*/
	for (iNumber = 0; iNumber < nSampleDoubles; iNumber++)
	{
		if (localDoubles[iNumber].bytes[0] ==
			localDoubles[iNumber].bytes[1])
		{
			localDoubles[iNumber].bytes[0]--;
		}
	} /*for (iNumber = 0; iNumber < nSampleDoubles; iNumber++)*/
#endif /*BASE_MICROSOFT*/

	/* find permutation for doubles*/
	for (i = 0; i < sizeof(double); i++)
	{
		c3 = localDoubles[0].bytes[i];
		for (j = 0; j < sizeof(double); j++)
		{
			if (foreignDoubles[0].bytes[j] == c3)
			{
				if (DoublePermutation[j] < sizeof(double))
				{
					/* duplicate byte; ambiguous permutation */
					DoPermute = duplicateDoubleByte;
					goto errorExit;
				}
				DoublePermutation[j] = i;
				break;
			}
		} /*for (j = 0; j < sizeof(double); j++)*/
		if (j == sizeof(double))
		{
			/* no matching byte found; no permutation exists */
#if (0)
			ALERT("i = %d, c3 = %02x",i,c3,0,0);
			ALERT("localDoubles[0].bits = %08x%08x, foreignDoubles[0].bits = %08x%08x",
				  localDoubles[0].bits[0],localDoubles[0].bits[1],
				  foreignDoubles[0].bits[0],foreignDoubles[0].bits[1]);
#endif /*0*/
			DoPermute = unmatchedDoubleByte;
			goto errorExit;
		} /*if (j == sizeof(double))*/
	} /*for (i = 0; i < sizeof(double); i++)*/
	
	/* find permutation for longs*/
	for (i = 0; i < sizeof(long); i++)
	{
		c3 = localLongs[0].bytes[i];
		for (j = 0; j < sizeof(long); j++)
		{
			if (foreignLongs[0].bytes[j] == c3)
			{
				if (LongPermutation[j] < sizeof(long))
				{
					/* duplicate byte; ambiguous permutation */
					DoPermute = duplicateLongByte;
					goto errorExit;
				}
				LongPermutation[j] = i;
				break;
			}
		} /*for (j = 0; j < sizeof(long); j++)*/
		if (j == sizeof(long))
		{
			/* no matching byte found; no permutation exists */
			DoPermute = unmatchedLongByte;
			goto errorExit;
		}
	} /*for (i = 0; i < sizeof(long); i++)*/

	/* Check to make sure permutations work for other numbers as well */
	for (iNumber = 1; iNumber < nSampleDoubles; iNumber++)
	{
		for (i = 0; i < sizeof(double); i++)
		{
			if (localDoubles[iNumber].bytes[i] !=
				foreignDoubles[iNumber].bytes[(int) DoublePermutation[i]])
			{
				DoPermute = badDoublePermutation;
				goto errorExit;
			}
		}
	} /*for (iNumber = 1; iNumber < nSampleDoubles; iNumber++)*/
	for (iNumber = 1; iNumber < nSampleLongs; iNumber++)
	{
		for (i = 0; i < sizeof(long); i++)
		{
			if (localLongs[iNumber].bytes[i] !=
				foreignLongs[iNumber].bytes[(int) LongPermutation[i]])
			{
				DoPermute = badLongPermutation;
				goto errorExit;
			}
		}
	} /*for (iNumber = 1; iNumber < nSampleLongs; iNumber++)*/

#if (0)
	sprintf(OUTSTR,"DoublePermutation = %d %d %d %d %d %d %d %d",
			DoublePermutation[0],DoublePermutation[1],DoublePermutation[2],
			DoublePermutation[3],DoublePermutation[4],DoublePermutation[5],
			DoublePermutation[6],DoublePermutation[7]);
	ALERT("%s",OUTSTR,0,0,0);
	sprintf(OUTSTR,"LongPermutation = %d %d %d %d",
			LongPermutation[0],LongPermutation[1],LongPermutation[2],
			LongPermutation[3]);
	ALERT("%s",OUTSTR,0,0,0);
	*OUTSTR = '\0';
#endif /*0*/

#if (1)
	DoPermute = yesPermute;
#else
	DoPermute = noPermute;
#endif
	/* fall through*/
  errorExit:
	;
} /*setPermutation()*/

/*
  980728  changed so that LIMITS is expected to be a CHARACTER vector
          when !Pre407
  001104  added argument foreignOK
*/
#define NLIMITS  4

static long checkMachine(Symbolhandle globals, int foreignOK)
{
	long                icomp;
	Symbolhandle        symh;
	long                limitsType = (Pre407Save) ? LONG : CHAR;
	long                retvalue = okPlatform;
	WHERE("checkMachine");

	for (icomp = 0;icomp < NCOMPS(globals);icomp++)
	{
		symh = COMPVALUE(globals,icomp);

		if (strcmp(NAME(symh), "LIMITS") == 0 && TYPE(symh) == limitsType &&
			isVector(symh) && DIMVAL(symh, 1) == NLIMITS)
		{
			long       limits[NLIMITS];
			long       maxdims, maxvars, wordsperterm, varsperword;

			if (Pre407Save)
			{
				symbol2Long(limits, "LIMITS", NLIMITS, symh);
			} /*if (Pre407Save)*/
			else
			{
				int        i;
				char      *place = STRINGPTR(symh);
				
				for (i = 0; i < NLIMITS; i++)
				{
					if (sscanf(place, "%ld", limits + i) != 1)
					{
						sprintf(OUTSTR,
								"ERROR: information on limits unreadable");
						goto errorExit;
					}
					place = skipStrings(place, 1);
				} /*for (i = 0; i < NLIMITS; i++)*/
			} /*if (Pre407Save){}else{}*/

			maxdims = limits[0];
			maxvars = limits[1];
			wordsperterm = limits[2];
			varsperword = limits[3];

			if (maxdims != MAXDIMS)
			{
				sprintf(OUTSTR,
						"ERROR: file was saved by incompatible version of MacAnova");
				goto errorExit;
			}
			else if (wordsperterm != WORDSPERTERM ||
					 varsperword != VARSPERWORD)
			{
				retvalue = badGlm;
			}
		}
		else if (strcmp(NAME(symh), "PLATFORM") == 0 && TYPE(symh) == CHAR &&
				 isScalar(symh))
		{
			char        *platform = STRINGPTR(symh);
			
			if (strcmp(platform,PLATFORM) != 0
#ifdef PLATFORM_ALT1
				&& strcmp(platform, PLATFORM_ALT1) != 0
#endif /*PLATFORM_ALT1*/
#ifdef PLATFORM_ALT2
				&& strcmp(platform, PLATFORM_ALT2) != 0
#endif /*PLATFORM_ALT2*/
#ifdef PLATFORM_ALT3
				&& strcmp(platform, PLATFORM_ALT3) != 0
#endif /*PLATFORM_ALT3*/
#ifdef PLATFORM_ALT4
				&& strcmp(platform, PLATFORM_ALT4) != 0
#endif /*PLATFORM_ALT4*/
#ifdef PLATFORM_ALT5
				&& strcmp(platform, PLATFORM_ALT5) != 0
#endif /*PLATFORM_ALT5*/
#ifdef PLATFORM_ALT6
				&& strcmp(platform, PLATFORM_ALT6) != 0
#endif /*PLATFORM_ALT6*/
#ifdef PLATFORM_ALT7
				&& strcmp(platform, PLATFORM_ALT7) != 0
#endif /*PLATFORM_ALT7*/
#ifdef PLATFORM_ALT8
				&& strcmp(platform, PLATFORM_ALT8) != 0
#endif /*PLATFORM_ALT8*/
				)
			{
				if (Binary)
				{
					if (!foreignOK)
					{
						sprintf(OUTSTR,
							"ERROR: Binary file saved by version of Macanova compiled for %s",
								platform);
						goto errorExit;
					}
					DoPermute = maybePermute;
				} /*if (Binary)*/
				sprintf(OUTSTR,
					"WARNING: File saved by version of Macanova compiled for %s",
						platform);
				putErrorOUTSTR();
				ForeignFile = 1;
			} /*if (strcmp(platform,PLATFORM) != 0)*/
		}
		else if (strcmp(NAME(symh), "NUMBERS") == 0 && TYPE(symh) == CHAR &&
				 isMatrix(symh, (long *) 0))
		{
			if (DoPermute == maybePermute)
			{
				setPermutation(symh);
				if (DoPermute < noPermute)
				{
					sprintf(OUTSTR,
							"WARNING: can't translate foreign numbers; error number = %d",
							DoPermute);
					putErrorOUTSTR();
					DoPermute = noPermute;
				}
				SwapNewlines = yesSwap;
			}
			else
			{
				SwapNewlines = noSwap;
			}
		}
	} /*for (icomp = 0;icomp < NCOMPS(globals);icomp++)*/
	return (retvalue);

  errorExit:
	putErrorOUTSTR();
	putOutMsg("       Restore aborted");

	return (badPlatform);
} /*checkMachine()*/
#endif /*PLATFORM*/

/* Modified from listObject in list.c */
static void listRestoredSymbol(Symbolhandle symh, int restored)
{
	char            *name;
	char            *action;
	int              type;
	long             i, nclass;
	WHERE("listRestoredSymbol");

	name = NAME(symh);
	type = TYPE(symh);

	action = (restored) ? "Restoring" : "Reading  ";
	sprintf(OUTSTR,"%s %-15s %-6s", action, name,typeName(type));
	myprint(OUTSTR);

	if (type == REAL || type == LOGIC || type == CHAR || type == STRUC)
	{
		for (i = 1; i <= NDIMS(symh); i++)
		{
			sprintf(OUTSTR, " %-5ld", DIMVAL(symh,i));
			myprint(OUTSTR);
		}
		if ((nclass = NCLASS(symh)) > 0)
		{
			sprintf(OUTSTR, " FACTOR with %ld levels",nclass);
			myprint(OUTSTR);
		}
	}
	if (isLocked(symh))
	{
		myprint(" (locked)");
	}
	myeol();
	*OUTSTR = '\0';
} /*listRestoredSymbol()*/


static int isFolderSymbol(Symbolhandle symh)
{
	int         i;
	char       *name = NAME(symh);
	long        type = TYPE(symh);

	for (i = 0; i < nFolderSymbols; i++)
	{
		if ((i < nFolderScalars && isCharOrString(symh) ||
			 i >= nFolderScalars && type == CHAR && isVector(symh)) &&
			strcmp(name, FolderSymbolNames[i]) == 0)
		{
			return (i);
		}
	}
	return (-1);
}

/*
	Make scratch copies of folder/file name variables and stash
	them in FolderSymbols;  they can now be deleted if necessary
*/
static int preserveFolderSymbols(void)
{
	int         i;

	for (i = 0; i < nFolderSymbols; i++)
	{
		Symbolhandle     folderSymbol = Lookup(FolderSymbolNames[i]);

		if (i < nFolderScalars && isCharOrString(folderSymbol) ||
			i >= nFolderScalars &&
				isVector(folderSymbol) && TYPE(folderSymbol) == CHAR)
		{
			FolderSymbols[i] = Install(SCRATCH, CHAR);
			if (FolderSymbols[i] == SymbolError ||
				!Copy(folderSymbol, FolderSymbols[i]))
			{
				return (0);
			}
			Setname(FolderSymbols[i], SCRATCH);
		}
		else
		{
			FolderSymbols[i] = SymbolZero;
		}
	} /*for (i = 0; i < nFolderSymbols; i++)*/
	return (1);
} /*preserveFolderSymbols()*/

/*
	For any FolderSymbol that is still in the list,
	reinstall it if there is no variable by that name; otherwise
	remove it
*/
static void restoreFolderSymbols(void)
{
	int        i;

	for (i = 0; i < nFolderSymbols; i++)
	{
		if (FolderSymbols[i] != SymbolZero)
		{
			if (Lookup(FolderSymbolNames[i]) == SymbolZero)
			{
				Cutsymbol(FolderSymbols[i]);
				Setname(FolderSymbols[i], FolderSymbolNames[i]);
				Addsymbol(FolderSymbols[i]);
			}
			else
			{
				Removesymbol(FolderSymbols[i]);
			}
			FolderSymbols[i] = SymbolZero;
		}
	} /*for (i = 0; i < nFolderSymbols; i++)*/
} /*restoreFolderSymbols()*/

/*
	010411 added to create structure with any folder variables
	       from the file being restored
*/
static int bundleFolderSymbols(int listRestored)
{
	int        ncomps = 0, i;

	for (i = 0; i < nFolderSymbols; i++)
	{
		if (SavedFolderSymbols[i] != SymbolZero)
		{
			ncomps++;
		}
	}
	if (ncomps > 0)
	{
		char             *savedFoldersName = FolderSymbolNames[nFolderSymbols];
		Symbolhandle      oldBundle = Lookup(savedFoldersName);
		Symbolhandle      bundle;
		int               jcomp = 0;
		
		bundle = StrucInstall(SCRATCH, ncomps);
		if (bundle == SymbolError)
		{
			return (0);
		}
		Removesymbol(oldBundle);
		Cutsymbol(bundle);
		Setname(bundle, savedFoldersName);
		Addsymbol(bundle);

		for (i = 0; i < nFolderSymbols; i++)
		{
			Symbolhandle    symh = SavedFolderSymbols[i];
			
			if (symh != SymbolZero)
			{
				Cutsymbol(symh);
				COMPVALUE(bundle, jcomp) = symh;
				jcomp++;
				Setname(symh, FolderSymbolNames[i]);
				SavedFolderSymbols[i] = SymbolZero;
			} /*if (symh != SymbolZero)*/
		} /*for (i = 0; i < nFolderSymbols; i++)*/
		if (listRestored)
		{
			sprintf(OUTSTR,
					"File/folder name variables from file saved in structure %s",
					savedFoldersName);
			putOUTSTR();
		} /*if (listRestored)*/
	} /*if (ncomps > 0)*/
	return (1);
} /*bundleFolderSymbols()*/

/*
   Usage: restore(fileName [,delete:T] [,list:T])
	delete:T suppresses deletion of symbols that are not being replaced
	list:T lists symbols as they are restored.  Primarily for debugging
	purposes

	960305 Modified so that original contents of Symbol "HOME" are saved
	are restored so that restoring of a save file made on a different
	system or with a different home directory will not change HOME.

	960419 Modified so that default value of delete is taken from global
	DEFAULTRESTOREDEL which is settable by setoptions.

	960502 Changed fopen() to fmyopen()
	960503 Changed macOpen() to macFindFile()
	970409 Modified so that a restore of an ascii file will work whether the
           files line separators are CR, LF or CR/LF.  New function
           restoregetc() to handle reading.
	970506 Used TEXTREADMODE and BINARYREADMODE to symbolize modes for fmyopen()
	980623 Recognizes symbol HISTRY and resets the history
	980715 Uses ASCIIHEADER7 and BINARYHEADER7 by default
	980821 When in batch mode, restore() does not change current prompt
	990325 On Macintosh, 'ttro' is now an acceptable type for a restore file
	990422 Replaced "STRMODEL" by STRMODELNAME
	990824 GRAPHWINDOWS not restore on Macintosh in non-interactive mode
    010411 Does not restore HOME, DATAPATHS or MACROFILES; originals are
	       kept;  any in the file are combined in structure SAVEDNAMES 
    010521 Changed calling sequence to macFindFile()
	010524 Changed macFindFile() to mvFindFile() and dropped wxFindFile()
*/

Symbolhandle    restore(Symbolhandle list)
{
	Symbolhandle    arg1, argKey, symh, next, symh1;
	char            instuff[HEADERLENGTH+1];
	long            i;
	long            nargs = NARGS(list);
	int             deleteCurrent = -1;
	int             restoreHistory = -1;
	int             listRestored = -1;
	int             foreignOK = -1;
	int             restoreOptions = -1;
	int             forceDelete = 0, restoreGlobs = 1;
	int             fullSave;
	char           *keyword;
	char            whenSaved[TIMELENGTH];
	char           *fname, name[NAMELENGTH+1];
	WHERE("restore");

	*OUTSTR = '\0';
	RESTOREFILE = (FILE *) 0;
	whenSaved[0] = '\0';
	DoPermute = noPermute;
	SwapNewlines = noSwap;
	ForeignFile = 0;

	arg1 = COMPVALUE(list, 0);

	if (nargs == 1 && arg1 == EmptyArgument)
	{
		badNargs(FUNCNAME, -1 + atLeastNArgs);
		goto errorExit;
	}

	if (!argOK(arg1, 0, 0))
	{
		goto errorExit;
	}
	if (!isCharOrString(arg1))
	{
		notCharOrString("workspace file name");
		goto errorExit;
	}

	for (i = 1; i < nargs; i++)
	{
		/*
		  Arg must be logical keyword ('delete', 'list', 'foreignok' or
		  'history' or simply T or F (equivalent to delete:T or delete:F)
		*/
		long        logValue;
		char       *conflict =
		  "ERROR: don't know which of conflicting values for '%s' to use";

		argKey = COMPVALUE(list,i);
		if (!argOK(argKey, 0, i+1))
		{
			goto errorExit;
		}
		keyword = isKeyword(argKey);
		if (keyword != (char *) 0)
		{
			if (strncmp(keyword, "del", 3) != 0 &&
#ifdef SAVEHISTORY
				strncmp(keyword, "his", 3) != 0 &&
#endif /*SAVEHISTORY*/
				strcmp(keyword, "foreignok") != 0 &&
				strncmp(keyword, "option",6) != 0 &&
				strncmp(keyword, "lis", 3) != 0)
			{
				badKeyword(FUNCNAME, keyword);
				goto errorExit;
			}
		}
		if (!isTorF(argKey))
		{
			notTorF((keyword) ? keyword : (char *) 0);
			goto errorExit;
		}
		
		logValue = (DATAVALUE(argKey, 0) != 0.0);
		if (!keyword || keyword[0] == 'd')
		{
			if (deleteCurrent >= 0 && deleteCurrent != logValue)
			{
				sprintf(OUTSTR, conflict, "delete");
				goto errorExit;
			}
			forceDelete = deleteCurrent = logValue;
		}
		else if (keyword[0] == 'l')
		{
			if (listRestored >= 0 && listRestored != logValue)
			{
				sprintf(OUTSTR, conflict, "list");
				goto errorExit;
			}
			listRestored = logValue;
		}
#ifdef SAVEHISTORY
		else if (keyword[0] == 'h')
		{
			if (restoreHistory >= 0 && restoreHistory != logValue)
			{
				sprintf(OUTSTR, conflict, keyword);
				goto errorExit;
			}
			restoreHistory = logValue;
		}
#endif /*SAVEHISTORY*/
		else if (keyword[0] == 'o')
		{
			if (restoreOptions >= 0 && restoreOptions != logValue)
			{
				sprintf(OUTSTR, conflict, keyword);
				goto errorExit;
			}
			restoreOptions = logValue;
		}
		else if (keyword[0] = 'f')
		{
			if (foreignOK >= 0 && foreignOK != logValue)
			{
				sprintf(OUTSTR, conflict, keyword);
				goto errorExit;
			}
			foreignOK = logValue;
		}
	} /*for (i = 1; i < nargs; i++)*/
	deleteCurrent = (deleteCurrent < 0) ? DEFAULTRESTOREDEL : deleteCurrent;
	listRestored = (listRestored < 0) ? 0 : listRestored;
	restoreHistory = (restoreHistory < 0) ? 1 : restoreHistory;
	restoreOptions = (restoreOptions < 0) ? 1 : restoreOptions;
	foreignOK = (foreignOK < 0) ? 0 : foreignOK;
	
#ifdef SAVEHISTORY
	if (!(ISATTY & ITTYIN))
	{
		restoreHistory = 0;
	}
#endif /*SAVEHISTORY*/

	fname = STRINGPTR(arg1);

#ifdef HASFINDFILE

	fname = mvFindFile(fname, "Specify the restore file", NullString,
					   mvReadit | mvTextFile | mvROTextFile |
					   mvAsciiSaveFile | mvSaveFile,
					   getMvDirInfo(restoreVol));
	if (fname == (char *) 0)
	{
		sprintf(OUTSTR, "WARNING: %s() cancelled", FUNCNAME);
		goto errorExit;
	}
	if (STRINGVALUE(arg1, 0) == '\0')
	{
		if (!isMvDirInfoSet(keepVol))
		{
			copyMvDirInfo(keepVol, getMvDirInfo(restoreVol));
		}
		if (!isMvDirInfoSet(windowVol))
		{
			copyMvDirInfo(windowVol, getMvDirInfo(restoreVol));
		}
	}
#endif /*HASFINDFILE*/

	fname = expandFilename(fname);
	if (fname == (char *) 0 || !isfilename(fname))
	{
		goto errorExit;
	}

	KEEPFILE = (FILE *) 0;
	mydisphandle(KEEPFILENAME);
	
	/* save the restore file name as save file name */
	if ((KEEPFILENAME = mygethandle(strlen(fname)+1)) != (char **) 0)
	{
		strcpy(*KEEPFILENAME, fname);
#ifdef MACINTOSH
		copyMvDirInfo(keepVol, getMvDirInfo(restoreVol));
#endif /*MACINTOSH*/
	}

	RESTOREFILE = fmyopen(fname, BINARYREADMODE);
	LastRestoreWasCr = 0;

	if (RESTOREFILE == (FILE *) 0 )
	{
		sprintf(OUTSTR,"ERROR: cannot open file %s for restore", fname);
		goto errorExit;
	}

	if (ModelStackDepth > 1)
	{
		/* Clear model information stack without recreating side effects */
		(void) doPopModelInfo(-1);
	}
	
	PadChar = ' ';
	if (!restoreNString(HEADERLENGTH, instuff))
	{
		sprintf(OUTSTR,
				"ERROR: Does not appear to be workspace file;header = %s",
				instuff);
		goto errorExitWithClose;
	}

	if (strcmp(instuff, BINARYHEADER1) == 0)
	{
		ASCII = SAVEBINARY1;
	}
	else if (strcmp(instuff, BINARYHEADER4) == 0)
	{
		ASCII = SAVEBINARY4;
	}
	else if (strcmp(instuff, BINARYHEADER5) == 0)
	{
		ASCII = SAVEBINARY5;
	}
	else if (strcmp(instuff, BINARYHEADER6) == 0)
	{
		ASCII = SAVEBINARY6;
	}
	else if (strcmp(instuff, BINARYHEADER7) == 0)
	{
		ASCII = SAVEBINARY7;
	}
	else if (strcmp(instuff, ASCIIHEADER1) == 0)
	{
		ASCII = SAVEASCII1;
	}
	else if (strcmp(instuff, ASCIIHEADER2) == 0)
	{
		ASCII = SAVEASCII2;
	}
	else if (strcmp(instuff, ASCIIHEADER3) == 0)
	{
		ASCII = SAVEASCII3;
	}
	else if (strcmp(instuff, ASCIIHEADER4) == 0)
	{
		ASCII = SAVEASCII4;
	}
	else if (strcmp(instuff, ASCIIHEADER5) == 0)
	{
		ASCII = SAVEASCII5;
	}
	else if (strcmp(instuff, ASCIIHEADER6) == 0)
	{
		ASCII = SAVEASCII6;
	}
	else if (strcmp(instuff, ASCIIHEADER7) == 0)
	{
		ASCII = SAVEASCII7;
	}
	else if (strncmp(instuff, ASCIIHEADER4, HEADERLENGTH-3) == 0 ||
			 strncmp(instuff, BINARYHEADER4, HEADERLENGTH-4) == 0)
	{
		sprintf(OUTSTR,
				"ERROR: Workspace file cannot be restored by this version of MacAnova");
	}
	else
	{
		sprintf(OUTSTR,"ERROR: file is not a MacAnova workspace file");
	}
	if (*OUTSTR)
	{
		goto errorExitWithClose;
	}
	
	if (!Binary && ASCII != SAVEASCII1)
	{
		/* close file and reopen in text mode */
		fclose(RESTOREFILE);
		RESTOREFILE = fmyopen(fname, TEXTREADMODE);
		if (RESTOREFILE == (FILE *) 0 )
		{ /*be paranoid*/
			sprintf(OUTSTR,"ERROR: cannot open file %s for restore", fname);
			goto errorExit;
		}
		LastRestoreWasCr = 0;
		/* skip identifying header */
		if(!restoreNString(HEADERLENGTH,instuff))
		{
			goto errorExitWithClose;
		}
		PadChar = '\n';
	} /*if (!Binary && ASCII != SAVEASCII1)*/

	if (!Pre407Save)
	{
		if (!restoreName(instuff))
		{
			goto errorExitWithClose;
		}
		if (strcmp(instuff, FULLSAVE) == 0)
		{
			fullSave = 1;
		}
		else if (strcmp(instuff, PARTIALSAVE) == 0)
		{
			fullSave = 0;
			deleteCurrent = forceDelete;
		}
		else
		{
			goto incompleteRestore;
		}
	}
	else
	{
		fullSave = 1;
	}
	
	sprintf(OUTSTR, "Restoring %sworkspace from file ",
			(fullSave) ? NullString : "partial ");
	myprint(OUTSTR);
	putOutMsg(fname);
	*OUTSTR = '\0';

	if (ASCII == SAVEASCII1)
	{
		sprintf(OUTSTR,
				"WARNING: MacAnova 2.4 or earlier asciisave file ; may include old style macros");
		putErrorOUTSTR();
	}

	/*
		Save copies of variables that are folder or directory names
		in scratch symbol table and in FolderSymbols
	*/
	if (!preserveFolderSymbols())
	{
		goto errorExitWithClose;
	}

	/* delete current workspace */
	if (deleteCurrent)
	{
		/* first remove items in the argument list */
		for (i = 0;i < nargs;i++)
		{
			RemoveLockedsymbol(COMPVALUE(list,i));
			COMPVALUE(list,i) = SymbolZero;
		}

		/*
			Clear regular symbol table
			Everything necessary to save, should have already been
			saved or copied
		*/
		symh = Firstsymbol(0);
		while (symh != SymbolZero)
		{
			next = Nextsymbol(symh,0);
			RemoveLockedsymbol(symh); /* will not remove builtins */
			symh = next;
		} /*while (symh != SymbolZero)*/

		/*
			Now clear scratch, UNDEF and keyword symbol table
			This contains any saved file/folder name variables as
			well as list so	you have to make sure they are not cleared
		*/
		symh = Firstsymbol(1);
		while (symh != SymbolZero)
		{
			next = Nextsymbol(symh,1);
			if (symh != list)
			{/* do not remove list; that is done in mainpars.y */
				int      j;

				for (j = 0; j < nFolderSymbols && FolderSymbols[j] != symh; j++)
				{
					;
				}
				if (j == nFolderSymbols)
				{
					/* Remove if not a saved folder symbol*/
					Removesymbol(symh);
				} /*if (j == nFolderSymbols)*/
			} /*if (symh != list)*/
			symh = next;
		} /*while (symh != SymbolZero)*/
		clearGlobals();
	} /*if (deleteCurrent)*/

/* Make sure correct info is in VERSION */
	if (VERSION_ID != (char **) 0)
	{
		char         *name = "VERSION";
		Symbolhandle  symhVersion  = Lookup(name);
		
		RemoveLockedsymbol(symhVersion);
		
		symh = Install(name, CHAR);
		TMPHANDLE = myduphandle(VERSION_ID);
		if(symh == SymbolError || TMPHANDLE == (char **) 0)
		{
			goto errorExitWithClose;
		}
		setSTRING(symh, TMPHANDLE);
		setNDIMS(symh,1);
		setDIM(symh,1,1);
		setLocked(symh);
	}  /*if (deleteCurrent){}else{}*/

	while (1)
	{
		int            ch;

		while ((ch = restoregetc(RESTOREFILE)) == ' ' || ch == PadChar)
		{
			;
		}
		if (ch == EOF)
		{
			break;
		}
		ungetc((char) ch, RESTOREFILE);

		symh = restoresym(); /* is already in Symbol table */
		if (symh == SymbolError)
		{
			fclose(RESTOREFILE);
			goto incompleteRestore;
		}

		strcpy(name, NAME(symh));
		if(strcmp(name, "T") == 0 || strcmp(name, "F") == 0 ||
		   strcmp(name, "HELP") == 0 || strcmp(name, "help") == 0 ||
		   strcmp(name, "VERSION") == 0)
		{ /* all but the last might be saved by MacAnova2.4 */
			RemoveLockedsymbol(symh);
		}
		else if (strcmp(name, OPTIONS) == 0)
		{
			if (!restoreOptions)
			{
				if (listRestored)
				{
					putOutMsg("Not restoring option values");
				}
			} /*if (!restoreOptions)*/
			else
			{
				promptType       currentPrompt;
#ifdef SCROLLABLEWINDOW
				long             currentMatchDelay = MatchDelay;
#endif /*SCROLLABLEWINDOW*/

#if (0) /* no longer any comment on unrecognized options*/
				long             nOptions = NCOMPS(symh);
			
				for (i = 0;i < nOptions;i++)
				{
					keyword = NAME(COMPVALUE(symh,i));
					if(!isOption(keyword))
					{
						break;
					}
				} /*for (i = 0;i < nOptions;i++)*/
				if (i < nOptions)
				{
					sprintf(OUTSTR,
							"WARNING: unrecognized option name %s found during restore",
							keyword);
					putErrorOUTSTR();
				}
#endif /*0*/
				if (listRestored)
				{
					putOutMsg("Restoring option values");
				}

				strcpy(currentPrompt, PROMPT);

				/* arg2 == 1 means don't complain about unrecognized options*/
				setOpt(symh, 1);

				/* if reading batch file, don't change the current prompt */
				if (BDEPTH > 0)
				{
					strcpy(PROMPT, currentPrompt);
				}
#ifdef SCROLLABLEWINDOW
				/* 
				   restore current MatchDelay when restoring from a
				   foreign file
				*/
				if (ForeignFile)
				{
					MatchDelay = currentMatchDelay;
				}
#endif /*SCROLLABLEWINDOW*/
			} /*if (!restoreOptions){}else{}*/
			Removesymbol(symh);
		}
#ifdef PLATFORM
		else if(strcmp(name,MACHINE) == 0)
		{
			int       platformStatus = checkMachine(symh, foreignOK);

			Removesymbol(symh);
			if (platformStatus == badPlatform)
			{
				goto errorExitWithClose;
			}
			if (platformStatus == badGlm)
			{
				restoreGlobs = 0;
			}
		}
#endif /*PLATFORM*/
		else if(strcmp(name, GLOBALS) == 0)
		{
			if(restoreGlobs && !deleteCurrent)
			{
				clearGlobals();
			}

			if(restoreGlobs)
			{
				restoreGlobals(symh);
				if (listRestored)
				{
					putOutMsg("Restoring private glm values");
				}

				if(NVARS < 0)
				{
					clearGlobals();
				}
			}
			else if(PRINTWARNINGS)
			{
				putOutErrorMsg("WARNING: incompatible internal representation of model on workspace file");
				putOutMsg("         Model information not restored");
			}
			Removesymbol(symh);
		}
		else if(strcmp(name, SAVETIME) == 0)
		{
			strncpy(whenSaved,STRINGPTR(symh),TIMELENGTH-1);
			whenSaved[TIMELENGTH-1] = '\0';
			Removesymbol(symh);
		}
#ifdef SAVEHISTORY
		else if (strcmp(name, HISTRY) == 0)
		{
			if (restoreHistory)
			{
				if (listRestored)
				{
					putOutMsg("Restoring history of commands");
				}
				setSomeHistory(DIMVAL(symh, 1), STRING(symh));
			}
			else if (listRestored)
			{
				putOutMsg("Not restoring history of commands");
			}
			Removesymbol(symh);
		}
#endif /*SAVEHISTORY*/
		else if (strcmp(name, GRPHWINDOWS) == 0)
		{
			int        reply;
			
			if (listRestored)
			{
#ifdef MACINTOSH
				if (!UseWindows)
				{
					putOutMsg("GRAPHWINDOWS not restored in non-interactive mode");
				}
				else
#endif /*MACINTOSH*/
				{
					putOutMsg("Restoring GRAPHWINDOWS");
				}
			} /*if (listRestored)*/
			
#ifdef MACINTOSH
			if (!UseWindows)
			{
				/* don't restore GRAPHWINDOWS in non-interactive mode */
				reply = 1;
			}
			else
#endif /*MACINTOSH*/
			{
				reply = restoreGraphWindows(symh);
			}
			
			Removesymbol(symh);
			Remove(LastGraphName);

#if defined(MACINTOSH) || defined (WXWIN) || defined(SKEL_CARAPACE)
#ifdef MACINTOSH
			if (UseWindows)
#endif /*MACINTOSH*/
			{
				showCurrentCmdWindow();
			}
#endif /*MACINTOSH || WXWIN || SKEL_CARAPACE*/
			if (!reply)
			{
				goto errorExitWithClose;
			}
		}
		else
		{ /* ordinary symbol */
			int      whichFolderSymbol;
			
			if(!deleteCurrent)
			{ /* delete symbol if it exists */
				Cutsymbol(symh); /* make sure it won't be found by Lookup */
				symh1 = Lookup(name);
				if (symh1 != SymbolZero)
				{
					if (isLocked(symh1))
					{
						if (TYPE(symh) != REAL || TYPE(symh1) != REAL ||
							!isScalar(symh) || !isScalar(symh1) ||
							DATAVALUE(symh, 0) != DATAVALUE(symh1, 0) ||
							NDIMS(symh) != 1 || NDIMS(symh1) != 1)
						{
							
							sprintf(OUTSTR,
									"WARNING: variable %s is locked; value from file not restored",
									name);
							putErrorOUTSTR();
						}
						Delete(symh);
						continue;
					}
					else
					{
						Removesymbol(symh1);
					}
				}
				Addsymbol(symh); /* put back in symbol table if not deleted */
			} /*if(!deleteCurrent)*/
			
			/* Make sure folder/file name variables don't get changed */
			whichFolderSymbol = isFolderSymbol(symh);
			if (whichFolderSymbol >= 0)
			{
				if (FolderSymbols[whichFolderSymbol] != SymbolZero)
				{
					/*
						This folder/file symbol previously existed and will not
						be replaced.  The version from the workspace file
						is saved in SavedFolderSymbols as a scratch variable
					*/
					Cutsymbol(symh);
					Setname(symh, SCRATCH);
					Addsymbol(symh);
					SavedFolderSymbols[whichFolderSymbol] = symh;

					/* Reinstall original FolderSymbol */
					symh = FolderSymbols[whichFolderSymbol];
					FolderSymbols[whichFolderSymbol] = SymbolZero;
					/* move from scratch symbol table to regular symbol table*/
					Cutsymbol(symh);
					Setname(symh, FolderSymbolNames[whichFolderSymbol]);
					Addsymbol(symh);
				} /*if (FolderSymbols[whichFolderSymbol] != SymbolZero)*/
				else
				{
					whichFolderSymbol = -1;
				}
			} /*if (whichFolderSymbol >= 0)*/

			if (listRestored)
			{
				listRestoredSymbol(symh, whichFolderSymbol < 0);
			}
		}
	} /* while (1) */
	fclose(RESTOREFILE);
	restoreFolderSymbols();
	if (!bundleFolderSymbols(listRestored))
	{
		goto errorExitWithClose;
	}
	if (*whenSaved)
	{
		sprintf(OUTSTR, "Workspace saved %s", whenSaved);
		putOUTSTR();
	}

#ifdef SEGMENTED
	UNLOADSEG(restoreGraph);
#endif /*SEGMENTED*/

	return (NULLSYMBOL);

  incompleteRestore:
	strcpy(OUTSTR, "ERROR: incomplete restore");

  errorExitWithClose:
	if (RESTOREFILE != (FILE *) 0)
	{
		fclose(RESTOREFILE);
	}
	
  errorExit:
	putErrorOUTSTR();
	
	restoreFolderSymbols();

#ifdef SEGMENTED
	UNLOADSEG(restoreGraph);
#endif /*SEGMENTED*/
	
	return (SymbolError);

} /*restore()*/


/*
  scans for integer followed by a PadChar; if the PadChar is not present, it is
  an error (return 0)
*/

long restoreLong(long * longValue)
{
	long          c;
	long          sign = 1, value = 0;

/*
	while((c = restoregetc(RESTOREFILE)) == ' ' || c == PadChar)
	{
		;
	}
*/
	c = restoregetc(RESTOREFILE);

	if(c == EOF || (!isdigit(c) && c != '-'))
	{
		return (0);
	}
	if(c == '-')
	{
		sign = -1;
	}
	else
	{
		value = c - '0';
	}
	while(isdigit((char) (c = restoregetc(RESTOREFILE))))
	{
		value = 10*value + (c - '0');
	}
	*longValue = sign*value;
	return (c == PadChar);
} /*restoreLong()*/

long restoreName(char * name)
{
	int         i, c;

/*
	while((c = restoregetc(RESTOREFILE)) == ' ' || c == PadChar)
	{
		;
	}
	if(c == EOF)
	{
		return (0);
	}
	*name++ = c;
*/
	for(i=0;i<NAMELENGTH;i++)
	{
		if((c = restoregetc(RESTOREFILE)) == EOF)
		{
			return (0);
		}

		if(c == PadChar)
		{
			break;
		}
		*name++ = (char) c;
	} /* for i */
	*name++ = '\0';
	return (i < NAMELENGTH || restoregetc(RESTOREFILE) == PadChar);
} /*restoreName()*/

/*
  restoreString() should be used only when one is sure the string to be
  read will fit in string[].  Otherwise, use restoreNString()
*/
long restoreString(char * string)
{
	long        length;

	*string = '\0';
	if(!restoreLong(&length) || !restoreNString(length,string))
	{
		return (0);
	}

	return (1);

} /*restoreString()*/

/*
  011002 modified so it recognizes '\n' and '\r'
*/
long restoreNString(long length, char * string)
{
	int            ch, nc, s;

	while(length-- > 0)
	{
		if((ch = restoregetc(RESTOREFILE)) == EOF)
		{
			return (0);
		}
		if(ASCII > SAVEASCII3 && ch == '\\')
		{
			if((ch = restoregetc(RESTOREFILE)) == EOF)
			{
				return (0);
			}
			if(ch == 't')
			{
				ch = '\t';
			}
			else if (ch == 'n')
			{
				ch = '\n';
			}
			else if (ch == 'r')
			{
				ch = '\r';
			}
			else if(isdigit(ch) && ch <= '3')
			{
				s = ch - '0';
				for(nc = 1; nc < 3; nc++)
				{
					if((ch = restoregetc(RESTOREFILE)) == EOF ||
					   !isdigit(ch) || ch > '7')
					{
						goto errorExit;
					}
					s = 8*s + (ch - '0');
				}
				ch = s;
			}
		} /*if(ASCII > SAVEASCII3 && ch == '\\')*/
		*string++ = ch;
	} /*while(length-- > 0)*/
	*string = '\0';

	return (restoregetc(RESTOREFILE) == PadChar);

  errorExit:
	return (0);
} /*restoreNString()*/

long restoreDouble(double * value)
{
	int         c, length = 0;
	char       *end;
	char        buffer[100];
	WHERE("restoreDouble");

	while((c = restoregetc(RESTOREFILE)) == ' ' || c == PadChar)
	{
		;
	}
	if(c == EOF)
	{
		return (0);
	}
	buffer[0] = c;

	length = 1;
	while((c = restoregetc(RESTOREFILE)) != EOF && c != PadChar)
	{
		buffer[length++] = c;
	}

	if(c == EOF)
	{
		return (0);
	}
	buffer[length] = '\0';
	if (ASCII >= SAVEASCII5 && buffer[0] == '?')
	{
		setMissing(*value);
		end = buffer + length;
	}
	else
	{
		*value = mystrtod(buffer, &end);
		if (ASCII < SAVEASCII5 && *value == OLDMISSING)
		{
			setMissing(*value);
		}
	}

	return (end - buffer == length);
} /*restoreDouble()*/

/*
  The handle restore must be attached to a symbol, even
  if it is a fake symbol.

  If the appropriate component (DATA or STRING) of symh is
  non-NULL, it is assumed that a handle was allocated in the
  calling program so no initial allocation is needed.

  990502 restoreHandle() created with code moved from restoresym()
*/
long restoreHandle(Symbolhandle symh, int binary)
{
	double            **doubledata;
	long              **longdata;
	char              **chardata;
	long                i, size, spaceleft, ch, nchar, needed;
	long                type = TYPE(symh) & TYPEMASK;
	long                nvals = symbolSize(symh);
	WHERE("restoreHandle");

	if (nvals > 0)
	{
		/*
		  Check of nvals allows for possibility of null symbol on save file
		  This shouldn't happen for type 5 files or later
		*/

		if (type == MACRO || type == CHAR)
		{
			size = (type == MACRO) ? MACROCHUNK : nvals*CHARCHUNK;

			chardata = STRING(symh);
			if (chardata != (char **) 0)
			{
				size = myhandlelength(chardata);
				if (size < 0)
				{
					/* something is wrong with the handle*/
					goto errorExit;
				}
			}
			else
			{
				setSTRING(symh, mygethandle(size));
				chardata = STRING(symh);
				if (chardata == (char **) 0)
				{
					goto errorExit;
				}
			}
	
			spaceleft = size;
			nchar = 0;
	
			if (binary || ASCII == SAVEASCII1)
			{
				i = 0;
				while (i < nvals && (ch = restoregetc(RESTOREFILE)) != EOF)
				{
					if (spaceleft < 2)
					{
						size = nchar + (nvals - i)*CHARCHUNK;
						chardata = mygrowhandle((char **) chardata, size);
						setSTRING(symh,chardata);
						spaceleft = size - nchar;
						if (chardata == (char **) 0)
						{
							goto errorExit;
						}
					}
					if (SwapNewlines == yesSwap && (ch == CR || ch == NL))
					{
						/*
							will only happen when reading a foreign file
							map any type of new line character to local
							newline character
						*/
						ch = '\n';
					}
					(*chardata)[nchar++] = (char) ch;
					spaceleft--;
					if (ch == '\0')
					{
						i++;
					}
				} /*while (i < nvals && (ch = restoregetc(RESTOREFILE)) != EOF)*/
			} /*if (binary || ASCII == SAVEASCII1)*/
			else
			{ /* ASCII > SAVEASCII1 */
				/*
				  For a macro, nvals should be 1 and an insufficient
				  amoung MACROCHUNK will have been allocated, probably
				  from the cache.  The exact amount needed will be
				  reallocated
	
				  For a CHARACTER vector, if there is not enough room,
				  enough is allocated for the item being read plus
				  CHARCHUNK for each item after that.  This means
				  that if we run out of room for the last item, exactly
				  the right length will be reallocated
				*/
				for (i = 0;i < nvals;i++)
				{
					long        length;
	
					if(!restoreLong(&length))
					{
						goto errorExit;
					}
					if (spaceleft < length + 1)
					{
						/* not enough room for this item */
	
						size = nchar + length + 1 + (nvals - i - 1)*CHARCHUNK;
	
						setSTRING(symh, mygrowhandle(chardata, size));
						spaceleft = size - nchar;
					}
					chardata = STRING(symh);
					if(chardata == (char **) 0)
					{
						goto errorExit;
					}
					if (!restoreNString(length, *chardata + nchar))
					{
						goto errorExit;
					}
					nchar += length + 1;
					spaceleft -= length + 1;
				} /*for (i = 0;i < nvals;i++)*/
			}
	
			/* if necessary, shrink to exact size */
			if (spaceleft > 0)
			{
				setSTRING(symh,(char **) 0);
				chardata = mygrowhandle(chardata, nchar);
				setSTRING(symh, chardata);
				if(chardata == (char **) 0)
				{
					goto errorExit;
				}
			}
		} /*if (type == MACRO || type == CHAR)*/
		else if (type == REAL || type == LOGIC)
		{
			needed = nvals * sizeof(double);
			
			doubledata = DATA(symh);
			if (doubledata != (double **) 0)
			{
				size = myhandlelength((char **) doubledata);
				if (size < 0)
				{
					/* something is wrong with handle */
					goto errorExit;
				}
				if (size != needed)
				{
					doubledata = (double **) mygrowhandle((char **) doubledata,
														  needed);
				}
			}
			else
			{
				doubledata = (double **) mygethandle(needed);
			}
			setDATA(symh, doubledata);
			if (doubledata == (double **) 0)
			{
				goto errorExit;
			}
			if (!binary)
			{	/* ascii file */
				for (i = 0; i < nvals; i++)
				{
					if(!restoreDouble(*doubledata + i))
					{
						goto errorExit;
					}
				}
			} /*if (!binary)*/
			else
			{	/* binary file */
				if(fread((void *) *doubledata, sizeof(double),
						 (size_t) nvals, RESTOREFILE) != nvals)
				{
					goto errorExit;
				}
				if (ASCII <= SAVEBINARY7 && DoPermute == yesPermute)
				{
					SaveDoublePtr   x = (SaveDoublePtr) *doubledata;
					SaveDouble      tempx;
					int             j;

					for (i = 0; i < nvals; i++, x++)
					{
						for (j = 0; j < sizeof(double); j++)
						{
							tempx.bytes[j] =
							  x->bytes[(int) DoublePermutation[j]];
						}
						x->value = tempx.value;
					} /*for (i = 0; i < nvals; i++, x++)*/
				} /*if (ASCII <= SAVEBINARY7 && DoPermute == yesPermute)*/
				if (ASCII > SAVEBINARY5)
				{
					double    *x = *doubledata;

					for (i = 0; i < nvals; i++)
					{
						if (x[i] == OLDMISSING)
						{
							setMissing(x[i]);
						}
					} /*for (i = 0; i < nvals; i++)*/
				} /*if (ASCII > SAVEBINARY5)*/
			} /*if (!binary){}else{}*/
		}
		else if (type == LONG)
		{	
			needed = nvals * sizeof(long);
			
			longdata = LONGDATA(symh);
			if (longdata != (long **) 0)
			{
				size = myhandlelength((char **) longdata);
				if (size < 0)
				{
					/* Something is wrong with handle */
					goto errorExit;
				}
				if (size != needed)
				{
					longdata = (long **) mygrowhandle((char **) longdata, needed);
				}
			}
			else
			{
				longdata = (long **) mygethandle(needed);
			}
			setLONGDATA(symh, longdata);
			if (longdata == (long **) 0)
			{
				goto errorExit;
			}
	
			if (!binary)
			{ /* ascii file */
				for (i = 0; i < nvals; i++)
				{
					if(!restoreLong(*longdata + i))
					{
						goto errorExit;
					}
				}
			} /*if (!binary)*/
			else
			{ /* binary file */
				if (fread((void *) *longdata, sizeof(long), (size_t) nvals,
						  RESTOREFILE) != nvals)
				{
					goto errorExit;
				}
				if (ASCII <= SAVEBINARY7 && DoPermute == yesPermute)
				{
					SaveLongPtr   x = (SaveLongPtr) *longdata;
					SaveLong      tempx;
					int           j;

					for (i = 0; i < nvals; i++, x++)
					{
						for (j = 0; j < sizeof(long); j++)
						{
							tempx.bytes[j] =
							  x->bytes[(int) LongPermutation[j]];
						}
						x->value = tempx.value;
					} /*for (i = 0; i < nvals; i++, x++)*/
				} /*if (ASCII <= SAVEBINARY7 && DoPermute == yesPermute)*/
			} /*if (!binary){}else{}*/
		}
		else
		{
			/* shouldn't happen*/
			putOutMsg("ERROR: (internal) unrecognized type in restoreHandle()");
			goto errorExit;
		}
	} /*if (nvals > 0)*/
	return (1);

  errorExit:
	return (0);
} /*restoreHandle()*/

/*
   skipRestoreVector() skips vector of indicated type.
*/
long skipRestoreVector(long itemType, long itemLength)
{
	double         itemDvalue;
	long           j, itemLvalue;
	char           itemCvalue[MAXCHARITEMSIZE + 1];

	for (j = 0; j < itemLength; j++)
	{
		if (itemType & SAVEREAL)
		{
			if (!restoreDouble(&itemDvalue))
			{
				return (0);
			}
		}
		else if (itemType & SAVELONG)
		{
			if (!restoreLong(&itemLvalue))
			{
				restore (0);
			}
		}
		else if (itemType & SAVECHAR)
		{
			if (!restoreString(itemCvalue) ||
				strlen(itemCvalue) > MAXCHARITEMSIZE)
			{
				return (0);
			}
		}
		else
		{
			return (0);
		}
	} /*for (j = 0; j < itemLength; j++)*/
	return (1);
} /*skipRestoreVector()*/

/*
  Get one item in ascii form.
  If the item is a scalar (itemType & SAVESCALAR != 0), its value is
  returned in itemDvalue, itemLvalue or itemCvalue.  It the item is a vector
  (itemType & SAVEVECTOR != 0), its length is returned in itemLength
  but no values are returned.  This can be called after reading a
  named restore item.
*/
int getRestoreItem(long itemType, long * itemLength,
				   double * itemDvalue, long * itemLvalue, char * itemCvalue)
{
	WHERE("getRestoreItem");
	
	if (itemType & SAVEVECTOR)
	{				
		if (!restoreLong(itemLength) && *itemLength < 1)
		{
			return (0);
		}
	}
	else if (itemType & SAVESCALAR)
	{
		if (itemType & SAVEREAL)
		{
			if (!restoreDouble(itemDvalue))
			{
				return (0);
			}
		}
		else if (itemType & SAVELONG)
		{
			if (!restoreLong(itemLvalue))
			{
				return (0);
			}
		}
		else if (itemType & SAVECHAR)
		{
			if (!restoreString(itemCvalue) ||
				strlen(itemCvalue) > MAXCHARITEMSIZE)
			{
				return (0);
			}
		}
		else
		{
			return (0);
		}
	} /*else if (itemType & SAVESCALAR)*/
	else if (!(itemType & SAVESYMBOL))
	{
		return (0);
	}
	return (1);
} /*getRestoreItem()*/

int getNextRestoreItem(char prefix [], int * itemNumber, int * itemType,
					   long * itemLength, double * itemDvalue,
					   long * itemLvalue, char * itemCvalue)
{
	char          itemName[MAXCHARITEMSIZE + 1];
	char          thisPrefix[3];
	WHERE("getNextRestoreItem");

	if (!restoreName(itemName) ||
		!decodeItem(itemName, thisPrefix, itemNumber, itemType) ||
		strcmp(prefix, thisPrefix) != 0)
	{
		return (0);
	}

	return (*itemNumber == EndItems ||
			getRestoreItem(*itemType, itemLength, itemDvalue,
						   itemLvalue, itemCvalue));

} /*getNextRestoreItem()*/
