/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.12 or later
*(C)*
*(C)* Copyright (c) 2001 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Rotfac
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/


#include "globals.h"
#include "blas.h"

#ifndef ROTMAXITER
#define ROTMAXITER  100
#endif /*ROTMAXITER*/

#ifndef ROTEPSILON
#define ROTEPSILON 1e-5
#endif /*ROTEPSILON*/

/*
  010614 reorganized code and added capability for quartimax and
         orthomax rotation
*/
enum rotationScratch
{
	GFNORM = 0,
	GKNORM,
	GLABELS,
	NTRASH
};

enum rotationMethods
{
	IVARIMAX = 0,
	IQUARTIMAX,
	IEQUIMAX,	
	IORTHOMAX,
	IOBLIMIN,
	NMETHODS,
	NOTAVAILABLE = IOBLIMIN /* first non-implemented recognized method*/
};

typedef struct methodName
{
	char      *name;
	short      length; /*minimum number of letters for recognition*/
} methodName;

static methodName     Methods[NMETHODS] =
{
	{"Varimax",    3},
	{"Quartimax",  5},
	{"Equimax",    4},
	{"Orthomax",   5},
	{"Oblimin",    7}
};

/*
  Code acquired from Doug Hawkins, University of Minnesota, 12/93

  Comments from Fortran original
	Routine to do a varimax rotation on the real array aload(nv,nnf).
	If nnf is positive, the routine feels free to reverse and reorder
	the factors;this is suppressed in nnf is entered as the negative
	of its actual value.  This suppression is desirable when doing
	a q-mode analysis.

  Translated to C by C. Bingham with the following modifications
     Argument nnf is now nf and is assumed positive and additional argument
     fnorm replaces the local variable fnorm.  If fnorm == (double *) 0,
	 reversing and reordering the factors is suppressed

	 Also argument itmax has been added to control the maximum number of
	 iterations and argument eps provides a convergence limit (originally
	 hard wired as 1e-4).
	 
	 Information on the iteration is printed if verbose != 0

	 varmx() returns a non-zero value if and only if fewer than itmax
	 iterations are required.

	 960919 Modified code to make it easier to add new rotation methods.
            except for slgihtly different error messages, it should have no
            effect on what it does.
     980303 Changed check before computation of rotaton angle to avoid
            atan2 domain error

     010614 added argument lambda to implement orthomax; lambda = 1 is
            varimax; lambda = 0 is quartimax
            Also computation of the criterion moved to separate function
            compcrit and names of certain variables were changed
*/

static double compcrit(double *loadings, long nv, long nf, double lambda)
{
	double       crit = 0.0, *loadingj = loadings;
	double       fnv = (double) nv;
	long         i, j;
	
	for (j = 0; j < nf ;j++)
	{
		double       s2 = 0.0;

		for (i = 0;i < nv ;i++)
		{
			double        sq = loadingj[i]*loadingj[i];

			s2 += sq;
			crit += sq*sq;
		}
		crit -= lambda*s2*s2/fnv;
		loadingj += nv;
	} /*for (j = 0;j < nf ;j++)*/

	return (crit);
	
} /*compcrit()*/

/*
  010614 added arguments method and param and modified code so that it
         finds optimal orthomax rotation with parameter lambda = params[0]
         lambda == 1 <==> variamx
         lambda == 2 <==> quartimax
*/
static long varmx(double *aload,long nv, long nf, long method, double *params,
				  double *fnorm,
				  long itmax, double eps, long verbose)
/*float aload[nv][1];*/
{
	double         crit, startCrit, fnv = (double) nv;
	double        *aloadj, *aloadk;
	double         denominator, numerator, angl, trot;
	double         eps1 = eps, eps2 = eps;
	double         lambda;
	long           inoim = 0, ict = 0, irot = 0;
	long           i, j, k, iflip, nf1 = nf - 1;
	WHERE("varmx");

	if (method <= IORTHOMAX)
	{
		lambda = params[0];
	}

	startCrit = crit = compcrit(aload, nv, nf, lambda);

	do /*while (inoim < 2 && ict < itmax && iflip);*/
	{
		double      oldCrit = crit;

		iflip = 0;
		aloadj = aload;

		for (j = 0;j < nf1 ;j++)
		{
			aloadk = aloadj + nv;
			for (k = j + 1;k < nf ;k++)
			{
				double      a = 0.0, b = 0.0, c = 0.0, d = 0.0, s = 0.0;
				
				for (i = 0;i < nv ;i++)
				{
					double    c2 = aloadj[i]*aloadj[i] - aloadk[i]*aloadk[i];
					double    s2 = 2.0*aloadj[i]*aloadk[i];

					a += c2;
					b += s2;
					c += c2*c2 - s2*s2;
					d += c2*s2;
				} /*for (i = 0;i < nv ;i++)*/

				denominator = fnv*c + lambda*(b*b - a*a);
				numerator = 2.0*(fnv*d - lambda*a*b);

				if (fabs(numerator) > eps1*fabs(denominator))
				{
					iflip = 1;
					irot++;
					angl = 0.25*atan2(numerator,denominator);
					
					c = cos(angl);
					s = sin(angl);
					for (i = 0;i < nv ;i++)
					{
						double   t = c*aloadj[i] + s*aloadk[i];

						aloadk[i] = -s*aloadj[i] + c*aloadk[i];
						aloadj[i] = t;
					} /*for (i = 0;i < nv ;i++)*/
				} /*if (fabs(numerator) >= eps1*fabs(denominator))*/
				aloadk += nv;
			} /*for (k = j + 1;k < nf ;k++)*/
			aloadj += nv;
		} /*for (j = 0;j < nf1 ;j++)*/
		ict++;

		crit = compcrit(aload, nv, nf, lambda);
		trot = (crit > 0.0) ? (crit - oldCrit)/crit : 0.0;
		inoim++;
		if (trot > eps2)
		{
			inoim = 0;
		}
	} while (inoim < 2 && ict < itmax && iflip);

	if (fnorm != (double *) 0)
	{
		aloadj = aload;
		for (j = 0;j < nf ;j++)
		{
			double     ssj = 0, sj = 0;
			
			for (i = 0;i < nv ;i++)
			{
				sj += aloadj[i];
				ssj += aloadj[i]*aloadj[i];
			} /*for (i = 0;i < nv ;i++)*/
			fnorm[j] = ssj;
			if (sj <= 0.0)
			{
				for (i = 0;i < nv ;i++)
				{
					aloadj[i] = -aloadj[i];
				}
			} /*if (sj <= 0.0)*/

			aloadk = aload;
			for (k = 0;k < j ;k++)
			{
				if (fnorm[k] < fnorm[j])
				{
					double       t = fnorm[k];

					fnorm[k] = fnorm[j];
					fnorm[j] = t;
					for (i = 0;i < nv ;i++)
					{
						t = aloadj[i];
						aloadj[i] = aloadk[i];
						aloadk[i] = t;
					} /*for (i = 0;i < nv ;i++)*/
				} /*if (fnorm[k] < fnorm[j])*/
				aloadk += nv;
			} /*for (k = 0;k < j ;k++)*/
			aloadj += nv;
		} /*for (j = 0;j < nf ;j++)*/
	} /*if (fnorm != (double *) 0)*/
	if (verbose)
	{
		char   *outstr = OUTSTR;
		
		outstr += formatChar(outstr, Methods[method].name, CHARASIS);
		outstr += formatChar(outstr, " starting criterion = ", CHARASIS);
		outstr += formatDouble(outstr, startCrit, DODEFAULT | TRIMLEFT);
		outstr += formatChar(outstr, ", final criterion = ", CHARASIS);
		outstr += formatDouble(outstr, crit, DODEFAULT | TRIMLEFT);
		putOUTSTR();
		sprintf(OUTSTR,	"%ld iterations and %ld rotations", ict, irot);
		putOUTSTR();
	} /*if (verbose)*/	
	return (ict < itmax);
} /*varmx()*/

/*
  011125 added argument knorm. 
         knorm != (double *) 0 signals Kaiser normalization with knorm
         providing scratch for row norms
*/
static long doRotation(long method, double *aload, long nv, long nf, 
					   double *params, double *fnorm, double *knorm,
					   long itmax, double eps, long verbose)
{
	int        reply = -1;

	if (method <= IORTHOMAX)
	{
		if (method < IORTHOMAX)
		{
			if (method == IVARIMAX)
			{
				params[0] = 1.0;
			}
			else if (method == IQUARTIMAX)
			{
				params[0] = 0.0;
			}
			else
			{
				/* equimax */
				params[0] = .5*(double) nf;
			}
		}
		if (knorm != (double *) 0)
		{
			long         i, j, k;
			double       s;

			for (i = 0; i < nv; i++)
			{
				k = i;
				s = 0.0;
				for (j = 0; j < nf; j++, k += nv)
				{
					s += aload[k]*aload[k];
				}
				knorm[i] = s = (s > 0) ? sqrt(s) : 1.0;

				k = i;
				for (j = 0; j < nf; j++, k += nv)
				{
					aload[k] /= s;
				}
			}
		} /*if (fcopy != (double *) 0)*/
		
		reply = varmx(aload, nv, nf, method, params, fnorm, itmax,
					  eps, verbose);
		if (knorm != (double *) 0)
		{
			long      i, j, k = 0;
			
			for (j = 0; j < nf; j++)
			{
				for (i = 0; i < nv; i++, k++)
				{
					aload[k] *= knorm[i];
				}
			}
		}
	} /*if (method <= IORTHOMAX)*/
	return (reply);
} /*doRotation()*/

/*
   Function to perform factor rotation.
   At present (940905) it can do only varimax rotation
   usage:
   	rotation(loadings [,method:"varimax"] 
		[,verbose:T] [,reorder:F [,maxiter:m] [,eps:value])
   The default values for verbose and reorder are F and T
   The default values for maxiter and eps are 100 and 1e-5

  980303 added check for MISSING values

  980730 added new argument to reuseArg() so that notes will not be kept.

  990226 Replaced most use of putOUTSTR() by putErrorOUTSTR()

  010614 Implemented quartimax rotation in addition to varimax

  011125 Added keyword phrase kaiser:T to specify Kaiser normalization
*/

Symbolhandle rotatefac(Symbolhandle list)
{
	Symbolhandle     loadings, result = SymbolZero;
	Symbolhandle     keySymh;
	long             verbose = 0, reorder = 0, maxiter = ROTMAXITER;
	long             converged;
	long             nargs = NARGS(list);
	double           params[4], lambda;
	double           eps = ROTEPSILON, value;
	double         **fnorm = (double **) 0;
	double         **knorm = (double **) 0;
	char            *keyword, *methodName;
	long             i, methodCode = IVARIMAX;
	long             nv, nf, dims[2];
	int              kaiser = 0;
	WHERE("rotatefac");
	TRASH(NTRASH, errorExit);
	
	OUTSTR[0] = '\0';
	setMissing(lambda);
	
	loadings = COMPVALUE(list,0);
	if (!argOK(loadings, 0, (nargs > 1) ? 1 : 0))
	{
		goto errorExit;
	}
	if (TYPE(loadings) != REAL || !isMatrix(loadings, dims) ||
		anyMissing(loadings))
	{
		sprintf(OUTSTR,
				"ERROR: argument 1 to %s() not a REAL matrix with no MISSING elements",
				FUNCNAME);
	}
	else if (dims[1] > dims[0])
	{
		sprintf(OUTSTR, "ERROR: argument 1 to %s() has more columns than rows",
				FUNCNAME);
	}
	if (*OUTSTR)
	{
		goto errorExit;
	}
	
	nv = dims[0];
	nf = dims[1];
	
	for (i = 1; i < nargs;i++)
	{
		keySymh = COMPVALUE(list,i);
		if (!argOK(keySymh, 0, i + 1))
		{
			goto errorExit;
		}
		if (!(keyword = isKeyword(keySymh)))
		{
			sprintf(OUTSTR,
					"ERROR: all except 1st argument to %s() must be keyword phrases",
					FUNCNAME);
			goto errorExit;
		}
		if (strncmp(keyword, "verb", 4) == 0 ||
			strncmp(keyword, "reord", 5) == 0 ||
			strncmp(keyword, "kais", 4) == 0)
		{
			if (!isTorF(keySymh))
			{
				notTorF(keyword);
				goto errorExit;
			}
			value = DATAVALUE(keySymh,0);
			if (keyword[0] == 'v')
			{
				verbose = (value != 0.0);
			}
			else if (keyword[0] == 'r')
			{
				reorder = (value != 0.0);
			}
			else
			{
				kaiser = (value != 0.0);
			}
		}
		else if (strncmp(keyword, "maxit", 5) == 0)
		{
			if (!isInteger(keySymh, POSITIVEVALUE))
			{
				notPositiveInteger(keyword);
				goto errorExit;
			}
			maxiter = (long) DATAVALUE(keySymh, 0);
        }
		else if (strncmp(keyword, "eps", 3) == 0)
		{
      		if (!isNumber(keySymh, POSITIVEVALUE))
         	{
				notPositiveReal(keyword);
				goto errorExit;
			}
			eps = DATAVALUE(keySymh, 0);
		}
		else if (strncmp(keyword, "lambda", 3) == 0)
		{
      		if (!isNumber(keySymh, NONNEGATIVEVALUE))
         	{
				notNonNegativeReal(keyword);
				goto errorExit;
			}
			lambda = DATAVALUE(keySymh, 0);
		}
		else if (strncmp(keyword, "meth", 4) == 0)
		{
			if (!isCharOrString(keySymh))
			{
				notCharOrString(keyword);
				goto errorExit;
			}
			methodName = STRINGPTR(keySymh);
			for (methodCode = 0; methodCode < NMETHODS; methodCode++)
			{
				if (mystrncmp(methodName, Methods[methodCode].name,
							  Methods[methodCode].length) == 0)
				{
					break;
				}
			} /*for (methodCode = 0; methodCode < NMETHODS; methodCode++)*/
			
			if (methodCode == NMETHODS)
			{
				sprintf(OUTSTR,
						"ERROR: %s is not a recognized rotation method for %s()",
						methodName, FUNCNAME);
			}
			else if (methodCode >= NOTAVAILABLE)
			{
				sprintf(OUTSTR,
						"ERROR: %s rotation is not yet implemented for %s()",
						Methods[methodCode].name, FUNCNAME);
			}
			if (*OUTSTR)
			{
				goto errorExit;
			}
		}
		else
		{
			badKeyword(FUNCNAME, keyword);
			goto errorExit;
		}
	} /*for (i = 1; i < nargs;i++)*/
	
	if (methodCode == IORTHOMAX)
	{
		if (isMissing(lambda))
		{
			sprintf(OUTSTR,
					"ERROR: lambda:value required with method:\"%s\" in %s()",
					methodName, FUNCNAME);
			goto errorExit;
		}
		params[0] = lambda;
	}
	else
	{
		if (!isMissing(lambda))
		{
			sprintf(OUTSTR,
					"WARNING: keyword lambda ignored with method:\"%s\" in %s()",
					methodName, FUNCNAME);
		}
		putErrorOUTSTR();
	}

	if (reorder && !getScratch(fnorm, GFNORM, nf, double))
	{
		goto errorExit;
	}

	if (kaiser && !getScratch(knorm, GKNORM, nv, double))
	{
		goto errorExit;
	}

	if (isscratch(NAME(loadings)))
	{ /* reuse argument */
		result = reuseArg(list, 0, 1, 0);
	}
	else
	{
		result = Install(SCRATCH, REAL);
		if (result == SymbolError || !Copy(loadings, result))
		{
			Removesymbol(result);
			goto errorExit;
		}
		setNAME(result, SCRATCH);
		clearNotes(result);
	}

	if (HASLABELS(result))
	{
		if (!fixupMatLabels(result, USEROWLABELS))
		{
			goto errorExit;
		}
	} /*if (HASLABELS(result))*/ 
	else
	{
		setNDIMS(result, 2);
		setDIM(result, 1, nv);
		setDIM(result, 2, nf);
	}
	
	setNCLASS(result, -1);
	
	converged = doRotation(methodCode, DATAPTR(result), nv, nf, params,
						   (fnorm != (double **) 0) ? *fnorm : (double *) 0,
						   (knorm != (double **) 0) ? *knorm : (double *) 0,
						   maxiter, eps, verbose);
	
	if (!converged)
	{
		sprintf(OUTSTR,
				"WARNING: %s rotation did not converge in %ld iterations",
				Methods[methodCode].name, maxiter);
		putErrorOUTSTR();
	}
	
	emptyTrash();
	return (result);

  errorExit:
	emptyTrash();
	putErrorOUTSTR();
	return (SymbolError);
	
} /*rotatefac()*/


