/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 4.13 or later
*(C)*
*(C)* Copyright (c) 2003 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#include "globals.h"

#include "mvIface.h"

#include "mvIo.h"

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Setoptio
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

/*
  981007 Minor change to an error message
  981118 Added new option 'findmacros' to control automatic
         searching for undefined macros
  981218 Added new option 'plotdelay' to control wait before plotting (Motif only)
  990212 Replaced most putOUTSTR() by putErrorOUTSTR()
  990822 For versions using Tektronix emulation, added option 'tekset'
         to allow setting of character strings that switch to and from
         Tektronix mode.  Its value is vector(totek, fromtek).  If
         a string is "", not attempt is made to switch.  When TERMINAL
         is "xterm", the default is appropriate to xterm;  otherwise, the
         default is appropriate for Versaterm, a Macintosh terminal emulator
  000115 Added option 'vecread' with values "oldstyle" and "newstyle" to provide
         default for vecread() keyword 'oldstyle'
  000120 Changed VecreadOldStyle to VecreadByfields
  001025 Added option 'maxlinelen' to set maximum input line length scanned
  001102 Added option 'matchdelay' to set timing loop in matching parentheses
  010515 BATCHECHO[0] not set in windowed versions
  010601 Include mvIface.h instead of macIface.h or wxIface.h
  010611 Added option 'keyboard' for Windows version to control recognition
         of characters generated by Alt Gr key combinations.  NOTE:  This
         option does not enable any behavior that is not already enabled
         for the keyboard used.
  010802 Added option 'quiet' (T or F)
  010811 Added option 'traceback' (T or F)
  020808 Added option 'minpvalue' (small minp > 0)
  020819 Added option 'enableitem' (CHARACTER vector)
*/

#define MAXSEED 0x7fffffffL

/*
   The option codes must be consecutive integers starting with 1
   When adding a new option, don't forget
     (a) Makesure NOPTIONS is still correct; will be so if option put
	     before history
     (b) Update structure OptionInfo
     (c) Add initialization code to setDefaultOptions()
     (d) Add code to setopt() and getopt()
*/
enum optionCodes
{
	ISEEDS = 1,
	INSIG,
	IFORMAT,
	IWFORMAT,
	IVECREAD,
	IANGLES,
	ILINES,
	IHEIGHT,
	IWIDTH,
	IERRORS,
	IPROMPT,
	IBATCHECHO,
	IRESTOREDEL,
	IDUMBPLOT,
	ISCROLLBACK,
	IUPDATE,
	IMISSING,
	IWARNINGS,
	IFSTATS,
	IPVALS,
	IMINPVALUE,
	IEXCLUDEM,
	IFONTSIZE,
	IFONT,
	IMAXWHILE,
	ILABELABOVE,
	ILABELSTYLE,
	IINLINE,
	IFINDMACROS,
	IQUIET,
	ITRACEBACK,
	IENABLEITEM,
	IPLOTDELAY,
	IMATCHDELAY,
	IKEYBOARD,
	ITEKSET,
	IMAXLINELEN,
	ISAVEHISTRY,
	IHISTORY,
	NOPTIONS = IHISTORY, /* number of retrievable options */
/* the remainder are not retrievable */
	IDEFAULTS, /*(NOPTIONS+1)*/
	IGUBED,    /*(NOPTIONS+2)*/
#ifdef MACINTOSH
	IPEELS,     /*(NOPTIONS+3)*/
#endif /*MACINTOSH*/
	NOPTIONSALL  /*number of all options*/
};
/*
  since "lines" and "height" are synonyms, actually there are actuall
  NOPTIONS - 1 retrievable options
*/
typedef struct optionInfo
{
	char   *name;
	short  length;
	short  code;
} optionInfo;

/*
   option names must not exceed 10 characters and in same order as codes
   0 < Options[j].code <= NOPTIONS  regular option in this version
   -NOPTIONS < Options[j].code < 0  regular option not in this version
   Options[j].code < -NOPTIONS, non-retrievable special option
   Options[j].code > NOPTIONS, retrievable special option
*/
static optionInfo   Options[] = 
{
	{"seeds",     4, ISEEDS}, /* random number seeds (vector(posInt1,PosInt2)) */
	{"nsig",      4, INSIG},  /* sets default format to %{nsig}+7).{nsig}g */
	{"format",    6, IFORMAT},/* default format and default format for print */
	{"wformat",   7, IWFORMAT},/* default format for write */
	{"vecread",   7, IVECREAD},/* value "byfields" or "notbyfields" affects vecread()*/
	{"angles",    5, IANGLES}, /* value "radians","cycles", "degrees" */
	{"lines",     4, ILINES},  /* lines of text on screen */
	{"height",    5, IHEIGHT}, /* synonym for "lines" */
	{"width",     5, IWIDTH},  /* number of characters per output line */
	{"errors",    5, IERRORS}, /* limit on number of errors allowed */
	{"prompt",    6, IPROMPT}, /* default prompt */
	{"batchecho", 9, IBATCHECHO}, /* should commands be echoed in batch mode */
	{"restoredel",10,IRESTOREDEL}, /*default for 'delete' on restore()*/
	{"dumbplot",  4, IDUMBPLOT}, /*default for 'dumb' on plotting commands*/
#ifdef SCROLLABLEWINDOW
	{"scrollback",10, ISCROLLBACK}, /*default for 'scrollback' on help(), etc*/
#else /*SCROLLABLEWINDOW*/
	{"scrollback",10, -ISCROLLBACK}, /*not available in this version*/
#endif /*SCROLLABLEWINDOW*/
#ifdef ACTIVEUPDATE
	{"update",    6, IUPDATE}, /*enables active updating of screen */
#else /*ACTIVEUPDATE*/
	{"update",    6, -IUPDATE}, /*not available in version */
#endif /*ACTIVEUPDATE*/
	{"missing",   7, IMISSING},  /*filler to be used for MISSING*/
	{"warnings",  7, IWARNINGS},  /* should WARNING: be printed */
	{"fstats",    5, IFSTATS},  /* default value for glm keyword 'fstat'*/
	{"pvals",     4, IPVALS},   /* default value for glm keyword 'pvals'*/
	{"minpvalue", 4, IMINPVALUE}, /*threshhold for special P-value output*/
	{"excludeM",  8, IEXCLUDEM}, /* default value for describe keyword 'excludeM'*/
#ifdef CANSETFONTS
	{"fontsize",  7, IFONTSIZE},  /*"fontName Size"*/
	{"font",      4, IFONT},  /*"fontName Size"*/
#else /*CANSETFONTS*/
	{"fontsize",  7, -IFONTSIZE},  /*not available in version*/
	{"font",      4, -IFONT},  /*not available in version*/
#endif /*CANSETFONTS*/
	{"maxwhile",  8, IMAXWHILE},  /* max number of while(){} repetitions */
	{"labelabove",6, ILABELABOVE}, /* last coord label above */
	{"labelstyle",6, ILABELSTYLE}, /* default paren for labelling */ 
	{"inline",    6, IINLINE},     /* default type of macro expansion */
	{"findmacros",7, IFINDMACROS}, /* controls automatic search for macros */
	{"quiet",     5, IQUIET},      /* controls suppression of output */
	{"traceback", 5, ITRACEBACK},  /* controls traceback of macro call chains */
#if defined(SKEL_WXWIN) 
	{"enableitem",6, IENABLEITEM}, /* controls certain menu items */
#else /*SKEL_WXWIN*/
	{"enableitem",6, -IENABLEITEM}, /* not avaiable in this version*/
#endif

#if defined(SKEL_WXWIN) && defined(BASE_UNIX)
	{"plotdelay", 7, IPLOTDELAY},  /* controls wait before opening graph window*/
#else /*SKEL_WXWIN && BASE_UNIX*/
	{"plotdelay", 7, -IPLOTDELAY},  /* not available in version*/
#endif /*SKEL_WXWIN && BASE_UNIX*/

#ifdef SCROLLABLEWINDOW
	{"matchdelay", 8, IMATCHDELAY}, /* controls wait used in matching parens*/
#else /*SCROLLABLEWINDOW*/
	{"matchdelay", 8, -IMATCHDELAY}, /* not available in this version*/
#endif /*SCROLLABLEWINDOW*/
#if defined(OS_WIN32) && defined(SKEL_WXWIN) 
	{"keyboard",  8, IKEYBOARD},   /* controls recognition of Windows Alt Gr key combinations*/
#else /*defined(OS_WIN32) && defined(SKEL_WXWIN)*/ 
	{"keyboard",  8, -IKEYBOARD},   /* not available in version*/
#endif /*defined(OS_WIN32) && defined(SKEL_WXWIN)*/ 
#ifdef TEK
	{"tekset",    6, ITEKSET},    /*Codes to switch to & from Tektronix mode*/
#else /*TEK*/
	{"tekset",    6, -ITEKSET},     /* not available in version */
#endif /*TEK*/
	{"maxlinelen",7, IMAXLINELEN},    /*maximum size of input line*/
#ifdef SAVEHISTORY
	{"savehistry",8, ISAVEHISTRY}, /* default value for save keyword history */
	{"history",   4, IHISTORY}, /*number of history lines to save*/
#else /*SAVEHISTORY*/
	{"savehistry",8, -ISAVEHISTRY},/* not available in version */
	{"history",   4, -IHISTORY}, /*not available in version*/
#endif /*SAVEHISTORY*/
	{"defaults",  7, IDEFAULTS},	/* reset all options to defaults*/
	{"GUBED",     5, -IGUBED},    /* DEBUG backwards, for debugging */
#ifdef MACINTOSH
	{"PEELS",     5, -IPEELS},    /* SLEEP backwards for testing waitNextEvent */
#endif /*MACINTOSH*/
	{(char *) NULL, 0, 0}
};

#ifdef CANSETFONTS
#define DEFAULTFONT "default"

enum fontConstants
{
#ifdef MACINTOSH
	FONTNAMESIZE    = 27,
	MINFONTSIZE     =  9, 
	MAXFONTSIZE     = 64,
#endif /*MACINTOSH*/
	OKFONTSPEC      =  0,
	BADFONTSIZE     = -1,
	FONTNAMETOOLONG = -2,
	BADFONTSPEC     = -3
};

/*
	Function to decode a font of the form "FontName size"
	If font is of form "FontName", set *fontsize to -1
*/
static int decodeFont(char *font, char **fontName, long *fontSize)
{
	static char     name[FONTNAMESIZE+1];
	long            length = strlen(font);
	char           *sizePtr = font + length - 1;
	WHERE("decodeFont");
	
	*fontSize = -1;
	while (sizePtr > font && *sizePtr == ' ')
	{ /* skip backwara over trailing spaces */
		length--;
		sizePtr--;
	}
	if (isdigit(*sizePtr))
	{ /* may be of form "Name Size" */
		while (sizePtr > font && isdigit(*sizePtr))
		{
			sizePtr--;
		}
		if (*sizePtr == ' ' || *sizePtr == '-' || *sizePtr == '+')
		{
			if (*sizePtr == ' ')
			{
				sizePtr++;
			}
			/* sizePtr should point to number to be decoded */
			if (*(sizePtr-1) != ' ')
			{ /* not of form "Name Number" */
				*fontSize = -1;
				length = strlen(font);
			}
			else
			{
				if (sscanf(sizePtr, "%ld", fontSize) != 1 ||
					*fontSize < MINFONTSIZE || *fontSize > MAXFONTSIZE)
				{
					return (BADFONTSIZE);
				}
				sizePtr--;
				while (sizePtr > font && *sizePtr == ' ')
				{
					sizePtr--;
				}
				sizePtr++;
				length = sizePtr - font;
			}
		} /*if (*sizePtr == ' ' || *sizePtr == '-' || *sizePtr == '+')*/
	} /*if (isdigit(*sizePtr))*/
	if (length > FONTNAMESIZE)
	{
		return (FONTNAMETOOLONG);
	}
	if (length == 0)
	{
		return (BADFONTSPEC);
	}
	strncpy(name, font, length);
	name[length] = '\0';
	*fontName = name;

	return (OKFONTSPEC);
} /*decodeFont()*/

/*
   Function to set the font for the current window and the default font used
   in new windows.  If newFontSize > 0,  the font size is also set.  Called
   from functions setDefaultOptions() and setOpt().

   Actual function called depends on platform.

   Note: setDefaultOptions() and hence setNewFont() is called from
   initialize() which might be (and is, on a Macintosh) before any
   windows have been created.  Hence special care needs to be taken in the
   code implementing the change.
*/

static void setNewFont(char * newFontName, short newFontSize)
{
#ifdef MACINTOSH
	macSetNewFont(newFontName, newFontSize);
#else /*MACINTOSH*/
	/* add calls needed on other platforms here */
#endif /*MACINTOSH*/
} /*setNewFont()*/

/*
	returns 1 if newFontName and newFontSize match current
	does not check newFontName when (char *) 0
	does not check newFontSize when <= 0
*/
static int    isCurrentFont(char * newFontName, short newFontSize)
{
	int       reply = 1;
#ifdef MACINTOSH
	if (newFontName != (char *) 0)
	{
		Str255        pNewFontName;
		
		c2pstrcpy(pNewFontName, newFontName);
		reply = (str255cmp(pNewFontName, CmdFontName) == 0);
	}
	if (reply && newFontSize > 0)
	{
		reply = (newFontSize == CmdFontSize);
	}
#endif /*MACINTOSH*/
	return (reply);
} /*isCurrentFont()*/

static void   setCurrentFontName(char * fontName)
{
#ifdef MACINTOSH
	p2cstrcpy(fontName, CmdFontName);
#endif /*MACINTOSH*/
} /*setCurrentFontName()*/
#endif /*CANSETFONTS*/

/*
   set default values of all options
   971230 set default for restoredel to T to conform with help file
   010515 BATCHECHO[0] not set in windowed versions
*/

#ifdef TEK

/*
  XTERM2TEK decimal codes are 27,91,63,51,56,104,27,56
            characters are    esc [ ? 3 8 h esc 8
  VTERM2TEK decimal codes are 29,27,56
            characters are    GS esc 8
*/
#define XTERM2TEK "\033\133\077\063\070\150\033\070" /* esc [ ? 3 8 h esc 8 */
#define VTERM2TEK "\035\033\070"                     /* GS esc 8 */

/*
  XTERM2VT decimal codes are 27, 3
           chracters are     esc, ^C
  VTERM2VT decimal codes are 27, 50
           characters are    esc, 2
*/
#define XTERM2VT  "\033\003"                 /* esc ^C */
#define VTERM2VT  "\033\062"

#endif /*TEK*/

/*
  031216 deeper WHILELIMITS are set to the default value of MAXWHILE
*/
void setDefaultOptions(void)
{
	int          i;
#ifndef MACINTOSH
	long         height, width;
#endif /*MACINTOSH*/
	WHERE("setDefaultOptions");

	/* ISEEDS */
	RANDS1 = 0;
	RANDS2 = 0;

	/* IHEIGHT and IWIDTH */
#if defined(SKEL_CARAPACE)
	BATCHECHO[0] = (ISATTY != (ITTYIN | ITTYOUT));
#endif /*CARAPACE*/
#ifndef SCROLLABLEWINDOW
	BATCHECHO[0] = (ISATTY != (ITTYIN | ITTYOUT));
	height = (ISATTY & ITTYOUT) ?  DEFAULTSCREENHEIGHT : 0;
	width = DEFAULTSCREENWIDTH;
	installScreenDims(width, height);
#else /*SCROLLABLEWINDOW*/
#ifdef MACINTOSH
	if (teCmd != (TEHandle) 0)
	{
		setCmdScreen();
	}
#endif /*MACINTOSH*/
#endif /*SCROLLABLEWINDOW*/
	/* INSIG and IFORMAT */
	PRINTFORMAT[0] = DEFAULTPRINTDEC+7;
	PRINTFORMAT[1] = DEFAULTPRINTDEC;
	PRINTFORMAT[2] = (long) 'g';
	installFormat(PRINTFORMAT[0],PRINTFORMAT[1],PRINTFORMAT[2]);
	saveFormat(); /* save copy of format and SCREENWIDTH and SCREENHEIGHT */

	/* IWFORMAT */
	WRITEFORMAT[0] = DEFAULTWRITEDEC+7;
	WRITEFORMAT[1] = DEFAULTWRITEDEC;
	WRITEFORMAT[2] = (long) 'g';

	/* IVECREAD */
	VecreadByfields = DEFAULTVECREADBYFIELDS;

	/* IANGLES */
	ANGLES2RADIANS = 1.0;  /* "radians" */

	/* IERRORS */
	MAXERRORS1 = MAXERRORS = 0;

	/* IPROMPT */
	for (i = 0; i <= BDEPTH; i++)
	{
		strcpy((char *) PROMPTS[i], (char *) DEFAULTPROMPTS[i]);
	}
	
	 /*IBATCHECHO*/
	DEFAULTBATCHECHO = 1;

	 /*IRESTOREDEL*/
	DEFAULTRESTOREDEL = 1;

	 /*IDUMBPLOT*/
#ifdef MULTIPLOTWINDOWS
	DEFAULTDUMBPLOT = !UseWindows;
#else /*MULTIPLOTWINDOWS*/
	DEFAULTDUMBPLOT = 0;
#endif /*MULTIPLOTWINDOWS*/

#ifdef SCROLLABLEWINDOW
	 /*ISCROLLBACK*/
	SCROLLBACK = DEFAULTSCROLLBACK;
#endif /*SCROLLABLEWINDOW*/

	/* IUPDATE */
#ifdef ACTIVEUPDATE
	ActiveUpdate = ACTIVEUPDATE;
#else /*ACTIVEUPDATE*/
	ActiveUpdate = 0;
#endif /*ACTIVEUPDATE*/

	/* IMISSING */
	strncpy(NAMEFORMISSING,DEFAULTNAMEFORMISSING,LENGTHMISSING);
	NAMEFORMISSING[LENGTHMISSING] = '\0';
#ifdef HASINFINITY
	strncpy(NAMEFORINFINITY,DEFAULTNAMEFORINFINITY,LENGTHINFINITY);
	NAMEFORINFINITY[LENGTHINFINITY] = '\0';
#endif /*HASINFINITY*/

	/* IWARNINGS */
	PRINTWARNINGS = 1;

	/* IFSTATS and IPVALS */
	PRINTFSTATS = DEFAULTPRINTFSTATS;
	PRINTPVALS = DEFAULTPRINTPVALS;

	/* IMINPVALUE */
	MINPVALUE = DEFAULTMINPVALUE;

	/* IEXCLUDEM */
	EXCLUDEMEDIAN = 0;

#ifdef CANSETFONTS
	/* IFONTSIZE and IFONT */
	setNewFont(DEFAULTFONT, -1);
#endif /*CANSETFONTS*/

	/* IMAXWHILE */
	MAXWHILE = DEFAULTMAXWHILE;
	for (i = WDEPTH; i < MAXWDEPTH; i++)
	{
		WHILELIMITS[i] = MAXWHILE;
	}

	
	/* ILABELABOVE */
	USECOLLABS = DEFAULTUSECOLLABS;
	
	/* ILABELSTYLE */
	LEFTBRACKET[0] = DEFAULTLEFTBRACKET[0];
	RIGHTBRACKET[0] = Rparens[strchr(Lparens, LEFTBRACKET[0]) - Lparens];
	
	/* IINLINE */
	EXPANDINLINE = DEFAULTEXPANDINLINE;

	/* IFINDMACROS */
	MACROSEARCH = DEFAULTMACROSEARCH;

	/* IQUIET */
	QUIET = DEFAULTQUIET;
	
	/* ITRACEBACK */
	TRACEBACK = DEFAULTTRACEBACK;
	
	/* IENABLEITEM */
	ENABLEDITEMS = DEFAULTENABLEDITEMS;

	/* IPLOTDELAY */
#if defined(SKEL_WXWIN) && defined(BASE_UNIX) 
	FocusDelay = DefaultFocusDelay;
#endif /*SKEL_WXWIN && BASE_UNIX*/

	/* IKEYBOARD */
#if defined(SKEL_WXWIN) && defined(OS_WIN32) 
	KeyboardUse = DefaultKeyboardUse;
#endif /*WXWIN && OS_WIN32*/

#ifdef SCROLLABLEWINDOW
	/* IMATCHDELAY */
	MatchDelay = DefaultMatchDelay;
#endif /*SCROLLABLEWINDOW*/

	/* IMAXLINELEN */
	MAXLINE = DEFAULTMAXLINE;

#ifdef SAVEHISTORY
	/* ISAVEHISTRY */
#ifdef SCROLLABLEWINDOW
	DEFAULTSAVEHISTORY = 1;
#else /*SCROLLABLEWINDOW*/
	DEFAULTSAVEHISTORY = ISATTY & ITTYIN;
#endif /*SCROLLABLEWINDOW*/
	/* IHISTORY */
#if !defined(SKEL_CARAPACE)
	initHistory(DEFAULTHISTORYLENGTH);
#endif /*CARAPACE*/
#endif /*SAVEHISTORY*/
	
#ifdef TEK
	/*
	  ToTek and FromTek are defined in globkb.h
	*/
	/* ITEKSET */
	if (strcmp(TERMINAL, "xterm") == 0)
	{
		strcpy(ToTek, XTERM2TEK);
		strcpy(FromTek, XTERM2VT);
	}
	else
	{
		strcpy(ToTek, VTERM2TEK);
		strcpy(FromTek, VTERM2VT);
	}
#endif /*TEK*/

} /*setDefaultOptions()*/

/*
  functions to set and retrieve various options using keywords
  usage:
  	setoptions(option1,...,optionk)
  where optionj is of the form keywordj:valuej, e.g., 'nsig:5'

    setoptions(str)
	where str is structure whose component names are option names

	setoptions(default:T) resets all options to their default values

	getoptions() or getoptions(all:T)
	  returns a structure with all current values of options

	getoptions(keyword1:T [,keyword2:T...])
	  returns the specified options.  If there are more than one specified,
	  a structure is returned.

	getoptions(all:T, keyword1:F, ...)
	  returns all except the indicated options

  For backward compatibility, for getoptions, you can use, for example,
  "nsig" instead of nsig:T

  In addition to the public options (id numbers > 0) there are non-public
  options that may be set but not retrieved.  Chief among these is GUBED
  (DEBUG backwards) which is used to control various features useful in
  debugging

  If a new option is added, you must do all of the following
   (a)  Add a new item to enum optionCodes (above)
   (b)  Add an new element to optionInfo (above)
   (c)  Add a new global for the option value, probably in globkb.h
   (d)  Modify code where the option has effect to use the global
   (e)  Add code to setDefaultOptions() to set default value (above)
   (f)  Add new case and code to getOpt() to retrieve the value (below)
   (g)  Add new case and code to setOpt() to set the value (below)

  960418 Added option restoredel to provide default value for keyword delete
         on restore()
  960419 Added option dumbplot to provide default value for keyword dumb
         on plotting commands such as boxplot(), plot() and showplot()
  970211 Fixed bug (dimension of whichOptions)
  970313 Add options labelstyle and labelabove to control default labelling
         of unlabeled vectors, matrices and arrays
  970619 Added option inline to control default expansion of macros.
  971226 Fixed handling of "all" so that option 'lines' not set; only 'height'
  980624 Add option savehistry to provide default value for save()/
         asciisave() option 'history'
  981219 Add option plotdelay for Motif version to provide default delay
         length before drawing to graph window to overcome failure to
         set focus problem
  020808 Added option minpvalue
  030503 Added keyword phrase badoptok:T to suppress error or warning
         when unrecognized or illegal option is specified.  This is to
         make it easier to write macros which use getoptions() or
         setoptions() to still work in older versions of MacAnova
  030615 Carapace modifications made
  031216 Bug fix for setting MAXWHILE; setoptions(MAXWHILE:n) didn't have
         an immediate effect in a macro

*/

Symbolhandle    setoptions(Symbolhandle list)
{
	Symbolhandle     option, arg;
	long             nOptions = 0;
	long             i, j, nargs = NARGS(list), ncomps;
	long             optionNo, argno;
	int              isStructure = 0, set, getit;
	int              badOptionOK = 0;
	short            whichOptions[NOPTIONSALL];
	char            *keyword;
	char            *notavailFmt =
	  "WARNING: option %s is not available in this version%s";
	WHERE("setoptions");

	OUTSTR[0] = '\0';
	set = (strncmp(FUNCNAME,"set",3) == 0);

	if (nargs > 0)
	{
		Symbolhandle      symhKey = COMPVALUE(list, nargs - 1);
		char             *keyword = isKeyword(symhKey);

		if (keyword != (char *) 0 &&
			mystrncmp(keyword,"badoptok", 0) == 0)
		{
			if (!isTorF(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			badOptionOK = (DATAVALUE(symhKey, 0) != 0.0);
			nargs--;
		}
	} /*if (nargs > 0)*/

	if (nargs == 1 && COMPVALUE(list, 0) == EmptyArgument)
	{
		nargs = 0;
	} /*if (nargs == 1 && COMPVALUE(list, 0) == EmptyArgument)*/
	if (nargs == 0)
	{
		if (!set)
		{
			/* empty argument list to getoptions(); get everything */
			option = getOpt((short *) 0);
			return (option);
		}
		if (badOptionOK)
		{
			return (NULLSYMBOL);
		}
		sprintf(OUTSTR,
				"ERROR: no options specified to be set by %s()",
				FUNCNAME);
		goto errorExit;
	} /*if (nargs == 0)*/
	
	ncomps = nargs;

	if (nargs == 1)
	{
		arg = COMPVALUE(list,0);
		if (set && TYPE(arg) == STRUC)
		{ /* single STRUCT argument to setoptions() */
			isStructure = 1;
			list = arg;
			ncomps = NCOMPS(list);
		}
	} /*if (nargs == 1)*/
	else if (set && (argno = findKeyword(list,"default",0)) >= 0 &&
			 isTorF(arg = COMPVALUE(list,argno)) && DATAVALUE(arg,0) != 0.0)
	{
		sprintf(OUTSTR,
				"ERROR: '%s:T' must be the only argument", isKeyword(arg));
		goto errorExit;
	}

	for (i = 0;i < NOPTIONSALL;i++)
	{
		whichOptions[i] = 0;
	}

	/*
	  At this point either list is the actual argument list (isStructure == 0)
	  or the command is setOptions(Struc) and list is Struc (isStructure != 0)

	  If the command is setOptions() and isStructure == 0, all arguments must be
	  keyword phrases.

	  If the command is getOptions() all arguments must be keyword phrases or
	  CHARACTER scalars containing option names.
	*/

	for (i = 0;i < ncomps;i++)
	{
		arg = COMPVALUE(list,i);
		if (arg == EmptyArgument)
		{
			sprintf(OUTSTR, "ERROR: missing argument to %s()", FUNCNAME);
		}
		else if (isStructure || (keyword = isKeyword(arg)) != (char *) 0)
		{
			if (isStructure)
			{
				keyword = NAME(arg);
			} /*if (isStructure)*/
			else
			{ /* argument is keyword*/
				if (mystrncmp(keyword, "badoptok", 0) == 0)
				{
					if (!isTorF(arg))
					{
						notTorF(keyword);
						goto errorExit;
					}
					badOptionOK = (DATAVALUE(arg, 0) != 0.0);
					continue;
				}

				if (!set && strcmp(keyword, "all") == 0)
				{
					if (!isTorF(arg) || DATAVALUE(arg, 0) == 0.0)
					{
						sprintf(OUTSTR,
								"ERROR: %s must have value T", keyword);
						goto errorExit;
					}

					nOptions = 0;
					for (j = 0;j < NOPTIONS;j++)
					{ /* set everything T*/
						getit = (Options[j].code > 0 && j != ILINES - 1) ?
						  1 : 0;
						whichOptions[j] = getit;
						nOptions += getit;
					} /*for (j = 0;j < NOPTIONS;j++)*/
					continue;
				} /*if (!set && strcmp(keyword, "all") == 0)*/
			} /*if (isStructure){}else{}*/

			optionNo = isOption(keyword);
			if (!optionNo || !set && optionNo < -NOPTIONS)
			{
				if (!badOptionOK)
				{
					sprintf(OUTSTR,"ERROR: %s is unrecognized %s %s()",
							keyword,
							(isStructure) ?
							"component name in argument to" : "keyword on",
							FUNCNAME);
				} /*if (!badOptionOK)*/
			}
			else if (-NOPTIONS <= optionNo && optionNo < 0)
			{ /* option unavailable in version*/
				if (!badOptionOK)
				{
					sprintf(OUTSTR, notavailFmt, Options[-optionNo-1].name,
							(set) ? "; not set" : "; ignored");
				}

			}
			else if (optionNo > NOPTIONS && (!set || optionNo != IDEFAULTS))
			{
				if (!badOptionOK)
				{
					sprintf(OUTSTR,
							"ERROR: option %s illegal on %s()",
							Options[optionNo-1].name, FUNCNAME);
				}
			}
			else if (!set)
			{ /* getoptions() with keyword argument */
				keyword = Options[optionNo-1].name;
				if (!isTorF(arg))
				{
					notTorF(keyword);
					goto errorExit;
				}

				nOptions -= whichOptions[optionNo - 1];
				getit = (DATAVALUE(arg, 0) != 0.0) ? 1 : 0;
				whichOptions[optionNo - 1] = getit;
				nOptions += getit;
			} 
			else if (optionNo > 0)
			{
				whichOptions[optionNo - 1] = 1;
			}
		}
		else if (set)
		{ /* setoptions() with non-keyword argument*/
			sprintf(OUTSTR,
					"ERROR: %ld%s argument to %s() not keyword phrase",
					i+1, n_th(i+1), FUNCNAME);
		}
		else
		{ /*getoptions() with non-keyword argument*/
			if (!isCharOrString(arg) ||
				!(optionNo = isOption(STRINGPTR(arg))) ||
				optionNo < -NOPTIONS)
			{ /* can get values of public options only */
				sprintf(OUTSTR,
						"ERROR: non-keyword argument to %s() must be CHARACTER option name",
						FUNCNAME);
			}
			else if (optionNo < 0)
			{
				if (!badOptionOK)
				{
					sprintf(OUTSTR, notavailFmt, Options[-optionNo-1].name,
							"; ignored"); 
				}
			}
			else
			{
				whichOptions[optionNo - 1] = 1;
				nOptions++;
			}
		}
		
		if (*OUTSTR)
		{
			if (OUTSTR[0] != 'W')
			{
				goto errorExit;
			}
			
			/* WARNING only, don't quit */
			putErrorOUTSTR();
		} /*if (*OUTSTR)*/
	} /*for (i = 0;i < ncomps;i++)*/

	if (!set && nOptions == 0)
	{
		if (badOptionOK)
		{
			return (NULLSYMBOL);
		}
		sprintf(OUTSTR,
				"ERROR: no legal options specified for %s()", FUNCNAME);
		goto errorExit;
	}

	if (whichOptions[ILINES-1] && whichOptions[IHEIGHT-1])
	{
		if (!badOptionOK)
		{
			strcpy(OUTSTR,
				   "WARNING: 'height' and 'lines' are synonyms; 'lines' ignored");
			putErrorOUTSTR();
		}
		whichOptions[ILINES-1] = 0;
	} /*if (whichOptions[ILINES-1] && whichOptions[IHEIGHT-1])*/

	if (!set)
	{ /* getoptions */
		option = getOpt(whichOptions);
	} /*if (!set)*/
	else
	{ /* setoptions */
		setOpt(list, 0);
		option = NULLSYMBOL;
	} /*if (!set){...}else{...} */
	
	return (option);

  errorExit:
	putErrorOUTSTR();
	
	return (0);
	
} /*setoptions()*/

#define TEMPSTRINGSIZE 255

Symbolhandle getOpt(short * optionlist)
{
	Symbolhandle       optionStructure = SymbolZero;
	Symbolhandle       symh = SymbolZero;
	char              *optionName;
	char               tempstring[TEMPSTRINGSIZE+1];
	long               i, jcomp, all = (optionlist == (short *) 0);
	long               nOptions;
#ifdef TEK
	long               tekLength;
#endif /*TEK*/
	long               getit, option;
	WHERE("getOpt");
	
	OUTSTR[0] = '\0';
	nOptions = 0;
	for (i = 0; i < NOPTIONS; i++)
	{
		getit = (all && Options[i].code > 0 || !all && optionlist[i]) ? 1 : 0;
		nOptions += getit;
	} /*for (i = 0; i < NOPTIONS; i++)*/
	if (all)
	{
		nOptions--; /*adjust for height and lines being the same */
	}
	
	/*
	   nOptions must be > 0 or we wouldn't be in getOpt()
	*/
	if (nOptions > 1)
	{ /* create structure for output */
		optionStructure = StrucInstall(SCRATCH,nOptions);
		if (optionStructure == SymbolError)
		{
			goto errorExit;
		}
	} /*if (nOptions > 1)*/
	else
	{ /* create SCRATCH variable for output */
		symh = Install(SCRATCH,REAL); /* may end up as CHAR or LOGIC */
		if (symh == SymbolError)
		{
			goto errorExit;;
		}
		setNDIMS(symh,1);
		setDIM(symh,1,1);
	} /*if (nOptions > 1){...}else if (nOptions == 1){...} */
	
	jcomp = 0;
	
	for (i = 0;i < NOPTIONS;i++)
	{
		option = Options[i].code;
		optionName = Options[i].name;
		if (all && option > 0 && option != ILINES || !all && optionlist[i])
		{
			if (nOptions > 1)
			{
				symh = COMPVALUE(optionStructure,jcomp++) = Makesymbol(REAL);
				/* Note symh may end up CHAR or LOGIC so don't use Makereal*/
				if (symh == EmptyArgument)
				{
					goto errorExit;
				}
				setNDIMS(symh,1);
				if (nOptions > 1)
				{
					setNAME(symh,optionName);
				}
			} /*if (nOptions > 1)*/
		
			switch (option)
			{
			  case ISEEDS:
				TMPHANDLE = mygethandle(2 * sizeof(double));
				setDATA(symh,(double **) TMPHANDLE);
				if (TMPHANDLE == (char **) NULL)
				{
					goto errorExit;
				}
				setTYPE(symh,REAL);
				setDIM(symh,1,2);
				DATAVALUE(symh,0) = (double) RANDS1;
				DATAVALUE(symh,1) = (double) RANDS2;
				break;

				/* single double value */
			  case INSIG:

				/* single LOGICAL values */
			  case IBATCHECHO:
			  case IRESTOREDEL:
			  case IDUMBPLOT:
#ifdef SCROLLABLEWINDOW
			  case ISCROLLBACK:
#endif /*SCROLLABLEWINDOW*/
#ifdef ACTIVEUPDATE
			  case IUPDATE:
#endif /*ACTIVEUPDATE*/
			  case IWARNINGS:
			  case IFSTATS:
			  case IPVALS:
			  case IMINPVALUE:
			  case IEXCLUDEM:
			  case ILABELABOVE:
			  case IINLINE:
			  case IQUIET:
			  case ITRACEBACK:
#ifdef SAVEHISTORY
			  case ISAVEHISTRY:
#endif /*SAVEHISTORY*/				
				TMPHANDLE = mygethandle(sizeof(double));
				setDATA(symh,(double **) TMPHANDLE);
				if (TMPHANDLE == (char **) NULL)
				{
					goto errorExit;
				}	
				setDIM(symh,1,1);
				setTYPE(symh, (option == INSIG || option == IMINPVALUE) ?
						REAL : LOGIC);
				if (option == INSIG)
				{
					DATAVALUE(symh,0) = (double) PRINTFORMAT[1];
				}
				else if (option == IBATCHECHO)
				{
					DATAVALUE(symh,0) = (DEFAULTBATCHECHO) ? 1.0 : 0.0;
				}
				else if (option == IRESTOREDEL)
				{
					DATAVALUE(symh,0) = (DEFAULTRESTOREDEL) ? 1.0 : 0.0;
				}
				else if (option == IDUMBPLOT)
				{
					DATAVALUE(symh,0) = (DEFAULTDUMBPLOT) ? 1.0 : 0.0;
				}
#ifdef SCROLLABLEWINDOW
				else if (option == ISCROLLBACK)
				{
					DATAVALUE(symh,0) = (SCROLLBACK) ? 1.0 : 0.0;
				}
#endif /*SCROLLABLEWINDOW*/
#ifdef ACTIVEUPDATE
				else if (option == IUPDATE)
				{
					DATAVALUE(symh,0) = (ActiveUpdate) ? 1.0 : 0.0;
				}
#endif /*ACTIVEUPDATE*/
				else if (option == IWARNINGS)
				{
					DATAVALUE(symh,0) = (PRINTWARNINGS) ? 1.0 : 0.0;
				}
				else if (option == IFSTATS)
				{
					DATAVALUE(symh,0) = (PRINTFSTATS) ? 1.0 : 0.0;
				}
				else if (option == IPVALS)
				{
					DATAVALUE(symh,0) = (PRINTPVALS) ? 1.0 : 0.0;
				}
				else if (option == IMINPVALUE)
				{
					DATAVALUE(symh,0) = MINPVALUE;
				}
				else if (option == IEXCLUDEM)
				{
					DATAVALUE(symh,0) = (EXCLUDEMEDIAN) ? 1.0 : 0.0;
				}
				else if (option == ILABELABOVE)
				{
					DATAVALUE(symh,0) = (USECOLLABS) ? 1.0 : 0.0;
				}
				else if (option == IINLINE)
				{
					DATAVALUE(symh,0) = (EXPANDINLINE) ? 1.0 : 0.0;
				}
				else if (option == IQUIET)
				{
					DATAVALUE(symh,0) = (SWALLOWOUTPUT) ? 1.0 : 0.0;
				}
				else if (option == ITRACEBACK)
				{
					DATAVALUE(symh,0) = (TRACEBACK) ? 1.0 : 0.0;
				}
#ifdef SAVEHISTORY
				else if (option == ISAVEHISTRY)
				{
					DATAVALUE(symh,0) = (DEFAULTSAVEHISTORY) ? 1.0 : 0.0;
				}
#endif /*SAVEHISTORY*/
				else
				{				/* should not happen */
					setMissing(DATAVALUE(symh,0));
				}
				break;
				
				/* single string values */
			  case IFORMAT:
			  case IWFORMAT:
			  case IVECREAD:
			  case IANGLES:
			  case IPROMPT:
			  case IMISSING:
			  case ILABELSTYLE:
			  case IFINDMACROS:
#ifdef CANSETFONTS
			  case IFONT:
#endif /*CANSETFONTS*/
				if (option == IFORMAT)
				{
					sprintf(tempstring,"%ld.%ld%c",
							PRINTFORMAT[0],PRINTFORMAT[1],(char) PRINTFORMAT[2]);
				}
				else if (option == IWFORMAT)
				{
					sprintf(tempstring,"%ld.%ld%c",
							WRITEFORMAT[0],WRITEFORMAT[1],(char) WRITEFORMAT[2]);
				}
				else if (option == IVECREAD)
				{
					strcpy(tempstring,
						   (!VecreadByfields) ? "notbyfields" : "byfields");
				}
				else if (option == IANGLES)
				{
					double angles2cycles = 8.0*MV_PI_4;
					double angles2degrees = MV_PI_4/45.0;

#ifdef COMP_DJGPP
/*
	971231
	Following line added to overcome problem arising from optimization
	Without it, the "degrees" value of ANGLES2RADIANS is being recognized
	It can also be fixed by compiling without -O flag

	010528 Replaced tests for exact equality with "fuzzy" tests; it then
           worked correctly even with -O file
*/
					fprintf(stdout,"");
#endif /*COMP_DJGPP*/
					if (fabs(ANGLES2RADIANS - 1.0) < 1e-10)
					{
						strcpy(tempstring,"radians");
					}
					else if (fabs(ANGLES2RADIANS - angles2cycles) < 1e-10)
					{
						strcpy(tempstring,"cycles");
					}
					else if (fabs(ANGLES2RADIANS - angles2degrees) < 1e-10)
					{
						strcpy(tempstring,"degrees");
					}
					else		/* should not happen */
					{
						strcpy(tempstring,"radians");
						ANGLES2RADIANS = 1.0;
					}
				}
				else if (option == IPROMPT)
				{
					strncpy(tempstring, (char *) PROMPTS[0], TEMPSTRINGSIZE);
				}
				else if (option == IMISSING)
				{
					strncpy(tempstring, NAMEFORMISSING, TEMPSTRINGSIZE);
				}
				else if (option == ILABELSTYLE)
				{
					strcpy(tempstring, LEFTBRACKET);
				}
				else if (option == IFINDMACROS)
				{
					if (MACROSEARCH == noMacroSearch)
					{
						strcpy(tempstring, "no");
					}
					else if (MACROSEARCH == verboseMacroSearch)
					{
						strcpy(tempstring, "yes");
					}
					else
					{
						strcpy(tempstring, "silent");
					}
				}				
#ifdef CANSETFONTS
#ifdef MACINTOSH
				else if (option == IFONT)
				{
					p2cstrcpy(tempstring, CmdFontName);
				}				
#endif /*MACINTOSH*/
#endif /*CANSETFONTS*/
				tempstring[TEMPSTRINGSIZE] = '\0';
				
				TMPHANDLE = mygethandle((strlen(tempstring)+1)*sizeof(char));
				setSTRING(symh, TMPHANDLE);
				if (TMPHANDLE == (char **) NULL)	
				{
					goto errorExit;
				}	
				setTYPE(symh, CHAR);
				setDIM(symh, 1, 1);
				strcpy(STRINGPTR(symh), tempstring);
				break;

				/* single integer values */
			  case ILINES:
			  case IHEIGHT:
			  case IWIDTH:
			  case IERRORS:
			  case IMAXLINELEN:
#ifdef CANSETFONTS
			  case IFONTSIZE:
#endif /*CANSETFONTS*/
			  case IMAXWHILE:
#if defined(SKEL_WXWIN) && defined(BASE_UNIX) 
			  case IPLOTDELAY:
#endif /*SKEL_WXWIN && BASE_UNIX*/
#if defined(SKEL_WXWIN) && defined(OS_WIN32)
			  case IKEYBOARD:
#endif /*SKEL_WXWIN && OS_WIN32*/
#ifdef SCROLLABLEWINDOW
			  case IMATCHDELAY:
#endif /*SCROLLABLEWINDOW*/
#ifdef SAVEHISTORY
			  case IHISTORY:
#endif /*SAVEHISTORY*/
			  case IGUBED:
#ifdef MACINTOSH
			  case IPEELS:
#endif /*MACINTOSH*/
				TMPHANDLE = mygethandle(sizeof(double));
				setDATA(symh,(double **) TMPHANDLE);
				if (TMPHANDLE == (char **) NULL)
				{
					goto errorExit;
				}	
				setTYPE(symh,REAL);
				setDIM(symh,1,1);

				if (option == ILINES || option == IHEIGHT)
				{
					DATAVALUE(symh,0) = (double) SCREENHEIGHT;
				}
				else if (option == IWIDTH)
				{
					DATAVALUE(symh,0) = (double) SCREENWIDTH;
				}
				else if (option == IERRORS)
				{
					DATAVALUE(symh,0) = (double) MAXERRORS;
				}
				else if (option == IMAXLINELEN)
				{
					DATAVALUE(symh,0) = (double) MAXLINE;
				}
#ifdef CANSETFONTS
#ifdef MACINTOSH
				else if (option == IFONTSIZE)
				{
					DATAVALUE(symh,0) = (double) CmdFontSize;
				}
#endif /*MACINTOSH*/
#endif /*CANSETFONTS*/
				else if (option == IMAXWHILE)
				{
					DATAVALUE(symh,0) = (double) MAXWHILE - 1;
				}
#if defined(SKEL_WXWIN) && defined(BASE_UNIX) 
				else if (option == IPLOTDELAY)
				{
					DATAVALUE(symh,0) = (double) FocusDelay;
				}
#endif /*SKEL_WXWIN && BASE_UNIX*/
#if defined(SKEL_WXWIN) && defined(OS_WIN32) 
				else if (option == IKEYBOARD)
				{
					DATAVALUE(symh,0) = (double) KeyboardUse;
				}
#endif /*SKEL_WXWIN && OS_WIN32*/
#ifdef SCROLLABLEWINDOW
				else if (option == IMATCHDELAY)
				{
					DATAVALUE(symh,0) = (double) MatchDelay;
				}
#endif /*SCROLLABLEWINDOW*/
#ifdef SAVEHISTORY
				else if (option == IHISTORY)
				{
					DATAVALUE(symh,0) = (double) HISTORY;
				}
#endif /*SAVEHISTORY*/
				else if (option == IGUBED)
				{
					DATAVALUE(symh,0) = (double) GUBED;
				}
#ifdef MACINTOSH
				else if (option == IPEELS)
				{
					DATAVALUE(symh,0) =
					  (double) (CmdSleep +
								100 * BackgroundSleep +
								10000 * FgInterruptInterval +
								1000000 * InterruptInterval);
				}
#endif /*MACINTOSH*/
				break;
				
#if defined(SKEL_WXWIN)
			  case IENABLEITEM:
				  {
					  char      *printWindow = "printwindow";
					  char      *printGraph = "printgraph";
					  char      *noWindow, *noGraph;
					  char      *place;
					  int        needed;
				/* CHARACTER vector*/
					  
					  setTYPE(symh, CHAR);
					  setNDIMS(symh, 1);
					  setDIM(symh, 1, 2);
					  if (!(ENABLEDITEMS & enablePrintWindow))
					  {
						  noWindow = "no";
					  }
					  else
					  {
						  noWindow = NullString;
					  }
					  
					  if (!(ENABLEDITEMS & enablePrintGraph))
					  {
						  noGraph = "no";
					  }
					  else
					  {
						  noGraph = NullString;
					  }
					  
					  needed = strlen(printWindow) + strlen(printGraph) + 
						strlen(noWindow) + strlen(noGraph) + 2;
					  TMPHANDLE = mygethandle(needed);
					  setSTRING(symh, TMPHANDLE);
					  if (TMPHANDLE == (char **) 0)
					  {
						  goto errorExit;
					  }
					  place = *TMPHANDLE;
					  sprintf(place, "%s%s", noWindow, printWindow);
					  place = skipStrings(place, 1);
					  sprintf(place, "%s%s", noGraph, printGraph);
				  } 
				  break;

#endif /*SKEL_WXWIN*/
#ifdef TEK
			  case ITEKSET:

				/* CHARACTER vector */
				setTYPE(symh, CHAR);
				TMPHANDLE = mygethandle(strlen(ToTek) + strlen(FromTek) + 2);
				setSTRING(symh, TMPHANDLE);
				if (TMPHANDLE == (char **) 0)
				{
					goto errorExit;
				}
				setNDIMS(symh, 1);
				setDIM(symh, 1, 2);
				strcpy(*TMPHANDLE, ToTek);
				strcpy(skipStrings(*TMPHANDLE, 1), FromTek);
				break;
#endif /*TEK*/


			  default:
				/* should not get here */
				sprintf(OUTSTR,"WARNING: unrecognized option '%s'",optionName);
				putErrorOUTSTR();
			} /*switch (option)*/
		} /*if (all && option > 0 || !all && optionlist[i])*/
	} /*for (i = 0;i < NOPTIONS;i++)*/

	return ((nOptions > 1) ? optionStructure : symh);

  errorExit:
	
	if (nOptions > 1)
	{
		Removesymbol(optionStructure);
	}
	else
	{
		Removesymbol(symh);
	}

	return  (0);
} /*getOpt()*/


/*
  Function to do the actual work of setting options, called from setoptions()
  or restore().  If called from setoptions() list should either be an
  argument list (type LIST) of keyword arguments, in which case their names
  have already been checked for validity, or a structure (type STRUC), the
  names of whose components are option names which have also been checked
  for validity.  If called from restore(), list is a structure whose names
  have not been checked.  In any case, if an unrecognized name is found and
  quiet == 0, a warning message is printed. 

  031216 setting MAXWHILE also sets deeper level WHILELIMITS so that
         setoptions(MAXWHILE:maxwhile) affects a following while loop
         in a macro

*/  

void           setOpt(Symbolhandle list, long quiet)
{
	long            icomp, k, ncomps, length;
	double          value[3];
	long            isStructure;
	long            fmt[3];
	long            option, logValue;
	Symbolhandle    arg;
	Symbolhandle    symhLines = SymbolZero;
	Symbolhandle    symhHeight = SymbolZero;
	char           *keyword;
	int             reply;
	long            newWidth = -1, newLines = -1;
	char           *mustbe = "WARNING: value for %s must be %s";
	char           *mustbeChar =
		"WARNING: value for %s must be CHARACTER scalar";
	char           *mustbe2 = "WARNING: value for %s must be integer >= %ld";
	char           *toolongChar =
		"WARNING: %s must have no more than %ld characters";
#if defined(SKEL_CARAPACE)
	cpc_io          io;
#endif /*CARAPACE*/
#ifdef CANSETFONTS
	int             changeFont = 0;
	long            newFontSize = -1, fontSize;
	char           *newFontName = (char *) 0;
#endif /*CANSETFONTS*/
	WHERE("setOpt");
	
/*
   it should be the case that all arguments should be valid option names
   Also, unless called from restore() (quiet != 0), any unavailable options
   should have been diagnosed
*/
	isStructure = (TYPE(list) == STRUC);
	ncomps = NCOMPS(list);
	for (icomp = 0;icomp < ncomps;icomp++)
	{
		arg = COMPVALUE(list,icomp);
		if (isStructure)
		{
			keyword = NAME(arg);
		}
		else
		{
			keyword = isKeyword(arg);
			/* should not need to check that non-null */
		}
		option = isOption(keyword);
		if (option < -NOPTIONS || option > 0)
		{
			OUTSTR[0] = '\0';
			option = (option < 0) ? -option  : option;
			keyword = Options[option-1].name;
			length = symbolSize(arg);

			switch (option)
			{
			  case ISEEDS:
				if (TYPE(arg) != REAL || length != 2)
				{
					sprintf(OUTSTR, mustbe, keyword, "vector of two integers");
				}
				else if ((value[0] = DATAVALUE(arg,0)) < 0 || 
						(value[1] = DATAVALUE(arg,1)) < 0 ||
						value[0] > MAXSEED ||
						value[1] > MAXSEED ||
						value[0] != floor(value[0]) ||
						value[1] != floor(value[1]))
				{
					sprintf(OUTSTR,
							"WARNING: %s must be positive integers <= %ld",
							keyword,MAXSEED);
				}
				else
				{
					RANDS1 = (long) value[0];
					RANDS2 = (long) value[1];
				}
				break;
			
			  case INSIG:
			
				if (!isInteger(arg, NONNEGATIVEVALUE) ||
					(value[0] = DATAVALUE(arg, 0)) > 20)
				{
					sprintf(OUTSTR, mustbe, keyword, "integer between 0 and 20");
				}
				else
				{
					PRINTFORMAT[0] = (long) value[0] + 7;
					PRINTFORMAT[1] = (long) value[0];
					PRINTFORMAT[2] = (long) 'g';
					installFormat(PRINTFORMAT[0],PRINTFORMAT[1],
								  PRINTFORMAT[2]);
				}
				break;

			  case IMINPVALUE:
				if (!isNumber(arg, NONNEGATIVEVALUE) ||
					(value[0] = DATAVALUE(arg, 0)) > 1.0001e-3)
				{
					sprintf(OUTSTR,
							mustbe, keyword, "number between 0 and .001");
				}
				else
				{
					/*round to nearest x.x*10^-d*/
					
					if (value[0] == 0.0)
					{
						MINPVALUE = 0.0;
					}
					else
					{
						double       pow10;

						pow10 = pow(10.0,floor(log10(value[0]*1.0000001)))/10.0;
						MINPVALUE = pow10*floor(value[0]/pow10 + .5);
						if (fabs(MINPVALUE/value[0] - 1.0) > 1e-8)
						{
							sprintf(OUTSTR,
									"WARNING: value of %s rounded to %.2g",
									keyword, MINPVALUE);
							putErrorOUTSTR();
						}
					}
				}
				break;
				
			  case IBATCHECHO:
			  case IRESTOREDEL:
#ifdef SCROLLABLEWINDOW
			  case ISCROLLBACK:
#endif /*SCROLLABLEWINDOW*/
			  case IDUMBPLOT:
#ifdef ACTIVEUPDATE
			  case IUPDATE:
#endif /*ACTIVEUPDATE*/
			  case IWARNINGS:
			  case IFSTATS:
			  case IPVALS:
			  case IEXCLUDEM:
			  case IDEFAULTS:
			  case ILABELABOVE:
			  case IINLINE:
			  case IQUIET:
			  case ITRACEBACK:
#ifdef SAVEHISTORY
			  case ISAVEHISTRY:
#endif /*SAVEHISTORY*/
				if (!isTorF(arg))
				{
					sprintf(OUTSTR, mustbe, keyword, "T of F");
				}
				else
				{
					logValue = DATAVALUE(arg,0) != 0.0;
					if (option == IBATCHECHO)
					{
						DEFAULTBATCHECHO = logValue;
					}
					else if (option == IRESTOREDEL)
					{
						DEFAULTRESTOREDEL = logValue;
					}
					else if (option == IDUMBPLOT)
					{
#ifdef MACINTOSH
						if (!UseWindows && !logValue)
						{
							sprintf(OUTSTR,
									"WARNING: %s:F illegal in non-interactive mode",
									keyword);
							logValue = 1;
						}
#endif /*MACINTOSH*/
						DEFAULTDUMBPLOT = logValue;
					}
#ifdef SCROLLABLEWINDOW
					else if (option == ISCROLLBACK)
					{
						SCROLLBACK = logValue;
					}
#endif /*SCROLLABLEWINDOW*/
#ifdef ACTIVEUPDATE
					else if (option == IUPDATE)
					{
						ActiveUpdate = logValue;
					}
#endif /*ACTIVEUPDATE*/
					else if (option == IWARNINGS)
					{
						PRINTWARNINGS = logValue;
					}
					else if (option == IFSTATS)
					{
						PRINTFSTATS = logValue;
					}
					else if (option == IPVALS)
					{
						PRINTPVALS = logValue;
					}
					else if (option == IEXCLUDEM)
					{
						EXCLUDEMEDIAN = logValue;
					}
					else if (option == ILABELABOVE)
					{
						USECOLLABS = logValue;
					}
					else if (option == IINLINE)
					{
						EXPANDINLINE = logValue;
					}
					else if (option == IQUIET)
					{	
						int        currentQuiet =
						  ((ACTIVEMACROS > 0) ? ThisMacroQuiet : QUIET);
						
						if (logValue)
						{
							currentQuiet |= windowQuiet;
						}
						else
						{
							currentQuiet &= ~windowQuiet;
						}
						
						if (ACTIVEMACROS > 0)
						{
							ThisMacroQuiet = currentQuiet;
						}
						else
						{
							QUIET = currentQuiet;
						}
					}
					else if (option == ITRACEBACK)
					{
						TRACEBACK = logValue;
					}
#ifdef SAVEHISTORY
					else if (option == ISAVEHISTRY)
					{
						DEFAULTSAVEHISTORY =
						  (ISATTY & ITTYIN) ?  logValue : 0;
					}
#endif /*SAVEHISTORY*/
					else if (logValue)
					{			/* set defaults, do nothing if F */
						setDefaultOptions();
					}
				}
			
				break;
			
			  case IFORMAT:
			  case IWFORMAT:
				if (!isCharOrString(arg))
				{
					sprintf(OUTSTR, mustbeChar, keyword);
				}
				else if (!(reply = setFormat(STRINGPTR(arg),fmt)))
				{
					sprintf(OUTSTR,"WARNING: '%s' is not proper value for %s",
							STRINGPTR(arg), keyword);
				}
				else
				{
					for (k = 0;k<3;k++)
					{
						if (option == IFORMAT)
						{
							PRINTFORMAT[k] = fmt[k];
						}
						else
						{
							WRITEFORMAT[k] = fmt[k];
						}
					}
					if (option == IFORMAT)
					{
						installFormat(PRINTFORMAT[0],PRINTFORMAT[1],
									  PRINTFORMAT[2]);
					}
					if (reply < 0)
					{
						reply = -reply;
						if (reply % 10)
						{
							sprintf(OUTSTR,
									"WARNING: %s width > %ld reduced to %ld",
									keyword, (long) MAXFMTWIDTH,
									(long) MAXFMTWIDTH);
							putErrorOUTSTR();
						} /*if (reply % 10)*/
						if (reply / 10)
						{
							sprintf(OUTSTR,
									"WARNING: %s %s > %ld reduced to %ld",
									keyword,
									(fmt[2] == 'g') ?
									"significant digits" : "decimals",
									(long) MAXFMTDIGITS, (long) MAXFMTDIGITS);
							putErrorOUTSTR();
						} /*if (reply / 10)*/
					} /*if (reply < 0)*/
				}
				break;

			  case IVECREAD:
				if (!isCharOrString(arg))
				{
					sprintf(OUTSTR, mustbeChar, keyword);
				}
				else if (mystrncmp(STRINGPTR(arg), "notbyf", 6) != 0 &&
						 mystrncmp(STRINGPTR(arg), "byf", 3) != 0)
				{
					sprintf(OUTSTR, mustbe, keyword,
							"\"notbyfields\" or \"byfields\"");
				}
				else
				{
					char       c = tolower(STRINGVALUE(arg, 0));
					
					VecreadByfields = (c == 'b') ? 1 : 0;
				}
				break;
				
			  case IANGLES:
				if (!isCharOrString(arg))
				{
					sprintf(OUTSTR, mustbeChar, keyword);
				}
				else if (strncmp(STRINGPTR(arg),"radian",6) == 0)
				{
					ANGLES2RADIANS = 1.0;
				}
				else if (strncmp(STRINGPTR(arg),"cycle",5) == 0)
				{
					ANGLES2RADIANS = 8.0*MV_PI_4;
				}
				else if (strncmp(STRINGPTR(arg),"degree",6) == 0)
				{
					ANGLES2RADIANS = MV_PI_4/45.0;
				}
				else
				{
					sprintf(OUTSTR, mustbe, keyword,
							"\"radians\", \"cycles\", or \"degrees\"");
				}
				break;

			  case IFINDMACROS:
				if (!isCharOrString(arg))
				{
					sprintf(OUTSTR, mustbeChar, keyword);
				}
				else if (strncmp(STRINGPTR(arg),"n",1) == 0)
				{
					MACROSEARCH = noMacroSearch;
				}
				else if (strncmp(STRINGPTR(arg),"y",1) == 0)
				{
					MACROSEARCH = verboseMacroSearch;
				}
				else if (strncmp(STRINGPTR(arg),"silent",6) == 0)
				{
					MACROSEARCH = quietMacroSearch;
				}
				else
				{
					sprintf(OUTSTR, mustbe, keyword,
							"\"yes\", \"no\", or \"silent\"");
				}
				break;

#ifdef CANSETFONTS
			  case IFONT:
				if (!isCharOrString(arg))
				{
					sprintf(OUTSTR, mustbeChar, keyword);
				} /*if (!isCharOrString(arg))*/
				else
				{
					reply = decodeFont(STRINGPTR(arg), &newFontName, &fontSize);
					if (reply == BADFONTSIZE)
					{
						sprintf(OUTSTR,
								"WARNING: %s specifies a bad font size",
								STRINGPTR(arg));
					}
					else if (reply == FONTNAMETOOLONG)
					{
						sprintf(OUTSTR,
								"WARNING: %s specifies too long a font name",
								STRINGPTR(arg));
					}
					else if (reply == BADFONTSPEC)
					{
						sprintf(OUTSTR,
								"WARNING: '%s' not a valid font specification",
								STRINGPTR(arg));
					}
					else
					{
						if (fontSize > 0)
						{
							newFontSize = fontSize;
						}
						if (!changeFont)
						{
							changeFont = !isCurrentFont(newFontName, newFontSize);
						}
					}					
				} /*if (!isCharOrString(arg)){}else{}*/
				break;
#endif /*CANSETFONTS*/
			
			  case IPROMPT:
				if (!isCharOrString(arg))
				{
					sprintf(OUTSTR, mustbeChar, keyword);
				}
				else if (strlen(STRINGPTR(arg)) > MAXPROMPT - 1)
				{
					sprintf(OUTSTR, toolongChar,
							keyword, (long) MAXPROMPT - 1);
				}
				else
				{
					strcpy(PROMPT, STRINGPTR(arg));
				}
				break;

			  case IMISSING:
				if (!isCharOrString(arg))
				{
					sprintf(OUTSTR, mustbeChar, keyword);
				}
				else if (strlen(STRINGPTR(arg)) > LENGTHMISSING)
				{
					sprintf(OUTSTR, toolongChar,
							keyword, (long) LENGTHMISSING);
				}
				else
				{
					strcpy(NAMEFORMISSING,STRINGPTR(arg));
				}
				break;

			  case ILABELSTYLE:
				if (!isCharOrString(arg))
				{
					sprintf(OUTSTR, mustbeChar, keyword);
				}
				else
				{
					char      c = STRINGVALUE(arg, 0);
					char     *pc = (c) ? strchr(Lparens+1, c) : (char *) 0;
					
					if (strlen(STRINGPTR(arg)) != 1 || pc == (char *) 0)
					{
						sprintf(OUTSTR, mustbe, keyword,
						"one of \"(\", \"[\", \"{\", \"<\", \"/\", or \"\\\"");
					}
					else
					{
						LEFTBRACKET[0] = c;
						RIGHTBRACKET[0] = Rparens[pc - Lparens];
					}
				}
				break;
				
				/*
				   integer valued options
				   */
			  case ILINES:
			  case IHEIGHT:
			  case IWIDTH:
			  case IERRORS:
			  case IFONTSIZE:
			  case IMAXWHILE:
#if defined(SKEL_WXWIN) && defined(BASE_UNIX) 
			  case IPLOTDELAY:
#endif /*SKEL_WXWIN && BASE_UNIX*/
#if defined(SKEL_WXWIN) && defined(OS_WIN32) 
			  case IKEYBOARD:
#endif /*SKEL_WXWIN && OS_WIN32*/
#ifdef SCROLLABLEWINDOW
			  case IMATCHDELAY:
#endif /*SCROLLABLEWINDOW*/
			  case IMAXLINELEN:
#ifdef SAVEHISTORY
			  case IHISTORY:
#endif /*SAVEHISTORY*/
			  case IGUBED:
#ifdef MACINTOSH
			  case IPEELS:
#endif /*MACINTOSH*/
				value[0] = DATAVALUE(arg,0);
				if (TYPE(arg) != REAL || length != 1 || value[0] != floor(value[0]))
				{
					sprintf(OUTSTR, mustbe, keyword, "single integer");
				}
				else if (option == ILINES || option == IHEIGHT)
				{
					if (value[0] != 0.0 && value[0] < MINSCREENHEIGHT)
					{
						sprintf(OUTSTR,
								"WARNING: value for %s must be 0 or an integer >= %ld",
								keyword, (long) MINSCREENHEIGHT);
					}
					else
					{
						newLines = (long) value[0];
						if (option == ILINES)
						{
							symhLines = arg;
						}
						else
						{
							symhHeight = arg;
						}
					}
				}
				else if (option == IWIDTH)
				{
					if (value[0] < MINSCREENWIDTH)
					{
						sprintf(OUTSTR, mustbe2,
								keyword, (long) MINSCREENWIDTH);
					}
					newWidth = (long) value[0];
				}
				else if (option == IERRORS)
				{
					if (value[0] < 0)
					{
						sprintf(OUTSTR, mustbe2, keyword, 0L);
					}
					else
					{
						if (BDEPTH == 0)
						{
							MAXERRORS1 = MAXERRORS = (long) value[0];
						}
						else
						{
							MAXERRORS1 = (long) value[0];
						}
					}
				}
#ifdef CANSETFONTS
				else if (option == IFONTSIZE)
				{
					newFontSize = (long) value[0];
					if (newFontSize < MINFONTSIZE || newFontSize > MAXFONTSIZE)
					{
						sprintf(OUTSTR,
								"WARNING: illegal font size %ld", newFontSize);
					}
					else if (!changeFont)
					{
						changeFont = !isCurrentFont((char *) 0, newFontSize);
					}
				}
#endif /*CANSETFONTS*/
				else if (option == IMAXWHILE)
				{
					long      newMaxwhile = (long) value[0];

					if (newMaxwhile < MINMAXWHILE)
					{
						sprintf(OUTSTR, mustbe2, keyword, MINMAXWHILE);
					}
					else
					{
						int          i;

 						MAXWHILE = newMaxwhile + 1;
						for (i = WDEPTH; i < MAXWDEPTH; i++)
						{
							WHILELIMITS[i] = MAXWHILE;
						}
					}
				}
#if defined(SKEL_WXWIN) && defined(BASE_UNIX) 
				else if (option == IPLOTDELAY)
				{
					long      newPlotdelay = (long) value[0];

					if (newPlotdelay < 0)
					{
						sprintf(OUTSTR, mustbe2, keyword, 0L);
					}
					else
					{
						FocusDelay = newPlotdelay;
					}
				}
#endif /*SKEL_WXWIN && BASE_UNIX*/
#if defined(SKEL_WXWIN) && defined(OS_WIN32) 
				else if (option == IKEYBOARD)
				{
					long      newKeyboardUse = (long) value[0];

					if (newKeyboardUse < 0)
					{
						sprintf(OUTSTR, mustbe2, keyword, 0L);
					}
					else
					{
						KeyboardUse = newKeyboardUse;
					}
				}
#endif /*SKEL_WXWIN && OS_WIN32*/
#ifdef SCROLLABLEWINDOW
				else if (option == IMATCHDELAY)
				{
					long      newMatchdelay = (long) value[0];

					if (newMatchdelay < 0)
					{
						sprintf(OUTSTR, mustbe2, keyword, 0L);
					}
					else
					{
						MatchDelay = newMatchdelay;
					}
				}
#endif /*SCROLLABLEWINDOW*/
#define MINLINE 80L
				else if (option == IMAXLINELEN)
				{
					long     newLinemax = (long) value[0];

					if (newLinemax < MINLINE)
					{
						sprintf(OUTSTR, mustbe2, keyword, MINLINE);
					}
					else
					{
						MAXLINE = newLinemax;
					}
				}
#ifdef SAVEHISTORY
				else if (option == IHISTORY)
				{
					long      newHistory = (long) value[0];

					if (newHistory < 0)
					{
						sprintf(OUTSTR, mustbe2, keyword, 0L);
					}
					else
					{
#if  !defined(SKEL_CARAPACE)
						initHistory(newHistory);
#endif /*CARAPACE*/
					}
				}
#endif /*SAVEHISTORY*/
				else if (option == IGUBED)
				{
					GUBED = (long) value[0];
				}

#ifdef MACINTOSH
				else if (option == IPEELS && value[0] >= 0)
				{
					long         isleep;
					/*
						CmdSleep +
						    100 * BackgroundSleep +
						  10000 * FgInterruptInterval +
						1000000 * InterruptInterval
						Sleep value of 99 means do not change
						Example:
						  setoptions(PEELS:99300299)
					 	    sets  BackgroundSleep to 2
							sets  FgInterruptInterval to 30
							does not change CmdSleep
							does not change BgInterruptInterval
					*/
#if (0)
					PRINT("CmdSleep = %d,BackgroundSleep = %d,FgInterruptInterval = %d,BgInterruptInterval = %d\n",
						  CmdSleep,BackgroundSleep,FgInterruptInterval,
						  BgInterruptInterval);
#endif /*(0)*/
					isleep = (long) value[0];
					if (isleep % 100 != 99)
					{
						CmdSleep = isleep % 100;
					}
					isleep /= 100;
					if (isleep % 100 != 99)
					{
						BackgroundSleep = isleep % 100;
					}
					isleep /= 100;
					if (isleep % 100 != 99)
					{
						FgInterruptInterval = isleep % 100;
					}
					isleep /= 100;
					if (isleep % 100 != 99)
					{
						BgInterruptInterval = isleep % 100;
					}
					SkelSetWaitTimes(CmdSleep, BackgroundSleep);
#if (0)
					PRINT("CmdSleep = %d,BackgroundSleep = %d,FgInterruptInterval = %d,BgInterruptInterval = %d\n",
						  CmdSleep,BackgroundSleep,FgInterruptInterval,
						  BgInterruptInterval);
#endif /*(0)*/
				}
#endif /*MACINTOSH*/
				break;

#if defined(SKEL_WXWIN) 
			  case IENABLEITEM:
				if (TYPE(arg) != CHAR || !isVector(arg) ||
					length > maxEnabledItems)
				{
					char          required[50];
					
					sprintf(required, "CHARACTER vector, length <= %d",
							maxEnabledItems);
					sprintf(OUTSTR, mustbe, keyword, required);
				}
				else
				{
					char           *pc = STRINGPTR(arg);
					int             iword;
					unsigned long   enabledItems = ENABLEDITEMS;
					
					for (iword = 0; iword < length; iword++)
					{
						int          enableIt;
						int          enableWhat;
						
						if (mystrncmp(pc, "no", 2) == 0)
						{
							enableIt = 0;
							pc += 2;
						}
						else
						{
							enableIt = 1;
						}
						if (mystrncmp(pc, "printwindow", 6) == 0)
						{
							enableWhat = enablePrintWindow;
						}
						else if (mystrncmp(pc, "printgraph", 6) == 0)
						{
							enableWhat = enablePrintGraph;
						}
						else
						{
							sprintf(OUTSTR,
									"WARNING: unrecognized value \"%s\" for %s",
									skipStrings(STRINGPTR(arg), iword), keyword);
							break;
						}
						if (enableIt)
						{
							enabledItems |= enableWhat;
						}
						else
						{
							enabledItems &= ~enableWhat;
						}
						pc = skipStrings(pc, 1);
					} /*for (iword = 0; iword < length; iword++)*/
					if (*OUTSTR == '\0')
					{
						ENABLEDITEMS = enabledItems;
					}
				}
				break;
#endif /*SKEL_WXWIN*/
					
#ifdef TEK
			  case ITEKSET:
				if (TYPE(arg) != CHAR || !isVector(arg) || length != 2)
				{
					sprintf(OUTSTR,
							mustbe, keyword, "CHARACTER vector of length 2");
				}
				else
				{
					char       *pc = STRINGPTR(arg);
					
					if (strlen(pc) >= TEKLENGTH ||
						strlen(skipStrings(pc, 1)) >= TEKLENGTH)
					{
						sprintf(OUTSTR,
								"WARNING: element of %s is longer than %d characters",
								keyword, TEKLENGTH - 1);
					}
					else
					{
						strcpy(ToTek, pc);
						pc = skipStrings(pc, 1);
						strcpy(FromTek, pc);
					}
				}
				break;
#endif /*TEK*/
			  default:
				/* should not happen */
				sprintf(OUTSTR,"WARNING: unrecognized option '%s'",keyword);
			
			} /*switch (option)*/

			if (*OUTSTR)
			{
				if (!quiet)
				{
					sprintf(OUTSTR + strlen(OUTSTR),"; %s not set",keyword);
					putErrorOUTSTR();
				}
				else
				{
					*OUTSTR = '\0';
				}
			} /*if (*OUTSTR)*/
		} /*if (option < -NOPTIONS || option > 0)*/
	} /*for (icomp = 0;icomp < ncomps;icomp++)*/

#ifdef CANSETFONTS
	if (changeFont)
	{
		setNewFont(newFontName, newFontSize);
	}
#endif /*CANSETFONTS*/
	/*
		set SCREENWIDTH and SCREENHEIGHT after setNewFont()
		since setNewFont() can change them
	*/
	if (symhLines != SymbolZero && symhHeight != SymbolZero)
	{
		newLines = (long) DATAVALUE(symhHeight, 0);
	}
	installScreenDims((newWidth > 0) ? newWidth : SCREENWIDTH,
					  (newLines >= 0) ? newLines : SCREENHEIGHT);
	saveFormat(); /* save copy of formats and screen dimensions*/

#ifdef ACTIVEUPDATE
	if (newLines >= 0)
	{
		activeUpdateInit();
	}
#endif /*ACTIVEUPDATE*/
}  /*setOpt()*/

/*
  Check keyword name to see if it is valid option name
*/

long isOption(char *keyword)
{
	optionInfo    *option;
	
	for (option = Options;option->length;option++)
	{
		if (strncmp(option->name, keyword, (int) option->length) == 0)
		{
			return (option->code);
		}
	} /*for (option = Options;option->length;option++)*/
	return (0);
} /*isOption()*/

#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Macmain
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

/*
   Implements function elapsedtime()
   Usage:
     gettime()                   prints time in seconds since start
	 gettime(quiet:F)            ditto
     gettime(quiet:T)            saves time since start for future reference
     gettime(keep:T)             returns time since start but prints nothing
	 gettime(keep:T,quiet:F)     returns and prints time since start
	 gettime(interval:T)         prints time in seconds since previous call
     gettime(interval:T,quiet:F) ditto
     gettime(interval:T,keep:T)  returns time since previous call, no print
     gettime(interval:T,keep:T,quiet:F) returns and prints time since start

   960306 Installed
   960307 Modified to add keyword interval
*/
#define TimeSinceLast   times[0]
#define TimeSinceFirst  times[1]

Symbolhandle elapsedTime(Symbolhandle list)
{
	Symbolhandle       result = SymbolZero, symhKey;
	long               nargs = NARGS(list);
	char              *keyword;
	long               reportit = -1, keepit = 0, interval = 0, i;
	double             times[2], timeWanted;
	
	getElapsedTime(times);
	if (nargs > 3)
	{
		badNargs(FUNCNAME, -3);
		goto errorExit;
	}
	
	if (nargs > 1 || COMPVALUE(list, 0) != EmptyArgument)
	{
		for (i = 0; i < nargs; i++)
		{
			symhKey = COMPVALUE(list, i);
			if (!argOK(symhKey, 0, (nargs > 1) ? i+1 : 0))
			{
				goto errorExit;
			}
			if (!(keyword = isKeyword(symhKey)))
			{
				sprintf(OUTSTR,
						"ERROR: argument %ld to %s() not keyword phrase",
						i+1, FUNCNAME);
				goto errorExit;
			}
			if (strcmp(keyword, "quiet") != 0 && strcmp(keyword, "keep") != 0
				&& strncmp(keyword, "int", 3) != 0)
			{				
				badKeyword(FUNCNAME, keyword);
				goto errorExit;
			}
			if (!isTorF(symhKey))
			{
				notTorF(keyword);
				goto errorExit;
			}
			if (keyword[0] == 'q')
			{
				reportit = (DATAVALUE(symhKey, 0) == 0.0);
			}
			else if (keyword[0] == 'i')
			{
				interval = (DATAVALUE(symhKey, 0) != 0.0);
			}
			else
			{
				keepit = (DATAVALUE(symhKey, 0) != 0.0);
			}
		} /*for (i = 0; i < nargs; i++)*/
		if (keepit && reportit < 0)
		{
			reportit = 0;
		}
	} /*if (nargs > 1 || COMPVALUE(list, 0) != EmptyArgument)*/

	timeWanted = (interval) ? TimeSinceLast : TimeSinceFirst;
	if (reportit)
	{
		char      *buffer = OUTSTR;
		
		strcpy(buffer,
			   (interval) ? "Elapsed time is " : "Time since start is ");
		buffer += strlen(buffer);
		buffer += formatDouble(buffer, timeWanted , TRIMLEFT | TRIMRIGHT);
		strcpy(buffer," seconds");
		putOUTSTR();
	} /*if (reportit)*/

	if (keepit)
	{
		result = RInstall(SCRATCH, 1);
		if (result == SymbolError)
		{
			goto errorExit;
		}
		DATAVALUE(result, 0) = timeWanted;
	} /*if (keepit)*/
	else
	{
		result = NULLSYMBOL;
	}
	
	return (result);

  errorExit:
	putErrorOUTSTR();
	return (0);

} /*elapsedTime()*/
