/*
*(C)* The file is part of the source distribution of MacAnova
*(C)* version 5.00 or later
*(C)*
*(C)* Copyright (c) 2005 by Gary Oehlert and Christopher Bingham
*(C)* unless indicated otherwise
*(C)*
*(C)* You may give out copies of this software;  for conditions see the
*(C)* file COPYING included with this distribution
*(C)*
*(C)* This file is distributed WITHOUT ANY WARANTEE; without even
*(C)* the implied warantee of MERCHANTABILITY or FITNESS FOR
*(C)* A PARTICULAR PURPOSE
*(C)*
*/


#ifdef SEGMENTED
#if defined(MPW) || defined(MW_CW)
#pragma segment Solve
#endif /*MPW||MW_CW*/
#endif /*SEGMENTED*/

#include "globals.h"
#include "mainpars.h"
#include "blas.h"   /* prototypes and macros for names */
#include "linpack.h"   /* prototypes and macros for names */

/* 
  Routine to compute solve(a), solve(a,b), and det(a), where a is square
  Change in check for singularity

  050613 added ludecomp(a [, pack:F])
*/

enum funcCodes
{
	IDET       = 0x01,
	ISOLVE     = 0x02,
	IRSOLVE    = 0x04,
	IINVERSE   = 0x08,
	ILUDCMP    = 0x10,
	ISOLVERS   = ISOLVE | IRSOLVE | IINVERSE | ILUDCMP,
	IALLOPS    = ISOLVERS | IDET
};

enum solveKeyCodes
{
	KMANTEXP,
	KQUIET,
	KSINGOK,
	KMINRCOND,
	KPACK,
	NSOLVEKEYS
};

static keywordList    SolveKeys[] =
{
	InitKeyEntry("mantexp", 0, IDET, LOGICSCALAR),
	InitKeyEntry("quiet", 0, IALLOPS, LOGICSCALAR),
	InitKeyEntry("singok", 4, ISOLVERS, LOGICSCALAR),
	InitKeyEntry("minrcond", 0, ISOLVERS, POSITIVESCALAR),
	InitKeyEntry("pack", 4, ILUDCMP, LOGICSCALAR)
};

#define         Mantexp      KeyLogValue(SolveKeys,KMANTEXP)
#define         Quiet        KeyLogValue(SolveKeys,KQUIET)
#define         SingOK       KeyLogValue(SolveKeys,KSINGOK)
#define         Pack         KeyLogValue(SolveKeys,KPACK)
#define         MinRcond     KeyRealValue(SolveKeys,KMINRCOND)

enum scratchCodes
{
	GIPVT = 0,
	GSCRATCH,
	NTRASH
};

/*
   Routine to implement solve(a), solve(a,b) (a %\% b),
   rsolve(a,b) (b %/% a), det(a), det(x,mantexp:T)
   960320 If a has labels, solve(a) has same labels)
   960329 If a and b have labels, solve(a,b) has labels
      structure(getlabels(a,1),getlabels(b,2) and rsolve(a,b) has
	  labels structure(getlabels(b,1),getlabels(a,2))	  
   971126 Modified to use append codes in labels.h
   980819 added keyword phrase mantexp:T to det().
   981227 added keyword phrase singok:T to solve().  When present,
          solve(a) and solve(a,x) return NULLSYMBOL when a is singular
          Fixed bug for det(a,mantexp:F) which produced a vector
          instead of a scalar.
          Added keyword phrase 'quiet:T' to det() to suppress warning
          message when argument is singular
          det(a) and det(a,mantexp:T) always return 0 when a is singular
   990710 Minor change to fix mis-identification of operands to %/% whose
          order is swapped in Arith() in Lang.c
   000407 'singularok:T' now accepted as synonym for 'sing:ok'
   020913 added keyword phrase minrcon:eps
   020916 overflow is now an error, not just a condition triggering a warning
   050613 added ludecomp() to return the LU decomposition of a square
          matrix; it is used like solve, but returns struction(LU,rowperm,sign)
          where LU is L (without the diagonal) and U in packed form,
          rowperm is permutation of rows of a so that a[rowperm,] = L %*% U
*/
Symbolhandle solve(Symbolhandle list)
{
	Symbolhandle     result = (Symbolhandle) 0;
	Symbolhandle     arg1, arg2;
	Symbolhandle     out1;
	Symbolhandle     symhKey = (Symbolhandle) 0;
	double         **scratchH;
	long           **ipvtH;
	long             op;
	long             nargs = NARGS(list);
	long             dim[2], nrows, ncols;
	long             needed;
	long             length, inc1 = 1;
	long            *ipvt;
	long             job;
	int              matop;
	int              argno1 = 1, argno2 = 2, startKey = nargs;
	double          *scratch, *a, *work, *x, *b, *tmprow;
	double           rcond;
	double           determ[2]; 
	char            *keyword, *leftRight;
	WHERE("solver");
	TRASH(NTRASH,errorExit);

	*OUTSTR = '\0';

	setLogicalFalse(Mantexp);
	setLogicalFalse(SingOK);
	unsetLogical(Quiet);
	unsetLogical(Pack);
	unsetReal(MinRcond);
	
	matop = FUNCNAME[0] == '%';

	if (strcmp(FUNCNAME,"det") == 0)
	{
		op = IDET;
		startKey = 1;
	}
	else if (strcmp(FUNCNAME,"ludecomp") == 0)
	{
		op = ILUDCMP;
		startKey = 1;
	}
	else if (strcmp(FUNCNAME,"solve") == 0 ||
			 strcmp(FUNCNAME, opName(DIVMAT)) == 0)
	{
		op = (nargs > 1 && isKeyword(COMPVALUE(list, 1)) == (char *) 0) ?
		  ISOLVE : IINVERSE;
		startKey = (op == IINVERSE) ? 1 : 2;
	}
	else
	{
		if (strcmp(FUNCNAME, opName(MATDIV)) == 0)
		{
			argno1 = 2;
			argno2 = 1;
		}

		op = IRSOLVE;
		startKey = 2;
	}

	if (matop)
	{
		leftRight = (op == ISOLVE) ? "left" : "right";
	}
	else if (nargs >= startKey)
	{
		int         jarg;
		
		for (jarg = startKey; jarg < nargs; jarg++)
		{
			Symbolhandle     symhKey = COMPVALUE(list,jarg);
			int              reply;

			keyword = isKeyword(symhKey);
			if (keyword == (char *) 0)
			{
				sprintf(OUTSTR,
						"ERROR: non-keyword after argument %d to %s()",
						startKey + 1, FUNCNAME);
				goto errorExit;
			}
			
			reply = getOneKeyValue(symhKey, op, SolveKeys, NSOLVEKEYS);

			if (!reply)
			{
				goto errorExit;
			}
			if (reply < 0)
			{
				badKeyword(FUNCNAME, keyword);
				goto errorExit;
			}
		}
	}

	SingOK = SingOK || op == IDET;
	
	if (!logicalIsSet(Quiet))
	{
		Quiet = SingOK && op != IDET;
	}
	
	if (!logicalIsSet(Pack))
	{
		setLogicalTrue(Pack);
	}
	
	if (!realIsSet(MinRcond))
	{
		double     rcondPlus1;
		
		MinRcond = .5;
		do
		{
			MinRcond /= 2.0;
			rcondPlus1 = MinRcond + 1.0;
		} while (rcondPlus1 > 1.0);
	}
	
	arg1 = COMPVALUE(list,0);
	if (!argOK(arg1,REAL,(nargs > 1) ? argno1 : 0))
	{
		goto errorExit;
	}

	if (!isMatrix(arg1, dim) || dim[0] != dim[1])
	{
		if (!matop)
		{
			sprintf(OUTSTR,
					"ERROR: %sargument to %s() not square matrix",
					(nargs > 1) ? "1st " : NullString, FUNCNAME);
		}
		else
		{
			sprintf(OUTSTR,
					"ERROR: %s operand of %s is not square matrix",
					leftRight, FUNCNAME);
		}
	} /*if (!isMatrix(arg1, dim) || dim[0] != dim[1])*/
	else if (anyMissing(arg1) && !matop)
	{
		sprintf(OUTSTR,"ERROR: missing values in 1st argument to %s()",
				FUNCNAME);
	}
	if (*OUTSTR)
	{
		goto errorExit;
	}

	nrows = dim[0];

	if (op == IDET)
	{
		needed = (Mantexp) ? 2 : 1;
		ncols = 1;
	} /*if (op == IDET)*/
	else if (op == ILUDCMP)
	{
		ncols = nrows;
		needed = nrows * ncols;
	}
	else
	{
		if (op == ISOLVE || op == IRSOLVE)
		{
			arg2 = COMPVALUE(list,1);
			if (!argOK(arg2, REAL, argno2))
			{
				goto errorExit;
			}
			if (!isMatrix(arg2,dim) ||
				nrows != ((op == ISOLVE) ? dim[0] : dim[1]))
			{
				if (matop)
				{
					sprintf(OUTSTR,
							"ERROR: dimensions do not match for %s",
							FUNCNAME);
				}
				else
				{
					sprintf(OUTSTR,
							"ERROR: 2nd argument to %s() not matrix with same number of %s as 1st",
							FUNCNAME, (op == ISOLVE) ? "rows" : "cols");
				}
			}
			else if (matop && (anyMissing(arg1) || anyMissing(arg2)))
			{
				sprintf(OUTSTR,
						"ERROR: missing values in operand of %s", FUNCNAME);
			}
			else if (anyMissing(arg2))
			{
				sprintf(OUTSTR,"ERROR: missing values in 2nd argument to %s()",
						FUNCNAME);
			}
			if (*OUTSTR)
			{
				goto errorExit;
			}
			ncols = (op == ISOLVE) ? dim[1] : dim[0];
		} /*if (op == ISOLVE || op == IRSOLVE)*/
		else
		{
			ncols = nrows;
		}
		needed = nrows * ncols;
	} /*if (op == IDET){}else{}*/

	if (op != ILUDCMP)
	{
		out1 = result = RInstall(SCRATCH, needed);
	}
	else
	{
		result = StrucInstall(SCRATCH, 3 + !Pack);
	}
	
	if (result == SymbolError)
	{
		goto errorExit;
	}
	
	if (op == ILUDCMP)
	{
		out1 = COMPVALUE(result, 0) = Makereal(nrows*nrows);
		if (out1 == SymbolError)
		{
			goto errorExit;
		}
	} /*if (op == ILUDCMP)*/
	
	if (op != IDET)
	{
		setNDIMS(out1,2);
		setDIM(out1,1, (op != IRSOLVE) ? nrows : ncols);
		setDIM(out1,2, (op != IRSOLVE) ? ncols : nrows);

		if (op != ILUDCMP && HASLABELS(arg1))
		{
			char      *labs1[2], *labs2[2];
			long       lengths1[2], lengths2[2];

			getMatLabels(arg1, labs1, lengths1);

			if (op == IINVERSE)
			{
				long          tmp = lengths1[0];

				lengths1[0] = lengths1[1];
				lengths1[1] = tmp;

				TMPHANDLE = createLabels(2, lengths1);
				if (TMPHANDLE == (char **) 0 || !setLabels(out1, TMPHANDLE))
				{
					mydisphandle(TMPHANDLE);
					goto errorExit;
				}
				/* Swap labels */
				appendLabels(out1, labs1[1], 0, dontExpand);
				appendLabels(out1, labs1[0], 1, dontExpand);
			} /*if (op == IINVERSE)*/
			else
			{
				char        numericLabel[2];
				char       *tmpLab;

				if (HASLABELS(arg2))
				{
					getMatLabels(arg2, labs2, lengths2);
				}
				else
				{
					numericLabel[0] = NUMERICLABEL;
					numericLabel[1] = '\0';
					labs2[0] = labs2[1] = numericLabel;
					lengths2[0] = lengths2[1] = 2*ncols;
				}

				if (op == ISOLVE)
				{ /* a %\% b*/
					lengths1[1] = lengths2[1];
				}
				else
				{ /* a %/% b*/
					lengths1[0] = lengths2[0];
				}

				TMPHANDLE = createLabels(2, lengths1);
				if (TMPHANDLE == (char **) 0 || !setLabels(out1, TMPHANDLE))
				{
					mydisphandle(TMPHANDLE);
					goto errorExit;
				}
				getMatLabels(arg1, labs1, lengths1);
				if (HASLABELS(arg2))
				{
					getMatLabels(arg2, labs2, lengths2);
				}

				tmpLab = (op == ISOLVE) ? labs1[1] : labs2[0];
				appendLabels(out1, tmpLab, 0,
							 (tmpLab == numericLabel) ? doExpand : dontExpand);
				tmpLab = (op == ISOLVE) ? labs2[1] : labs1[0];
				appendLabels(out1, tmpLab, 1,
							 (tmpLab == numericLabel) ? doExpand : dontExpand);
			} /*if (op == IINVERSE)else{}*/
		} /*if (HASLABELS(arg1))*/	
	} /*if (op != IDET)*/

	needed = (op == IINVERSE || op == ILUDCMP) ? nrows : nrows * (nrows + 1);

	if (!getScratch(ipvtH,GIPVT,nrows,long) ||
		!getScratch(scratchH,GSCRATCH,needed,double))
	{
		goto errorExit;
	}

	ipvt = *ipvtH;
	scratch = *scratchH;
	work = scratch;
	scratch += nrows;

	a = (op == IINVERSE || op == ILUDCMP) ? DATAPTR(out1) : scratch;

	x = DATAPTR(arg1);

	if (op != IRSOLVE)
	{
		length = nrows * nrows;
		DCOPY(&length, x, &inc1, a, &inc1);
	} /*if (op != IRSOLVE)*/
	else
	{ /* copy transpose(arg1) */
		long         j;
		
		for (j = 0; j < nrows; j++)
		{
			DCOPY(&nrows, x + j, &nrows, a + j*nrows, &inc1);
		}
	} /*if (op != IRSOLVE){}else{}*/


	DGECO(a, &nrows, &nrows, ipvt, &rcond, work);
	if (interrupted(DEFAULTTICKS) != INTNOTSET)
	{
		goto errorExit;
	}

	if (rcond <= MinRcond)
	{
		char        *howSing = (rcond) ? "apparently " : NullString;

		if (!Quiet || !SingOK)
		{
			if (!matop)
			{
				sprintf(OUTSTR,
						"%s: argument to %s() is %ssingular",
						(op == IDET || SingOK) ? "WARNING" : "ERROR",
						FUNCNAME, howSing);
			} /*if (!matop)*/
			else
			{
				sprintf(OUTSTR,
						"ERROR: %s operand to %s is %ssingular",
						leftRight, FUNCNAME, howSing);
			} /*if (!matop){}else{}*/		
			putErrorOUTSTR();
		} /*if (!Quiet || !SingOK)*/

		if (op == IDET)
		{
			rcond = 0.0;
			DATAVALUE(result,0) = 0.0;
			if (Mantexp)
			{
				DATAVALUE(result, 1) = 0.0;
			}
		}
		else if (!SingOK)
		{
			goto errorExit;
		}
		else
		{
			Removesymbol(result);
			result = NULLSYMBOL;
		}
		goto normalExit;
	} /*if (rcond <= MinRcond)*/

	if (op == IINVERSE || op == IDET && rcond != 0.0)
	{
		job = (op == IDET) ? 10 : 1;
		DGEDI(a, &nrows, &nrows, ipvt, determ, work, &job);
		if (interrupted(DEFAULTTICKS) != INTNOTSET)
		{
			goto errorExit;
		}
		if (op == IDET)
		{
			if (!Mantexp)
			{
				DATAVALUE(result,0) = determ[0] * intpow(10.0, determ[1]);
			}
			else
			{
				DATAVALUE(result, 0) = determ[0];
				DATAVALUE(result, 1) = (determ[0] != 0.0) ? determ[1] : 0.0;
			}
		} /*if (op == IDET)*/		
	} /*if (op == IINVERSE || op == IDET && rcond != 0.0)*/
	else if (op == ISOLVE || op == IRSOLVE)
	{
		long         j;

		b = DATAPTR(arg2);
		x = DATAPTR(result);
		/* solve column by column (row by row for rsolve() )*/
		for (j = 0;j < ncols;j++)
		{
			tmprow = (op == ISOLVE) ? x : work;
			DCOPY(&nrows, b, (op == ISOLVE) ? &inc1 : &ncols, tmprow, &inc1);
			job = 0;
			DGESL(a, &nrows, &nrows, ipvt, tmprow, &job);
			if (interrupted(DEFAULTTICKS) != INTNOTSET)
			{
				goto errorExit;
			}
			if (op == ISOLVE)
			{
				x += nrows;
				b += nrows;
			}
			else
			{ /* put in row of output */
				DCOPY(&nrows, tmprow, &inc1, x, &ncols);
				x++;
				b++;
			}
		} /*for(j=0;j<ncols;j++)*/
	}
	else if (op == ILUDCMP)
	{
		Symbolhandle    out2, out3, out4;
		double         *lij, *uij;
		long            j;
		int             sign = 1, icomp = 1;
		
		if (!Pack)
		{
			out2 = COMPVALUE(result, icomp) = Makereal(nrows*nrows);
			if (out2 == SymbolError)
			{
				goto errorExit;
			}
			icomp++;
		}
		out3 = COMPVALUE(result, icomp) = Makereal(nrows);
		icomp++;
		out4 = COMPVALUE(result, icomp) = Makereal(1);
		
		if (out3 == SymbolError || out4 == SymbolError)
		{
			goto errorExit;
		}
		
		if (Pack)
		{
			setNAME(out1, "LU");
		}
		else
		{
			setNAME(out1, "L");
			setNAME(out2, "U");
			setNDIMS(out2, 2);
			setDIM(out2, 1, nrows);
			setDIM(out2, 2, nrows);
			symFill(out2, 0.0);
		}
		
		setNAME(out3, "rowperm");
		setNAME(out4, "sign");

		for (j = 0; j < nrows; j++)
		{
			/* initialize component 'rowperm' */
			DATAVALUE(out3, j) = (double) (j + 1);
		}
		
		lij = DATAPTR(out1);
		if (!Pack)
		{
			uij = DATAPTR(out2);
		}
		
		for (j = 0; j < nrows; j++, lij += nrows)
		{
			long      i;

			if (!Pack)
			{
				/* Copy diag and above from LU to U and set it to 0 or 1 */
				for (i = 0; i <= j; i++)
				{
					uij[i] = lij[i];
					lij[i] = 0.0;
				}
				lij[j] = 1.0;
				uij += nrows;
			} /*if (!Pack)*/			
				
			for (i = j+1; i < nrows; i++)
			{
				/* 
				   Reorder and change sign of column j of L taking
				   into account the conventions of DGECO()
				*/
				double    t = lij[i];

				if (i != ipvt[i])
				{
					lij[i] = -lij[ipvt[i]];
					lij[ipvt[i]] = t;
				}
				else
				{
					lij[i] = -t;
				}
			} /*for (i = j+1; i < nrows; i++)*/

			if (ipvt[j] != j)
			{
				double      t = DATAVALUE(out3, j);
				
				/* Update components 'rowperm' and 'sign' */ 
				DATAVALUE(out3, j) = DATAVALUE(out3, ipvt[j]);
				DATAVALUE(out3, ipvt[j]) = t;
				sign = -sign;
			} /*if (ipvt[j] != j)*/
		} /*for (j = 0; j < nrows; j++, lij += nrows)*/
		DATAVALUE(out4, 0) = (double) sign;
	}
	
  normalExit:
	emptyTrash();

#ifdef SEGMENTED
	UNLOADSEG(DDOT);
	UNLOADSEG(DGECO);
#endif /*SEGMENTED*/

#if defined(HASNAN) || defined(HASINFINITY)
	if (result != NULLSYMBOL && anyNaN(result))
	{
		/* Note: anyNaN() replaces NaN's by MISSING*/
		sprintf(OUTSTR,
				"ERROR: unrecoverable overflow in computations of %s%s",
				FUNCNAME, (!matop) ? "()" : NullString);
		goto errorExit;
	}
#endif /*HASNAN || HASINFINITY*/

	return(result);

  errorExit:
	putErrorOUTSTR();
	emptyTrash();
	Removesymbol(result);

#ifdef SEGMENTED
	UNLOADSEG(DDOT);
	UNLOADSEG(DGECO);
#endif /*SEGMENTED*/

	return (0);
} /*solve()*/
